# $Id: bindings.rb,v 1.7 2003/09/21 04:54:06 pervago Exp $

module Kahakai
    # Wraps callback to be attached to Action 
    # * attached to Action#menufunc, Action#winfunc, etc
    #   by Screen#bind (actually Screen::bind, Window::bind, etc)
    class Reaction
        def initialize( type, command, &callback )
            puts "-- Reaction(#{type}, #{command}, #{callback})"
            # TODO accept parameters
            if command
                if command =~ /^\{(.*)\}$/
                    commandName = $1
                    @proc = Proc::new do
                        info = currentScreen.displaystring.split '='    
                        fork do # child
                            Process::setsid
                            ENV[ info[0] ] = info[1]
                            exec commandName
                            exit 0
                        end
                    end
                else
                  funcname = "#{type}Ev#{command}".sub("Kahakai::", "")
                  @proc = Kahakai.const_get(funcname).new()
                end     
            elsif callback
                puts "Assigning callback #{callback}"
                @proc = callback
            end 
        end

        def call( *args )
            puts "--- Reaction.call() #{@proc}"
            @proc.call( *args )
        end
    end

    class Screen
        ACTIONS = []
        # Central point-of-access to set bindings
        # Examples:
        #
        # Setting up mousewheel desktop switching:
        #   screen.bind ButtonPress::new('Button4'), 'PreviousDesktop'
        #   screen.bind ButtonPress::new('Button5'), 'NextDesktop'
        #
        # 
        #
        def Screen::processFunction(command, &callback)
            # superclass check is ugly but we need because we need to know
            # when command is a real C++ object.
            if command.class.superclass == ScreenFunction
                command
            elsif command.instance_of?(String) && !command.include?('{')
                return Kahakai.const_get("ScreenEv#{command}").new()
            else
                Kahakai::createScreenFunction(Reaction.new( Screen, command, 
                                                           &callback ))
            end
        end

        def bind( fulltarget, *args, &callback )
            case fulltarget
            when Event
                Screen::bind( self, fulltarget, *args, &callback )
            else
                fulltarget = fulltarget.to_s.downcase
                if fulltarget == 'all'
                    ["edge", "window", "menu"].each do |target|
                        bind("#{target}.all", *args, &callback)
                    end
                elsif fulltarget =~ /^(edge|window|menu)[._](.+)$/
                    klass = Kahakai::const_get ($1.capitalize)
                    if $2 == 'all'
                        klass::TARGETS.each_key do |target|
                                klass::bind( self, target, *args, &callback )
                        end
                    else
                        klass::bind( self, $2, *args, &callback )
                    end
                end
            end
        end


        def bindList( fulltarget, list )
            # Can be way more efficient by evaluating fulltarget just once
            list.each {|item| bind(fulltarget, *item)}
        end

        # To be called from Screen#bind
        def Screen::bind(screen, event, command=nil, param=nil, 
                         edge=false, &callback )
            action = Action::new
            action.param = param.to_s if param 
                
            if event.set action
                action.rootfunc = processFunction(command, &callback)
                screen.config.rootacts.push action unless edge

                puts action.rootfunc
            end
            action
        end
    end

    # Dummy Edge class
    # * merely used for ease-of-binding
    class Edge
        TARGETS = { 
            :n  => :neacts, 
            :s  => :seacts, 
            :e  => :eeacts, 
            :w  => :weacts 
        }
        # Bind an event/action to an edge
        # To be called from Screen#bind
        def Edge::bind( screen, direction, event, command=nil, param=nil, &callback )
            direction = direction.to_sym
            if Edge::TARGETS.key? direction
                action = Action.new
                action.rootfunc = Screen::processFunction(command, &callback)
                action.param = param.to_s if param
                action_list = screen.config.send Edge::TARGETS[ direction ]
                action_list.push action
            end
            action
        end
    end

    # Window class
    class Window
        TARGETS = {
            :frame         => :frameacts,
            :title         => :titleacts,
            :label         => :labelacts,
            :handle        => :handleacts,
            :activeclient  => :awinacts,
            :focused       => :awinacts,
            :passiveclient => :pwinacts,
            :unfocused     => :pwinacts,
            :leftgrip      =>  :lgacts,
            :rightgrip     => :rgacts,
        }

        def Window.processFunction(command, &callback)
            if [WindowFunction, ScreenFunction].include? command.class.superclass
                command
            elsif command.instance_of? String
                if Kahakai.const_defined? "WindowEv#{command}"
                    Kahakai.const_get("WindowEv#{command}").new()
                else
                    Kahakai.const_get("ScreenEv#{command}").new()
                end
            else
                Kahakai::createWindowFunction(Reaction.new( Screen, command, 
                                                           &callback ))
            end
        end
        
        # To be called from Screen#bind
        def Window.bind(screen, target, event, command=nil, param=nil, 
                         &callback)
            puts "Window.bind #{screen}  #{target} #{event} #{command}"

            target = target.to_s.downcase.to_sym
            action = Action::new
            if event.set action
                if target.to_s =~ /^button\.(.+)$/
                    action_list = screen.config.bacts[ $1 ]
                    unless action_list
                        warn "Invalid button ID: #{$1}"
                        return
                    end
                else
                    action_list = screen.config.send Window::TARGETS[ target ]
                end
                if action_list
                    action.winfunc = processFunction(command, &callback)
                    action.param = param.to_s if param                    
                    action_list.push action
                else
                    warn "Unknown/Unsupported window target: #{target}"
                    return
                end     
            end
            action
        end
        
    end
    
    class Menu
        TARGETS = {            
            :title    => :mtacts,
            :item     => :miacts,
            :sub      => :msacts,
            :checkbox => :mcbacts
        }

        def Menu.processFunction(command, &callback)
            if [WindowFunction, ScreenFunction, MenuFunction].include? command.class.superclass
                command
            elsif command.instance_of? String
                if Kahakai.const_defined? "MenuEv#{command}"
                    Kahakai.const_get("MenuEv#{command}").new()
#                elsif Kahakai.const_defined? "WindowEv#{command}"
#                    Kahakai.const_get("WindowEv#{command}").new()
                else
                    Kahakai.const_get("ScreenEv#{command}").new()
                end
            else
                MenuReaction.new( Window, command, &callback )
            end
        end

        # To be called from Screen#bind
        def Menu::bind( screen, target, event, command=nil, param=nil, &callback )
            target = target.to_s.downcase.to_sym
            action = Action::new
            if event.set action
                action_list = screen.config.send Menu::TARGETS[ target ]
                if action_list
                    func = processFunction(command, &callback)
                    puts func
                    action.menufunc = func
                    action.param = param.to_s if param
                    action_list.push action
                else
                    warn "Unknown/Unsupported menu target: #{target}"
                end
            end
            action
        end
    end
    
end

