/* $Id: ChildWindow.cc,v 1.16 2003/11/29 04:47:29 pervago Exp $ */

/**
 * @file   ChildWindow.cc
 * @date   31 Aug 2003
 *
 * @brief Impleemnts KaChildWindow class which used for grips, handle
 * bar, buttons, etc.
 *
 */

#include "ChildWindow.hh"
#include "Window.hh"
#include "Screen.hh"
#include "Resources.hh"

#include "Imlib2pp.hh"

#include "DrawOp.hh"
#include "ImageControl.hh"
#include "ExpressionEvaluator.hh"
#include "Texture.hh"

#include <boost/shared_ptr.hpp>
using boost::shared_ptr;

using std::make_pair;

/**
 * @fn    KaChildWindow(KaWindow *ka_win, Window parent, int bw, int type) :
 *        WindowObject(0, type)
 * @brief Constructor for KaChildWindow class
 *
 * Creates a child window, could be of one of these types: FrameType,
 * TitleType, LabelType, HandleType, CButtonType, IButtonType, MButtonType,
 * LGripType, RGripType.
 *
 * @param ka_win KaWindow who wish to use the child window
 * @param parent Parent window to child window
 * @param type Type of window
 */
KaChildWindow::KaChildWindow(KaWindow *ka_win, Window parent, int type) :
    WindowObject(0, type) {

    ka = ka_win;
    kascreen = ka->kascreen;
    display = ka->display;

    attrib.setX(0);
    attrib.setY(0);
    attrib.setWidth(1);
    attrib.setHeight(1);
    
    control = new ImageControl(kascreen, kascreen->display, 
                               kascreen->screen_number, 
                               kascreen->ic->getVisual(), 
                               kascreen->ic->getColormap());
}

/**
 * @fn    ~KaChildWindow()
 * @brief Destructor for KaChildWindow class
 *
 * Destroys the window and removes it from the window_table hash_map.
 */
KaChildWindow::~KaChildWindow(void) {
    delete control;

    ka->kahakai->window_table.erase(id);
    XDestroyWindow(display, id);
}

void KaChildWindow::clearCache() {
    control->clearCache();
}

/** 
 * Checks whether child window is at least partly visible.
 * 
 * @return True if any part of window is visible, false otherwise.
 */
bool KaChildWindow::isVisible() {
    int wx = ka->GetFrameLeft(), wy = ka->GetFrameTop();

    int workx, worky, workw, workh;
    kascreen->GetWorkareaSize(&workx, &worky, &workw, &workh);

    return intersects(wx + attrib.getLeft(), wy  + attrib.getTop(),
                      wx + attrib.getRight(), wy  + attrib.getBottom(),
                      workx, worky, workx + workw, worky + workh);
}


/**
 * @fn    Render(void)
 * @brief Render KaChildWindow background
 *
 * Renders KaChildWindow background pixmap for the current window state.
 */
void KaChildWindow::render() {
    XSetWindowBackgroundPixmap(display, id, ParentRelative);
    XClearWindow(display, id);
}

/**
 * @fn    createWindow(Window parent, int create_mask, 
 *                     XSetWindowAttributes attrib_set)
 * @brief Creates window using child window's dimensions and specified 
 *        attributes
 *
 */
void KaChildWindow::createWindow(Window parent, int create_mask,
                                 XSetWindowAttributes* attrib_set) {

    if (create_mask == 0)
        create_mask = CWOverrideRedirect | CWBorderPixel | CWEventMask |
            CWColormap;

    XSetWindowAttributes default_attrib_set;

    if (!attrib_set) {
        default_attrib_set.border_pixel = 
            ka->kascreen->style.window.borderColor.getPixel();

        default_attrib_set.event_mask = ButtonPressMask | ButtonReleaseMask |
            EnterWindowMask | LeaveWindowMask;

        default_attrib_set.colormap = ka->kascreen->colormap;
        default_attrib_set.override_redirect = true;

        attrib_set = &default_attrib_set;
    }

    id = XCreateWindow(display, parent, attrib.getX(), attrib.getY(),
                       attrib.getWidth(), attrib.getHeight(), 0, 
                       CopyFromParent,
                       CopyFromParent, CopyFromParent, create_mask,
                       attrib_set);

    ka->kahakai->window_table.insert(make_pair(id, this));    
}

KaWindowFrame::KaWindowFrame(KaWindow* kawin, Window parent) 
    : KaChildWindow(kawin, parent, FrameType)
{
    XSetWindowAttributes attrib_set;

    int create_mask = CWOverrideRedirect | CWBorderPixel | CWEventMask |
        CWColormap;
    attrib_set.border_pixel = 
        ka->kascreen->style.window.borderColor.getPixel();

    attrib_set.colormap = ka->kascreen->colormap;
    attrib_set.override_redirect = true;
    attrib_set.event_mask = ButtonPressMask | ButtonReleaseMask |
        EnterWindowMask | LeaveWindowMask;
    attrib_set.event_mask |= SubstructureRedirectMask;
    create_mask |= CWBackPixmap;
    attrib_set.background_pixmap = ParentRelative;

    attrib.setX(ka->GetFrameLeft());
    attrib.setY(ka->GetFrameTop());
    attrib.setWidth(ka->GetFrameWidth());
    attrib.setHeight(ka->GetFrameHeight());  

    createWindow(parent, create_mask, &attrib_set);
}

void KaChildWindow::update() {
    focusedRenderedTexture = TexturePtr();
    unfocusedRenderedTexture = TexturePtr();

    control->cleanup();

    render();
}


KaWindowLabel::KaWindowLabel(KaWindow* kawin, Window parent) 
    : KaChildWindow(kawin, parent, LabelType)
{   
    XSetWindowAttributes attrib_set;

    int create_mask = CWOverrideRedirect | CWBorderPixel | CWEventMask |
        CWColormap;
    attrib_set.border_pixel = 
        ka->kascreen->style.window.borderColor.getPixel();

    attrib_set.event_mask = ButtonPressMask | ButtonReleaseMask |
        EnterWindowMask | LeaveWindowMask;

    attrib_set.colormap = ka->kascreen->colormap;
    attrib_set.override_redirect = true;
    attrib_set.event_mask |= ExposureMask;

    createWindow(parent, create_mask, &attrib_set);
}

KaWindowLabel::~KaWindowLabel() {
}


KaWindowHandle::KaWindowHandle(KaWindow* kawin, Window parent) 
    : KaChildWindow(kawin, parent, HandleType)
{
    createWindow(parent);

    focusedTexture = kawin->kascreen->style.window.focused.handleTexture;
    unfocusedTexture = kawin->kascreen->style.window.unfocused.handleTexture;

    pixmap = None;

    focusedRenderedTexture = draw(true);
    unfocusedRenderedTexture = draw(false);
}

KaWindowHandle::~KaWindowHandle() {    
    if (pixmap != None) imlib_free_pixmap_and_mask(pixmap);
}

TexturePtr KaWindowHandle::draw(bool focused) {
    if (!focusedTexture || !unfocusedTexture) return TexturePtr();

    DrawOps ops;
    ops.push_back(shared_ptr<DrawOp>
                  (new DrawOpTexture(focused ? 
                                     focusedTexture : 
                                     unfocusedTexture,
                                     "0", "0", "width", "height")));
    
    if (!kascreen->style.window.usesDrawing && 
        kascreen->style.window.borderThickness >= 1) {
        KaColor borderKaColor = kascreen->style.window.borderColor;

        ImColor borderColor(borderKaColor.getRed(), 
                            borderKaColor.getGreen(), 
                            borderKaColor.getBlue());

        std::string thicknessStr = 
            to_string(kascreen->style.window.borderThickness);

        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpFillRect(borderColor,
                                          "0", "height - " + thicknessStr,
                                          "width", "height")));
    }
    return TexturePtr(new Texture(ops));
}


void KaWindowHandle::render() {
    if (!isVisible()) return;

    if (pixmap != None && pixmap != ParentRelative) 
        imlib_free_pixmap_and_mask(pixmap);

    if ((focusedTexture && !focusedRenderedTexture) || 
        (unfocusedTexture && !unfocusedRenderedTexture)) { 
        focusedRenderedTexture = draw(true);
        unfocusedRenderedTexture = draw(false);
    }

    pixmap = control->render(id, ka->has_focus ? 
                             focusedRenderedTexture : unfocusedRenderedTexture,
                             ka->GetFrameLeft() + attrib.getX(),
                             ka->GetFrameTop() + attrib.getY(),
                             attrib.getWidth(), 
                             attrib.getHeight());    

    XSetWindowBackgroundPixmap(display, id, pixmap);
    XClearWindow(display, id);    
}

KaWindowButton::KaWindowButton(KaWindow* kawin, Window parent) 
    : KaChildWindow(kawin, parent, ButtonType)
{
    XSetWindowAttributes attrib_set;

    attrib_set.border_pixel = 
        ka->kascreen->style.window.borderColor.getPixel();
    attrib_set.colormap = ka->kascreen->colormap;
    attrib_set.override_redirect = true;
    attrib_set.event_mask = ButtonPressMask | ButtonReleaseMask |
        EnterWindowMask | LeaveWindowMask | ExposureMask;

    createWindow(parent, 0, &attrib_set);

    pressed = false;
}

bool KaWindowButton::isActive() {
    switch (bstyle->cb) {
    case MaxCBoxType: 
        return ka->flags.max;
    case ShadeCBoxType: 
        return ka->flags.shaded;
    default:
        return false;
    }
}

ButtonStateStyle * KaWindowButton::getButtonStateStyle() {
    if (isActive()) {        
        if (pressed)
            return &bstyle->pressedActive;
        if (ka->has_focus)
            return &bstyle->focusedActive;
        
        return &bstyle->unfocusedActive;
    }

    if (pressed)
        return &bstyle->pressed;
    if (ka->has_focus)
        return &bstyle->focused;

    return &bstyle->unfocused;
}


KaWindowGrip::KaWindowGrip(KaWindow* kawin, Window parent, Side side) 
    : KaChildWindow(kawin, parent, side == Left ? LGripType : RGripType)
{   
    XSetWindowAttributes attrib_set;

    attrib_set.border_pixel = 
        ka->kascreen->style.window.borderColor.getPixel();
    attrib_set.colormap = ka->kascreen->colormap;
    attrib_set.override_redirect = true;
    attrib_set.event_mask = ButtonPressMask | ButtonReleaseMask |
        EnterWindowMask | LeaveWindowMask;

    attrib_set.cursor = side == Left ?
        ka->kahakai->resizesw_cursor :ka->kahakai->resizese_cursor;

    createWindow(parent, 
                 CWOverrideRedirect | CWBorderPixel | CWEventMask | 
                 CWColormap | CWCursor,
                 &attrib_set);

    WindowStyle& wstyle = kawin->kascreen->style.window;

    if (side == Left) {
        focusedTexture = wstyle.focused.leftGripTexture;
        unfocusedTexture = wstyle.unfocused.leftGripTexture;
    } else {
        focusedTexture = wstyle.focused.rightGripTexture;
        unfocusedTexture = wstyle.unfocused.rightGripTexture;
    }

    focusedRenderedTexture = draw(true);
    unfocusedRenderedTexture = draw(false);

    pixmap = None;
}

KaWindowGrip::~KaWindowGrip() {    
    if (pixmap != None) imlib_free_pixmap_and_mask(pixmap);
}

TexturePtr KaWindowGrip::draw(bool focused) {
    if (!focusedTexture || !unfocusedTexture) return TexturePtr();

    DrawOps ops;
    ops.push_back(shared_ptr<DrawOp>
                  (new DrawOpTexture(focused ? 
                                     focusedTexture : unfocusedTexture,
                                     "0", "0", "width", "height")));

    if (!kascreen->style.window.usesDrawing &&  
        kascreen->style.window.borderThickness >= 1) {
        KaColor borderKaColor = kascreen->style.window.borderColor;

        ImColor borderColor(borderKaColor.getRed(), 
                            borderKaColor.getGreen(), 
                            borderKaColor.getBlue());

        std::string thicknessStr = 
            to_string(kascreen->style.window.borderThickness);

        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpFillRect(borderColor,
                                          "0", "height - " + thicknessStr,
                                          "width", "height")));
                      
        if (type == LGripType)
            ops.push_back(shared_ptr<DrawOp>
                          (new DrawOpFillRect(borderColor, 
                                              "0", "0", thicknessStr, 
                                              "height")));
        else
            ops.push_back(shared_ptr<DrawOp>
                          (new DrawOpFillRect(borderColor, 
                                             "width-" + thicknessStr, "0", 
                                              "width", "height")));
    }
    return TexturePtr(new Texture(ops));
}

void KaWindowGrip::render() {
    if (!isVisible()) return;

    if (pixmap != None && pixmap != ParentRelative) 
        imlib_free_pixmap_and_mask(pixmap);

    if ((focusedTexture && !focusedRenderedTexture) || 
        (unfocusedTexture && !unfocusedRenderedTexture)) { 
        focusedRenderedTexture = draw(true);
        unfocusedRenderedTexture = draw(false);
    }
        
    pixmap = control->render(id, 
                             ka->has_focus ? 
                             focusedRenderedTexture : 
                             unfocusedRenderedTexture, 
                             ka->GetFrameLeft() + attrib.getX(),
                             ka->GetFrameTop() + attrib.getY(),
                             attrib.getWidth(), 
                             attrib.getHeight());    

    XSetWindowBackgroundPixmap(display, id, pixmap);
    XClearWindow(display, id);    
}


KaWindowEdge::KaWindowEdge(KaWindow* kawin, Window parent, Side side) 
    : KaChildWindow(kawin, parent, side == Left ? LeftEdgeType : RightEdgeType)
{   
    XSetWindowAttributes attrib_set;

    attrib_set.border_pixel = 
        ka->kascreen->style.window.borderColor.getPixel();
    attrib_set.colormap = ka->kascreen->colormap;
    attrib_set.override_redirect = true;
    attrib_set.event_mask = ButtonPressMask | ButtonReleaseMask |
        EnterWindowMask | LeaveWindowMask;

    //    attrib_set.cursor = side == Left ?
    //        ka->kahakai->resizew_cursor :ka->kahakai->resizee_cursor;

    createWindow(parent, 
                 CWOverrideRedirect | CWBorderPixel | CWEventMask | 
                 CWColormap /*| CWCursor*/,
                 &attrib_set);

    const WindowStyle& wstyle = kawin->kascreen->style.window;

    focusedTexture = side == Left ? wstyle.focused.leftEdgeTexture : 
        wstyle.focused.rightEdgeTexture;
    unfocusedTexture = side == Left ? wstyle.unfocused.leftEdgeTexture :
        wstyle.unfocused.rightEdgeTexture;

    focusedRenderedTexture = draw(true);
    unfocusedRenderedTexture = draw(false);

    pixmap = None;
}

KaWindowEdge::~KaWindowEdge() {    
    if (pixmap != None) imlib_free_pixmap_and_mask(pixmap);
}

TexturePtr KaWindowEdge::draw(bool focused) {
    if (!focusedTexture || !unfocusedTexture) return TexturePtr();

    DrawOps ops;
    ops.push_back(shared_ptr<DrawOp>
                  (new DrawOpTexture(focused ? 
                                    focusedTexture : unfocusedTexture,
                                     "0", "0", "width", "height")));

    if (!kascreen->style.window.usesDrawing && 
        kascreen->style.window.borderThickness >= 1) {
        KaColor borderKaColor = kascreen->style.window.borderColor;

        ImColor borderColor(borderKaColor.getRed(), 
                            borderKaColor.getGreen(), 
                            borderKaColor.getBlue());

        std::string thicknessStr = 
            to_string(kascreen->style.window.borderThickness);

        //        if (type == LeftEdgeType)
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpFillRect(borderColor, 
                                         "0", "0", thicknessStr, 
                                          "height")));
        //        else
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpFillRect(borderColor, 
                                         "width-" + thicknessStr, "0", 
                                          "width", "height")));        
    }
    return TexturePtr(new Texture(ops));
}

void KaWindowEdge::render() {
    if (!isVisible()) return;

    if (pixmap != None && pixmap != ParentRelative) 
        imlib_free_pixmap_and_mask(pixmap);

    if ((focusedTexture && !focusedRenderedTexture) || 
        (unfocusedTexture && !unfocusedRenderedTexture)) { 
        focusedRenderedTexture = draw(true);
        unfocusedRenderedTexture = draw(false);
    }
        
    pixmap = control->render(id, ka->has_focus ? 
                             focusedRenderedTexture : 
                             unfocusedRenderedTexture,
                             ka->GetFrameLeft() + attrib.getX(),
                             ka->GetFrameTop() + attrib.getY(),
                             attrib.getWidth(), 
                             attrib.getHeight());   

    XSetWindowBackgroundPixmap(display, id, pixmap);
    XClearWindow(display, id);
}

