/* $Id: Dockapp.cc,v 1.18 2003/12/01 00:18:18 pervago Exp $ */

/**
 * @file   Dockapp.cc
 * @date   29-Nov-2001 22:13:22
 *
 * @brief Implementation of DockappHandler class
 *
 * This class handles docking of 'dockapp' programs.
 *
 */

#include "Dockapp.hh"
#include "Screen.hh"
#include "Resources.hh"
#include "SimpleMatch.hh"

#include <list>

using std::make_pair;
using std::list;
using std::vector;

/**
 * @fn    DockappHandler(void)
 * @brief Constructor for DockappHandler class
 *
 * Creates the dockapp handler window.
 *
 * @param scrn KaScreen to create dockapp handler window on.
 * @param ds Style structure to use when creating this dockapphandler
 */
DockappHandler::DockappHandler(KaScreen *scrn, DockStyle *ds) :
    WindowObject(0, DockHandlerType) {
    XSetWindowAttributes attrib_set;

    style = ds;
    kascreen = scrn;
    kahakai = kascreen->kahakai;
    display = kahakai->display;
    pixmap = None;
    hidden = true;
    mapped = false;
    x = 1;
    y = 1;
    if (style->geometry & XValue ||
        style->geometry & YValue) {
        if (style->geometry & XValue) {
            x = style->x;
        }
        if (style->geometry & YValue) {
            y = style->y;
        }
    } else
        style->geometry = XValue | YValue | XNegative;
    
    width = 0;
    height = 0;
    
    attrib_set.border_pixel = style->style.borderColor.getPixel();
    attrib_set.colormap = kascreen->colormap;
    attrib_set.override_redirect = true;
    attrib_set.event_mask = SubstructureRedirectMask | ButtonPressMask |
        EnterWindowMask | LeaveWindowMask;

    id = XCreateWindow(display, kascreen->id, 0, 0, 1, 1,
                       style->style.borderThickness, kascreen->screen_depth,
                       CopyFromParent, kascreen->visual, CWOverrideRedirect |
                       CWEventMask | CWColormap | CWBorderPixel,
                       &attrib_set);

    DESKTOP_LOOP_MASK(kascreen, style->desktop_mask)
        if (style->stacking == AlwaysOnTop)
            (*__dit)->aot_stacking_list.push_back(id);
        else
            (*__dit)->aab_stacking_list.push_back(id);
        

    if (! style->inworkspace) {
        wm_strut = new WMstrut;
        wm_strut->window = id;
        wm_strut->left = 0;
        wm_strut->right = 0;
        wm_strut->top = 0;
        wm_strut->bottom = 0;
        kascreen->strut_list.push_back(wm_strut);
    }
    kahakai->window_table.insert(make_pair(id, this));

    control = new ImageControl(kascreen, display, 
                               kascreen->screen_number,
                               kascreen->visual, 
                               kascreen->colormap);
}

/**
 * @fn    ~DockappHandler(void)
 * @brief Destructor for DockappHandler class
 *
 * Removes all dockapps and destroys the dockapp handler window.
 */
DockappHandler::~DockappHandler(void) {
    DESKTOP_LOOP_MASK(kascreen, style->desktop_mask)
        if (style->stacking == AlwaysOnTop)
            (*__dit)->aot_stacking_list.remove(id);
        else
            (*__dit)->aab_stacking_list.remove(id);
    XDestroyWindow(display, id);
    if (! style->inworkspace) {
        kascreen->strut_list.remove(wm_strut);
        delete wm_strut;
    }
    kahakai->window_table.erase(id);

    delete control;
}

/**
 * @fn    Show()
 * @brief Show the dockapp handler
 *
 * Shows & redraws the dock app handler.
 */

void DockappHandler::Show() {
    hidden = false;

    if (!dockapp_list.empty()) {
        XMapWindow(display, id);
        mapped = true;
        Render();
    }
}

/**
 * @fn    Hide()
 * @brief Hide the dockapp handler
 *
 */

void DockappHandler::Hide() {
    if (mapped) {
        XUnmapWindow(display, id);
    }

    hidden = true;
}


/**
 * @fn    Update(void)
 * @brief Update the dockapp handler
 *
 * Repositions all dockapps. Moves and resizes dockapp handler.
 */
void DockappHandler::Update() {
    if (dockapp_list.empty()) {
        if (! style->inworkspace) {
            wm_strut->left = 0;
            wm_strut->right = 0;
            wm_strut->top = 0;
            wm_strut->bottom = 0;
            kascreen->UpdateWorkarea();
        }
        XUnmapWindow(display, id);
        return;
    }
    map_x = x;
    map_y = y;
    width = style->gridspace;
    height = style->gridspace;

    reorderDockapps();
    updatePositions();
}

/** 
 * Reorders dock apps based on dock order specified in ~/.kahakai/config
 * 
 */

void DockappHandler::reorderDockapps() {
    list<Dockapp *>::iterator it = dockapp_list.begin();
    for (; it != dockapp_list.end(); ++it)
        (*it)->added = false;

    list<Dockapp *>::reverse_iterator d_it;
    vector<SimpleMatch *>::reverse_iterator reg_it = style->order.rbegin();

    for (; reg_it != style->order.rend(); ++reg_it) {        
        for (d_it=dockapp_list.rbegin(); d_it != dockapp_list.rend(); ++d_it) {
            Dockapp* dockApp = *d_it;            
            if (dockApp->c_hint && !dockApp->added && 
                ((*reg_it)->matches(dockApp->c_hint->res_name))) {
                dockApp->added = true;
                dockapp_list.remove(dockApp);
                dockapp_list.push_front(dockApp);
            }
        }
    }
}

/** 
 * Updates dockapp positions, struts and resizes the dock window.
 * 
 */
void DockappHandler::updatePositions() {
    int dock_x = style->gridspace;
    int dock_y = style->gridspace;    

    list<Dockapp *>::iterator it;
    
    it = dockapp_list.begin();
    for (; it != dockapp_list.end(); ++it) {
        switch (style->direction) {
            case VerticalDock:
                if (((*it)->width + style->gridspace * 2) > width)
                    width = (*it)->width + style->gridspace * 2;
                break;
            case HorizontalDock:
                if (((*it)->height + style->gridspace * 2) > height)
                    height = (*it)->height + style->gridspace * 2;
                break;
        }
    }
    it = dockapp_list.begin();
    XGrabServer(display);
    for (; it != dockapp_list.end(); ++it) {
        if (validatedrawable((*it)->id)) {
            switch (style->direction) {
                case VerticalDock:
                    dock_y = height;
                    height += (*it)->height + style->gridspace;
                    dock_x = (((width - style->gridspace * 2) -
                               (*it)->width) / 2) + style->gridspace;
                    break;
                case HorizontalDock:
                    dock_x = width;
                    width += (*it)->width + style->gridspace;
                    dock_y = (((height - style->gridspace * 2) -
                               (*it)->height) / 2) + style->gridspace;
                    break;
            }
            (*it)->x = dock_x;
            (*it)->y = dock_y;
            XMoveWindow(display, (*it)->id, dock_x, dock_y);
        }
    }
    XUngrabServer(display);

    if (! style->inworkspace)
        wm_strut->left = wm_strut->right = wm_strut->top =
            wm_strut->bottom = 0;
    if (style->geometry & XNegative) {
        map_x = kascreen->width - style->style.borderThickness * 2 -
            width + x;
        if (! style->inworkspace)
            wm_strut->right = kascreen->width - map_x;
    } else {
        if (! style->inworkspace)
            wm_strut->left = map_x + style->style.borderThickness * 2 + width;
    }
    
    if (style->geometry & YNegative) {
        map_y = kascreen->height - style->style.borderThickness * 2 -
            height + y;
        if (style->direction == HorizontalDock && (! style->inworkspace)) {
            wm_strut->bottom = kascreen->height - map_y;
            wm_strut->right = wm_strut->left = 0;
        }
    } else
        if (style->direction == HorizontalDock && (! style->inworkspace)) {
            wm_strut->top = map_y + style->style.borderThickness * 2 + height;
            wm_strut->right = wm_strut->left = 0;
        }

    if (style->centered) {
        switch (style->direction) {
            case VerticalDock: map_y = kascreen->height / 2 - height / 2;
                break;
            case HorizontalDock: map_x = kascreen->width / 2 - width / 2;
                break;
        }
    }
    XResizeWindow(display, id, width, height);
    XMoveWindow(display, id, map_x, map_y);
    if (style->desktop_mask & (1L << kascreen->current_desktop->number)) {
        XMapWindow(display, id);
        hidden = false;
        Render();
        kascreen->UpdateWorkarea();
    }
}

/**
 * @fn    Render(void)
 * @brief Render background
 *
 * Renders background for dockapp holder.
 */
void DockappHandler::Render() {    
    if (pixmap != None && pixmap != ParentRelative) 
        imlib_free_pixmap_and_mask(pixmap);
        
    pixmap = control->render(id, style->style.texture,
                             map_x + style->style.borderThickness,
                             map_y + style->style.borderThickness,
                             width, height);

    XSetWindowBackgroundPixmap(display, id, pixmap);
    XClearWindow(display, id);
}
    
/**
 * @fn    Dockapp(Window win, DockappHandler *dhand)
 * @brief Constructor for Dockapp class
 *
 * Reparents window to dockapp handler window and adds it to the window_table
 * hash_map.
 */
Dockapp::Dockapp(Window win, DockappHandler *dhand) :
    WindowObject(win, DockAppType) {
    XWindowAttributes attrib;
    dh = dhand;
    client_id = win;
    display = dh->display;
    deleted = false;
    c_hint = NULL;
    title = NULL;

    XWMHints *wmhints = XGetWMHints(display, win);
    if (wmhints) {
        if ((wmhints->flags & IconWindowHint) &&
            (wmhints->icon_window != None)) {
            XUnmapWindow(display, client_id);
            icon_id = wmhints->icon_window;
            id = icon_id;
        } else {
            icon_id = None;
            id = client_id;
        }
        XFree(wmhints);
    } else {
        icon_id = None;
        id = client_id;
    }    
    XGrabServer(display);
    if (validatedrawable(id)) {
        if (XGetWindowAttributes(display, id, &attrib)) {
            width = attrib.width;
            height = attrib.height;
        } else
            width = height = 64;
        
        XSetWindowBorderWidth(display, id, 0);
        XReparentWindow(display, id, dh->id, dh->width, dh->height);
        XMapRaised(display, id);
        XSelectInput(display, id, StructureNotifyMask |
                     SubstructureNotifyMask);
    } else {
        XUngrabServer(display);
        delete this;
        return;
    }
    XUngrabServer(display);
    dh->kahakai->window_table.insert(make_pair(id, this));
    dh->dockapp_list.push_back(this);
}

/**
 * @fn    ~Dockapp(void)
 * @brief Destructor for Dockapp class
 *
 * Reparents dockapp window back to root if it still exists and removes it
 * from the window_table hash_map and dockapp list.
 */
Dockapp::~Dockapp(void) {
    dh->dockapp_list.remove(this);
    dh->kahakai->window_table.erase(id);
    if (! deleted) {
        XGrabServer(display);
        if (validatedrawable(id)) {
            if (icon_id) XUnmapWindow(display, id);
            XReparentWindow(display, id, dh->kascreen->id,
                            dh->map_x + x, dh->map_y + y);
            XMapWindow(display, client_id);
        }
        XUngrabServer(display);
    }
    if (c_hint) {
        XFree(c_hint->res_name);
        XFree(c_hint->res_class);
        XFree(c_hint);
    }
    if (title) XFree(title);

}
