/**
 * @file   DrawOp.hh
 * @date   Sat Oct  4 23:07:39 2003
 * 
 * @brief Definition of DrawOp classes which operate on ImImage such as
 *        drawing lines, rectangles, gradients, and interlacing.
 * 
 */

#ifndef __DrawOp_hh
#define __DrawOp_hh

#include "Imlib2pp.hh"
#include "ExpressionEvaluator.hh"

#include <string>
#include <map>

#include <boost/shared_ptr.hpp>

using boost::shared_ptr;

class Texture;

// Base class for drawing operations
class DrawOp {
public:
    DrawOp();
    virtual ~DrawOp();
    virtual void draw(ImImage& img) {};
protected:
    std::map<std::string, float> variables;
    ExpressionEvaluator evaluator;
    int eval(const ImImage& img, std::string expr);
};

typedef std::vector<shared_ptr<DrawOp> > DrawOps;

// Sample class, draw a rectangle
class DrawOpRect : public DrawOp {
public:
    DrawOpRect(ImColor c, int _x, int _y, int _w, int _h) :
        dynamic(false), x(_x), y(_y), width(_w), height(_h), color(c) {};

    DrawOpRect(ImColor c, std::string _x, std::string _y,
               std::string _w, std::string _h) :
        dynamic(true), 
        expr_x(_x), expr_y(_y), expr_width(_w), expr_height(_h), color(c) {};

    virtual void draw(ImImage& img);
protected:
    bool dynamic;

    int x, y, width, height;
    std::string expr_x, expr_y, expr_width, expr_height;
    ImColor color;
};

// Another sample class, fill a rectangle
class DrawOpFillRect : public DrawOpRect {
public:
    DrawOpFillRect(ImColor c, int _x, int _y, int _w, int _h) :
        DrawOpRect(c, _x, _y, _w, _h) {};
    DrawOpFillRect(ImColor c, std::string _x, 
                   std::string _y, std::string _w, std::string _h) :
        DrawOpRect(c, _x, _y, _w, _h) {};

    virtual void draw(ImImage& img);
};

// Another sample class, fill a rectangle
class DrawOpLine : public DrawOp {
public:
    DrawOpLine(ImColor c, int _x, int _y, int _x2, int _y2) :
        dynamic(false), color(c), x(_x), y(_y), x2(_x2), y2(_y2) {};

    DrawOpLine(ImColor c, 
               std::string _x, std::string _y, 
               std::string _x2, std::string _y2) :
        dynamic(true), color(c), expr_x(_x), 
        expr_y(_y), expr_x2(_x2), expr_y2(_y2) {};

    virtual void draw(ImImage& img);
private:
    bool dynamic;
    ImColor color;

    int x, y, x2, y2;
    std::string expr_x, expr_y, expr_x2, expr_y2;
};



// Another sample class, draw a point
class DrawOpPoint : public DrawOpRect {
public:
    DrawOpPoint(ImColor c, int x, int y) : 
        DrawOpRect(c, x, y, 1, 1) {};
};

enum ImageDrawType {
    DrawScaled,
    DrawStretched,
    DrawTiled
};

class DrawOpImage : public DrawOp {
public:
    DrawOpImage(std::string filename, int _x, int _y, 
                unsigned char _alpha=255, ImageDrawType _drawType=DrawTiled,
                unsigned int borderLeft=0, 
                unsigned int borderTop=0, 
                unsigned int borderRight=0, 
                unsigned int borderBottom=0);
    void draw(ImImage& img);
private:
    void setBorders(ImImage& image);
                    
    std::string filename;
    int x, y;
    unsigned char alpha;
    ImageDrawType drawType;
    unsigned int borderLeft, borderTop, borderRight, borderBottom; 
};

class DrawOpBevel : public DrawOp {
public:
    DrawOpBevel(ImColor _color, unsigned int _width, int _bevelType): 
        color(_color), width(_width), bevelType(_bevelType) {};
    void draw(ImImage& img);
private:
    ImColor color;
    unsigned int width;
    int bevelType;
};

class DrawOpInterlace : public DrawOp {
public:
    DrawOpInterlace(unsigned char _whiteAlpha, unsigned char _blackAlpha,
                    const ImColor& baseColor);
    ~DrawOpInterlace();
    void draw(ImImage& img);
private:
    ImImage* tile;
    unsigned char whiteAlpha, blackAlpha;
};

class DrawOpGradient : public DrawOp {
public:
    DrawOpGradient(const std::vector<std::pair<ImColor, int> >& _colorRange,
                   int _x, int _y, 
                   int _width, int _height, double _angle ):
        colorRange(_colorRange), dynamic(false), x(_x), y(_y),
        width(_width), height(_height), angle(_angle)  {};

    DrawOpGradient(const std::vector<std::pair<ImColor, int> >& _colorRange,
                   std::string _x, std::string _y, 
                   std::string _width, std::string _height, double _angle ):
        colorRange(_colorRange), dynamic(true), expr_x(_x), expr_y(_y),
        expr_width(_width), expr_height(_height), angle(_angle)  {};
    
    void draw(ImImage& img);
private:
    ImColorRange colorRange;

    bool dynamic;
    std::string expr_x, expr_y, expr_width, expr_height;    

    int x, y;
    unsigned int width, height;
    double angle;
};

// Draws a Texture.
class DrawOpTexture : public DrawOp {
public:
    DrawOpTexture(shared_ptr<Texture> _texture, 
                  int _x, int _y, int _w, int _h) :
        dynamic(false), texture(_texture), x(_x), y(_y), width(_w), 
        height(_h) {};
    DrawOpTexture(shared_ptr<Texture> _texture, 
                  std::string _x, std::string _y, 
                  std::string _w, std::string _h) :
        dynamic(true), texture(_texture), expr_x(_x), expr_y(_y), expr_width(_w), 
        expr_height(_h) {};
        
    virtual void draw(ImImage& img);
private:
    bool dynamic;
    shared_ptr<Texture> texture;

    int x, y;
    unsigned int width, height;    
    std::string expr_x, expr_y, expr_width, expr_height;
};

#endif // __DrawOp_hh
