/* $Id: Event.cc,v 1.39 2004/02/07 04:56:42 pervago Exp $ */

/**
 * @file   Event.cc
 * @date   27-May-2003
 *
 * @brief Implementation of EventHandler class
 *
 * Eventloop function and functions for handling XEvents.
 *
 */

#include "ChildWindow.hh"
#include "Dockapp.hh"
#include "Event.hh"
#include "Kahakai.hh"
#include "Menu.hh"
#include "MenuItem.hh"
#include "Net.hh"
#include "Timer.hh"
#include "TitleBar.hh"
#include "Window.hh"

extern "C" {
#include <X11/extensions/shape.h>
#include <X11/extensions/Xrandr.h>
#include <X11/Xatom.h>
}

#include <iostream>
#include <list>
#include <map>
#include <set>
#include <vector>

using std::list;
using std::make_pair;
using std::map;
using std::set;
using std::vector;

/**
 * @fn    EventHandler(Kahakai *ka)
 * @brief Constructor for EventHandler class
 *
 * Sets kahakai and rh pointers. Creates menu move return mask, window
 * move/resize return mask and empty return mask sets.
 *
 * @param ka Pointer to kahakai object
 */
EventHandler::EventHandler(Kahakai *ka) {
    kahakai = ka;
    rh = kahakai->rh;
    focused = last_click_win = (Window) 0;
    move_resize = EndMoveResizeType;
    last_button = 0;

    empty_return_mask = new set<int>;

    moveresize_return_mask = new set<int>;
    moveresize_return_mask->insert(MotionNotify);
    moveresize_return_mask->insert(ButtonPress);
    moveresize_return_mask->insert(ButtonRelease);
    moveresize_return_mask->insert(KeyPress);
    moveresize_return_mask->insert(KeyRelease);
    moveresize_return_mask->insert(MapRequest);
    moveresize_return_mask->insert(UnmapNotify);
    moveresize_return_mask->insert(DestroyNotify);
    moveresize_return_mask->insert(EnterNotify);
    moveresize_return_mask->insert(LeaveNotify);
    moveresize_return_mask->insert(ConfigureRequest);

    menu_viewport_move_return_mask = new set<int>;
    menu_viewport_move_return_mask->insert(MotionNotify);
    menu_viewport_move_return_mask->insert(ButtonPress);
    menu_viewport_move_return_mask->insert(ButtonRelease);
    menu_viewport_move_return_mask->insert(KeyPress);
    menu_viewport_move_return_mask->insert(KeyRelease);
    menu_viewport_move_return_mask->insert(MapRequest);
    menu_viewport_move_return_mask->insert(EnterNotify);
    menu_viewport_move_return_mask->insert(LeaveNotify);
}

/**
 * @fn    ~EventHandler(void)
 * @brief Destructor for EventHandler class
 *
 * Deletes return mask lists
 */
EventHandler::~EventHandler(void) {
    MAPPTRCLEAR(empty_return_mask);
    MAPPTRCLEAR(moveresize_return_mask);
    MAPPTRCLEAR(menu_viewport_move_return_mask);
}

/**
 * @fn    EventLoop(set<int> *return_mask, XEvent *event)
 * @brief Eventloop
 *
 * Infinite loop waiting for an event to occur. This function can be called
 * from move and resize functions the return_mask set is then used for
 * deciding if an event should be processed as normal or returned to the
 * function caller.
 *
 * @param return_mask set to use as return_mask
 * @param event Pointer to allocated event structure
 */
void EventHandler::EventLoop(set<int> *return_mask, XEvent *event) {
    for (;;)
    {
        XNextEvent(kahakai->display, event);
        
        if (return_mask->find(event->type) != return_mask->end())
            return;

        HandleEvent(event);
    }
}

/**
 * @fn    HandleEvent(XEvent *event);
 * @brief Eventloop
 *
 * Executes a matching function for an event. If what to do for an
 * event is controlled by an action list we set etype, edetail and emod
 * variables and call the EvAct() function.
 *
 * @param event Pointer to allocated event structure
 */
void EventHandler::HandleEvent(XEvent *event) {
    Window w;
    int i, rx, ry;

    EventDetail *ed = new EventDetail;

    switch (event->type) {
        case MappingNotify :
            /*
             * Receiving a MappingKeyboard event means the keyboard layout
             * was changed, we must update the grabs of the focused
             * window, so it doesn't filter keys based on the old
             * keycodes.
             */
            if (event->xmapping.request == MappingKeyboard) {
                map<Window, WindowObject *>::iterator mi =
                    kahakai->window_table.begin();

                for (; mi != kahakai->window_table.end(); mi++) {
                    if (mi->second->type == WindowType) {
                        KaWindow &win = *(KaWindow *)(mi->second);

                        if (win.has_focus) {
                            win.UpdateGrabs();
                            break;
                        }
                    }
                }
            }
            break;
        case ConfigureRequest:
            EvConfigureRequest(&event->xconfigurerequest);
            break;
        case Expose:
            if (event->xexpose.count == 0) {
                while (XCheckTypedWindowEvent(kahakai->display,
                                              event->xexpose.window,
                                              Expose, event));
                EvExpose(&event->xexpose);
            }
            break;
        case PropertyNotify:
            EvProperty(&event->xproperty);
            break;
        case UnmapNotify:
            if(event->xunmap.event != event->xunmap.window)
                return;
        case DestroyNotify:
            EvUnmapDestroy(event);
            break;
        case FocusOut:
        case FocusIn:
            EvFocus(&event->xfocus);
            break;
        case LeaveNotify:
        case EnterNotify:
            if (event->xcrossing.mode == NotifyGrab)
              break;
            ed->type = event->type;
            ed->mod = event->xcrossing.state;
            ed->detail = 0;
            EvAct(event, event->xcrossing.window, ed);
            break;
        case KeyPress:
        case KeyRelease:
            ed->type = event->type;
            ed->mod = event->xkey.state;
            ed->detail = XKeycodeToKeysym(kahakai->display, event->xkey.keycode, 0);
            EvAct(event, event->xkey.window, ed);
            break;
        case ButtonPress:
            ed->type = ButtonPress;
            CheckDoubleClick(event, ed);
        case ButtonRelease:
            if (event->type == ButtonRelease)
                ed->type = ButtonRelease;
            ed->mod = event->xbutton.state;
            ed->detail = event->xbutton.button;
            EvAct(event, event->xbutton.window, ed);
            break;
        case ColormapNotify:
            EvColormap(&event->xcolormap);
            break;
        case MapRequest:
            EvMapRequest(&event->xmaprequest);
            ed->type = event->type;
            XQueryPointer(kahakai->display, event->xmaprequest.parent,
                          &w, &w, &rx, &ry, &i, &i, &(ed->mod));
            ed->detail = 0;
            event->xbutton.x_root = rx;
            event->xbutton.y_root = ry;
            EvAct(event, event->xmaprequest.window, ed);
            break;
        case ClientMessage:
            EvClientMessage(event, ed);
            break;

        default:
            if (event->type == kahakai->shape_event) {
                XShapeEvent *e = (XShapeEvent *) event;
                KaWindow *ww = (KaWindow *)
                    kahakai->FindWin(e->window, WindowType);
                if (ww && kahakai->shape)
                    ww->ShapeEvent(e->window);
            }

#ifdef XRANDR
            if (kahakai->randr && event->type == kahakai->randr_event) {
                XRRScreenChangeNotifyEvent *e =
                    (XRRScreenChangeNotifyEvent *) event;
                KaScreen *ws = (KaScreen *)
                    kahakai->FindWin(e->window, RootType);
                if (ws) {
                    ws->width = e->width;
                    ws->height = e->height;
                    ws->RRUpdate();
                }
            }
#endif // XRANDR

    }
    delete ed;
}

void EventHandler::CheckDoubleClick(XEvent * event, EventDetail * ed) { 

    last_button = event->xbutton.button;

    struct timeval click_time;
    gettimeofday(&click_time, NULL);

    // if it's the same button as last, and on the same window,
    // and it's less than $double_click ms ago
    if (last_button == event->xbutton.button &&
        last_click_win == event->xbutton.window &&
        msDifference(last_click, click_time) <= (signed)kahakai->double_click) {

            // it's a double click!
            ed->type = DoubleClick;
            last_click_win = (Window) 0;
            return; 

    }

    // nope, last click was too long ago.  record this click.

    last_click_win = event->xbutton.window;
    last_click.tv_sec = click_time.tv_sec;
    last_click.tv_usec = click_time.tv_usec;
}

void EventHandler::PossiblyDemoteDoubleClick(KaWindow * win, EventDetail * ed) {

    // test to see whether we should make a distinction between buttonpresses
    // and doubleclicks for this action.  if we click button 1 two times in a
    // row quickly, and there's no doubleclick action registered for button 1,
    // then we should just treat it as two clicks, not a doubleclick.
    
    vector<KaAction *>::iterator it = win->actionlist->begin();
    
    if (ed->type != DoubleClick)
        return;

    for(it = win->actionlist->begin(); it != win->actionlist->end(); ++it)
        if (eventmatch(win->display, *it, ed, win->kascreen->imod_mask))
            return; // found matching double click action

    // never found matching double click action, demote it
    ed->type = ButtonPress;
}


/**
 * @fn    EvProperty(XPropertyEvent *e)
 * @brief PropertyEvent handler
 *
 * We receive a property event when a window want us to update some window
 * info. Unless the state of the property event is PropertyDelete, we try
 * to find the KaWindow managing the the window who sent the event. If a
 * KaWindow was found, we update the stuff indicated by the event. If the
 * name should be updated we also redraw the label foreground for the
 * KaWindow. If atom is _NET_WM_STRUT we update the strut list and workarea.
 *
 * @param e The PropertyEvent
 */
void EventHandler::EvProperty(XPropertyEvent *e) {
    KaWindow *ww;

    if (e->state == PropertyDelete) {
        if (e->atom == kahakai->net->net_wm_strut) {
            if ((ww = (KaWindow *) kahakai->FindWin(e->window, WindowType))) {
                list<WMstrut *>::iterator s_it =
                    ww->kascreen->strut_list.begin();
                for (; s_it != ww->kascreen->strut_list.end(); ++s_it) {
                    if ((*s_it)->window == e->window) {
                        ww->kascreen->strut_list.remove(*s_it);
                        free(*s_it);
                        ww->kascreen->UpdateWorkarea();
                    }
                }
            }
        }
    } else if (e->atom == kahakai->net->net_wm_strut) {
        if ((ww = (KaWindow *) kahakai->FindWin(e->window, WindowType)))
            kahakai->net->GetWmStrut(ww);
    } else if (e->atom == XA_WM_NAME) {
        if ((ww = (KaWindow *) kahakai->FindWin(e->window, WindowType))) {
            if (!kahakai->net->GetNetName(ww))
                kahakai->net->GetXaName(ww);

            ww->title->render();
            ww->label->render();
        }
    }
    else if (e->atom == kahakai->net->xrootpmap_id) {
        if (KaScreen *ws = (KaScreen *) kahakai->FindWin(e->window, RootType)) {
            kahakai->net->GetXRootPMapId(ws);
            ws->ic->setXRootPMapId((ws->xrootpmap_id)? true: false);

            list<DockappHandler *>::iterator dock_it = ws->docks.begin();
            for (; dock_it != ws->docks.end(); ++dock_it)
                if (!(*dock_it)->dockapp_list.empty()) (*dock_it)->Render();
            vector<KaWindow *>::iterator win_it = ws->kawindow_list.begin();
            for (; win_it != ws->kawindow_list.end(); ++win_it) {
                (*win_it)->updateDecorations();

                if ((*win_it)->title_w) (*win_it)->DrawTitlebar();
                if ((*win_it)->handle_w) (*win_it)->DrawHandlebar();
                if ((*win_it)->left_w || (*win_it)->right_w) 
                    (*win_it)->DrawEdges();
            }
            vector<KaMenu *>::iterator menu_it = ws->kamenu_list.begin();
            for (; menu_it != ws->kamenu_list.end(); ++menu_it) {
                (*menu_it)->clearCache();
                if ((*menu_it)->mapped) (*menu_it)->Render();
            }
        }
    }
}

/**
 * @fn    EvExpose(XExposeEvent *e)
 * @brief ExposeEvent handler
 *
 * We receive an expose event when a windows foreground has been exposed
 * for some change. If the event is from one of our windows with
 * foreground, we redraw the foreground for this window.
 *
 * @param e The ExposeEvent
 */
void EventHandler::EvExpose(XExposeEvent *e) {
    if (WindowObject *wo = kahakai->FindWin(e->window, LabelType | ButtonType |
                                           MenuTitleType | MenuItemType |
                                           MenuSubType | MenuCBItemType))
        switch (wo->type) {
            case LabelType:
                break;
            case ButtonType:
                ((KaChildWindow *) wo)->render(); break;
            case MenuTitleType:
            case MenuItemType:
            case MenuSubType:
            case MenuCBItemType:                
                KaMenuItem* mi = (KaMenuItem*) wo;
                mi->refresh();
                break;
        }
}

/**
 * @fn    EvFocus(XFocusChangeEvent *e)
 * @brief FocusChangeEvent handler
 *
 * We receive a focus change event when a window gets the keyboard focus.
 * If a KaWindow is managing the window pointed to by the event then we change
 * the KaWindows decorations to represent a focused window, and we change the
 * before focused KaWindows decorations to represent an unfocused window.
 *
 * @param e The FocusChangeEvent
 */
void EventHandler::EvFocus(XFocusChangeEvent *e) {
    KaWindow *ww = NULL, *ww2;
    KaScreen *ws = NULL;

    if (e->type == FocusIn && e->window != focused) {
        ww  = (KaWindow *) kahakai->FindWin(e->window, WindowType);
        ww2 = (KaWindow *) kahakai->FindWin(focused, WindowType);

        // Do nothing if focus stayed in the same window
        if (ww == ww2) return;

        if (ww) {
            ww->actionlist = &ww->kascreen->config.awinacts;
            ww->UpdateGrabs();
            ww->FocusWin();
            ww->net->SetActiveWindow(ww->kascreen, ww);
        } else if ((ws = (KaScreen *) kahakai->FindWin(e->window, RootType)))
            ws->focus = true;

        if (ww2) {
            ww2->actionlist = &ww2->kascreen->config.pwinacts;
            ww2->UpdateGrabs();
            ww2->UnFocusWin();
            if (! ww) kahakai->net->SetActiveWindow(ww2->kascreen, NULL);
        }

        focused = e->window;
    }
}

/**
 * @fn    EvConfigureRequest(XConfigureRequestEvent *e)
 * @brief ConfigureRequestEvent handler
 *
 * When we receive this event a window wants to be reconfigured (raised,
 * lowered, moved, resized). We try find a matching KaWindow managing the
 * window, if found we update that KaWindow with the values from the configure
 * request event. If we don't found a KaWindow managing the window who wants
 * to be reconfigured, we just update that window with the values from the
 * configure request event.
 *
 * @param e The ConfigureRequestEvent
 */
void EventHandler::EvConfigureRequest(XConfigureRequestEvent *e) {
    WindowObject *wo;
    KaWindow *ww;
    Dockapp *da;
    XWindowChanges wc;

    wc.x = e->x;
    wc.y = e->y;
    wc.width = e->width;
    wc.height = e->height;
    wc.sibling = e->above;
    wc.stack_mode = e->detail;
    wc.border_width = e->border_width;

    if ((wo = kahakai->FindWin(e->window, WindowType | DockAppType))) {
        if (wo->type == WindowType) {
            ww = (KaWindow *) wo;

            if (ww->ign_config_req)
              return;

            ww->Gravitate(RemoveGravity);

            if (e->value_mask & CWX) ww->attrib.setX(e->x);
            if (e->value_mask & CWY) ww->attrib.setY(e->y);
            if (e->value_mask & CWWidth) ww->attrib.setWidth(e->width);
            if (e->value_mask & CWHeight) ww->attrib.setHeight(e->height);

            // update size increments etc.
            ww->net->GetWMNormalHints(ww);

            ww->Gravitate(ApplyGravity);

            if (e->value_mask & CWStackMode) {
                switch (e->detail) {
                    case Above:
                        ww->kascreen->RaiseWindowAllDesktops(ww);
                        break;
                    case Below:
                        ww->kascreen->LowerWindowAllDesktops(ww);
                        break;
                    case TopIf:
                        ww->AlwaysontopOn();
                        break;
                    case BottomIf:
                        ww->AlwaysatbottomOn();
                        break;
                    case Opposite:
                        if (ww->flags.alwaysontop)
                            ww->AlwaysatbottomOn();
                        else if (ww->flags.alwaysatbottom)
                            ww->AlwaysontopOn();
                }
            }

            ww->RedrawWindow();
            return;

        } else if (wo->type == DockAppType) {
            da = (Dockapp *) wo;
            if (e->value_mask & CWWidth) da->width = e->width;
            if (e->value_mask & CWHeight) da->height = e->height;
            XGrabServer(e->display);
            if (validatedrawable(da->id))
                XConfigureWindow(e->display, da->id, e->value_mask, &wc);
            XUngrabServer(e->display);
            da->dh->Update();
        }
    }
    XGrabServer(e->display);
    if (validatedrawable(e->window))
        XConfigureWindow(e->display, e->window, e->value_mask, &wc);
    XUngrabServer(e->display);
}

/**
 * @fn    EvColormap(XColormapEvent *e)
 * @brief ColormapEvent handler
 *
 * A window wants to install a new colormap, so we do it.
 *
 * @param e The ColormapEvent
 */
void EventHandler::EvColormap(XColormapEvent *e) {
    XInstallColormap(e->display, e->colormap);
}

/**
 * @fn    EvMapRequest(XMapRequestEvent *e)
 * @brief MapRequestEvent handler
 *
 * We receive this event then a window wants to be mapped. If the window
 * isn't in our window hash_map already it's a new window and we create
 * a KaWindow for it. If the window already is managed we just set its
 * state to NormalState.
 *
 * @param e The MapRequestEvent
 */
void EventHandler::EvMapRequest(XMapRequestEvent *e) {
    XWindowAttributes attr;
    XWMHints *wm_hints;

    if (KaWindow *ww = (KaWindow *) kahakai->FindWin(e->window, WindowType)) {
        if (ww->flags.hidden)
            ww->RaiseFocus();
        return;

    } else
    if (KaScreen *ws = (KaScreen *) kahakai->FindWin(e->parent, RootType)) {


        if (ws->net->IsSystrayWindow(e->window)) {
            if (! kahakai->FindWin(e->window, SystrayType)) {
                XGrabServer(ws->display);
                if (validatedrawable(e->window)) {
                    XSelectInput(ws->display, e->window, StructureNotifyMask);
                }
                XUngrabServer(ws->display);
                SystrayWindow *stw = new SystrayWindow(e->window, ws);
                kahakai->window_table.insert(make_pair(e->window, stw));
                ws->systray_window_list.push_back(e->window);
                ws->net->SetSystrayWindows(ws);
            }

        } else {
            XGetWindowAttributes(e->display, e->window, &attr);
            wm_hints = NULL;

            if (! attr.override_redirect) {
                if ((wm_hints = XGetWMHints(e->display, e->window)) &&
                    (wm_hints->flags & StateHint) &&
                    (wm_hints->initial_state == WithdrawnState)) {
                    ws->AddDockapp(e->window);
                } else {
                    new KaWindow(e->window, ws);
                    ws->net->SetClientList(ws);
                    ws->net->SetClientListStacking(ws);
                }
            }
            if (wm_hints) XFree(wm_hints);
        }
    }
}

/**
 * @fn    EvUnmapDestroy(XEvent *e)
 * @brief UnmapEvent, DestroyEvent and ReparentEvent handler
 *
 * We receive this event then a window has been unmapped, destroyed or
 * reparented. If we can find a KaWindow for this window then we delete that
 * KaWindow. If we couldn't find a KaWindow we check if the window is a
 * dockapp window and if it is, we update the dockapp handler holding the
 * dockapp.
 *
 * @param e The XEvent
 */
void EventHandler::EvUnmapDestroy(XEvent *e) {
    DockappHandler *dh;
    WindowObject *wo;

    if ((wo = kahakai->FindWin((e->type == UnmapNotify)?
                              e->xunmap.window:
                              (e->type == DestroyNotify)?
                              e->xdestroywindow.window:
                              e->xreparent.window,
                              WindowType | DockAppType | SystrayType))) {
        if (wo->type == WindowType) {
            if (e->type == DestroyNotify)
                ((KaWindow *) wo)->deleted = true;
            delete ((KaWindow *) wo);
        }
        else if (wo->type == DockAppType) {
            if (e->type == DestroyNotify)
                ((Dockapp *) wo)->deleted = true;
            dh = ((Dockapp *) wo)->dh;
            delete ((Dockapp *) wo);
            dh->Update();
        }
        else if (wo->type == SystrayType && (e->type == DestroyNotify)) {
            SystrayWindow *stw = (SystrayWindow *) wo;
            kahakai->window_table.erase(stw->id);
            XGrabServer(stw->ws->display);
            if (validatedrawable(stw->id)) {
                XSelectInput(stw->ws->display, stw->id, NoEventMask);
            }
            XUngrabServer(stw->ws->display);
            stw->ws->systray_window_list.remove(stw->id);
            stw->ws->net->SetSystrayWindows(stw->ws);
            delete stw;
        }
    }
}

/**
 * @fn    EvClientMessage(XEvent *e, EventDetail *ed)
 * @brief ClientMessageEvent handler
 *
 * This function handles all client message events received.
 *
 * @param e The XEvent
 * @param ed Structure containing event details
 */
void EventHandler::EvClientMessage(XEvent *e, EventDetail *ed) {
    Window w;
    int i, rx, ry;
    KaWindow *ww;

    if (e->xclient.message_type == kahakai->net->net_active_window) {
        if ((ww = (KaWindow *) kahakai->FindWin(e->xclient.window,
                                               WindowType)))
            ww->RaiseFocus();
    }
    else if (e->xclient.message_type == kahakai->net->net_wm_name) {
        if ((ww = (KaWindow *) kahakai->FindWin(e->xclient.window,
                                               WindowType))) {
            kahakai->net->GetNetName(ww);
            ww->title->render();
            ww->label->render();
        }
    }
    else if (e->xclient.message_type == kahakai->net->wm_change_state) {
        if ((ww=(KaWindow *) kahakai->FindWin(e->xclient.window, WindowType))) {
            if ((unsigned int) e->xclient.data.l[0] == IconicState)
                ww->Minimize();
            else if ((unsigned int) e->xclient.data.l[0] == NormalState)
                ww->RaiseFocus();
            else if ((unsigned int) e->xclient.data.l[0] == WithdrawnState)
                delete ww;
        }
    }
    else if (e->xclient.message_type == kahakai->net->net_wm_desktop) {
        if ((ww=(KaWindow *) kahakai->FindWin(e->xclient.window, WindowType))) {
            if ((unsigned int) e->xclient.data.l[0] == 0xffffffff ||
                (unsigned int) e->xclient.data.l[0] == 0xfffffffe) {
                ww->desktop_mask = (1L << 16) - 1;
                ww->UpdateStack();
                ww->Show();
                ww->net->SetDesktop(ww);
                ww->net->SetDesktopMask(ww);
            }
            else if ((unsigned int) e->xclient.data.l[0] <
                ww->kascreen->config.desktops) {
                ww->desktop_mask = 1L << (unsigned int) e->xclient.data.l[0];
                ww->UpdateStack();
                if (ww->desktop_mask &
                    (1L << ww->kascreen->current_desktop->number))
                    ww->Show();
                else
                    ww->Hide();
                ww->net->SetDesktop(ww);
                ww->net->SetDesktopMask(ww);
            }
        }
    }
    else if (e->xclient.message_type ==
             kahakai->net->kahakai_net_wm_desktop_mask) {
        if ((ww = (KaWindow *) kahakai->FindWin(e->xclient.window,
                                               WindowType))) {
            if (e->xclient.data.l[0] <
                ((1L << ww->kascreen->config.desktops) - 1) &&
                e->xclient.data.l[0] >= 0) {
                ww->desktop_mask = e->xclient.data.l[0];
                ww->UpdateStack();
                if (ww->desktop_mask &
                    (1L << ww->kascreen->current_desktop->number))
                    ww->Show();
                else
                    ww->Hide();
                ww->net->SetDesktop(ww);
                ww->net->SetDesktopMask(ww);
            }
        }
    }
    else if (e->xclient.message_type == kahakai->net->net_wm_state) {
        if ((ww = (KaWindow *) kahakai->FindWin(e->xclient.window,
                                               WindowType))) {
            bool max_done = false;
            for (int i = 1; i < 3; i++) {
                if ((unsigned long) e->xclient.data.l[i] ==
                    kahakai->net->net_wm_state_sticky) {
                    switch (e->xclient.data.l[0]) {
                        case _NET_WM_STATE_REMOVE:
                            ww->UnSticky(); break;
                        case _NET_WM_STATE_ADD:
                            ww->Sticky(); break;
                        case _NET_WM_STATE_TOGGLE:
                            ww->ToggleSticky(); break;
                    }
                } else if ((unsigned long) e->xclient.data.l[i] ==
                           kahakai->net->net_wm_state_shaded) {
                    switch (e->xclient.data.l[0]) {
                        case _NET_WM_STATE_REMOVE:
                            ww->UnShade(); break;
                        case _NET_WM_STATE_ADD:
                            ww->Shade(); break;
                        case _NET_WM_STATE_TOGGLE:
                            ww->ToggleShade(); break;
                    }
                } else if ((unsigned long) e->xclient.data.l[i] ==
                           kahakai->net->net_wm_state_hidden) {
                    switch (e->xclient.data.l[0]) {
                        case _NET_WM_STATE_REMOVE:
                            ww->Minimize(); break;
                        case _NET_WM_STATE_ADD:
                            ww->RaiseFocus(); break;
                        case _NET_WM_STATE_TOGGLE:
                            ww->ToggleMinimizeActive(); break;
                    }
                } else if ((unsigned long) e->xclient.data.l[i] ==
                           kahakai->net->net_wm_maximized_vert ||
                           (unsigned long) e->xclient.data.l[i] ==
                           kahakai->net->net_wm_maximized_horz) {
                    if (max_done) break;
                    max_done = true;
                    switch (e->xclient.data.l[0]) {
                        case _NET_WM_STATE_REMOVE:
                            ww->UnMaximize(); break;
                        case _NET_WM_STATE_ADD:
                            ww->Maximize(); break;
                        case _NET_WM_STATE_TOGGLE:
                            ww->ToggleMaximize(); break;
                    }
                } else if ((unsigned long) e->xclient.data.l[i] ==
                           kahakai->net->net_wm_state_above ||
                           (unsigned long) e->xclient.data.l[i] ==
                           kahakai->net->net_wm_state_stays_on_top) {
                    switch (e->xclient.data.l[0]) {
                        case _NET_WM_STATE_REMOVE:
                            ww->AlwaysontopOff(); break;
                        case _NET_WM_STATE_ADD:
                            ww->AlwaysontopOn(); break;
                        case _NET_WM_STATE_TOGGLE:
                            ww->AlwaysontopToggle(); break;
                    }
                } else if ((unsigned long) e->xclient.data.l[i] ==
                           kahakai->net->net_wm_state_below ||
                           (unsigned long) e->xclient.data.l[i] ==
                           kahakai->net->net_wm_state_stays_at_bottom) {
                    switch (e->xclient.data.l[0]) {
                        case _NET_WM_STATE_REMOVE:
                            ww->AlwaysatbottomOff(); break;
                        case _NET_WM_STATE_ADD:
                            ww->AlwaysatbottomOn(); break;
                        case _NET_WM_STATE_TOGGLE:
                            ww->AlwaysatbottomToggle(); break;
                    }
                } else if ((unsigned long) e->xclient.data.l[i] ==
                           kahakai->net->net_wm_state_skip_taskbar) {
                    switch (e->xclient.data.l[0]) {
                        case _NET_WM_STATE_REMOVE:
                            ww->flags.tasklist = true; break;
                        case _NET_WM_STATE_ADD:
                            ww->flags.tasklist = false; break;
                        case _NET_WM_STATE_TOGGLE:
                            ww->flags.tasklist = !ww->flags.tasklist; break;
                    }
                } else if ((unsigned long) e->xclient.data.l[i] ==
                           kahakai->net->net_wm_state_fullscreen) {
                    switch (e->xclient.data.l[0]) {
                        case _NET_WM_STATE_REMOVE:
                            ww->FullscreenOff();
                            ww->UnMaximize();
                            ww->AlwaysontopOff();
                            ww->DecorAllOn();
                            break;
                        case _NET_WM_STATE_ADD:
                            ww->DecorAllOff();
                            ww->AlwaysontopOn();
                            ww->FullscreenOn();
                            ww->Maximize();
                            break;
                        case _NET_WM_STATE_TOGGLE:
                            if (ww->flags.fullscreen) {
                                ww->FullscreenOff();
                                ww->UnMaximize();
                                ww->AlwaysontopOff();
                                ww->DecorAllOn();
                            } else {
                                ww->DecorAllOff();
                                ww->AlwaysontopOn();
                                ww->FullscreenOn();
                                ww->Maximize();
                            }
                            break;
                    }
                } else if ((unsigned long) e->xclient.data.l[i] ==
                           kahakai->net->kahakai_net_wm_state_decor) {
                    switch (e->xclient.data.l[0]) {
                        case _NET_WM_STATE_REMOVE:
                            ww->DecorAllOff(); break;
                        case _NET_WM_STATE_ADD:
                            ww->DecorAllOn(); break;
                        case _NET_WM_STATE_TOGGLE:
                            if (ww->flags.all) ww->DecorAllOff();
                            else ww->DecorAllOn();
                            break;
                    }
                } else if ((unsigned long) e->xclient.data.l[i] ==
                           kahakai->net->kahakai_net_wm_state_decortitle) {
                    switch (e->xclient.data.l[0]) {
                        case _NET_WM_STATE_REMOVE:
                            ww->DecorTitleOff(); break;
                        case _NET_WM_STATE_ADD:
                            ww->DecorTitleOn(); break;
                        case _NET_WM_STATE_TOGGLE:
                            ww->DecorTitleToggle(); break;
                    }
                } else if ((unsigned long) e->xclient.data.l[i] ==
                           kahakai->net->kahakai_net_wm_state_decorhandle) {
                    switch (e->xclient.data.l[0]) {
                        case _NET_WM_STATE_REMOVE:
                            ww->DecorHandleOff(); break;
                        case _NET_WM_STATE_ADD:
                            ww->DecorHandleOn(); break;
                        case _NET_WM_STATE_TOGGLE:
                            ww->DecorHandleToggle(); break;
                    }
                } else if ((unsigned long) e->xclient.data.l[i] ==
                           kahakai->net->kahakai_net_wm_state_decorborder) {
                    switch (e->xclient.data.l[0]) {
                        case _NET_WM_STATE_REMOVE:
                            ww->DecorBorderOff(); break;
                        case _NET_WM_STATE_ADD:
                            ww->DecorBorderOn(); break;
                        case _NET_WM_STATE_TOGGLE:
                            ww->DecorBorderToggle(); break;
                    }
                }
            }
        }
    }
    else if (e->xclient.message_type == kahakai->net->xdndenter ||
             e->xclient.message_type == kahakai->net->xdndleave) {
        if (e->xclient.message_type == kahakai->net->xdndenter) {
            e->type = EnterNotify;
            ed->type = EnterNotify;
        } else {
            e->type = LeaveNotify;
            ed->type = LeaveNotify;
        }

        if (KaScreen *ws = (KaScreen *) kahakai->FindWin(e->xclient.window,
                                                        RootType)) {
            XQueryPointer(ws->display, ws->id, &w, &w, &rx, &ry, &i, &i,
                          &(ed->mod));
        } else {
            rx = 0;
            ry = 0;
        }
        ed->detail = 0;
        e->xcrossing.x_root = rx;
        e->xcrossing.y_root = ry;

        EvAct(e, e->xclient.window, ed);
    }
    else if (e->xclient.message_type == kahakai->net->net_desktop_viewport) {
        if (KaScreen *ws = (KaScreen *) kahakai->FindWin(e->xclient.window,
                                                        RootType))
            ws->MoveViewportTo(e->xclient.data.l[0], e->xclient.data.l[1]);
    }
    else if (e->xclient.message_type == kahakai->net->net_close_window) {
        if ((ww = (KaWindow *) kahakai->FindWin(e->xclient.window, WindowType))) {
            ww->Close();
        }
    }
    else if (e->xclient.message_type == kahakai->net->net_current_desktop) {
        if (KaScreen *ws = (KaScreen *) kahakai->FindWin(e->xclient.window,
                                                        RootType))
            ws->GoToDesktop(e->xclient.data.l[0]);
    }
    else if (e->xclient.message_type == kahakai->net->kahakai_net_virtual_pos) {
        if ((ww = (KaWindow *) kahakai->FindWin(e->xclient.window,
                                               WindowType))) {
            ww->attrib.x = e->xclient.data.l[0] - ww->kascreen->v_x;
            ww->attrib.y = e->xclient.data.l[1] - ww->kascreen->v_y;
            if (ww->flags.sticky) {
                ww->attrib.x = ww->attrib.x % ww->kascreen->width;
                ww->attrib.y = ww->attrib.y % ww->kascreen->height;
            }
            ww->RedrawWindow();
        }
    }
    else if (e->xclient.message_type == kahakai->net->net_moveresize_window) {
        if ((ww = (KaWindow *) kahakai->FindWin(e->xclient.window,
                                               WindowType))) {

            char gravity = (char) e->xclient.data.l[0];
            if (gravity == 0) gravity = (char) ww->size.win_gravity;

            int x = ww->attrib.x;
            int y = ww->attrib.y;
            int width = ww->attrib.width;
            int height = ww->attrib.height;

            if (e->xclient.data.l[0] & (1L << 8)) x = e->xclient.data.l[1];
            if (e->xclient.data.l[0] & (1L << 9)) y = e->xclient.data.l[2];
            if (e->xclient.data.l[0] & (1L << 10))
                width = e->xclient.data.l[3];
            if (e->xclient.data.l[0] & (1L << 11))
                height = e->xclient.data.l[4];

            ww->IncSizeCheck(width, height, &ww->attrib.width,
                             &ww->attrib.height);

            if (gravity != StaticGravity)
                ww->Gravitate(RemoveGravity);
            if (gravity == NorthEastGravity ||
                gravity == EastGravity ||
                gravity == SouthEastGravity)
                ww->attrib.x = ww->kascreen->width - x - ww->attrib.width;
            else ww->attrib.x = x;

            if (gravity == SouthWestGravity ||
                gravity == SouthGravity ||
                gravity == SouthEastGravity)
                ww->attrib.y = ww->kascreen->height - y - ww->attrib.height;
            else ww->attrib.y = y;

            if (gravity == NorthGravity ||
                gravity == SouthGravity ||
                gravity == CenterGravity)
                ww->attrib.x -= ww->attrib.width / 2;

            if (gravity == EastGravity ||
                gravity == WestGravity ||
                gravity == CenterGravity)
                ww->attrib.y -= ww->attrib.height / 2;

            if (gravity != StaticGravity)
                ww->Gravitate(ApplyGravity);

            ww->RedrawWindow();
        }
    }
    else if (e->xclient.message_type == kahakai->net->net_wm_moveresize) {
        if ((ww = (KaWindow *) kahakai->FindWin(e->xclient.window,
                                               WindowType))) {
            if (e->xclient.data.l[2] == _NET_WM_MOVERESIZE_MOVE ||
                e->xclient.data.l[2] == _NET_WM_MOVERESIZE_MOVE_KEYBOARD)
                ww->MoveOpaque(e);
            else if (e->xclient.data.l[2] == _NET_WM_MOVERESIZE_SIZE_TOPLEFT ||
                     e->xclient.data.l[2] ==
                     _NET_WM_MOVERESIZE_SIZE_BOTTOMLEFT ||
                     e->xclient.data.l[2] == _NET_WM_MOVERESIZE_SIZE_LEFT)
                ww->ResizeLeftOpaque(e);
            else
                ww->ResizeRightOpaque(e);
        }
    }
    else if (e->xclient.message_type == kahakai->net->kahakai_net_restart) {
        restart(NULL);
    }
    else if (e->xclient.message_type == kahakai->net->kahakai_net_shutdown) {
        quit(EXIT_SUCCESS);
    }
    else if (e->xclient.message_type == kahakai->net->net_startup_info_begin ||
             e->xclient.message_type == kahakai->net->net_startup_info) {
        //        printf("STARTUP INFO!\n");
        //        printf("DATA: %s\n", &e->xclient.data.b[0]);
        //        printf("Window ID: %x\n", e->xclient.window);
        if ((ww = (KaWindow *) kahakai->FindWin(e->xclient.window, WindowType))) {
            //  printf("Window name: %s\n", ww->name);
        }
    }
}

/**
 * @fn    EvAct(XEvent *e, Window win)
 * @brief Event action handler
 *
 * The eventloop will call this function whenever a event that could
 * be controlled by the action lists occur. This function finds the
 * WindowObject for the window, gets the action list for that type of
 * WindowObject and calls the WindowObjects EvAct function.
 *
 * @param e The Event
 * @param win The window we should use in the WindowObject search
 * @param ed Structure containing event details
 */
void EventHandler::EvAct(XEvent *e, Window win, EventDetail *ed) {
    WindowObject *wo;
    KaWindow *ka;

    map<Window, WindowObject *>::iterator it;
    if ((it = kahakai->window_table.find(win)) != kahakai->window_table.end()) {
        wo = (*it).second;

        kahakai->timer->ValidateInterrupts(e);

        switch (wo->type) {
            case WindowType:
                ka = (KaWindow *) wo;
                ka->EvAct(e, ed, wo->actionlist, wo->type);
                break;
            case FrameType:
            case TitleType:
            case LabelType:
            case HandleType:
            case LGripType:
            case RGripType:
                ka = ((KaChildWindow *) wo)->ka;
                ka->EvAct(e, ed, wo->actionlist, wo->type);
                break;
            case ButtonType: {
                ka = ((KaChildWindow *) wo)->ka;
                ka->EvAct(e, ed, wo->actionlist, wo->type);
                if (ed->type == ButtonPress)
                    ka->ButtonPressed((KaWindowButton *) wo);
            } break;
            case MenuTitleType:
            case MenuItemType:
            case MenuCBItemType:
            case MenuSubType:
                ((KaMenuItem *) wo)->EvAct(e, ed, wo->actionlist);
                break;
            case WEdgeType:
            case EEdgeType:
            case NEdgeType:
            case SEdgeType:
                (((ScreenEdge *) wo)->ka)->EvAct(e, ed, wo->actionlist);
                break;
            case RootType:
                ((KaScreen *) wo)->EvAct(e, ed, wo->actionlist);
                break;
        }
    }
}

/**
 * @fn    eventmatch(KaAction *act, EventDetail *ed)
 * @brief Event to action matcher
 *
 * Checks if action type, detail and modifiers are correct.
 *
 * @param act Action to use for matching
 * @param ed Structure containing event details
 *
 * @return True if match, otherwise false
 */
Bool eventmatch(Display *display, KaAction *act, EventDetail *ed,
               unsigned int imod_mask) {

    /* the event types need to match */
    if (ed->type != act->type)
        return false;

    /* If the ed->detail (keysym) is zero, an unknown key was pressed, so
       we ignore it */
    if ((ed->type == KeyPress || ed->type == KeyRelease) && !ed->detail)
        return false;
    
    /* both detail's must be non-zero.  If they are then they must be equal */
    if (act->detail && ed->detail && act->detail != ed->detail)
        return false;

    /* make sure the first 13 bits of the KaAction modifier that are
     * required are on set on the EventDetail */
    if (((ed->mod & ~imod_mask) & (act->mod & 0x1fff)) != (act->mod & 0x1fff))
        return false;

    /* if the KaAction requires that the MoveResizeMask be set then
     * the EventDetail must also have the MoveResizeMask be set. */
    if (act->mod & MoveResizeMask)
        if (! (ed->mod & MoveResizeMask))
            return false;

    /* if the KaAction has any modifiers that cannot be set then the
     * EventDetail must not have them set */
    if ((ed->mod & (act->nmod & 0x1fff)))
        return false;

    /* if the action cannot have the MoveResizeMask,
     * then make sure the EventDetail doesn't have it set. */
    if (act->nmod & MoveResizeMask)
        if (ed->mod & MoveResizeMask)
            return false;

    /* if all the previous tests passed then we are good */
    return true;
}
