/* $Id: ExpressionEvaluator.cc,v 1.6 2003/11/29 05:03:09 pervago Exp $ */

/**
 * @file   ExpressionEvaluator.cc
 * @date   20-Aug-2003
 *
 * @brief  Class for evaluation of expressions used in the new style format.
           (e.g. baseFontSize*1.1 + 3)
 *
 */

#include "ExpressionEvaluator.hh"

#include <boost/spirit/core.hpp>
#include <boost/spirit/symbols/symbols.hpp>
#include <stack>
#include <functional>
#include <string>
#include <map>

using namespace std;
using namespace boost::spirit;

///////////////////////////////////////////////////////////////////////////////
//
//  Semantic actions
//
///////////////////////////////////////////////////////////////////////////////
struct push_num
{
    push_num(stack<double>& eval_)
        : eval(eval_) {}

    void operator()(double n) const
    {
        eval.push(n);
    }

    stack<double>& eval;
};

template <typename op>
struct do_op
{
    do_op(op const& the_op, stack<double>& eval_)
        : m_op(the_op), eval(eval_) {}

    void operator()(char const*, char const*) const
    {
        double rhs = eval.top(); eval.pop();
        double lhs = eval.top(); eval.pop();

        eval.push(m_op(lhs, rhs));
    }

    op m_op;
    stack<double>& eval;
};

template <class op>
do_op<op>
make_op(op const& the_op, stack<double>& eval)
{
    return do_op<op>(the_op, eval);
}

struct do_negate
{
    do_negate(stack<double>& eval_)
        : eval(eval_) {}

    void operator()(char const*, char const*) const
    {
        double lhs = eval.top();
        eval.pop();

        eval.push(-lhs);
    }

    stack<double>& eval;
};

struct get_var
{
    get_var(stack<double>& eval_)
        : eval(eval_) {}

    void operator()(double n) const
    {
        eval.push(n);
    }

    stack<double>& eval;
};

///////////////////////////////////////////////////////////////////////////////
//
//  Our calculator grammar
//
///////////////////////////////////////////////////////////////////////////////
struct calculator : public grammar<calculator>
{
    calculator(stack<double>& eval_)
        : eval(eval_) {}
        
    template <typename ScannerT>
    struct definition
    {
        definition(calculator const& self)
        {
            factor =
                real_p[push_num(self.eval)]
                |   self.vars[get_var(self.eval)]
                |   '(' >> expression >> ')'
                |   ('-' >> factor)[do_negate(self.eval)]
                ;
                
            term =
                factor
                >> *(   ('*' >> factor)[make_op(multiplies<double>(), self.eval)]
                        |   ('/' >> factor)[make_op(divides<double>(), self.eval)]
                        )
                ;
                
            expression =
                term
                >> *(  ('+' >> term)[make_op(plus<double>(), self.eval)]
                       |   ('-' >> term)[make_op(minus<double>(), self.eval)]
                       )
                ;
                
        }
            
        rule<ScannerT>  expression, term, factor;
            
        rule<ScannerT> const& start() const { return expression; }
    };
        
    symbols<double> vars;       
    stack<double>&  eval;
};

ExpressionEvaluator::ExpressionEvaluator(const map<string, float>& variables_):
    variables(variables_) {}

ExpressionEvaluator::~ExpressionEvaluator() {}

float ExpressionEvaluator::evaluate(const std::string& expression) {
    map<string, float>::const_iterator it;
    std::stack<double>   eval_stack;

    calculator calc(eval_stack);

    for (it = variables.begin(); it != variables.end(); ++it) {
        calc.vars.add(it->first.c_str(), it->second);
    }

    parse_info<> info = parse(expression.c_str(), calc, space_p);
    if (!info.full) {
        throw parse_exception(expression, info.stop);
    } else {
        return calc.eval.top();
    }
}

