/* $Id: Font.cc,v 1.20 2003/12/26 07:09:23 pervago Exp $ */

/**
 * @file   Font.cc
 * @date   08-Oct-2002 11:05:01
 *
 * @brief Implementation of KaFont class
 *
 * Kahakai's font implementation. Includes both Xft and X core fonts.
 *
 */

#include "Font.hh"
#include "Kahakai.hh"
#include "Image.hh"
#include "Logger.hh"

#include <iostream>
#include <vector>
#include <string>

using std::endl;
using std::vector;
using std::string;

/**
 * @fn    KaFont()
 * @brief Constructor for KaFont class
 *
 * Clears all class members.
 */
KaFont::KaFont() {
    xft = font_ok = false;
    shadow_off_x = shadow_off_y = 0;
    xfont = NULL;
    diff = 0;
    gc = s_gc = NULL;

    xftfont = NULL;
    color = s_color = NULL;
}

/**
 * @fn    KaFont(const KaFont& font)
 * @brief Copy constructor for KaFont class
 *
 * Copy constructor.
 */
KaFont::KaFont(const KaFont& other) {
    xft = other.xft;
    font_ok = other.font_ok;
    shadow_off_x = other.shadow_off_x;
    shadow_off_y = other.shadow_off_y;

    fontNames = other.fontNames;

    //    xfont = new XFontStruct;
    //    memcpy(xfont, other.xfont, sizeof(xfont));
    xfont = other.xfont;
    diff = other.diff;
    gc = other.gc;
    s_gc = other.s_gc;

    xftfont = new XftFont;
    memcpy(xftfont, other.xftfont, sizeof(xftfont));
    color = new XftColor;
    memcpy(color, other.color, sizeof(color));
    s_color = new XftColor;
    memcpy(s_color, other.s_color, sizeof(color));
}

/**
 * @fn    Open(Display *dpy, int screen_number, KaFont *default_font)
 * @brief Opens font
 *
 * Opens the for dpy and screen_number. default_font is used if open fails.
 *
 * @param dpy Display connection
 * @param screen_number Screen to open font for
 * @param default_font Font used if open fails
 *
 * @return Returns height of the opened font
 */
int KaFont::Open(Display *dpy, int screen_number, KaFont *default_font) {
    vector<string>::iterator it = fontNames.begin();
    bool found = false;

    for (; it != fontNames.end(); ++it) {
        string fontName = *it;
        unsigned int xft_index = lower(fontName).find("[xft]");

        if (xft_index != string::npos) {
            string font = trim(fontName.substr(0, xft_index));

            if (! (xftfont = XftFontOpenName(dpy, screen_number, 
                                             font.c_str()))) {
                logger << warn << "failed loading font pattern `" << font << "'" <<
                    endl;
            } else {
                font_ok = true;
                diff = xftfont->ascent - xftfont->descent;
                found = true;
                xft = true;
                break;
            }
        } else {
            if (! (xfont = XLoadQueryFont(dpy, fontName.c_str()))) {
                logger << warn << "failed loading font `" << fontName << "'" << endl;
            } else {
                font_ok = true;
                diff = xfont->ascent - xfont->descent;
                found = true;
                xft = false;
                break;
            }
        }
    }

    if (!found) {
        if (! default_font) return -1;
        xftfont = default_font->xftfont;
        xft = default_font->xft;
        xfont = default_font->xfont;
        diff = default_font->diff;
    }

    if (xft) 
        return xftfont->height;
    else      
        return xfont->ascent + xfont->descent;
}

/**
 * @fn    AllocColor(Display *dpy, Window id, KaColor *kac, KaColor *skac)
 * @brief Allocate colors
 *
 * Creates colors used for font rendering.
 *
 * @param dpy Display connection
 * @param id Window id used for GC creation
 * @param kac Font color
 * @param kac Font shadow color
 */
void KaFont::AllocColor(Display *dpy, Window id, KaColor *kac, KaColor *skac) {
    XGCValues gcv;

    if (xft) {
        color = kac->getXftColor();
        if (skac) s_color = skac->getXftColor();
        return;
    }

    gcv.foreground = kac->getPixel();
    gcv.font = xfont->fid;
    gc = XCreateGC(dpy, id, GCForeground | GCFont, &gcv);
    if (skac) {
        gcv.foreground = skac->getPixel();
        gcv.font = xfont->fid;
        s_gc = XCreateGC(dpy, id, GCForeground | GCFont, &gcv);
    }
}

/**
 * @fn    Draw(Display *dpy, Window id, XftDraw *xftdraw, int x, int y,
 *             string s, int len)
 * @brief Draw text
 *
 * Draws text string on drawable at a given position.
 *
 * @param dpy Display connection
 * @param id Drawable used for X core font rendering
 * @param xftdraw XftDrawable used for Xft font rendering
 * @param x X position to draw text at
 * @param y Y position to draw text at
 * @param s Text string to draw
 * @param len Length of text string s
 */

void KaFont::Draw(Display *dpy, Drawable id, XftDraw *xftdraw, int x, int y,
                  string s, int len) {    

    if (s == "") return;

    if (xft) {
        if (shadow_off_x || shadow_off_y) {
            s_color->color.red = s_color->color.green = s_color->color.blue = 0;
            s_color->color.alpha = 0x3FFF; // 0xFFFF * 25%
            XftDrawStringUtf8(xftdraw, s_color, xftfont,
                              x + 1, y + 2, (unsigned char *) s.c_str(), len);

            s_color->color.alpha = 0x4CCC; // 0xFFFF * 30%
            XftDrawStringUtf8(xftdraw, s_color, xftfont,
                              x + 2, y + 2, (unsigned char *) s.c_str(), len);

            s_color->color.alpha = 0x9999; // 0xFFFF * 60%;
            XftDrawStringUtf8(xftdraw, s_color, xftfont,
                              x + 1, y + 1, (unsigned char *) s.c_str(), len);
        }

        /*        if (shadow_off_x || shadow_off_y)
            XftDrawStringUtf8(xftdraw, s_color, xftfont,
                              x + shadow_off_x, y + shadow_off_y,
                              (unsigned char *) s.c_str(), len);*/
        XftDrawStringUtf8(xftdraw, color, xftfont, x, y,
                          (unsigned char *) s.c_str(), len);
        return;
    }
    
    if (shadow_off_x || shadow_off_y)
        XDrawString(dpy, (Drawable) id, s_gc,
                    x + shadow_off_x, y + shadow_off_y, s.c_str(), len);
    XDrawString(dpy, (Drawable) id, gc, x, y, s.c_str(), len);
}

/**
 * @fn    Width(Display *dpy, string s, int len)
 * @brief Returns text width
 *
 * Calculates text width for a given string.
 *
 * @param dpy Display connection
 * @param s Text string to calculate width for
 *
 * @return Calculated width
 */ 
int KaFont::Width(Display *dpy, string s) {
    if (s == "") return 0;

    if (xft) {
        XGlyphInfo extents;

        char* buf = new char[s.length() + 1];
        strncpy(buf, s.c_str(), s.length());
        buf[s.length()] = '\0';

        XftTextExtentsUtf8(dpy, xftfont, (FcChar8*)buf, strlen(buf), &extents);

        delete[] buf;

        return extents.width;
    }
    
    return XTextWidth(xfont, s.c_str(), s.length());
}
