/* $Id: FrameGeometry.hh,v 1.2 2003/09/11 22:42:21 faugn Exp $ */

/**
 * @file   FrameGeometry.hh
 * @date   31 Aug 2003
 *
 * @brief Class used by KaWindow for resizing/moving/snapping.
 *
 */
#ifndef __FrameGeometry_hh
#define __FrameGeometry_hh

# include "Types.hh"

# include "Window.hh"

class FrameGeometry
{
public:
    virtual void setLeft(int value)=0;
    virtual void setRight(int value)=0;
    virtual void setTop(int value)=0;
    virtual void setBottom(int value)=0;

    virtual int getLeft()=0;
    virtual int getRight()=0;
    virtual int getTop()=0;
    virtual int getBottom()=0;

    virtual int getWidth()  { return getRight()  - getLeft(); };
    virtual int getHeight() { return getBottom() - getTop(); };

    void copyToWindow(KaWindow* w) {
        w->SetFrameLeft(getLeft());
        w->SetFrameTop(getTop());
        w->SetFrameWidth(getWidth());
        w->SetFrameHeight(getHeight());
    }
};

class WindowFrameGeometry: public FrameGeometry
{
public:
    WindowFrameGeometry(KaWindow* window) {
        this->window = window;
    }

    inline int getLeft()   { return window->GetFrameLeft(); };
    inline int getRight()  { return window->GetFrameRight(); };
    inline int getTop()    { return window->GetFrameTop(); };
    inline int getBottom() { return window->GetFrameBottom(); };

    inline void setLeft(int value) { window->SetFrameLeft(value); };
    inline void setRight(int value) { window->SetFrameRight(value); };
    inline void setTop(int value) { window->SetFrameTop(value); };
    inline void setBottom(int value) { window->SetFrameBottom(value); };
private:
    KaWindow* window;
};

class MoveFrameGeometry: public FrameGeometry
{
public:
    MoveFrameGeometry(int nx, int ny, int nwidth, int nheight): 
        x(nx), y(ny), width(nwidth), height(nheight) {};
    
    inline int getLeft()   { return x; };
    inline int getRight()  { return x + width; };
    inline int getTop()    { return y; };
    inline int getBottom() { return y + height; };    

    inline void setLeft(int value)   { x = value; };
    inline void setRight(int value)  { x = value - width; };
    inline void setTop(int value)    { y = value; };
    inline void setBottom(int value) { y = value - height; };
private:
    int x, y, width, height;
};


class ResizeFrameGeometry: public FrameGeometry
{
public:
    ResizeFrameGeometry(KaWindow* w, int nx, int ny, int nwidth, int nheight, 
                        int nhdir, int nvdir): 
        x(nx), y(ny), width(nwidth), height(nheight), 
        hdir(nhdir), vdir(nvdir), window(w) 
    {
    }

    inline int getLeft()   { return x; };
    inline int getRight()  { return x + width; };
    inline int getTop()    { return y; };
    inline int getBottom() { return y + height; };

    inline int getWidth()  { return width;  }
    inline int getHeight() { return height; }

    inline void setLeft(int value)   { 
        if (hdir < 0) {
            int dx  = value - x;

            int n_w, n_h;
            if (window->IncFrameSizeCheck(width - dx, height, &n_w, &n_h)) {
                x     -= n_w - width;
                width  = n_w;
            }
        }
    }

    inline void setRight(int value)  { 
        if (hdir > 0) {            
            int dx  = (value - width) - x;

            int n_w, n_h;
            if (window->IncFrameSizeCheck(width + dx, height, &n_w, &n_h)) {
                width = n_w;
            }
        }
    }

    inline void setTop(int value) { 
        if (vdir < 0) {
            int dy  = value - y;

            int n_w, n_h;
            if (window->IncFrameSizeCheck(width, height - dy, &n_w, &n_h)) {
                y      -= n_h - height;
                height  = n_h;
            }
        }
    }

    inline void setBottom(int value) { 
        if (vdir > 0) {
            int dy  = (value - height) - y;

            int n_w, n_h;
            if (window->IncFrameSizeCheck(width, height + dy, &n_w, &n_h)) {
                height = n_h;
            }
        }
    }

    inline void setWidth(int newWidth) {
        if (hdir == WestType)  x -= newWidth - width;
        width = newWidth;
    }
    inline void setHeight(int newHeight) {
        if (vdir == NorthType) y -= newHeight - height;
        height = newHeight;
    }

    inline void copyPositionFromWindow(KaWindow* w) {
        x = w->GetFrameLeft();
        y = w->GetFrameTop();

        if (hdir == WestType)  x -= width  - w->GetFrameWidth();
        if (vdir == NorthType) y -= height - w->GetFrameHeight();
    }
private:
    int x, y, width, height;
    int hdir, vdir;

    KaWindow* window;
};

#endif
