/* $Id: Image.cc,v 1.16 2003/11/29 01:27:01 pervago Exp $ */

/**
 * @file   Image.cc
 * @date   18-Jul-2001 01:45:32
 *
 * @brief Implementation of KaImage and KaImageControl classes
 *
 * These classes are the Image and ImageControl classes from
 * Blackbox Window Manager. Some minor changes has been made to make them
 * fit into the Kahakai Window Manager project. Except these small
 * changes the KaImage class and the KaImageControl classes are the same as
 * the Image class and the ImageControl class from Blackbox 0.61.1.
 *
 * Thanks Brad!
 * Copyright (c) 1997 - 2000 Brad Hughes (bhughes@tcac.net)
 *
 */

#include "Image.hh"
#include "Kahakai.hh"
#include "Screen.hh"
#include "Logger.hh"

extern "C" {
#ifdef    HAVE_SYS_TYPES_H
#  include <sys/types.h>
#endif // HAVE_SYS_TYPES_H

#ifndef u_int32_t
#  ifdef uint_32_t
typedef uint32_t u_int32_t;
#  else
#    ifdef __uint32_t
typedef __uint32_t u_int32_t;
#    else
typedef unsigned int u_int32_t;
#    endif
#  endif
#endif

#ifdef    STDC_HEADERS
#  include <stdlib.h>
#  include <string.h>
#endif // STDC_HEADERS

#ifdef    HAVE_STDIO_H
#  include <stdio.h>
#endif // HAVE_STDIO_H

#ifdef    HAVE_CTYPE_H
#  include <ctype.h>
#endif // HAVE_CTYPE_H
}

#include <iostream>
#include <list>

using std::cerr;
using std::cout;
using std::endl;
using std::list;

bool have_root_pmap = true;

void KaColor::setRGB(unsigned short r, unsigned short g, unsigned short b) {

    xrenderc.red = r;
    xrenderc.green = g;
    xrenderc.blue = b;
    xrenderc.alpha = 0xFFFF;

    if (r == 65535) red = 0xff;
    else red = (r / 0xff);
    if (g == 65535) green = 0xff;
    else green = (g / 0xff);
    if (b == 65535) blue = 0xff;
    else blue = (b / 0xff);        
}

void KaColor::setXftTransparency(unsigned char t) {
    xftc.color.red = xrenderc.red * (100 - t) / 100;
    xftc.color.green = xrenderc.green * (100 - t) / 100;
    xftc.color.blue = xrenderc.blue * (100 - t) / 100;
    xftc.color.alpha = 0xFFFF * (100 - t) / 100;
    xftc.pixel = getPixel();        
}

KaImageControl::KaImageControl(Display *dpy, KaScreen *scrn, bool _dither,
                               int _cpc, unsigned long cmax) {
    display       = dpy;
    screen_number = scrn->screen_number;
    screen_depth  = DefaultDepth(display, screen_number);
    window        = RootWindow(display, screen_number);
    colormap      = DefaultColormap(display, screen_number);
    visual        = DefaultVisual(display, screen_number);
    kascreen      = scrn;
    
    setDither(_dither);
    setColorsPerChannel(_cpc);
}


KaImageControl::~KaImageControl(void) {
    XSync(kascreen->display, false);
    XSync(kascreen->display, false);
    XSync(kascreen->pdisplay, false);
}



void KaImageControl::installRootColormap(void) {
    XSync(kascreen->display, false);
    
    bool install = true;
    int i = 0, ncmap = 0;
    Colormap *cmaps =
        XListInstalledColormaps(display, window, &ncmap);

    if (cmaps) {
        for (i = 0; i < ncmap; i++)
            if (*(cmaps + i) == colormap)
                install = false;
        
        if (install)
            XInstallColormap(display, colormap);
        
        XFree(cmaps);
    }   
    XSync(kascreen->display, false);
    XSync(kascreen->pdisplay, false);
}

void KaImageControl::setColorsPerChannel(int cpc) {
    if (cpc < 2) cpc = 2;
    if (cpc > 6) cpc = 6;
    
    colors_per_channel = cpc;
}

void KaImageControl::parseColor(KaColor *color, const char *c) {
    if (! color) return;

    if (color->isAllocated()) {
        unsigned long pixel = color->getPixel();

        XFreeColors(display, colormap, &pixel, 1, 0);

        color->setPixel(0l);
        color->setRGB(0, 0, 0);
        color->setAllocated(false);
    }
    if (c) {
        unsigned short r, g, b;

        color->setPixel(getColor(c, &r, &g, &b));
        color->setRGB(r, g, b);
        color->setAllocated(true);
    }
}

unsigned long KaImageControl::getColor(const char *colorname,
                                       unsigned short *r, unsigned short *g,
                                       unsigned short *b) {
    XColor color;
    color.pixel = 0;

    if (! XParseColor(display, colormap, colorname, &color))
        logger << warn << "color parse error: \"" << colorname << "\"" << endl;
    else if (! XAllocColor(display, colormap, &color))
        logger << warn << "color alloc error: \"" << colorname << "\"" << endl;

    *r = color.red;
    *g = color.green;
    *b = color.blue;

    return color.pixel;
}

void KaImageControl::setXRootPMapId(bool hrp) { have_root_pmap = hrp; }

