#include "Imlib2pp.hh"
#include "Logger.hh"


/* Imlib2 images */

ImColor::ImColor(int r, int g, int b, int a) {
    red = r;
    green = g;
    blue = b;
    alpha = a;
}

void ImColor::set() const {
    imlib_context_set_color(red, green, blue, alpha);
}

ImColor ImColor::get_hilite() const {
    int r, g, b;
    r = red + (red >> 1);
    g = green + (green >> 1);
    b = blue + (blue >> 1);
    if (r > 0xff)
        r = 0xff;
    if (g > 0xff)
        g = 0xff;
    if (b > 0xff)
        b = 0xff;
    return ImColor(r, g, b, alpha);
}

ImColor ImColor::get_lolite() const {
    int r, g, b;
    r = (red >> 2) + (red >> 1);
    g = (green >> 2) + (green >> 1);
    b = (blue >> 2) + (blue >> 1);
    return ImColor(r, g, b, alpha);
}

/* Imlib2 color ranges */
ImColorRange::ImColorRange(const std::vector<std::pair<ImColor, int> >& 
                           _range) {
    colorRange = imlib_create_color_range();
    set();

    std::vector<std::pair<ImColor, int> >::const_iterator it;
    for (it = _range.begin(); it != _range.end(); ++it) {
        it->first.set();
        imlib_add_color_to_color_range(it->second);
    }
}

ImColorRange::~ImColorRange() {
    set();
    imlib_free_color_range();
}

void ImColorRange::set() const {
    imlib_context_set_color_range(colorRange);
}

/* Imlib2 images */

ImImage::ImImage() {
    image = NULL;
}

ImImage::ImImage(int width, int height) {
    if (width && height)
        image = imlib_create_image(width, height);
    else
        image = NULL;
}

ImImage::ImImage(const std::string& fn) {
    image = imlib_load_image(fn.c_str());
    if (!image) {
        logger << warn << "failed loading image `" << fn << "'\n";
    }
}

ImImage::ImImage(int width, int height, DATA32 *data) {
    image = imlib_create_image_using_copied_data(width, height, data);
}

ImImage::ImImage(Drawable drawable, Pixmap mask, 
                 int x, int y, int width, int height, char grab_X) {
    imlib_context_set_drawable(drawable);
    image = imlib_create_image_from_drawable(mask, x, y, width, height, grab_X);
}

ImImage::ImImage(Drawable drawable, Pixmap mask,
                 int source_x, int source_y, 
                 int source_width, int source_height,
                 int dest_width, int dest_height,
                 char grab_X, char shape_mask) {
    imlib_context_set_drawable(drawable);
    image = imlib_create_scaled_image_from_drawable(mask,
           source_x, source_y, source_width, source_height,
           dest_width, dest_height, grab_X, shape_mask);
}

ImImage::~ImImage() {
    if (image) {
        set();
        imlib_free_image();
        image = 0;
   }
}

bool ImImage::valid() const {
    return image != 0;
}

void ImImage::set() const {
    imlib_context_set_image(image);
}

void ImImage::set_has_alpha(bool has_alpha) {
    set();
    imlib_image_set_has_alpha(has_alpha);
}

void ImImage::clear() {
    set();
    imlib_image_clear();
}

void ImImage::clear(const ImColor& color) {
    set();
    imlib_image_clear_color(color.red, color.green, color.blue, color.alpha);
}

int ImImage::get_width() const {
    set();
    return imlib_image_get_width();
}

int ImImage::get_height() const {
    set();
    return imlib_image_get_height();
}

ImImage ImImage::crop(int x, int y, int width, int height) {
    ImImage img;
    set();
    img.image = imlib_create_cropped_image(x, y, width, height);
    return img;
}

void ImImage::set_border(int left, int top, int right, int bottom) {
    Imlib_Border border;
    border.top = top;
    border.left = left;
    border.right = right;
    border.bottom = bottom;
    set();
    imlib_image_set_border(&border);
}

void ImImage::set_border(int width) {
    set_border(width, width, width, width);
}

void ImImage::get_border(int& left, int& top, int& right, int& bottom) {
    Imlib_Border border;
    set();
    imlib_image_get_border(&border);
    left = border.left;
    top = border.top;
    right = border.right;
    bottom = border.bottom;
}

void ImImage::blend_part_scaled(const ImImage &img, 
                                bool alpha, int s_x, int s_y, int s_w, int s_h,
                                int d_x, int d_y, int d_w, int d_h) {
    set();
    imlib_blend_image_onto_image(img.image, 
                                 alpha, s_x, s_y, s_w, s_h, 
                                 d_x, d_y, d_w, d_h);
}

void ImImage::blend_scaled(const ImImage& img, 
                           bool alpha, int d_x, int d_y, int d_w, int d_h) {
    blend_part_scaled(img, alpha,
                     0, 0, img.get_width(), img.get_height(),
                     d_x, d_y, d_w, d_h);
}

void ImImage::blend(const ImImage& img, bool alpha, int d_x, int d_y) {
    int s_w = img.get_width(),
        s_h = img.get_height();
    blend_part_scaled(img, alpha,
                     0, 0, s_w, s_h,
                     d_x, d_y, s_w, s_h);
}

void ImImage::draw_line(ImColor color, int x1, int y1, int x2, int y2) {
    color.set();
    set();
    imlib_image_draw_line(x1, y1, x2, y2, 0);
}

void ImImage::draw_rectangle(ImColor color, 
                             int x, int y, int width, int height) {
    color.set();
    set();
    imlib_image_draw_rectangle(x, y, width, height);
}

void ImImage::fill_rectangle(ImColor color, 
                             int x, int y, int width, int height) {
    color.set();
    set();
    imlib_image_fill_rectangle(x, y, width, height);
}

void ImImage::fill_color_range_rectangle(const ImColorRange& colorRange, 
                                         int x, int y, int width, int height,
                                         double angle) {
    colorRange.set();
    set();
    imlib_image_fill_color_range_rectangle(x, y, width, height, angle);
}

void ImImage::draw_polygon(ImPolygon polygon, ImColor color, bool closed) {
    color.set();
    imlib_image_draw_polygon(polygon.polygon, closed);
}

void ImImage::fill_polygon(ImPolygon polygon, ImColor color) {
    color.set();
    imlib_image_fill_polygon(polygon.polygon);
}

void ImImage::draw_bevel(ImColor color, 
                         unsigned int width, int bevel_type) {
    ImColor c1, c2;
    int w = get_width(),
        h = get_height();

    if (bevel_type == 0) {
        c1 = color.get_hilite();
        c2 = color.get_lolite();
    } else {
        c1 = color.get_lolite();
        c2 = color.get_hilite();
    }

    for (unsigned int i = 0; i < width; i++) {
        draw_line(c1, 0, i, w - i - 1, i);
        draw_line(c1, i, 1, i, h - i - 1);
        draw_line(c2, i+1, h - i - 1, w - 1, h - i - 1);
        draw_line(c2, w - i - 1, i + 1, w - i - 1, h - width);
    }
}

void ImImage::apply_fixed_alpha(unsigned char alpha) {
    set();
    
    // Create alpha-override mask
    set_has_alpha(true);
    
    Imlib_Color_Modifier modifier = imlib_create_color_modifier ();
    imlib_context_set_color_modifier(modifier);
    
    DATA8 red[256], green[256], blue[256], alph[256];
    imlib_get_color_modifier_tables (red, green, blue, alph);
    
    for (int i = 0; i < 256; i++)
        alph[i] = (DATA8) (i * alpha / 256);
    
    imlib_set_color_modifier_tables (red, green, blue, alph);
    
    imlib_apply_color_modifier ();
    imlib_free_color_modifier ();
}

void ImImage::set_format(const std::string& format) {
    set();
    imlib_image_set_format(format.c_str());
}

void ImImage::save(const std::string& fn) {
    set();
    imlib_save_image(fn.c_str());
}

void ImImage::render_pixmaps(Drawable parent, 
                             Pixmap *pixmap_return, 
                             Pixmap *mask_return) const {
    set();
    imlib_context_set_drawable(parent);
    imlib_render_pixmaps_for_whole_image(pixmap_return, mask_return);
}

void ImImage::render_pixmaps(Drawable parent, 
                             Pixmap *pixmap_return, 
                             Pixmap *mask_return, int width, int height) const {
    set();
    imlib_context_set_drawable(parent);
    imlib_render_pixmaps_for_whole_image_at_size(pixmap_return, 
                                                 mask_return, width, height);
}

void ImImage::free_pixmap_and_mask(Pixmap pixmap) {
    imlib_free_pixmap_and_mask(pixmap);
}

void ImImage::render_on_drawable(Drawable drawable, int x, int y) const {
    set();
    imlib_context_set_drawable(drawable);
    imlib_render_image_on_drawable(x, y);
}

void ImImage::render_on_drawable(Drawable drawable, int x, int y, int width, int height) {
    set();
    imlib_context_set_drawable(drawable);
    imlib_render_image_on_drawable_at_size(x, y, width, height);
}

void ImImage::render_on_drawable(Drawable drawable,
        int source_x, int source_y, int source_width, int source_height,
        int dest_x, int dest_y, int dest_width, int dest_height) {
    set();
    imlib_context_set_drawable(drawable);
    imlib_render_image_part_on_drawable_at_size(source_x, source_y, 
                                                source_width, source_height,
                                                dest_x, dest_y, 
                                                dest_width, dest_height);
}


/* Imlib2 polygons */

ImPolygon::ImPolygon() {
    polygon = imlib_polygon_new();
}

ImPolygon::~ImPolygon() {
    imlib_polygon_free(polygon);
}

void ImPolygon::add_point(int x, int y) {
    imlib_polygon_add_point(polygon, x, y);
}

void ImPolygon::get_bounds(int& x1, int& y1, int& x2, int& y2) {
    imlib_polygon_get_bounds(polygon, &x1, &y1, &x2, &y2);
}

bool ImPolygon::contains(int x, int y) {
    return imlib_polygon_contains_point(polygon, x, y);
}


/* Imlib2 contexts */

ImContext::ImContext(Display *display, Visual *visual, Colormap colormap) {
    context = imlib_context_new();
    push();
    imlib_context_set_display(display);
    imlib_context_set_visual(visual);
    imlib_context_set_colormap(colormap);
    pop();
}

ImContext::~ImContext() {
    imlib_context_free(context);
}

void ImContext::push() {
    imlib_context_push(context);
}

void ImContext::pop() {
    imlib_context_pop();
}


/* int main(int argc, char **argv) {
    Display *disp = XOpenDisplay(NULL);
    int screen = DefaultScreen(disp);
    Colormap cmap = DefaultColormap(disp, screen);
    Visual *visual = DefaultVisual(disp, screen);
    
    ImContext *ctx = new ImContext(disp, visual, cmap);
    ctx->push();
    
    ImColor *clr1 = new ImColor(255, 80, 80),
            *clr2 = new ImColor(80, 80, 255),
            *hi = clr2->get_hilite(),
            *lo = clr2->get_lolite();
            
    ImImage *img1 = new ImImage(400, 300),
            *img2 = new ImImage(200, 100);

    img1->fill_rectangle(clr1, 0, 0, 400, 300);
    img2->fill_rectangle(clr2, 0, 0, 200, 100);
    img2->draw_bevel(clr2, 5, 0);

    img1->blend(img2, 0, 100, 100);

    img1->set_format("png");
    img1->save("test.png");

    delete img1;
    delete img2;
    delete clr1;
    delete clr2;
    delete hi;
    delete lo;

    ctx->pop();
    delete ctx;

    XCloseDisplay(disp);
    
} */
