/* $Id: Kahakai.cc,v 1.27 2004/02/04 21:53:28 brailsmt Exp $ */

/**
 * @file   Kahakai.cc
 * @date   02-May-2001 00:48:03
 *
 * @brief Implementation of Kahakai and WindowObject classes
 *
 * Kahakai Window Manager.
 *
 */

#include "Kahakai.hh"
#include "Resources.hh"
#include "Net.hh"
#include "Screen.hh"
#include "Event.hh"
#include "Timer.hh"
#include "Window.hh"
#include "Logger.hh"

extern "C" {
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/extensions/shape.h>

#ifdef    XRANDR
# include <X11/extensions/Xrandr.h>
#endif // XRANDR

#ifdef    HAVE_STDIO_H
# include <stdio.h>
#endif // HAVE_STDIO_H

#ifdef    STDC_HEADERS
# include <stdlib.h>
#endif // STDC_HEADERS

#ifdef    HAVE_UNISTD_H
# include <sys/types.h>
# include <sys/wait.h>
# include <sys/stat.h>
# include <unistd.h>
#endif // HAVE_UNISTD_H

#ifdef    HAVE_SIGNAL_H
# include <signal.h>
#endif // HAVE_SIGNAL_H
}

#include <iostream>
#include <sstream>
#include <string>
#include <map>

using std::stringstream;
using std::string;
using std::endl;
using std::map;

Kahakai *kahakai;
char **argv;
bool hush;
int errors;

Kahakai* Kahakai::_instance = NULL;


/**
 * @fn    Kahakai(char **av)
 * @brief Constructor for Kahakai class
 *
 * Here we open a connection to the display and set signal and xerror
 * handler functions. A list for all KaWindow is created and one list for all
 * KaMenus. An hash_map is also created for effective window searching. Then
 * we load the configuration file, menu file and actions file. We create one
 * KaScreen for the displays default screen and an eventhandler for handling
 * all events.
 *
 * @param av Vector of command line arguments
 * @param _options Parsed command line options
 */
Kahakai::Kahakai(char **av, struct kaoptions *_options) {
    struct sigaction action;
    int dummy;

    argv = av;
    options = _options;
    XSetErrorHandler((XErrorHandler) xerrorhandler);
    if (! (display = XOpenDisplay(options->display))) {
        logger << "error: can't open display: " << options->display << endl;
        exit(1);
    }
    kahakai = this;
    hush = wmerr = false;
    errors = 0;
    eh = NULL;
    timer = NULL;


    _instance = this;

#ifdef XINERAMA
    xinerama = new Xinerama();
#endif

    action.sa_handler = signalhandler;
    action.sa_mask = sigset_t();
    action.sa_flags = SA_NOCLDSTOP | SA_NODEFER;

    sigaction(SIGTERM, &action, NULL);
    sigaction(SIGINT, &action, NULL);
    sigaction(SIGCHLD, &action, NULL);
    sigaction(SIGHUP, &action, NULL);

    session_cursor = XCreateFontCursor(display, XC_left_ptr);
    move_cursor = XCreateFontCursor(display, XC_fleur);
    resizesw_cursor = XCreateFontCursor(display, XC_ll_angle);
    resizese_cursor = XCreateFontCursor(display, XC_lr_angle);
    resizenw_cursor = XCreateFontCursor(display, XC_ul_angle);
    resizene_cursor = XCreateFontCursor(display, XC_ur_angle);
    resizew_cursor = XCreateFontCursor(display, XC_left_side);
    resizee_cursor = XCreateFontCursor(display, XC_right_side);

    shape = XShapeQueryExtension(display, &shape_event, &dummy);

#ifdef XRANDR
    randr = XRRQueryExtension(display, &randr_event, &dummy);
#endif // XRANDR

    rh = new ResourceHandler(this, options);
    net = new NetHandler(this);

    rh->LoadConfig(this);

    int i, screens = 0;
    KaScreen *ws;

    for (i = 0; i < ScreenCount(display); ++i) {
        if (screenmask & (1L << i)) {
            ws = new KaScreen(display, i, this);
            if (! wmerr) {
                kascreen_list.push_back(ws);
                screens++;
            }
            else wmerr = false;
        }
    }
    if (! screens) {
        logger << err << "no managable screens found on display " <<
            DisplayString(display) << endl;
        exit(1);
    }

    eh = new EventHandler(this);
    timer = new Timer(this);

    // Load autostart file
    kaexec_if_exists((string(getenv("HOME")) + "/.kahakai/autostart").c_str(),
                     DisplayString(display));
}

/**
 * @fn    ~Kahakai(void)
 * @brief Destructor for Kahakai class
 *
 * Deletes all KaScreens. Closes the connection to the display.
 */
Kahakai::~Kahakai(void) {
    XSetErrorHandler(NULL);
    LISTDEL(kascreen_list);
    delete net;
    delete rh;
    MAPCLEAR(window_table);
    if (eh) delete eh;
    if (timer) delete timer;

    delete [] pathenv;

    XSync(display, false);
    XCloseDisplay(display);
	//close_log();
}

void Kahakai::RunEventLoop() {
  XEvent e;
  eh->EventLoop(eh->empty_return_mask, &e);
}

Kahakai* Kahakai::Instance() {
  if (!_instance) {
    printf("WARNING: No Kahakai instance exists!");
    return NULL;
  } else {
    return _instance;
  }
}

/**
 * @fn    FindWin(Window id, int mask)
 * @brief Find WindowObject
 *
 * Returns WindowObject matching id and mask. NULL if no match was found.
 *
 * @param id Window id to use for matching
 * @param mask Window type mask to use for matching
 *
 * @return Matching WindowObject
 */
WindowObject *Kahakai::FindWin(Window id, int mask) {
    map<Window, WindowObject *>::iterator it;
    if ((it = window_table.find(id)) != window_table.end()) {
        if (((*it).second)->type & mask)
            return (*it).second;
    }
    return NULL;
}


/**
 * @fn    validatedrawable(Drawable d, unsigned int *w, unsigned int *h)
 * @brief Validates if a drawable exists
 *
 * Tries to get geometry for the drawable. The drawable is valid
 * if no XError is generated.
 *
 * @param d Resource ID used for drawable validation
 * @param w Return the drawable's width
 * @param h Return the drawable's height
 *
 * @return True if drawable is valid, otherwise false
 */
bool validatedrawable(Drawable d, unsigned int *w, unsigned int *h) {
    int _d;
    unsigned int _ud;
    Window _wd;

    XSync(kahakai->display, false);
    errors = 0;
    hush = 1;

    if (w == NULL)
        XGetGeometry(kahakai->display, d, &_wd, &_d,&_d, &_ud,&_ud,&_ud,&_ud);
    else
        XGetGeometry(kahakai->display, d, &_wd, &_d, &_d, w, h, &_ud, &_ud);

    XSync(kahakai->display, false);
    hush = 0;

    bool ret = (bool) (!errors);
    errors = 0;
    return ret;
}

/**
 * @fn    validateclient_mapped(Window id)
 * @brief Validates if a window exist and is mapped
 *
 * First we check that the window exists, if it exists we check the event
 * queue so that there's no UnmapNotify event from the window in it. If there's
 * no UnmapNotify event in the event queue then the window exists and is
 * mapped.
 *
 * @param id Resource ID used for window validation
 *
 * @return True if window is valid, otherwise false
 */
const bool validateclient_mapped(Window id) {
    XFlush(kahakai->display);

    XEvent e;
    if (validatedrawable((Drawable) id)) {
        if (XCheckTypedWindowEvent(kahakai->display, id, UnmapNotify, &e)) {
            XPutBackEvent(kahakai->display, &e);
            return false;
        }
        return true;
    }
    return false;
}


/**
 * @fn    kaexec(const char *command, char *displaystring)
 * @brief Executes a command line
 *
 * Executes a command line with /bin/sh
 *
 * @param format Command line to execute
 * @param displaystring Displaystring to export prior to execution
 */
void kaexec(const char *command, char *displaystring) {
    if (! fork()) {
        setsid();
        putenv(displaystring);
        execl("/bin/sh", "/bin/sh", "-c", command, NULL);
        exit(0);
    }
}

void kaexec_if_exists(const char *file, char* displaystring) {
    struct stat st;

    if (stat(file, &st) == 0 && st.st_mode & S_IXUSR)
        kaexec(file, displaystring);
}

/**
 * @fn    xerrorhandler(Display *d, XErrorEvent *e)
 * @brief X error handler function
 *
 * Prints error message then a X error occurs.
 *
 * @param d X display
 * @param e X error event
 *
 * @return always 0
 */
int xerrorhandler(Display *d, XErrorEvent *e) {
    char buff[128];
    map<Window, WindowObject *>::iterator it;

    errors++;

    if (! hush) {
        XGetErrorDatabaseText(d, "XlibMessage", "XError", "", buff, 128);
        logger << buff;

        XGetErrorText(d, e->error_code, buff, 128);
        logger << ":  " << buff << endl;

        XGetErrorDatabaseText(d, "XlibMessage", "MajorCode", "%d", buff, 128);
        logger << "  ";
        fprintf(stderr, buff, e->request_code);
        sprintf(buff, "%d", e->request_code);

        XGetErrorDatabaseText(d, "XRequest", buff, "%d", buff, 128);
        logger << " (" << buff << ")" << endl;

        XGetErrorDatabaseText(d, "XlibMessage", "MinorCode", "%d", buff, 128);
        logger << "  ";
        fprintf(stderr, buff, e->minor_code);
        logger << endl;

        XGetErrorDatabaseText(d, "XlibMessage", "ResourceID", "%d", buff, 128);
        logger << "  ";
        fprintf(stderr, buff, e->resourceid);

        if ( ((it = kahakai->window_table.find(e->resourceid)) !=
              kahakai->window_table.end()) &&
            (((*it).second)->type == WindowType)) {

            logger << " (" << ((KaWindow *) (*it).second)->name << ")";
        }

        logger << endl;
    }
    return 0;
}

/**
 * @fn    wmrunningerror(Display *d, XErrorEvent *)
 * @brief X error handler function
 *
 * X error handler function used when setting input mask of root window.
 * If X error occurs another window manager is running on that screen.
 *
 * @param d X display
 *
 * @return 0
 */
int wmrunningerror(Display *d, XErrorEvent *) {
    kahakai->wmerr = true;
    return 0;
}

/**
 * @fn    signalhandler(int sig)
 * @brief Signal handler function
 *
 * When one of the signals we handle arrives this function is called. Depending
 * on what type of signal we received we do something, ex. restart, exit.
 *
 * @param sig The signal we received
 */
void signalhandler(int sig) {
    switch(sig) {
        case SIGINT:
        case SIGTERM:
            quit(EXIT_SUCCESS);
            break;
        case SIGHUP:
            restart(NULL);
            break;
        case SIGCHLD:
            // Wait for all children that exited
            while (waitpid(-1, NULL, WNOHANG | WUNTRACED) > 0) {};
            break;
        default:
            quit(EXIT_FAILURE);
    }
}

/**
 * @fn    restart(char *command)
 * @brief Restarts program
 *
 * Deletes the kahakai object and restarts window manager.
 *
 * @param command Program name to execute when restarting
 */
void restart(char *command) {
    char *tmp_argv[128];
    char *__m_wastrdup_tmp;

    if (command && *command != '\0') {
        commandline_to_argv(__m_wastrdup(command), tmp_argv);
        delete kahakai;
        execvp(*tmp_argv, tmp_argv);
        perror(*tmp_argv);
        exit(EXIT_FAILURE);
    } else
        delete kahakai;

    execvp(argv[0], argv);
    perror(argv[0]);
    exit(EXIT_FAILURE);
}

/**
 * @fn    quit(void)
 * @brief Ends program execution
 *
 * Deletes the kahakai object and then ends program execution.
 *
 * @param status Return status
 */
void quit(int status) {
    delete kahakai;
    exit(status);
}

/**
 * @fn    commandline_to_argv(char *s)
 * @brief Parse command line to arguments
 *
 * Parses a command line string into an vector of arguments. Characters
 * between '"' are parsed as one argument. Dangerous because it destroys
 * command line string given as parameter.
 *
 * @param s Command line string
 * @param tmp_argv Vector that should point to the different parameters
 *
 * @return argument vector
 */
char **commandline_to_argv(char *s, char **tmp_argv) {
    int i;

    for (i = 0;;) {
        for (; *s == ' ' || *s == '\t'; s++);
        if (*s == '"') {
            tmp_argv[i++] = ++s;
            for (; *s != '"' && *s != '\0'; s++);
            if (*s == '\0') break;
            *s = '\0';
            s++;
            if (*s == '\0') break;
        }
        else {
            tmp_argv[i++] = s;
            for (; *s != ' ' && *s != '\t' && *s != '\0'; s++);
            if (*s == '\0') break;
            *s = '\0';
            s++;
        }
    }
    tmp_argv[i] = NULL;
    return tmp_argv;
}

/**
 * @fn    replaceAll(string src, string substring, string replacement)
 * @brief Replaces all instances of substring in src with replacement
 */
string replaceAll(string src, string substring, string replacement)
{
    string::size_type last = src.find(substring);
    string replaced = src;

    while (last != string::npos && last <= src.length() - 1) {
        replaced  = replaced.replace(last, substring.length(), replacement);
        last      = replaced.find(substring);
    }
    return replaced;
}



/**
 * @fn    expand(string org, KaWindow *w)
 * @brief Window info expansion
 *
 * Expands a special window info characters in a string. Special info
 * characters are:
 * '\p' replaced with _NET_WM_PID hint for window
 * '\n' replaced with name part of windows WM_CLASS hint
 * '\c' replaced with class part of windows WM_CLASS hint
 * '\h' replaced with WM_CLIENT_MACHINE hint
 *
 * @param org Original string used as source for expansion
 * @param w KaWindow to get expansion info from
 *
 * @return Expanded version of original string
 */
string expand(string org, KaWindow *w) {
    map<string, string> expansions;

    expansions["\\p"] = w->pid ? w->pid : "";
    expansions["\\h"] = w->host ? w->host : "";
    expansions["\\n"] = w->classhint->res_name ? w->classhint->res_name : "";
    expansions["\\c"] = w->classhint->res_class ? w->classhint->res_class : "";

    string expanded = org;

    map<string, string>::iterator e_it = expansions.begin();
    for (; e_it != expansions.end(); e_it++) {
        expanded = replaceAll(expanded, e_it->first, e_it->second);
    }

    return expanded;
}

char* expand(char* orig, KaWindow* w) {
    if (!orig) return NULL;

    string expanded = expand(string(orig), w);
    char* buf = new char[expanded.length() + 1];
    strcpy(buf, expanded.c_str());
    return buf;
}

string trim(string str) {
    char* whitespace = " \n\r\t";

    string::size_type start = str.find_first_not_of(whitespace);
    if (start == string::npos) return "";

    return string(str.begin() + start,
                  str.begin() + str.find_last_not_of(whitespace)+1);
}

/**
 * @fn     viewportStrToXY(string viewport_str, int * x, int * y)
 * @brief  Converts a viewport coordinate string to x and y integer coordinates.
 */
bool viewportStrToXY(string viewport_str, int * ret_x, int * ret_y) {

    // must match "%dx%d", and each %d must be only one digit.

    if (viewport_str.length() != 3 || viewport_str.find("x") != 1 ||
        viewport_str.find_first_of("0123456789") != 0 ||
        viewport_str.find_last_of("0123456789") != 2) {

        logger << warn << "viewportStrToXY: bad input: " << viewport_str <<endl;
        return false;
    }

    *ret_x = atoi(viewport_str.substr(0).c_str());
    *ret_y = atoi(viewport_str.substr(2).c_str());

    return true;
}

/**
 * @fn     desktopStrToNum(string desktop_str)
 * @brief  Converts a desktop string to a desktop number.
 *
 * Returns 0-15, corresponding to each desktop, 16 for all desktops,
 * and -1 on error.
 */
int desktopStrToNum(string desktop_str) {

    int len = desktop_str.length();

    if (!(4 > len && len > 0)) {
        logger << warn << "desktopStrToNum: bad input: " << desktop_str << endl;
        return -1;
    }

    if (lower(desktop_str) == "all")
        return 16;

    stringstream dss(desktop_str);
    int num;
    dss >> num;

    if(16 > num && num >= 0)
        return num;

    logger << warn << "desktopStrToNum: bad input: " << desktop_str << endl;
    return -1;
}

unsigned int desktopMaskStrToMask(const char * mask, unsigned int oldmask,
                                                     unsigned int allowed) {
    if (! mask) {
        logger << warn << "desktopMaskStrToMask: bad input: " << mask << endl;
        return (unsigned int) 0;
    }

    if (! strncasecmp("all", mask, 3))
        return (1L << 16) - 1;

    char *in = new char[strlen(mask)+1];
    sprintf(in, "%s", mask);

    char *token = strtok(in, " \t");
    while (token) {
        unsigned int desk = (unsigned int) atoi(token);
        if (desk < allowed)
            oldmask |= 1L << desk;
        token = strtok(NULL, " \t");
    }

    delete in;

    return oldmask; // new mask now
}


/**
 * @fn     lower(string a)
 * @brief  Returns lowercase version of the string
 */
string lower(string a) {
    for (unsigned int i=0; i < a.length(); i++)
        a[i] = tolower(a[i]);

    return a;
}

int msDifference(struct timeval first, struct timeval second) {
    return (second.tv_usec - first.tv_usec) / 1000 +
           (second.tv_sec - first.tv_sec) * 1000;
}

