/* $Id: Menu.cc,v 1.52 2004/02/03 01:05:30 hyriand Exp $ */

/**
 * @file   Menu.cc
 * @date   02-Aug-2001 22:40:20
 *
 * @brief Implementation of KaMenu, TaskSwitcher classes
 *
 * Implementation of menu system. A menu consists of a list of menu items
 * each menu item can be one of these three kinds Title, Item or Sub. The only
 * thing separating these items are the looks of them and action lists. All
 * three types can perform the same actions. Possible actions are execution of
 * program, call to function and mapping of menu as submenu.
 *
 */

#include "ImageControl.hh"
#include "Menu.hh"
#include "MenuItem.hh"
#include "Resources.hh"
#include "Texture.hh"
#include "Window.hh"

extern "C" {
#include <X11/extensions/shape.h>
#ifdef    HAVE_STDIO_H
#  include <stdio.h>
#endif // HAVE_STDIO_H

#ifdef    STDC_HEADERS
#  include <string.h>
#endif // STDC_HEADERS
}

#include <algorithm>
#include <iostream>
#include <string>

using std::max;
using std::map;
using std::cerr;
using std::cout;
using std::endl;
using std::list;
using std::vector;
using std::make_pair;

/**
 * @fn    KaMenu(string n) : WindowObject((Window) 0, MenuType)
 * @brief Constructor for KaMenu class
 *
 * Creates a new menu with no items
 *
 * @param n Name of menu
 */
KaMenu::KaMenu(string n) : WindowObject((Window) 0, MenuType) {
    name = n;

    height = 0;
    width = 0;
    mapped = has_focus = built = dynamic = dynamic_root =
        ignore = db = need_delete = false;
    ext_type = NoExtMenuType;
    root_menu = NULL;
    root_item = NULL;
    wf = (Window) 0;
    rf = NULL;
    mf = NULL;

    pixmap = None;
    render_if_transparency = false;

    dirty = false;

    iconsPresent = false;
}

/**
 * @fn    ~KaMenu(void)
 * @brief Destructor for KaMenu class
 *
 * Deletes all menu items in the menu and removes the frame.
 */
KaMenu::~KaMenu(void) {
    LISTDELITEMS(item_list);
    if (built) {
        //        delete control;

        XDestroyWindow(display, frame);

        if (pixmap != None && pixmap != ParentRelative) {
            XSync(display, false);
        }
    }
}

/**
 * @fn    AddItem(KaMenuItem *item)
 * @brief Adds item to menu
 *
 * Adds item to menu by inserting a KaMenuItem object to the menus item list.
 *
 * @param item Item to add
 */
void KaMenu::AddItem(KaMenuItem *item) {
    item->menu = this;
    item->hilited = false;
    item_list.push_back(item);
}


/**
 * @fn    Build(Kascreen *screen)
 * @brief Builds Menu
 *
 * Calculates size of menu, finds submenus sets pointers to them, renders
 * graphics and creates windows. If menu has already been built, when we just
 * resize the windows instead of creating new ones.
 *
 * @param screen Screen to create graphics and windows in
 */
void KaMenu::Build(KaScreen *screen) {
    printf("Menu::Build(screen=%d)\n", screen->screen_number);

    MenuStyle& mstyle = screen->style.menu;

    height = 0;
    width = 0;

    if (! built) {
        kascreen = screen;
        kahakai = kascreen->kahakai;
        display = kascreen->display;
        ic = kascreen->ic;
    }
    bullet_width = cb_width = 0;

    f_height = kascreen->style.menu.itemHeight;
    t_height = kascreen->style.menu.titleHeight;

    list<KaMenuItem *>::iterator it = item_list.begin();
    for (; it != item_list.end(); ++it) {
        if ((*it)->func_mask & MenuSubMask) {
            if ((*it)->sub.find('!') != string::npos) {
                (*it)->sdyn = (*it)->sdyn1 = true;
            } else {
                (*it)->submenu = kascreen->GetMenuNamed((*it)->sub);

                if (!(*it)->submenu) {
                    delete *it;
                    it = item_list.begin();
                }
            }
        }
        if ((*it)->func_mask2 & MenuSubMask) {
            if ((*it)->sub2.find('!') != string::npos) {
                (*it)->sdyn2 = true;
            } else {
                (*it)->submenu2 = kascreen->GetMenuNamed((*it)->sub2);

                if (!(*it)->submenu2) {
                    delete *it;
                    it = item_list.begin();
                }
            }
        }
    }


    bullet_width =  mstyle.bullet.font.Width(display, mstyle.bullet.label);

    int checkboxTrueWidth = mstyle.checkboxTrue.font.Width(display,
                                                            mstyle.checkboxTrue.label);
    int checkboxFalseWidth = mstyle.checkboxTrue.font.Width(display,
                                                             mstyle.checkboxFalse.label);
    cb_width = max(checkboxTrueWidth, checkboxFalseWidth);

    for (it = item_list.begin(); it != item_list.end(); ++it) {
        if ((*it)->type == MenuCBItemType) {
            (*it)->cb_width2 = checkboxTrueWidth;
            (*it)->cb_width1 = checkboxFalseWidth;
            (*it)->cb_width  = cb_width;
        }
    }
    extra_width = (bullet_width >= cb_width) ? bullet_width: cb_width;

    determineItemDimensions();

    if (width > (kascreen->width / 2)) width = kascreen->width / 2;

    control = new ImageControl(kascreen, kascreen->display, 
                               kascreen->screen_number, 
                               kascreen->visual, 
                               kascreen->colormap);

    createWindows();
    built = true;
}

/**
 * @fn    createWindows(KaScreen* screen)
 * @brief Create/resize menu window and item windows
 *
 * Private function (used by Build method).
 */
void KaMenu::createWindows() {
    MenuStyle &mstyle = kascreen->style.menu;

    XSetWindowAttributes attrib_set;

    attrib_set.border_pixel = mstyle.borderColor.getPixel();
    attrib_set.colormap = kascreen->colormap;
    attrib_set.override_redirect = true;
    attrib_set.event_mask = NoEventMask;

    if (! built)
        frame = XCreateWindow(display, kascreen->id, 0, 0,
                              width ? width : 1, height ? height : 1,
                              mstyle.borderThickness,
                              kascreen->screen_depth,
                              CopyFromParent, kascreen->visual,
                              CWOverrideRedirect | CWEventMask | 
                              CWColormap | CWBorderPixel,
                              &attrib_set);
    else 
        XResizeWindow(display, frame, width ? width : 1, height ? height : 1);

    int lasttype;
    int y, bw;
    list<KaMenuItem*>::iterator it = item_list.begin();
    for (y = 0, lasttype = 0; it != item_list.end(); ++it) {
        bw = 0;
        if ((*it)->type == MenuTitleType) {
            bw = mstyle.borderThickness;
            if ((y == 0) || (lasttype == MenuTitleType))
                y -= bw;
        }
        (*it)->CreateWindow(kascreen, y);
        y += (*it)->height + bw * 2;
        lasttype = (*it)->type;
    }
}

void KaMenu::Shape(void) {

    // A lot of credit to metacity 

    int new_window_width = width;
    int new_window_height = height;

    Region window_xregion;

    bool rounded_topleft = kascreen->style.menu.roundedTopLeft, 
         rounded_topright = kascreen->style.menu.roundedTopRight,
         rounded_bottomleft = kascreen->style.menu.roundedBottomLeft, 
         rounded_bottomright = kascreen->style.menu.roundedBottomRight;

    bool any_rounded = rounded_topleft || rounded_topright || 
                       rounded_bottomleft || rounded_bottomright;

    
    if (!any_rounded)
        return;
    
    Region corners_xregion = XCreateRegion();

    XRectangle xrect;
    if (rounded_topleft) {
            xrect.x = 0;
            xrect.y = 0;
            xrect.width = 5;
            xrect.height = 1;
      
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);

            xrect.y = 1;
            xrect.width = 3;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
      
            xrect.y = 2;
            xrect.width = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);

            xrect.y = 3;
            xrect.width = 1;
            xrect.height = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
        }

    if (rounded_topright) {            
            xrect.x = new_window_width - 5;
            xrect.y = 0;
            xrect.width = 5;
            xrect.height = 1;
      
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);

            xrect.y = 1;
            xrect.x = new_window_width - 3;
            xrect.width = 3;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
      
            xrect.y = 2;
            xrect.x = new_window_width - 2;
            xrect.width = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);

            xrect.y = 3;
            xrect.x = new_window_width - 1;
            xrect.width = 1;
            xrect.height = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
        }

    if (rounded_bottomleft) {
            xrect.x = 0;
            xrect.y = new_window_height - 1;
            xrect.width = 5;
            xrect.height = 1;
      
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
      
            xrect.y = new_window_height - 2;
            xrect.width = 3;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
      
            xrect.y = new_window_height - 3;
            xrect.width = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);

            xrect.y = new_window_height - 5;
            xrect.width = 1;
            xrect.height = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
        }

    if (rounded_bottomright) {
            xrect.x = new_window_width - 5;
            xrect.y = new_window_height - 1;
            xrect.width = 5;
            xrect.height = 1;
      
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
      
            xrect.y = new_window_height - 2;
            xrect.x = new_window_width - 3;
            xrect.width = 3;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
      
            xrect.y = new_window_height - 3;
            xrect.x = new_window_width - 2;
            xrect.width = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);

            xrect.y = new_window_height - 5;
            xrect.x = new_window_width - 1;
            xrect.width = 1;
            xrect.height = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
        }

    window_xregion = XCreateRegion();

    xrect.x = 0;
    xrect.y = 0;
    xrect.width = width;
    xrect.height = height;

    XUnionRectWithRegion (&xrect, window_xregion, window_xregion);
    XSubtractRegion (window_xregion, corners_xregion, window_xregion);
    XDestroyRegion (corners_xregion);
    XShapeCombineRegion(display, frame, ShapeBounding, 0, 0,
                            window_xregion, ShapeSet);
    XDestroyRegion(window_xregion);
}

/**
 * @fn    determineItemDimensions()
 * @brief Determine width/height of the items in the menu
 *
 * Private function (used by Build method).
 */
void KaMenu::determineItemDimensions() {
    MenuStyle &mstyle = kascreen->style.menu;

    unsigned int i;
    int lasttype = 0;
    list<KaMenuItem *>::iterator it = item_list.begin();

    iconsPresent = false;

    for (i = 1; it != item_list.end(); ++it, ++i) {
        KaMenuItem* item = *it;
        KaFont *kafont;
        if ((*it)->type == MenuTitleType)
            kafont = &mstyle.title.font;
        else
            kafont = &mstyle.frame.font;

        string l = item->e_label != "" ? item->e_label: item->label;
        item->width = kafont->Width(display, l) + 20;

        if (item->type == MenuCBItemType) {
            l = item->e_label2 != "" ? item->e_label2: item->label2;
            int cb_width = kafont->Width(display, l) + 20;
            if ((cb_width + 20) > item->width)
                item->width = cb_width + 20;
        }

        if (item->type == MenuTitleType) {
            height += t_height;
            item->height = item->realheight = t_height;
            height += mstyle.borderThickness * 2;
            item->realheight += mstyle.borderThickness * 2;

            if ((lasttype == MenuTitleType) || (i == 1)) {
                height -= mstyle.borderThickness;
                item->realheight -= mstyle.borderThickness;
            }
            if (i == item_list.size()) {
                height -= mstyle.borderThickness;
                item->realheight -= mstyle.borderThickness;
            }
        } else {
            height += f_height;
            item->height = item->realheight = f_height;
        }

        if (item->icon) {
            iconsPresent = true;

            float aspect     = 
                item->icon->get_width() / item->icon->get_height();
            int   iconHeight = item->height - 4; // Give some padding
            int   iconWidth =  int(aspect * iconHeight);
            item->width += iconWidth + 6;
        }

        if ((item->width + extra_width) > width)
            width = item->width + extra_width;

        lasttype = item->type;
    }

    // If there are icons, re-calculate max width as if there were an icon
    // on every item.
    if (iconsPresent) {
        for (it = item_list.begin(); it != item_list.end(); ++it) {
            if (!(*it)->icon) {
                width = max(width, (*it)->width + extra_width + 16 + 6);
            }
        }
    }
}

/** 
 * Clears the image rendering cache. Useful when root pixmap changes.
 * 
 */
void KaMenu::clearCache() {
    control->clearCache();
}


/**
 * @fn    Render(void)
 * @brief Render menu background
 *
 * Renders frame background and all item backgrounds.
 */
void KaMenu::Render() {   
    if ((x + width < 0) || (y + height < 0)) 
        return;

    const MenuStyle& mstyle = kascreen->style.menu;
    
    struct timeval tv, tv2;
    gettimeofday(&tv, NULL);

    
    if (pixmap != None && pixmap != ParentRelative) 
        imlib_free_pixmap_and_mask(pixmap);

    shared_ptr<ImImage> background = control->renderToImage
        (mstyle.frame.texture, 
         x + mstyle.borderThickness, 
         y + mstyle.borderThickness, 
         width + mstyle.borderThickness * 2, 
         height + mstyle.borderThickness * 2);

    gettimeofday(&tv2, NULL);

    double elapsed = 
        (tv2.tv_sec + tv2.tv_usec * 1e-6) - 
        (tv.tv_sec + tv.tv_usec * 1e-6);
    
    list<KaMenuItem *>::iterator it;
    for (it = item_list.begin(); it != item_list.end(); ++it) {
        (*it)->drawBackground(background);
    }

    pixmap = control->renderImage(frame, background);

    for (it = item_list.begin(); it != item_list.end(); ++it) {
        (*it)->drawForeground(pixmap);
    }

    XSetWindowBackgroundPixmap(display, frame, pixmap);
    XClearWindow(display, frame);
    for (it = item_list.begin(); it != item_list.end(); ++it) {
        XClearWindow(display, (*it)->id);
    }

    XSync(display, false);
    gettimeofday(&tv2, NULL);

    elapsed = 
        (tv2.tv_sec + tv2.tv_usec * 1e-6) - 
        (tv.tv_sec + tv.tv_usec * 1e-6);
    
    //    printf("Menu: Rendered in %.4f ms (%.2f fps)\n", 
    //           elapsed*1000.0, 1.0/elapsed);

    dirty = false;
}

/**
 * @fn    Map(int mx, int my)
 * @brief Maps menu
 *
 * Maps menu at specified position if not already mapped. If menu is already
 * mapped we do nothing.
 *
 * @param mx X coordinate to map menu at
 * @param my Y coordinate to map menu at
 */
void KaMenu::Map(int mx, int my) {
    if (ext_type && item_list.size() < 1) return;
    if (mapped) return;
    if (width == 0 || height == 0) return;

    DESKTOP_LOOP(kascreen) {
        if (kascreen->config.menu_stacking == AlwaysAtBottom)
            (*__dit)->aab_stacking_list.push_front(frame);
        else if (kascreen->config.menu_stacking == AlwaysOnTop)
            (*__dit)->aot_stacking_list.push_front(frame);
        else
            (*__dit)->stacking_list.push_front(frame);
    }
    kascreen->RaiseWindow(frame);
    x = mx;
    y = my;
    mapped = true;
    has_focus = false;
    XMoveWindow(display, frame, x, y);
    Render();
    XMapSubwindows(display, frame);
    XMapWindow(display, frame);
    XUngrabPointer(display, CurrentTime);
    Shape();
}

/**
 * @fn    ReMap(int mx, int my)
 * @brief Remaps menu
 *
 * Maps menu at specified position if not already mapped. If menu is already
 * mapped then we move it to the position we want to remap it to
 *
 * @param mx X coordinate to remap menu at
 * @param my Y coordinate to remap menu at
 */
void KaMenu::ReMap(int mx, int my) {
    if (ext_type && item_list.size() < 2) return;
    if(width == 0 || height == 0) return;

    if (mapped) Move(mx - x, my - y);
    else {
        DESKTOP_LOOP(kascreen) {
            if (kascreen->config.menu_stacking == AlwaysAtBottom)
                (*__dit)->aab_stacking_list.push_front(frame);
            else if (kascreen->config.menu_stacking == AlwaysOnTop)
                (*__dit)->aot_stacking_list.push_front(frame);
            else
                (*__dit)->stacking_list.push_front(frame);
        }
        kascreen->RaiseWindow(frame);
    }
    x = mx;
    y = my;
    mapped = true;
    has_focus = false;
    XMoveWindow(display, frame, x, y);
    Render();
    XMapSubwindows(display, frame);
    XMapWindow(display, frame);
    XUngrabPointer(display, CurrentTime);
    Shape();
}

/**
 * @fn    Move(int dx, int dy, bool render)
 * @brief Moves menu
 *
 * Moves menu and all linked submenus to a specified position.
 *
 * @param dx Move in x coordinate relative to old position
 * @param dy Move in y coordinate relative to old position
 * @param render True if menu menu should be redrawn
 */
void KaMenu::Move(int dx, int dy, bool render) {
    list<KaMenuItem *>::iterator it = item_list.begin();
    for (; it != item_list.end(); ++it) {
        if (((*it)->func_mask & MenuSubMask) && (*it)->submenu &&
            (*it)->submenu->root_menu &&
            (*it)->submenu->mapped) {
            if (! (*it)->submenu->ignore)
                (*it)->submenu->Move(dx, dy, render);
        }
    }
    x += dx;
    y += dy;
    XMoveWindow(display, frame, x, y);

    if (render) {
        render_if_transparency = true;
        Render();
        render_if_transparency = false;
    }
    Shape();
}

/**
 * @fn    Unmap(bool focus)
 * @brief Unmaps menu
 *
 * Unmaps menu and dehilites root menu item if menu was mapped as a submenu.
 *
 * @param focus True if we should set focus to root item
 */
void KaMenu::Unmap(bool focus)
{
  XEvent e;

  DESKTOP_LOOP(kascreen) {
      if (kascreen->config.menu_stacking == AlwaysOnTop)
        (*__dit)->aot_stacking_list.remove(frame);
      else if (kascreen->config.menu_stacking == AlwaysAtBottom)
       (*__dit)->aab_stacking_list.remove(frame);
      else
        (*__dit)->stacking_list.remove(frame);
  }
  XUnmapWindow(display, frame);

  root_menu = NULL;

  UnmapSubmenus(dynamic ? false : focus);

  list<KaMenuItem *>::iterator	it = item_list.begin();

  for (; it != item_list.end(); ++it) {
      if ((*it)->hilited) {
          if ((*it)->func_mask & MenuSubMask) {
              if ((!(*it)->submenu) || (!((*it)->submenu->root_menu &&
                                          (*it)->submenu->mapped)))
                  (*it)->DeHilite();
          }
          else
              (*it)->DeHilite();
      }
  }

  if (focus) {
    XSync(display, false);
    while (XCheckTypedEvent(display, EnterNotify, &e));
  }
  
  if (root_item) {
    if (focus)
      root_item->Focus();
    else
      root_item->DeHilite();
  } else {
    vector<KaWindow *>::iterator it = kascreen->kawindow_list.begin();
    for (; it != kascreen->kawindow_list.end(); ++it) {
        if ((*it)->has_focus)
            break;

        if (! (*it)->hidden && ! (*it)->flags.hidden) {
            (*it)->Focus(false);
            break;
        }
    }
  }

  if (dynamic_root) {
      kascreen->kamenu_list.erase(remove(kascreen->kamenu_list.begin(),
                                         kascreen->kamenu_list.end(),
                                         this), kascreen->kamenu_list.end());
      
      if (root_item)
          root_item->submenu = NULL;

      it = item_list.begin();
      for (; it != item_list.end(); ++it) {
          if ((*it)->submenu) {
              (*it)->submenu->root_item = NULL;
              if ((*it)->submenu->dynamic) {
                  (*it)->submenu->dynamic_root = true;
                  (*it)->submenu->UnmapSubmenus(false);
              }
          }
      }
      if (root_item) {
          if (focus)
              root_item->Focus();
          else
              root_item->DeHilite();
      }
      need_delete = true;

      delete this;
  } else {
    root_item = NULL;
    mapped = false;
  }
}

/**
 * @fn    UnmapSubmenus(bool focus)
 * @brief Unmaps submenus
 *
 * Recursive function for unmapping submenu trees. Unmaps all submenus in the
 * current subtree that are still linked.
 *
 * @param focus True if we should set focus to root item
 */
void KaMenu::UnmapSubmenus(bool focus)
{
  ignore = true;
  list<KaMenuItem *>::iterator it = item_list.begin();
  for (; it != item_list.end(); ++it) {
      if ((*it)->func_mask & MenuSubMask) {
          if (! (*it)->submenu)
              (*it)->DeHilite();
          else if ((*it)->submenu->mapped &&
                   (*it)->submenu->root_menu &&
                   !(*it)->submenu->ignore)
              (*it)->submenu->Unmap(focus);
      }
  }
  ignore = false;
}

/**
 * @fn    UnmapTree(void)
 * @brief Unmaps menu tree
 *
 * Recursive function for unmapping complete menu trees. Unmaps all menus that
 * are still linked in the current menu tree.
 */
void KaMenu::UnmapTree(void)
{
  KaMenu *tmp = NULL;

  if (root_menu)
  {
    tmp = root_menu;
    root_menu = NULL;
  }
  Unmap(false);
  if (tmp)
    tmp->UnmapTree();
}

/**
 * @fn    CreateOutline(void)
 * @brief Creates outline
 *
 * Creates four windows used for displaying an outline when doing
 * non opaque moving and resizing of the menu.
 */
void KaMenu::CreateOutline(void) {
    XSetWindowAttributes attrib_set;

    int create_mask = CWOverrideRedirect | CWBackPixel | CWEventMask |
        CWColormap;
    attrib_set.background_pixel = kascreen->style.window.outlineColor.getPixel();
    attrib_set.colormap = kascreen->colormap;
    attrib_set.override_redirect = true;
    attrib_set.event_mask = NoEventMask;

    o_west = XCreateWindow(display, kascreen->id, -10, -10, 1, 1, 0,
                           CopyFromParent, CopyFromParent,
                           kascreen->visual, create_mask, &attrib_set);
    o_east = XCreateWindow(display, kascreen->id, -10, -10, 1, 1, 0,
                           CopyFromParent, CopyFromParent,
                           kascreen->visual, create_mask, &attrib_set);
    o_north = XCreateWindow(display, kascreen->id, -10, -10, 1, 1, 0,
                            CopyFromParent, CopyFromParent,
                            kascreen->visual, create_mask, &attrib_set);
    o_south = XCreateWindow(display, kascreen->id, -10, -10, 1, 1, 0,
                            CopyFromParent, CopyFromParent,
                            kascreen->visual, create_mask, &attrib_set);
    DESKTOP_LOOP(kascreen) {
        (*__dit)->aot_stacking_list.push_front(o_west);
        (*__dit)->aot_stacking_list.push_front(o_east);
        (*__dit)->aot_stacking_list.push_front(o_north);
        (*__dit)->aot_stacking_list.push_front(o_south);
    }
    XMapWindow(display, o_west);
    XMapWindow(display, o_east);
    XMapWindow(display, o_north);
    XMapWindow(display, o_south);

    kascreen->RestackWindows(o_west);

    list<KaMenuItem *>::iterator it = item_list.begin();
    for (; it != item_list.end(); ++it) {
        if (((*it)->func_mask & MenuSubMask) && (*it)->submenu &&
            (*it)->submenu->root_menu && (*it)->submenu->mapped) {
            (*it)->submenu->CreateOutline();
        }
    }
}

/**
 * @fn    DestroyOutline(void)
 * @brief Destroys window outline
 *
 * Destorys the four outline windows.
 */
void KaMenu::DestroyOutline(void) {
    list<KaMenuItem *>::iterator it = item_list.begin();
    for (; it != item_list.end(); ++it) {
        if (((*it)->func_mask & MenuSubMask) && (*it)->submenu &&
            (*it)->submenu->root_menu && (*it)->submenu->mapped) {
            (*it)->submenu->DestroyOutline();
        }
    }
    DESKTOP_LOOP(kascreen) {
        (*__dit)->aot_stacking_list.remove(o_west);
        (*__dit)->aot_stacking_list.remove(o_east);
        (*__dit)->aot_stacking_list.remove(o_north);
        (*__dit)->aot_stacking_list.remove(o_south);
    }
    XDestroyWindow(display, o_west);
    XDestroyWindow(display, o_east);
    XDestroyWindow(display, o_north);
    XDestroyWindow(display, o_south);
}

/**
 * @fn    DrawOutline(int dx, int dy)
 * @brief Draws menu outline
 *
 * Recursive function for drawing a outline for a menu tree. Draws outline for
 * all menu in the current menu subtree that are still linked. This function
 * is used for none opaque moving.
 *
 * @param dx X position relative to menu position to draw outline
 * @param dy Y position relative to menu position to draw outline
 */
void KaMenu::DrawOutline(int dx, int dy) {
    list<KaMenuItem *>::iterator it = item_list.begin();

    for (; it != item_list.end(); ++it) {
        if (((*it)->func_mask & MenuSubMask) && (*it)->submenu &&
            (*it)->submenu->root_menu && (*it)->submenu->mapped) {
            (*it)->submenu->DrawOutline(dx, dy);
        }
    }

    int bw = kascreen->style.menu.borderThickness;

    bw = bw ? bw : 2;

    XResizeWindow(display, o_west, bw, bw * 2 + height);
    XResizeWindow(display, o_east, bw, bw * 2 + height);
    XResizeWindow(display, o_north, width + bw * 2, bw);
    XResizeWindow(display, o_south, width + bw * 2, bw);

    int xx = x + dx;
    int yy = y + dy;

    XMoveWindow(display, o_west, xx, yy);
    XMoveWindow(display, o_east, xx + width + bw, yy);
    XMoveWindow(display, o_north, xx, yy);
    XMoveWindow(display, o_south, xx, yy + bw + height);
}

/**
 * @fn    Raise(void)
 * @brief Raise menu window in display stack
 *
 * Raises the menu frame to the top of the stacking layer.
 */
void KaMenu::Raise(void) {
    kascreen->RaiseWindow(frame);
    list<KaMenuItem *>::iterator it = item_list.begin();
    for (; it != item_list.end(); ++it) {
        //        if (! (*it)->db) (*it)->Draw();
    }
}

/**
 * @fn    Lower(void)
 * @brief Lowers menu window in display stack
 *
 * Lowers the menu frame to the bottom of the stacking layer.
 */
void KaMenu::Lower(void) {
    kascreen->LowerWindow(frame);
}

/**
 * @fn    FocusItem(int n)
 * @brief Focus nth zero-based selectable item in menu
 *
 * Set input focus to nth zero-based element of type [item] or [sub] in menu.
 */
void KaMenu::FocusItem(int n) {
    XEvent e;

    XSync(display, false);
    list<KaMenuItem *>::iterator it = item_list.begin();

    for (; it != item_list.end(); ++it)
        while (XCheckTypedWindowEvent(display, (*it)->id, EnterNotify, &e));

    it = item_list.begin();

    int current_pos = -1;
    for (; it != item_list.end(); ++it) {
        if ((*it)->type != MenuTitleType && ++current_pos == n) {
            (*it)->Focus();
            break;
        }
    }

}

/**
 * @fn    FocusFirst(void)
 * @brief Focus first selectable item in menu
 *
 * Set input focus to first element of type [item] or [sub] in menu.
 */
void KaMenu::FocusFirst(void) {
    FocusItem(0);
}
