/* $Id: MenuItem.cc,v 1.9 2004/02/01 20:21:01 hyriand Exp $ */

/**
 * @file   MenuItem.cc
 * @date   11-Sep-2003
 *
 * @brief Implementation of KaMenuItem
 *
 */

#include "DesktopMenu.hh"
#include "Event.hh"
#include "ImageControl.hh"
#include "Imlib2pp.hh"
#include "Menu.hh"
#include "MenuItem.hh"
#include "Resources.hh"
#include "Texture.hh"
#include "Timer.hh"
#include "WindowMenu.hh"

extern "C" {
#ifdef    HAVE_STDIO_H
# include <stdio.h>
#endif // HAVE_STDIO_H

#ifdef    STDC_HEADERS
# include <string.h>
#endif // STDC_HEADERS
}

#include <algorithm>
#include <iostream>
#include <string>

using namespace std;

/**
 * @fn    KaMenuItem(char *s) : WindowObject(0, 0)
 * @brief Constructor for KaMenuItem class
 *
 * Creates a new menu item
 *
 * @param s Menu item label
 */
KaMenuItem::KaMenuItem(std::string s) : WindowObject(0, 0) {
    label = label1 = s;
    id = (Window) 0;
    func_mask = func_mask1 = func_mask2 = height = width = dy =
        realheight = cb = 0;
    wfunc = wfunc2 = NULL;
    rfunc = rfunc2 = NULL;
    mfunc = mfunc2 = NULL;
    wf = (Window) 0;
    submenu = submenu1 = submenu2 = NULL;
    move_resize = sdyn = sdyn1 = sdyn2 = db = false;

    xftdraw = (Drawable) 0;

    pixmap = None;
}

/**
 * @fn    ~KaMenuItem()
 * @brief Destructor for KaMenuItem class
 *
 * Deletes xftdrawable if compiled with xft support and destroys menu item
 * window.
 */
KaMenuItem::~KaMenuItem() {
    menu->item_list.remove(this);

    if (xftdraw) XftDrawDestroy(xftdraw);

    if (pixmap != None)
        XFreePixmap(menu->display, pixmap);

    if (id) {
        menu->kahakai->window_table.erase(id);
        XDestroyWindow(menu->display, id);
    }
}

/**
 * @fn    CreateWindow(KaScreen* screen, int y)
 * @brief Creates X window for the item
 *
 * Creates X window for the item under menu->frame using (0, y) as origin.
 */

void KaMenuItem::CreateWindow(KaScreen* screen, int y)
{
    MenuStyle& mstyle = screen->style.menu;
    XSetWindowAttributes attrib_set;

    //    attrib_set.background_pixmap = ParentRelative;
    attrib_set.border_pixel = mstyle.borderColor.getPixel();
    attrib_set.colormap = screen->colormap;
    attrib_set.override_redirect = true;
    attrib_set.event_mask = ButtonPressMask | ButtonReleaseMask |
        EnterWindowMask | LeaveWindowMask | KeyPressMask |
        KeyReleaseMask | ExposureMask | FocusChangeMask;

    int x, bw;
    x = bw = 0;
    if (type == MenuTitleType) {
        bw = mstyle.borderThickness;
        x = -bw;
    }

    id = XCreateWindow(screen->display, menu->frame, x, y, menu->width,
                       height, bw,
                       screen->screen_depth, CopyFromParent,
                       screen->visual, CWOverrideRedirect |
                       CWEventMask | CWColormap,
                       &attrib_set);

    menu->kahakai->window_table.insert(make_pair(id, this));
    dy = y;
    xftdraw = XftDrawCreate(screen->display, (Drawable) id,
                            screen->visual,
                            screen->colormap);
    
    pixmap = XCreatePixmap(screen->pdisplay, screen->id, menu->width,
                           height, screen->screen_depth);
    
    if (type == MenuTitleType) {
        actionlist = &screen->config.mtacts;
        texture = mstyle.title.texture;
    } else {
        switch (type) {
        case MenuItemType:
            actionlist = &screen->config.miacts; break;
        case MenuCBItemType:
            actionlist = &screen->config.mcbacts; break;
        case MenuSubType:
            actionlist = &screen->config.msacts; break;
        }
        texture = mstyle.frame.texture;
    }
}


/**
 * @fn    drawForeground(Drawable drawable)
 * @brief Draws menu item foreground
 *
 * Draws menu item text.
 *
 * @param drawable Drawable to draw on
 */
void KaMenuItem::drawForeground(Drawable drawable) {
    drawLabel(drawable, dy);
}


/** 
 * Draws highlight/title background on menu's background.
 * 
 * @param background ImImage to draw background on
 */
void KaMenuItem::drawBackground(shared_ptr<ImImage> background) {
    if (hilited || type == MenuTitleType) {
        shared_ptr<ImImage> img = menu->control->getTextureImage(texture,
                                                                 menu->width,
                                                                 height);
        background->blend(*img, true, 0, dy);
    }
}

void KaMenuItem::refresh() {
     XSetWindowBackgroundPixmap(menu->display, id, ParentRelative);
     XClearWindow(menu->display, id);
}

/** 
 * Draws label on a pixmap.
 * 
 * @param pixmap 
 * @param y Position to draw label
 */
void KaMenuItem::drawLabel(Pixmap pixmap, int y) {
    MenuStyle& mstyle = menu->kascreen->style.menu;
    
    KaFont *kafont = hilited ?
        &mstyle.frame.highlightFont : &mstyle.frame.font;

    KaFont *kafont_b = hilited ?
        &mstyle.bullet.highlightFont : &mstyle.bullet.font;    
    
    if (cb) UpdateCBox();

    if (type == MenuTitleType)
        kafont = &mstyle.title.font;

    const string& lbl = e_label != "" ? e_label : label;
    int justify;
    int x = 0;
    int org_y = y;

    width = kafont_b->Width(menu->display, lbl) + 20;

    if (type == MenuTitleType)
        justify = mstyle.title.justify;
    else
        justify = mstyle.frame.justify;

    if (menu->width <= width) 
        justify = LeftJustify;

    switch (justify) {
        case LeftJustify: x += 10; break;
        case CenterJustify:
            if (type == MenuTitleType)
                x += (menu->width / 2) - ((width - 10) / 2);
            else if (type == MenuCBItemType)
                x += ((menu->width - menu->cb_width) / 2) - ((width - 10) / 2);
            else x += ((menu->width - menu->extra_width) / 2) -
                     ((width - 10) / 2);
            break;
        default:
            if (type == MenuTitleType)
                x += menu->width - (width - 10);
            else if (type == MenuCBItemType)
                x += (menu->width - menu->cb_width) - (width - 10);
            else x += (menu->width - menu->extra_width) - (width - 10);
    }

    if (type == MenuTitleType)
        y += mstyle.title.yPosition;
    else
        y += mstyle.frame.yPosition;

    if (icon) {
        float aspect     = icon->get_width() / icon->get_height();
        int   iconHeight = 16;
        int   iconWidth =  int(aspect * iconHeight);
        
        icon->render_on_drawable(pixmap, x + 2, org_y + 2, 
                                 iconWidth, iconHeight);
        x += iconWidth + 6;
    } else if (menu->iconsPresent) {
        x += 16 + 6;
    }

    XftDrawChange(xftdraw, pixmap);

    kafont->Draw(menu->display, pixmap,
                 xftdraw,
                 x, y, lbl, lbl.length());

    if (type == MenuSubType) {
        y = org_y + mstyle.bullet.yPosition;
        kafont_b->Draw(menu->display, pixmap, 
                       xftdraw,
                       menu->width - (menu->bullet_width + 5), y,
                       mstyle.bullet.label, mstyle.bullet.label.length());
    }
    else if (type == MenuCBItemType) {
        if (kafont_cb == &mstyle.checkboxTrue.highlightFont)
            kafont_cb = &mstyle.checkboxTrue.font;
        else if (kafont_cb == &mstyle.checkboxFalse.highlightFont)
            kafont_cb = &mstyle.checkboxFalse.font;

        kafont_cb->Draw(menu->display, pixmap,
                        xftdraw,
                        menu->width - (cb_width + 5), org_y + cb_y, cbox,
                        cbox.length());
    }
}

/**
 * @fn    Hilite()
 * @brief Hilite menu item
 *
 * Sets menu item look to hilited look.
 */
void KaMenuItem::Hilite() {
    if (type == MenuTitleType || hilited) return;

    menu->dirty = true;

    list<KaMenuItem *>::iterator it = menu->item_list.begin();
    for (; it != menu->item_list.end(); ++it) {
        if ((*it)->hilited && menu->has_focus)
            if (!(((*it)->func_mask & MenuSubMask) && (*it)->submenu &&
                  (*it)->submenu->mapped))
                (*it)->DeHilite();
    }
    hilited = true;
    texture = menu->kascreen->style.menu.frame.highlightTexture;
    //    Render();
    menu->Render();
}

/**
 * @fn    DeHilite()
 * @brief DeHilite menu item
 *
 * Sets menu item look to dehilited look.
 */
void KaMenuItem::DeHilite() {
    if (type == MenuTitleType || !hilited) return;
    hilited = false;
    texture = menu->kascreen->style.menu.frame.texture;

    // if menu->dirty, then Hilite() is in process. Avoid double rendering.
    if (!menu->dirty) menu->Render();
}

/**
 * @fn    UnmapMenu(XEvent *, KaAction *, bool focus)
 * @brief Unmaps menu
 *
 * Unmaps the menu holding the menu item.
 *
 * @param bool True if we should focus root item
 */
void KaMenuItem::UnmapMenu(XEvent *, KaAction *, bool focus) {
    if (! in_window) return;
    if (menu->kahakai->eh->move_resize != EndMoveResizeType) return;
    menu->Unmap(focus);
}

/**
 * @fn    MapSubmenu(XEvent *, KaAction *, bool focus, bool only)
 * @brief Maps Submenu
 *
 * Maps menu items submenu, if there is one, at a good position close to the
 * menu item. If the submenu is already mapped then we do nothing.
 *
 * @param focus True if we should focus first item in submenu
 * @param only True if we should unmap all other submenus before mapping this
 */
void KaMenuItem::MapSubmenu(XEvent *, KaAction *, bool focus, bool only) {
    int skip, x, y, diff;

    if (! in_window) return;
    if (menu->kahakai->eh->move_resize != EndMoveResizeType) return;
    if (! (func_mask & MenuSubMask)) return;

    Hilite();
    if (only) {
        list<KaMenuItem *>::iterator it = menu->item_list.begin();
        for (; it != menu->item_list.end(); ++it) {
            if ((*it)->hilited && *it != this) {
                if (((*it)->func_mask & MenuSubMask) && (*it)->submenu &&
                    (*it)->submenu->mapped && (*it)->submenu->root_menu) {
                    (*it)->submenu->Unmap(false);
                    (*it)->DeHilite();
                }
            }
        }
    }
    if (sdyn && (! submenu)) {
        XSync(menu->display, false);
        if (! (submenu = menu->kascreen->GetMenuNamed(e_sub != ""? e_sub : sub)))
            return;
    }
    if (!submenu) return;
    if (submenu->mapped) return;

    switch (submenu->ext_type) {
    case TaskExtMenuType:
        ((WindowMenu *) submenu)->Build(menu->kascreen);
        break;
    case DesktopExtMenuType:
        ((DesktopMenu *) submenu)->Build(menu->kascreen, menu->wf);
        break;
    case MinimizedExtMenuType:
        ((MinimizedWindowMenu *) submenu)->Build(menu->kascreen);
        break;
    case NoExtMenuType:
        // pacify the warnings.
        break;
    }

    submenu->root_menu = menu;
    submenu->root_item = this;
    submenu->wf = menu->wf;
    submenu->rf = menu->rf;
    submenu->mf = menu->mf;
    submenu->ftype = menu->ftype;
    list<KaMenuItem *>::iterator it = submenu->item_list.begin();
    for (skip = 0; it != submenu->item_list.end(); ++it) {
        if ((*it)->type == MenuTitleType)
            skip += (*it)->realheight;
        else break;
    }

    int workx, worky, workw, workh;
    menu->kascreen->GetWorkareaSize(&workx, &worky, &workw, &workh);

    unsigned int border_w = menu->kascreen->style.menu.borderThickness;

    x = menu->x + menu->width + border_w;
    y = menu->y + dy - skip;
    diff = (y + submenu->height + border_w * 2) - (workh + worky);
    if (diff > 0) y -= diff;
    if (y < 0) y = 0;
    if ((x + submenu->width + border_w * 2) >
        (unsigned int) (workw + workx))
        x = menu->x - submenu->width - border_w;

    submenu->Map(x, y);
    if (focus) submenu->FocusFirst();
}

/**
 * @fn    RemapSubmenu(XEvent *, KaAction *, bool focus)
 * @brief Remaps Submenu
 *
 * Maps menu items submenu, if there is one, at a good position close to the
 * menu item. If the submenu is already mapped then we just move it to
 * the position we want to remap it to.
 *
 * @param focus True if we should focus first item in submenu
 */
void KaMenuItem::RemapSubmenu(XEvent *, KaAction *, bool focus) {
    int skip, x, y, diff;

    if (! in_window) return;
    if (! (func_mask & MenuSubMask)) return;
    if (submenu && (submenu == menu)) return;
    if (menu->kahakai->eh->move_resize != EndMoveResizeType) return;

    Hilite();
    if (sdyn) {
        XSync(menu->display, false);
        if (submenu) {
            hilited = false;
            submenu->Unmap(submenu->has_focus);
            hilited = true;
        }
        if (! (submenu = menu->kascreen->GetMenuNamed(e_sub != "" ? e_sub : sub)))
            return;
    }

    switch (submenu->ext_type) {
    case TaskExtMenuType:
        ((WindowMenu *) submenu)->Build(menu->kascreen);
        break;
    case DesktopExtMenuType:
        ((DesktopMenu *) submenu)->Build(menu->kascreen, menu->wf);
        break;
    case MinimizedExtMenuType:
        ((MinimizedWindowMenu *) submenu)->Build(menu->kascreen);
        break;
    case NoExtMenuType:
        // pacify the warnings.
        break;
    }

    submenu->root_menu = menu;
    submenu->root_item = this;
    submenu->wf = menu->wf;
    submenu->rf = menu->rf;
    submenu->mf = menu->mf;
    submenu->ftype = menu->ftype;

    unsigned int border_w = menu->kascreen->style.menu.borderThickness;

    list<KaMenuItem *>::iterator it = submenu->item_list.begin();
    for (skip = 0; it != submenu->item_list.end(); ++it) {
        if ((*it)->type == MenuTitleType)
            skip += (*it)->height + border_w;
        else break;
    }

    int workx, worky, workw, workh;
    menu->kascreen->GetWorkareaSize(&workx, &worky, &workw, &workh);

    x = menu->x + menu->width + border_w;
    y = menu->y + dy - skip;
    diff = (y + submenu->height + border_w * 2) -
        (workh + worky);
    if (diff > 0) y -= diff;
    if (y < 0) y = 0;
    if ((x + submenu->width + border_w * 2) >
        (unsigned int) (workw + workx))
        x = menu->x - submenu->width - border_w;
    menu->ignore = true;
    submenu->ReMap(x, y);
    menu->ignore = false;
    if (focus) submenu->FocusFirst();
}

/**
 * @fn    UnLinkMenu(XEvent *, KaAction *)
 * @brief Unlink menu
 *
 * Removes link to menu tree from menu. Menu will not be apart of the menu tree
 * mapping the menu any longer after this. Unmapsubmenu and Unmapmenutree
 * functions applied somewhere in the old menu tree will not unmap this menu.
 */
void KaMenuItem::UnLinkMenu(XEvent *, KaAction *) {
    if (! in_window) return;
    menu->root_menu = NULL;
}

/**
 * @fn    Exec(XEvent *, KaAction *)
 * @brief Execute program
 *
 * This function executes menu items command line, if there is one.
 */
void KaMenuItem::Exec(XEvent *, KaAction *) {
    if (cb) UpdateCBox();
    if (! in_window) return;
    if (! (func_mask & MenuExecMask)) return;

    kaexec(exec.c_str(), menu->kascreen->displaystring);
}

/**
 * @fn    Func(XEvent *e, KaAction *ac)
 * @brief Call function
 *
 * This function calls the function stored in menu items function pointer, if
 * there is a function pointer and the menu has been linked to an object of
 * same type as the function is a member of.
 *
 * @param e Event causing function call
 * @param ac KaAction object
 */
void KaMenuItem::Func(XEvent *e, KaAction *ac) {
    map<Window, WindowObject *>::iterator it;
    Window func_win;
    char* tmp_param;
    char* __m_wastrdup_tmp;

    if (! in_window) return;
    if (cb) UpdateCBox();
    if (param != "") {
        tmp_param = ac->param;
        ac->param = __m_wastrdup(param.c_str());
    }

    if (wf) func_win = wf;
    else func_win = menu->wf;
    if ((func_mask & MenuWFuncMask) &&
        ((menu->ftype == MenuWFuncMask) || wf)) {
        if ((it = menu->kahakai->window_table.find(func_win)) !=
            menu->kahakai->window_table.end()) {
            if (((*it).second)->type == WindowType) {
                (*wfunc)((KaWindow *) it->second, e, ac);
            }
        }
    }
    else if ((func_mask & MenuRFuncMask) && (menu->ftype == MenuRFuncMask))
        (*rfunc)(menu->rf, e, ac);
    else if ((func_mask & MenuMFuncMask) && (menu->ftype == MenuMFuncMask))
        (*mfunc)(menu->mf, e, ac);

    if (param != "") ac->param = tmp_param;
}

/**
 * @fn    Focus()
 * @brief Focus menu
 *
 * Sets input focus to the menu item window.
 */
void KaMenuItem::Focus() {
    XSetInputFocus(menu->display, id, RevertToPointerRoot, CurrentTime);
    menu->has_focus = true;
    Hilite();
}

/**
 * @fn    Move(XEvent *e, KaAction *)
 * @brief Moves the menu items menu
 *
 * Non-opaque moving of menu.
 *
 * @param e Event causing function call
 */
void KaMenuItem::Move(XEvent *e, KaAction *) {
    XEvent event;
    int px, py, i;
    list<XEvent *> *maprequest_list;
    bool started = false;
    int nx = menu->x;
    int ny = menu->y;
    Window w;
    unsigned int ui;

    if (menu->kahakai->eh->move_resize != EndMoveResizeType) return;
    menu->kahakai->eh->move_resize = MoveType;
    move_resize = true;

    XQueryPointer(menu->display, menu->kascreen->id, &w, &w, &px,
                  &py, &i, &i, &ui);

    maprequest_list = new list<XEvent *>;
    if (XGrabPointer(menu->display, id, true, ButtonReleaseMask |
                     ButtonPressMask | PointerMotionMask | EnterWindowMask |
                     LeaveWindowMask, GrabModeAsync, GrabModeAsync,
                     menu->kascreen->id, menu->kahakai->move_cursor,
                     CurrentTime) != GrabSuccess) {
        move_resize = false;
        menu->kahakai->eh->move_resize = EndMoveResizeType;
        return;
    }
    if (XGrabKeyboard(menu->display, id, true, GrabModeAsync, GrabModeAsync,
                      CurrentTime) != GrabSuccess) {
        move_resize = false;
        menu->kahakai->eh->move_resize = EndMoveResizeType;
        return;
    }
    for (;;) {
        menu->kahakai->eh->EventLoop(
            menu->kahakai->eh->menu_viewport_move_return_mask, &event);
        switch (event.type) {
            case MotionNotify:
                while (XCheckTypedWindowEvent(menu->display,
                                              event.xmotion.window,
                                              MotionNotify, &event));
                if (! started) {
                    menu->CreateOutline();
                    started = true;
                }
                nx += event.xmotion.x_root - px;
                ny += event.xmotion.y_root - py;
                px  = event.xmotion.x_root;
                py  = event.xmotion.y_root;
                menu->DrawOutline(nx - menu->x, ny - menu->y);
                break;
            case LeaveNotify:
            case EnterNotify:
                if (menu->kascreen->west->id == event.xcrossing.window ||
                    menu->kascreen->east->id == event.xcrossing.window ||
                    menu->kascreen->north->id == event.xcrossing.window ||
                    menu->kascreen->south->id == event.xcrossing.window) {
                    menu->kahakai->eh->HandleEvent(&event);
                } /*  [by hyriand] what's this for? this breaks stuff...
                    else if (event.type == EnterNotify &&
                           event.xany.window != id) {
                    int cx, cy;
                    XQueryPointer(menu->display, menu->kascreen->id,
                                  &w, &w, &cx,
                                  &cy, &i, &i, &ui);
                    nx += cx - px;
                    ny += cy - py;
                    px = cx;
                    py = cy;
                    menu->DrawOutline(nx - menu->x, ny - menu->y);
                } */
                break;
            case MapRequest:
                maprequest_list->push_front(&event); break;
            case ButtonPress:
            case ButtonRelease:
                event.xbutton.window = id;
            case KeyPress:
            case KeyRelease:
                if (event.type == KeyPress || event.type == KeyRelease)
                    event.xkey.window = id;
                menu->kahakai->eh->HandleEvent(&event);
                if (menu->kahakai->eh->move_resize != EndMoveResizeType) break;
                if (started) menu->DestroyOutline();
                menu->Move(nx - menu->x, ny - menu->y);
                while (! maprequest_list->empty()) {
                    XPutBackEvent(menu->display, maprequest_list->front());
                    maprequest_list->pop_front();
                }
                delete maprequest_list;
                move_resize = false;
                XUngrabKeyboard(menu->display, CurrentTime);
                XUngrabPointer(menu->display, CurrentTime);
                return;
        }
    }
}

/**
 * @fn    MoveOpaque(XEvent *e, KaAction *)
 * @brief Moves the menu items menu
 *
 * Opaque moving of menu.
 *
 * @param e Event causing function call
 */
void KaMenuItem::MoveOpaque(XEvent *e, KaAction *) {
    XEvent event, *map_ev;
    int px, py, i;
    list<XEvent *> *maprequest_list;
    int nx = menu->x;
    int ny = menu->y;
    Window w;
    unsigned int ui;

    if (menu->kahakai->eh->move_resize != EndMoveResizeType) return;
    menu->kahakai->eh->move_resize = MoveType;
    move_resize = true;

    XQueryPointer(menu->display, menu->kascreen->id, &w, &w, &px,
                  &py, &i, &i, &ui);

    maprequest_list = new list<XEvent *>;
    if (XGrabPointer(menu->display, id, true, ButtonReleaseMask |
                     ButtonPressMask | PointerMotionMask | EnterWindowMask |
                     LeaveWindowMask, GrabModeAsync, GrabModeAsync,
                     menu->kascreen->id, menu->kahakai->move_cursor,
                     CurrentTime) != GrabSuccess) {
        move_resize = false;
        menu->kahakai->eh->move_resize = EndMoveResizeType;
        return;
    }
    if (XGrabKeyboard(menu->display, id, true, GrabModeAsync, GrabModeAsync,
                      CurrentTime) != GrabSuccess) {
        move_resize = false;
        menu->kahakai->eh->move_resize = EndMoveResizeType;
        return;
    }
    for (;;) {
        menu->kahakai->eh->EventLoop(
            menu->kahakai->eh->menu_viewport_move_return_mask, &event);
        switch (event.type) {
            case MotionNotify:
                while (XCheckTypedWindowEvent(menu->display,
                                              event.xmotion.window,
                                              MotionNotify, &event));
                nx += event.xmotion.x_root - px;
                ny += event.xmotion.y_root - py;
                px = event.xmotion.x_root;
                py = event.xmotion.y_root;
                menu->Move(nx - menu->x, ny - menu->y,
                           !menu->kascreen->config.lazy_trans);
                break;
            case LeaveNotify:
            case EnterNotify:
                if (menu->kascreen->west->id == event.xcrossing.window ||
                    menu->kascreen->east->id == event.xcrossing.window ||
                    menu->kascreen->north->id == event.xcrossing.window ||
                    menu->kascreen->south->id == event.xcrossing.window) {
                    menu->kahakai->eh->HandleEvent(&event);
                } else if (event.type == EnterNotify &&
                           event.xany.window != id) {
                    int cx, cy;
                    XQueryPointer(menu->display, menu->kascreen->id,
                                  &w, &w, &cx,
                                  &cy, &i, &i, &ui);
                    nx += cx - px;
                    ny += cy - py;
                    px = cx;
                    py = cy;
                    menu->Move(nx - menu->x, ny - menu->y,
                               !menu->kascreen->config.lazy_trans);
                }
                break;
            case MapRequest:
                map_ev = new XEvent;
                *map_ev = event;
                maprequest_list->push_front(map_ev); break;
            case ButtonPress:
            case ButtonRelease:
                event.xbutton.window = id;
            case KeyPress:
            case KeyRelease:
                if (event.type == KeyPress || event.type == KeyRelease)
                    event.xkey.window = id;
                menu->kahakai->eh->HandleEvent(&event);
                if (menu->kahakai->eh->move_resize != EndMoveResizeType) break;
                menu->Move(0, 0, true);
                while (! maprequest_list->empty()) {
                    XPutBackEvent(menu->display, maprequest_list->front());
                    delete maprequest_list->front();
                    maprequest_list->pop_front();
                }
                delete maprequest_list;
                move_resize = false;
                XUngrabKeyboard(menu->display, CurrentTime);
                XUngrabPointer(menu->display, CurrentTime);
                return;
        }
    }
}

/**
 * @fn    EndMoveResize(XEvent *e, KaAction *)
 * @brief Ends move
 *
 * Ends menu moving process.
 */
void KaMenuItem::EndMoveResize(XEvent *, KaAction *) {
    menu->kahakai->eh->move_resize = EndMoveResizeType;
}

/**
 * @fn    NextItem(XEvent *e, KaAction *ac)
 * @brief Hilite and focus next item
 *
 * Hilites and sets input focus to menu item beneath this menu item.
 *
 * @param e X event that have occurred
 * @param ed Event details
 */
void KaMenuItem::NextItem(XEvent *e, KaAction *ac) {
    list<KaMenuItem *>::iterator it = menu->item_list.begin();
    for (; it != menu->item_list.end(); ++it) {
        if (*it == this) {
            for (++it; it != menu->item_list.end() &&
                     (*it)->type == MenuTitleType; ++it);
            if (it == menu->item_list.end()) {
                it = menu->item_list.begin();
                for (; *it != this && (*it)->type == MenuTitleType; ++it);
                if (*it != this) {
                    (*it)->Focus();
                    return;
                }
            } else {
                (*it)->Focus();
                return;
            }
        }
    }
}

/**
 * @fn    PreviousItem(XEvent *e, KaAction *ac)
 * @brief Hilite and focus previous item
 *
 * Hilites and sets input focus to menu item above this menu item.
 *
 * @param e X event that have occurred
 * @param ed Event details
 */
void KaMenuItem::PreviousItem(XEvent *e, KaAction *ac) {
    list<KaMenuItem *>::reverse_iterator it = menu->item_list.rbegin();
    for (; it != menu->item_list.rend(); ++it) {
        if (*it == this) {
            for (++it; it != menu->item_list.rend() &&
                     (*it)->type == MenuTitleType; ++it);
            if (it == menu->item_list.rend()) {
                it = menu->item_list.rbegin();
                for (; *it != this && (*it)->type == MenuTitleType; ++it);
                if (*it != this) {
                    (*it)->Focus();
                    return;
                }
            } else {
                (*it)->Focus();
                return;
            }
        }
    }
}

/**
 * @fn    EvAct(XEvent *e, EventDetail *ed, vector<KaAction *> *acts)
 * @brief Calls menu item function
 *
 * Tries to match an occurred X event with the actions in an action list.
 * If we have a match then we execute that action.
 *
 * @param e X event that have occurred
 * @param ed Event details
 * @param acts List with actions to match event with
 */
void KaMenuItem::EvAct(XEvent *e, EventDetail *ed, vector<KaAction *> *acts) {
    Window w;
    unsigned int ui;
    int xp, yp, i;
    in_window = true;

    if (e->type == ButtonPress || e->type == ButtonRelease ||
        e->type == DoubleClick) {
        XQueryPointer(menu->display, id, &w, &w, &i, &i, &xp, &yp, &ui);
        if (xp < 0 || yp < 0 || xp > menu->width || yp > height)
            in_window = false;
    }

    if (ed->type == EnterNotify) {
        if (XCheckTypedWindowEvent(menu->display, e->xany.window,
                                   LeaveNotify, e)) {
            XPutBackEvent(menu->display, e);
            return;
        }
        Hilite();
        if (menu->has_focus && type != MenuTitleType) Focus();
        XSync(menu->display, false);
    }

    if (menu->kahakai->eh->move_resize != EndMoveResizeType)
        ed->mod |= MoveResizeMask;

    vector<KaAction *>::iterator it = acts->begin();
    for (; it != acts->end(); ++it) {
        if (eventmatch(menu->display, *it, ed)) {
            if ((*it)->delay.tv_sec || (*it)->delay.tv_usec) {
                Interrupt *i = new Interrupt(*it, e, id);
                menu->kahakai->timer->AddInterrupt(i);
            } else {
                if ((*it)->exec)
                    kaexec((*it)->exec, menu->kascreen->displaystring);
                else
                    (*((*it)->menufunc))(this, e, *it);
            }
        }
    }

    if (menu != NULL && menu->need_delete)
        delete menu;

    if (ed->type == LeaveNotify) {
        if (func_mask & MenuSubMask) {
            if ((! submenu) || (! submenu->mapped))
                DeHilite();
        } else
            DeHilite();
    }
}

/**
 * @fn    UpdateCBox()
 * @brief Update Checkbox
 *
 * Reads checkbox value and updates pointers so that checkbox is drawn and
 * handled correct.
 */
void KaMenuItem::UpdateCBox() {
    map<Window, WindowObject *>::iterator it;
    Window func_win;
    bool true_false = false;
    KaWindow *ww;
    MenuStyle& mstyle = menu->kascreen->style.menu;

    if (cb) {
        if (wf) func_win = wf;
        else func_win = menu->wf;
        if ((func_mask & MenuWFuncMask) &&
            ((menu->ftype == MenuWFuncMask) || wf)) {
            if ((it = menu->kahakai->window_table.find(func_win)) !=
                menu->kahakai->window_table.end()) {
                if (((*it).second)->type == WindowType) {
                    ww = (KaWindow *) (*it).second;
                    switch (cb) {
                        case MaxCBoxType:
                            true_false = ww->flags.max; break;
                        case MinCBoxType:
                            true_false = ww->flags.hidden; break;
                        case ShadeCBoxType:
                            true_false = ww->flags.shaded; break;
                        case StickCBoxType:
                            true_false = ww->flags.sticky; break;
                        case TitleCBoxType:
                            true_false = ww->flags.title; break;
                        case HandleCBoxType:
                            true_false = ww->flags.handle; break;
                        case BorderCBoxType:
                            true_false = ww->flags.border; break;
                        case AllCBoxType:
                            true_false = ww->flags.all; break;
                        case AOTCBoxType:
                            true_false = ww->flags.alwaysontop; break;
                        case AABCBoxType:
                            true_false = ww->flags.alwaysatbottom; break;
                        case FsCBoxType:
                            true_false = ww->flags.fullscreen;
                    }
                    if (true_false) {
                        if (hilited)
                            kafont_cb = &mstyle.checkboxTrue.highlightFont;
                        else
                            kafont_cb = &mstyle.checkboxTrue.font;

                        cb_y = mstyle.checkboxTrue.yPosition;
                        if (cbox != mstyle.checkboxTrue.label)
                            menu->cb_db_upd = true;
                        cbox = mstyle.checkboxTrue.label;
                        label = label2;
                        sub = sub2;
                        wfunc = wfunc2;
                        rfunc = rfunc2;
                        mfunc = mfunc2;
                        func_mask = func_mask2;
                        cb_width = cb_width2;
                        param = param2;
                        sdyn = sdyn2;
                        e_label = e_label2;
                        e_sub = e_sub2;
                    }
                    else {
                        if (hilited)
                            kafont_cb = &mstyle.checkboxFalse.highlightFont;
                        else
                            kafont_cb = &mstyle.checkboxFalse.font;

                        cb_y = mstyle.checkboxFalse.yPosition;
                        if (cbox != mstyle.checkboxFalse.label)
                            menu->cb_db_upd = true;
                        cbox = mstyle.checkboxFalse.label;
                        label = label1;
                        sub = sub1;
                        wfunc = wfunc1;
                        rfunc = rfunc1;
                        mfunc = mfunc1;
                        func_mask = func_mask1;
                        cb_width = cb_width1;
                        param = param1;
                        sdyn = sdyn1;
                        e_label = e_label1;
                        e_sub = e_sub1;
                    }
                }
            }
        }
    }
}

/**
 * @fn    ExpandAll(KaWindow *w)
 * @brief Window info expansion for KaMenuItem
 *
 * Expands label and sub strings for KaMenuItem.
 *
 * @param w KaWindow to get expansion info from
 *
 * @return 1 if item label was expanded otherwise 0
 */
int KaMenuItem::ExpandAll(KaWindow *w) {
    e_label1 = expand(label1, w);
    e_label2 = expand(label2, w);
    e_sub1 = expand(sub1, w);
    e_sub2 = expand(sub2, w);

    e_label = e_label1;
    e_sub = e_sub1;

    if (e_label1 != "" || e_label2 != "") return 1;
    else return 0;
}

void KaMenuItem::UnmapSubmenus(XEvent *, KaAction *) {
    if (! in_window) return;
    menu->UnmapSubmenus(false);
}
void KaMenuItem::UnmapTree(XEvent *, KaAction *) {
    if (! in_window) return;
    menu->UnmapTree();
}
void KaMenuItem::Raise(XEvent *, KaAction *) {
    if (! in_window) return;
    menu->Raise();
}
void KaMenuItem::Lower(XEvent *, KaAction *) {
    if (! in_window) return;
    menu->Lower();
}
