/* $Id: MenuLoader.cc,v 1.30 2003/12/26 02:51:35 pervago Exp $ */

/**
 * @file   MenuLoader.cc
 * @date   16 Jun 2003
 *
 * @brief Implementation of MenuLoader class which handles loading menu files
 *
 */

#include "MenuLoader.hh"
#include "Menu.hh"
#include "MenuItem.hh"
#include "Window.hh"
#include "Resources.hh"
#include "Logger.hh"

extern "C" {
#ifdef    HAVE_STDIO_H
#  include <stdio.h>
#endif // HAVE_STDIO_H

#ifdef    STDC_HEADERS
#  include <string.h>
#endif // STDC_HEADERS
}

#ifdef    HAVE_UNISTD_H
#  include <sys/stat.h>
#  include <unistd.h>
#endif // HAVE_UNISTD_H

#include <algorithm>
#include <iostream>
#include <string>

using std::map;
using std::endl;
using std::string;

#define SCREEN_FUNC(name) screenFunctions[#name] = new ScreenEv##name();
#define WINDOW_FUNC(name) windowFunctions[#name] = new WindowEv##name();
#define MENU_FUNC(name)   menuFunctions[#name]   = new MenuEv##name();

#define SFUNC_ALIAS(alias, original) screenFunctions[#alias] = screenFunctions[#original]    
#define WINFUNC_ALIAS(alias, original) windowFunctions[#alias] = windowFunctions[#original]
#define MFUNC_ALIAS(alias, original) menuFunctions[#alias] = menuFunctions[#original]


std::map<const char*, ScreenFunction*, ltcase> screenFunctions;
std::map<const char*, WindowFunction*, ltcase> windowFunctions;
std::map<const char*, MenuFunction*,   ltcase> menuFunctions;

inline string basename(string s) {
    string::size_type slash_index = s.rfind('/');

    if (slash_index == string::npos)
        return s;
    else
        return s.substr(slash_index + 1);
}

MenuLoader::MenuLoader(string filename):
    menu_file(filename), linenr(0)
{
    if (screenFunctions.empty()) {
        SCREEN_FUNC(GoToDesktop);
        SCREEN_FUNC(NextDesktop);
        SCREEN_FUNC(PreviousDesktop);
        
        SCREEN_FUNC(MoveViewportLeft);
        SCREEN_FUNC(MoveViewportRight);
        SCREEN_FUNC(MoveViewportUp);

        SCREEN_FUNC(MenuMap);
        SCREEN_FUNC(MenuMapFocused);
        SCREEN_FUNC(MenuRemap);
        SCREEN_FUNC(MenuRemapFocused);
        SCREEN_FUNC(MenuUnmap);
        SCREEN_FUNC(MenuUnmapFocus);

        SCREEN_FUNC(ViewportMove);
        SCREEN_FUNC(ViewportRelativeMove);
        SCREEN_FUNC(ViewportFixedMove);

        SCREEN_FUNC(EndMoveResize);
        SCREEN_FUNC(Focus);
        SCREEN_FUNC(Restart);
        SCREEN_FUNC(Exit);
        SCREEN_FUNC(TaskSwitcher);
        SCREEN_FUNC(NextTask);
        SCREEN_FUNC(PreviousTask);
        SCREEN_FUNC(PointerFixedWarp);
        SCREEN_FUNC(PointerRelativeWarp);

        SCREEN_FUNC(Nop);
        
        SFUNC_ALIAS(ViewportDown,  MoveViewportDown);
        SFUNC_ALIAS(ViewportUp,    MoveViewportUp);
        SFUNC_ALIAS(ViewportLeft,  MoveViewportLeft);
        SFUNC_ALIAS(ViewportRight, MoveViewportRight);
        
        SFUNC_ALIAS(ViewportWrapDown,  MoveViewportWrapDown);
        SFUNC_ALIAS(ViewportWrapUp,    MoveViewportWrapUp);
        SFUNC_ALIAS(ViewportWrapLeft,  MoveViewportWrapLeft);
        SFUNC_ALIAS(ViewportWrapRight, MoveViewportWrapRight);

        SFUNC_ALIAS(StartViewportMove, ViewportMove);

        windowFunctions.insert(screenFunctions.begin(), screenFunctions.end());
        menuFunctions.insert(screenFunctions.begin(), screenFunctions.end());
        
        
        WINDOW_FUNC(Focus);
        WINDOW_FUNC(FocusVis);
        WINDOW_FUNC(Close);
        WINDOW_FUNC(Raise);
        WINDOW_FUNC(Lower);
        WINDOW_FUNC(MoveResize);
        WINDOW_FUNC(MoveResizeVirtual);
        WINDOW_FUNC(MoveWindowToPointer);
        WINDOW_FUNC(MoveWindowBy);
        WINDOW_FUNC(MoveWindowToSmartPlace);
        WINDOW_FUNC(MoveWindowToViewport);
        WINDOW_FUNC(MoveWindowToCurrentViewport);

        WINDOW_FUNC(DesktopMask);
        WINDOW_FUNC(JoinDesktop);
        WINDOW_FUNC(PartCurrentJoinDesktop);
        WINDOW_FUNC(PartCurrentDesktop);
        WINDOW_FUNC(JoinAllDesktops);
        WINDOW_FUNC(PartAllDesktopsExceptCurrent);
        WINDOW_FUNC(PartDesktop);

        WINDOW_FUNC(Move);
        WINDOW_FUNC(MoveOpaque);
        WINDOW_FUNC(RaiseFocus);
        WINDOW_FUNC(Shade);
        WINDOW_FUNC(UnShade);
        WINDOW_FUNC(ToggleShade);

        WINDOW_FUNC(MenuMap);
        WINDOW_FUNC(MenuMapFocused);
        WINDOW_FUNC(MenuRemap);
        WINDOW_FUNC(MenuRemapFocused);
        WINDOW_FUNC(MenuUnmap);
        WINDOW_FUNC(MenuUnmapFocus);

        WINDOW_FUNC(ResizeLeft);
        WINDOW_FUNC(ResizeLeftOpaque);
        WINDOW_FUNC(ResizeRight);
        WINDOW_FUNC(ResizeRightOpaque);
        WINDOW_FUNC(EndMoveResize);
        WINDOW_FUNC(Maximize);
        WINDOW_FUNC(UnMaximize);
        WINDOW_FUNC(ToggleMaximize);
        WINDOW_FUNC(Kill);
        WINDOW_FUNC(CloseKill);

        WINDOW_FUNC(Sticky);
        WINDOW_FUNC(UnSticky);
        WINDOW_FUNC(ToggleSticky);

        WINDOW_FUNC(Minimize);
        WINDOW_FUNC(UnMinimize);
        WINDOW_FUNC(ToggleMinimize);
        WINDOW_FUNC(ToggleMinimizeActive);

        WINDOW_FUNC(FullscreenOn);
        WINDOW_FUNC(FullscreenOff);
        WINDOW_FUNC(FullscreenToggle);

        WINDOW_FUNC(DecorTitleOn);
        WINDOW_FUNC(DecorHandleOn);
        WINDOW_FUNC(DecorBorderOn);
        WINDOW_FUNC(DecorAllOn);

        WINDOW_FUNC(DecorTitleOff);
        WINDOW_FUNC(DecorHandleOff);
        WINDOW_FUNC(DecorBorderOff);
        WINDOW_FUNC(DecorAllOff);

        WINDOW_FUNC(DecorTitleToggle);
        WINDOW_FUNC(DecorHandleToggle);
        WINDOW_FUNC(DecorBorderToggle);

        // Create some backward compatibility aliases
        WINFUNC_ALIAS(StartMove,              Move);
        WINFUNC_ALIAS(StartOpaqueMove,        MoveOpaque);
        WINFUNC_ALIAS(StartResizeRight,       ResizeRight);
        WINFUNC_ALIAS(StartResizeLeft,        ResizeLeft);
        WINFUNC_ALIAS(StartOpaqueResizeLeft,  ResizeLeftOpaque);
        WINFUNC_ALIAS(StartOpaqueResizeRight, ResizeRightOpaque);
        WINFUNC_ALIAS(MoveToSmartPlace,       MoveWindowToSmartPlace);

        WINDOW_FUNC(AlwaysontopOn);
        WINDOW_FUNC(AlwaysontopOff);
        WINDOW_FUNC(AlwaysontopToggle);

        WINDOW_FUNC(AlwaysatbottomOn);
        WINDOW_FUNC(AlwaysatbottomOff);
        WINDOW_FUNC(AlwaysatbottomToggle);

        WINDOW_FUNC(AcceptConfigRequestOn);
        WINDOW_FUNC(AcceptConfigRequestOff);
        WINDOW_FUNC(AcceptConfigRequestToggle);

        MENU_FUNC(UnLinkMenu);
        MENU_FUNC(MapSubmenu);
        MENU_FUNC(MapSubmenuOnly);
        MENU_FUNC(MapSubmenuFocused);
        MENU_FUNC(MapSubmenuFocusedOnly);
        MENU_FUNC(RemapSubmenu);
        MENU_FUNC(RemapSubmenuFocused);
        MENU_FUNC(UnmapMenu);
        MENU_FUNC(UnmapMenuFocus);
        MENU_FUNC(Exec);
        MENU_FUNC(Func);
        MENU_FUNC(Move);
        MENU_FUNC(MoveOpaque);
        MENU_FUNC(EndMoveResize);
        MENU_FUNC(Focus);
        MENU_FUNC(UnmapSubmenus);
        MENU_FUNC(UnmapTree);
        MENU_FUNC(Raise);
        MENU_FUNC(Lower);
        MENU_FUNC(NextItem);
        MENU_FUNC(PreviousItem);
        MENU_FUNC(Nop);

        MFUNC_ALIAS(MapSub,          MapSubmenu);
        MFUNC_ALIAS(MapSubOnly,      MapSubmenuOnly);
        MFUNC_ALIAS(MapSubFocused,   MapSubmenuFocused);
        MFUNC_ALIAS(UnmapSub,        UnmapSubmenu);
        MFUNC_ALIAS(RemapSub,        RemapSubmenu);
        MFUNC_ALIAS(RemapSubFocused, RemapSubmenuFocused);
        MFUNC_ALIAS(UnmapSubs,       UnmapSubmenus);
        MFUNC_ALIAS(Unlink,          UnLinkMenu);
        MFUNC_ALIAS(Unmap,           UnmapMenu);
        MFUNC_ALIAS(UnmapFocused,    UnmapMenuFocus);
        MFUNC_ALIAS(StartOpaqueMove, MoveOpaque);
        MFUNC_ALIAS(StartMove,       Move);
    }
}

MenuLoader::~MenuLoader()
{
}


inline string::size_type findFirstNonBackslashed(string str, char sep,
                                                 string::size_type start=0)
{
    if (start > str.length() - 1) return string::npos;

    for (unsigned int i=start;i<str.length();i++)
        if (str[i] == sep && (i == 0 || str[i-1] != '\\')) return i;

    return string::npos;
}

inline string unescape(string str) {
    for (unsigned int i=0; i < str.size(); i++) {
        if (str[i] == '\\') {
            str.replace(i, 1, "");
        }
    }
    return str;
}

inline string stringWithin(const string& str, char left, char right)
{
    string::size_type start = findFirstNonBackslashed(str, left);
    if (start == string::npos || start == str.length() - 1) return "";
    string::size_type end   = findFirstNonBackslashed(str, right, start+1);
    if (end == string::npos) return "";

    if (start > end) return "";

    return unescape(string(str.begin() + start + 1, str.begin() + end));
}

/**
 * @fn    environment_expansion(string s)
 * @brief Expand '~' and environment variables
 *
 * Replaces all '~' characters with HOME environment variable and all
 * $ENVIRONMENT_VARIABLE with environment variable value.
 *
 * @param s String to expand
 * @return Expanded string
 */

inline string expandEnvironment(string s) {
    string result = s;
    char*  homeDir = getenv("HOME");

    if (homeDir)
        result = replaceAll(result, "~", homeDir);

    string::size_type last = s.find('$');
    while (last != string::npos && last <= s.length() - 1) {
        string::size_type i = last + 1;

        while (i < s.length() && (isalnum(s[i]) || s[i] == '_'))
            i++;

        int begin = last; 
        int end   = i - 1;    // just before the other char or end of string

        string name = s.substr(begin + 1,  end - begin);
        char* value = getenv(name.c_str());

        if (value) 
            result = replaceAll(result,
                                s.substr(begin, end - begin + 1),
                                value);
        
        last = s.find('$', end);
    }
    return result;
}


/**
 * @fn    loadMenus(KaScreen *kascreen)
 * @brief Reads menu file
 *
 * Creates menus by parsing the menu file.
 *
 * @param kascreen KaScreen to load menus for
 */
void MenuLoader::loadMenus(KaScreen *kascreen) {
    FILE *file;

    if (! (file = fopen(kascreen->config.menu_file, "r"))) {
        logger << warn << "can't open menufile `" << kascreen->config.menu_file << 
            "' for reading" << endl;
        kascreen->config.menu_file_mtime = -1;
        return;
    }
    while (! feof(file)) parseMenu(NULL, file, kascreen);
    fclose(file);

    if (kascreen->config.menu_file_mtime != -1) {
        struct stat _s;
        kascreen->config.menu_file_mtime = -1;
        if (stat(kascreen->config.menu_file, &_s) == 0)
            kascreen->config.menu_file_mtime = _s.st_mtime;
    }
}

inline void MenuLoader::processInclude(string line,
                                       KaMenu* menu,
                                       KaScreen* kascreen)
{
    FILE *include_file;
    string tmp_mf;
    int tmp_linenr;

    string s = stringWithin(line, '(', ')');

    if (s != "") {
        s = environment_expansion(const_cast<char*>(s.c_str()));
        if (! (include_file = fopen(s.c_str(), "r"))) {
            logger << warn << "can't open menufile `" << s <<
                "' for reading" << endl;
            return;
        }
        tmp_mf = menu_file;
        tmp_linenr = linenr;
        menu_file = s;
        while (! feof(include_file))
            parseMenu(menu, include_file, kascreen);
        menu_file = tmp_mf;
        linenr    = tmp_linenr;
        fclose(include_file);
    } else {
        reportError("missing menufile name");
    }
}

inline void MenuLoader::processSubmenu(string line,
                                       FILE*  file,
                                       KaMenu* menu,
                                       KaScreen* kascreen)
{
    KaMenuItem* m;

    string s = stringWithin(line, '(', ')');

    if (s != "") {
        if (menu) {
            m = new KaMenuItem(s);
            m->type = MenuSubType;
            m->func_mask |= MenuSubMask;
            m->func_mask1 |= MenuSubMask;
            m->sub = m->sub1 = s;
            menu->AddItem(m);
        }
        KaMenu* tmp_menu;
        tmp_menu = new KaMenu(s);
        m = new KaMenuItem(s);
        m->type = MenuTitleType;
        tmp_menu->AddItem(m);
        if (menu) {
            if (menu->dynamic) {
                tmp_menu->dynamic = true;
                if (parseMenu(tmp_menu, file, kascreen))
                    tmp_menu->Build(kascreen);
            } else
                parseMenu(tmp_menu, file, kascreen);
        }
        else menu = tmp_menu;
    } else
        reportError("missing menu name [" + line + "]");
}

/**
 * @fn    reportError(string message)
 * @brief Reports an error in the menu file as a warning
 */

void MenuLoader::reportError(string message) {
    logger << warn << "(" << basename(menu_file) << ":" << linenr << "): " 
            << message << endl;
}



/**
 * @fn    param_eval(const char *action, char *param, KaScreen *kascreen)
 * @brief Evaluate parameter string
 *
 * Replaces special parameter characters and returns new parameter string.
 *
 * @param action Action name for parameter
 * @param param Parameter string
 * @param kascreen KaScreen object pointer
 *
 * @return New parameter
 */
string param_eval(const string& action, const string& param, 
                  KaScreen *kascreen) {
    string p = param;

    if (action == "viewport" || action == "moveresize") {
        p = replaceAll(p, "W", to_string(kascreen->width));
        p = replaceAll(p, "w", to_string(kascreen->width));
        p = replaceAll(p, "H", to_string(kascreen->height));
        p = replaceAll(p, "h", to_string(kascreen->height));
    }

    return p;
}
/** 
 * Looks up a window function by name either in the global window
 * functions table or the windowFunctions in Kahakai object (custom functions)
 * 
 * @param name 
 * 
 * @return 
 */
WindowFunction* MenuLoader::getWindowFunction(const char* name, Kahakai* ka) {
    if (windowFunctions.count(name))
        return windowFunctions[name];
    else
        return ka->getWindowFunction(name);
}

ScreenFunction* MenuLoader::getScreenFunction(const char* name, Kahakai* ka) {
    if (screenFunctions.count(name))
        return screenFunctions[name];
    else
        return ka->getScreenFunction(name);
}

MenuFunction* MenuLoader::getMenuFunction(const char* name, Kahakai* ka) {
    if (menuFunctions.count(name))
        return menuFunctions[name];
    else
        return ka->getMenuFunction(name);
}


/**
 * @fn    parseMenu(KaMenu *menu, FILE *file, KaScreen &kascreen)
 * @brief Parses a menu file
 *
 * Parses a menu section of the menu file and creates a menu object for the
 * menu. If a [start] or [begin] statement is found when parsing a menu, we
 * make a recursive function call to this function. This makes it possible to
 * to define a submenu within a the menu itself.
 *
 * @param menu Menu to add items to
 * @param file File descriptor for menu file
 *
 * @return Pointer to new menu if menu was sucessfully parsed,
 *         otherwise NULL
 */
KaMenu *MenuLoader::parseMenu(KaMenu *menu, FILE *file,
                              KaScreen* kascreen) {
    char line[8192];        
    string line1, line2;
    KaMenuItem *m;
    int i, type, cb;
    KaMenu *tmp_menu;
    string str;
    Kahakai* ka = kascreen->kahakai;

    while (fgets(line, 8192, file)) {
        linenr++;
        for (i = 0; line[i] == ' ' || line[i] == '\t'; i++);
        if (line[i] == '\n') continue;
        if (line[i] == '#') continue;
        if (line[i] == '!') continue;

        cb = 0;

        string str;
        
        if ((str = stringWithin(line, '[', ']')) == "") {
            reportError("missing tag");
            continue;
        }

        str = lower(str);

        if (str == "include") {
            processInclude(line, menu, kascreen);
            continue;
        }

        if (!(str == "start" || str == "begin")) {
            if (menu == NULL) {
                reportError("bad tag, expected [start], [begin] or [include]");
                continue;
            }
        }
        if (str == "start") {
            string label = expandEnvironment(stringWithin(line, '(', ')'));

            if (label != "") {
                tmp_menu = new KaMenu(label);
                if (menu) {
                    if (menu->dynamic) {
                        tmp_menu->dynamic = true;
                        if (parseMenu(tmp_menu, file, kascreen))
                            tmp_menu->Build(kascreen);
                    } else
                        parseMenu(tmp_menu, file, kascreen);
                }
                else menu = tmp_menu;
            } else
                reportError("missing menu name");
            continue;
        }
        else if (str == "submenu" || str == "begin") {
            processSubmenu(line, file, menu, kascreen);
            continue;
        }
        else if (str == "restart") {
            string label = expandEnvironment(stringWithin(line, '(', ')'));

            m = new KaMenuItem(label);

            string param;
            if ((param = stringWithin(line, '{', '}')) != "") {
                m->param1 = m->param = param;
            }
            
            m->type = MenuItemType;            
            m->func_mask = MenuRFuncMask | MenuWFuncMask | MenuMFuncMask;
            m->rfunc = new ScreenEvRestart();
            m->wfunc = new ScreenEvRestart();
            m->mfunc = new ScreenEvRestart();
            
            menu->AddItem(m);
            continue;
        }
        else if (str == "exit") {
            string label = expandEnvironment(stringWithin(line, '(', ')'));

            m = new KaMenuItem(label);
            m->type = MenuItemType;
            m->func_mask = MenuRFuncMask | MenuWFuncMask | MenuMFuncMask;
            m->rfunc = new ScreenEvExit(); 
            m->wfunc = new ScreenEvExit();
            m->mfunc = new ScreenEvExit();
            
            menu->AddItem(m);
            continue;
        }
        else if (str == "exec") {
            string label = expandEnvironment(stringWithin(line, '(', ')'));

            m = new KaMenuItem(label);
            m->type = MenuItemType;

            string param;
            if ((param = stringWithin(line, '{', '}')) != "") {                
                m->exec = m->exec1 = param;
                m->func_mask  |= MenuExecMask;
                m->func_mask1 |= MenuExecMask;
            }
            menu->AddItem(m);
            continue;
        }
        else if (str == "nop") {
            string label = expandEnvironment(stringWithin(line, '(', ')'));

            m = new KaMenuItem(label);
            m->type = MenuItemType;
            menu->AddItem(m);
            continue;
        }
        else if (str == "end") {
            if (menu->item_list.empty()) {
                reportError("no elements in menu '" + menu->name + "'");
                delete menu;
                return NULL;
            }
            kascreen->kamenu_list.push_back(menu);
            return menu;
        }
        else if (str.find("checkbox") == 0) {
            map<string, int, ltcase> checkboxes;
            
            checkboxes["maximized"]      = MaxCBoxType;
            checkboxes["minimized"]      = MinCBoxType;
            checkboxes["shaded"]         = ShadeCBoxType;
            checkboxes["sticky"]         = StickCBoxType;
            checkboxes["decortitle"]     = TitleCBoxType;
            checkboxes["decorhandle"]    = HandleCBoxType;
            checkboxes["decorborder"]    = BorderCBoxType;
            checkboxes["decorall"]       = AllCBoxType;
            checkboxes["alwaysontop"]    = AOTCBoxType;
            checkboxes["alwaysatbottom"] = AABCBoxType;
            checkboxes["fullscreen"]     = FsCBoxType;

            string name = string(str.begin() + 9, str.end());

            if (checkboxes.count(name) > 0) {
                type = MenuCBItemType;
                cb   = checkboxes[name];                
            } else {
                reportError("'" + name + "' - unknown checkbox");
                continue;
            }

            string lstr = string(line);

            string::size_type false_pos = lstr.find("@FALSE");
            if (false_pos == string::npos) {
                reportError("No '@FALSE' line part for checkbox item");
                continue;
            }

            string::size_type true_pos = lstr.find("@TRUE", false_pos);
            if (true_pos == string::npos) {
                reportError("No '@TRUE' line part for checkbox item " 
                            "(should come after @FALSE)");
                continue;                
            }
            

            int false_part_start = false_pos + strlen("@FALSE");
            int false_part_size  = (true_pos - 1) - false_part_start + 1;

            int true_part_start  = true_pos + strlen("@TRUE");


            line1 = lstr.substr(false_part_start, false_part_size);
            line2 = lstr.substr(true_part_start);
        }
        else if (str == "title") {
            type = MenuTitleType;
        }
        else if (str == "item") {
            type = MenuItemType;
        }
        else if (str == "sub") {
            type = MenuSubType;
        }
        else {
            reportError("bad tag [" + str + "]");
            continue;
        }
        if (! cb) line1 = line;

        string label = expandEnvironment(stringWithin(line1, '(', ')'));
        m = new KaMenuItem(label);
        m->label1 = m->label;
        m->type = type;
        m->cb = cb;

        if ((str = stringWithin(line1, '{', '}')) != "") {
            m->exec = m->exec1 = str;
            m->func_mask |= MenuExecMask;
            m->func_mask1 |= MenuExecMask;
        }
        else if ((str = stringWithin(line1, '<', '>')) != "") {
            m->sub = m->sub1 = str;
            m->func_mask |= MenuSubMask;
            m->func_mask1 |= MenuSubMask;
        }
        else if ((str = stringWithin(line1, '"', '"')) != "") {
            str = expandEnvironment(str);            

            size_t lparen_idx = str.find('(');
            
            if (lparen_idx != string::npos) {
                string param = stringWithin(str, '(', ')');
                if (param == "") {
                    reportError("missing ')'");
                    continue;
                }

                str = str.substr(0, lparen_idx);
                                           
                m->param1 = m->param = param_eval(str, param, kascreen);
            }

            if ((m->wfunc1 = m->wfunc = getWindowFunction(str.c_str(), ka))) {
                m->func_mask  |= MenuWFuncMask;
                m->func_mask1 |= MenuWFuncMask; 
            }

            if ((m->rfunc1 = m->rfunc = getScreenFunction(str.c_str(), ka))) {
                m->func_mask  |= MenuRFuncMask;
                m->func_mask1 |= MenuRFuncMask; 
            }

            if ((m->mfunc1 = m->mfunc = getMenuFunction(str.c_str(), ka))) {
                m->func_mask  |= MenuMFuncMask;
                m->func_mask1 |= MenuMFuncMask; 
            }

            if (! (m->wfunc || m->rfunc || m->mfunc)) {
                reportError("function '" + str + "' not available");
                continue;
            }
        }
        
        if (cb) {
            if ((str = stringWithin(line2, '(', ')')) != "")
                m->label2 = expandEnvironment(str);
            else
                m->label2 = "";
                        
            if ((str = stringWithin(line2, '{', '}')) != "") {
                m->exec2 = str;
                m->func_mask2 |= MenuExecMask;
            }            
            else if ((str = stringWithin(line2, '<', '>')) != "") {
                m->sub2 = str;
                m->func_mask2 |= MenuSubMask;
            }
            else if ((str = stringWithin(line2, '"', '"')) != "") {                
                size_t lparen_idx = str.find('(');
                
                if (lparen_idx != string::npos) {
                    string param = stringWithin(str, '(', ')');
                    if (param == "") {
                        reportError("missing ')'");
                        continue;
                    }
                    
                    str = str.substr(0, lparen_idx);
                    
                    m->param2 = param_eval(str, param, kascreen);
                }
                
                if ((m->wfunc2 = getWindowFunction(str.c_str(), ka))) {
                    m->func_mask2 |= MenuWFuncMask;
                }

                if ((m->rfunc2 = getScreenFunction(str.c_str(), ka))) {
                    m->rfunc2 = m->rfunc = screenFunctions[str.c_str()];
                    m->func_mask2 |= MenuRFuncMask;
                }

                if ((m->mfunc2 = getMenuFunction(str.c_str(), ka))) {
                    m->mfunc2 = m->mfunc = menuFunctions[str.c_str()];
                    m->func_mask2 |= MenuMFuncMask;
                }

                if (! (m->wfunc2 || m->rfunc2 || m->mfunc2)) {
                    reportError("function '" + str + "' not available");
                    continue;
                }
            }
        }
        menu->AddItem(m);
    }
    if (menu) {
        if (menu->item_list.empty()) {
            reportError("no elements in menu '" + menu->name + "'");
            delete menu;
            return NULL;
        }
        kascreen->kamenu_list.push_back(menu);
        return menu;
    }
    return NULL;
}
