/* $Id: Net.cc,v 1.33 2004/02/02 21:47:25 brailsmt Exp $ */

/**
 * @file   Net.cc
 * @date   11-Oct-2001 22:36:12
 *
 * @brief Implementation of NetHandler class
 *
 * Functions for reading and writing window hints.
 *
 */

#include "Net.hh"
#include "Kahakai.hh"
#include "Window.hh"
#include "Event.hh"
#include "ChildWindow.hh"

extern "C" {
#include <X11/Xatom.h>
}

#include <algorithm>
#include <iostream>
#include <vector>
#include <list>
#include <map>

using std::cout;
using std::endl;
using std::vector;
using std::list;
using std::map;
using std::max;
using std::min;

/**
 * @fn    NetHandler(Kahakai *ka)
 * @brief Constructor for NetHandler class
 *
 * Creates atom identifiers, allocates wm_hints and size_hints.
 *
 * @param ka Kahakai object
 */
NetHandler::NetHandler(Kahakai *ka) {
    kahakai = ka;
    display = kahakai->display;

    wm_hints = NULL;
    size_hints = XAllocSizeHints();
    utf8_string = XInternAtom(display, "UTF8_STRING", false);

    mwm_hints_atom =
        XInternAtom(display, "_MOTIF_WM_HINTS", false);
    
    wm_state = XInternAtom(display, "WM_STATE", false);
    wm_change_state = XInternAtom(display, "WM_CHANGE_STATE", false);

    net_supported = XInternAtom(display, "_NET_SUPPORTED", false);
    net_supported_wm_check =
        XInternAtom(display, "_NET_SUPPORTING_WM_CHECK", false);

    net_client_list = XInternAtom(display, "_NET_CLIENT_LIST", false);
    net_client_list_stacking =
        XInternAtom(display, "_NET_CLIENT_LIST_STACKING", false);
    net_active_window = XInternAtom(display, "_NET_ACTIVE_WINDOW", false);

    char *atomname = new char[45];
    for(int i = 0; i < 16; i++) {
        sprintf(atomname, "_KAHAKAI_NET_CLIENT_LIST_STACKING_DESKTOP_%d", i);
        kahakai_net_client_list_stacking_desktop[i] =
            XInternAtom(display, atomname, false);
    }
    delete [] atomname;
	atomname = NULL;

    net_desktop_viewport =
        XInternAtom(display, "_NET_DESKTOP_VIEWPORT", false);
    net_desktop_geometry =
        XInternAtom(display, "_NET_DESKTOP_GEOMETRY", false);
    net_current_desktop = XInternAtom(display, "_NET_CURRENT_DESKTOP", false);
    net_number_of_desktops =
        XInternAtom(display, "_NET_NUMBER_OF_DESKTOPS", false);
    net_desktop_names = XInternAtom(display, "_NET_DESKTOP_NAMES", false);
    net_workarea = XInternAtom(display, "_NET_WORKAREA", false);
    
    net_wm_desktop = XInternAtom(display, "_NET_WM_DESKTOP", false);
    net_wm_name = XInternAtom(display, "_NET_WM_NAME", false);
    net_wm_visible_name = XInternAtom(display, "_NET_WM_VISIBLE_NAME", false);
    net_wm_strut = XInternAtom(display, "_NET_WM_STRUT", false);
    net_wm_pid = XInternAtom(display, "_NET_WM_PID", false);
    
    net_wm_state = XInternAtom(display, "_NET_WM_STATE", false);
    net_wm_state_sticky = XInternAtom(display, "_NET_WM_STATE_STICKY", false);
    net_wm_state_shaded = XInternAtom(display, "_NET_WM_STATE_SHADED", false);
    net_wm_state_hidden = XInternAtom(display, "_NET_WM_STATE_HIDDEN", false);
    net_wm_maximized_vert =
        XInternAtom(display, "_NET_WM_STATE_MAXIMIZED_VERT", false);
    net_wm_maximized_horz =
        XInternAtom(display, "_NET_WM_STATE_MAXIMIZED_HORZ", false);
    net_wm_state_above = XInternAtom(display, "_NET_WM_STATE_ABOVE", false);
    net_wm_state_below = XInternAtom(display, "_NET_WM_STATE_BELOW", false);
    net_wm_state_stays_on_top =
        XInternAtom(display, "_NET_WM_STATE_STAYS_ON_TOP", false);
    net_wm_state_stays_at_bottom =
        XInternAtom(display, "_NET_WM_STATE_STAYS_AT_BOTTOM", false);
    net_wm_state_fullscreen =
        XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", false);
    net_wm_state_skip_taskbar =
        XInternAtom(display, "_NET_WM_STATE_SKIP_TASKBAR", false);

    net_wm_allowed_actions =
        XInternAtom(display, "_NET_WM_ALLOWED_ACTIONS", false);
    net_wm_action_move = XInternAtom(display, "_NET_WM_ACTION_MOVE", false);
    net_wm_action_resize =
        XInternAtom(display, "_NET_WM_ACTION_RESIZE", false);
    net_wm_action_minimize =
        XInternAtom(display, "_NET_WM_ACTION_MINIMIZE", false);
    net_wm_action_shade = XInternAtom(display, "_NET_WM_ACTION_SHADE", false);
    net_wm_action_stick = XInternAtom(display, "_NET_WM_ACTION_STICK", false);
    net_wm_action_maximize_horz =
        XInternAtom(display, "_NET_WM_ACTION_MAXIMIZE_HORZ", false);
    net_wm_action_maximize_vert =
        XInternAtom(display, "_NET_WM_ACTION_MAXIMIZE_VERT", false);
    net_wm_action_fullscreen =
        XInternAtom(display, "_NET_WM_ACTION_FULLSCREEN", false);
    net_wm_action_change_desktop =
        XInternAtom(display, "_NET_WM_ACTION_CHANGE_DESKTOP", false);
    net_wm_action_close =
        XInternAtom(display, "_NET_WM_ACTION_CLOSE", false);

    net_wm_window_type = XInternAtom(display, "_NET_WM_WINDOW_TYPE", false);
    net_wm_window_type_desktop =
        XInternAtom(display, "_NET_WM_WINDOW_TYPE_DESKTOP", false);
    net_wm_window_type_dock =
        XInternAtom(display, "_NET_WM_WINDOW_TYPE_DOCK", false);
    net_wm_window_type_toolbar =
        XInternAtom(display, "_NET_WM_WINDOW_TYPE_TOOLBAR", false);
    net_wm_window_type_menu =
        XInternAtom(display, "_NET_WM_WINDOW_TYPE_MENU", false);
    net_wm_window_type_splash =
        XInternAtom(display, "_NET_WM_WINDOW_TYPE_SPLASH", false);
    net_wm_window_type_normal =
        XInternAtom(display, "_NET_WM_WINDOW_TYPE_NORMAL", false);
    net_wm_window_type_dialog =
        XInternAtom(display, "_NET_WM_WINDOW_TYPE_DIALOG", false);
    net_wm_window_type_utility =
        XInternAtom(display, "_NET_WM_WINDOW_TYPE_UTILITY", false);

    net_close_window = XInternAtom(display, "_NET_CLOSE_WINDOW", false);
    net_moveresize_window =
        XInternAtom(display, "_NET_MOVERESIZE_WINDOW", false);
    net_wm_moveresize = XInternAtom(display, "_NET_WM_MOVERESIZE", false);

    net_startup_info_begin = XInternAtom(display, "_NET_STARTUP_INFO_BEGIN", false);
    net_startup_info       = XInternAtom(display, "_NET_STARTUP_INFO", false);

    net_wm_icon = XInternAtom(display, "_NET_WM_ICON", false);

    kahakai_net_wm_state_decor =
        XInternAtom(display, "_WAIMEA_NET_WM_STATE_DECOR", false);
    kahakai_net_wm_state_decortitle =
        XInternAtom(display, "_WAIMEA_NET_WM_STATE_DECOR_TITLE", false);
    kahakai_net_wm_state_decorhandle =
        XInternAtom(display, "_WAIMEA_NET_WM_STATE_DECOR_HANDLE", false);
    kahakai_net_wm_state_decorborder =
        XInternAtom(display, "_WAIMEA_NET_WM_STATE_DECOR_BORDER", false);

    kahakai_net_maximized_restore =
        XInternAtom(display, "_WAIMEA_NET_MAXIMIZED_RESTORE", false);
    kahakai_net_virtual_pos =
        XInternAtom(display, "_WAIMEA_NET_VIRTUAL_POS", false);
    kahakai_net_wm_desktop_mask =
        XInternAtom(display, "_WAIMEA_NET_WM_DESKTOP_MASK", false);

    kahakai_net_restart = XInternAtom(display, "_WAIMEA_NET_RESTART", false);
    kahakai_net_shutdown = XInternAtom(display, "_WAIMEA_NET_SHUTDOWN", false);

    kahakai_net_wm_maximize_restore_handlebar = 
    XInternAtom(display, "_KAHAKAI_NET_WM_MAXIMIZE_RESTORE_HANDLEBAR", false);

    xdndaware = XInternAtom(display, "XdndAware", false);
    xdndenter = XInternAtom(display, "XdndEnter", false);
    xdndleave = XInternAtom(display, "XdndLeave", false);

    kde_net_system_tray_windows =
        XInternAtom(display, "_KDE_NET_SYSTEM_TRAY_WINDOWS", false);
    kde_net_wm_system_tray_window_for =
        XInternAtom(display, "_KDE_NET_WM_SYSTEM_TRAY_WINDOW_FOR", false);

    xrootpmap_id =
        XInternAtom(display, "_XROOTPMAP_ID", false);

    event.type = ClientMessage;
    event.xclient.display = display;
    event.xclient.format = 32;
    event.xclient.data.l[0] = event.xclient.data.l[1] =
        event.xclient.data.l[2] = event.xclient.data.l[3] =
        event.xclient.data.l[4] = 0l;
}

/**
 * @fn    ~NetHandler(void)
 * @brief Destructor for NetHandler class
 *
 * Frees wm_hints and size_hints.
 */
NetHandler::~NetHandler(void) {
    XFree(wm_hints);
    XFree(size_hints);
}

/**
 * @fn    GetWMHints(KaWindow *ww)
 * @brief Read WM hints
 *
 * Reads KaWindows WM hints.
 *
 * @param ww KaWindow object
 */
void NetHandler::GetWMHints(KaWindow *ww) {
    XTextProperty text_prop;
    char **list;
    int num;
    char *__m_wastrdup_tmp;
 
    ww->state = NormalState;
    XGrabServer(display);

    if (wm_hints) XFree(wm_hints);

    if (validatedrawable(ww->id)) {
        if ((wm_hints = XGetWMHints(display, ww->id))) {
            if (wm_hints->flags & StateHint)
                ww->state = wm_hints->initial_state;
        }
        if (ww->classhint) {
            XFree(ww->classhint->res_name);
            XFree(ww->classhint->res_class);
        }

        ww->classhint = XAllocClassHint();
        XGetClassHint(ww->display, ww->id, ww->classhint);
        if (XGetWMClientMachine(ww->display, ww->id, &text_prop)) {
            if (XTextPropertyToStringList(&text_prop, &list, &num)) {
                XFree(text_prop.value);
                ww->host = __m_wastrdup(*list);
                XFreeStringList(list);
            }
        }
    } else ww->deleted = true;
    XUngrabServer(display);
}

/**
 * @fn    GetMWMHints(KaWindow *ww)
 * @brief Read MWM Hints
 *
 * Reads KaWindows MWM hints.
 *
 * @param ww KaWindow object
 */
void NetHandler::GetMWMHints(KaWindow *ww) {
    Window trans;
    int status;
    ww->flags.title = ww->flags.border = ww->flags.handle = true;
    
    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        status = XGetWindowProperty(display, ww->id, mwm_hints_atom, 0L, 20L,
                                    false, mwm_hints_atom, &real_type,
                                    &real_format, &items_read, &items_left,
                                    (unsigned char **) &mwm_hints);
    } else WW_DELETED;
    XUngrabServer(display);
    
    if (status == Success && items_read >= PropMotifWmHintsElements) {
        if (mwm_hints->flags & MwmHintsDecorations
            && !(mwm_hints->decorations & MwmDecorAll)) {
            ww->flags.title  =
                (mwm_hints->decorations & MwmDecorTitle)  ? true: false;
            ww->flags.border =
                (mwm_hints->decorations & MwmDecorBorder) ? true: false;
            ww->flags.handle =
                (mwm_hints->decorations & MwmDecorHandle) ? true: false;
        }
    }
    if (ww->kascreen->config.transient_above) {
        XGrabServer(display);
        if (validatedrawable(ww->id)) {
            status = XGetTransientForHint(display, ww->id, &trans);
        } else WW_DELETED;
        XUngrabServer(display);
        if (status && trans && (trans != ww->id)) {
            if (trans == ww->kascreen->id) {
                vector<KaWindow *>::iterator it =
                    ww->kascreen->kawindow_list.begin();
                for (;it != ww->kascreen->kawindow_list.end(); ++it)
                    (*it)->transients.push_back(ww->id);
                ww->want_focus = true;
            } else {
                map<Window, WindowObject *>::iterator it;
                if ((it = kahakai->window_table.find(trans)) !=
                    kahakai->window_table.end()) {
                    if (((*it).second)->type == WindowType) {
                        ww->transient_for = trans;
                        ((KaWindow *)
                         (*it).second)->transients.push_back(ww->id);
                        if (kahakai->eh && trans == kahakai->eh->focused)
                            ww->want_focus = true;
                    }
                }
            }
        }
    }
    ww->flags.all = ww->flags.title && ww->flags.handle && ww->flags.border;
}

/**
 * @fn    GetWMNormalHints(KaWindow *ww)
 * @brief Read size hints
 *
 * Reads KaWindows size hints.
 *
 * @param ww KaWindow object
 */
void NetHandler::GetWMNormalHints(KaWindow *ww) {
    long dummy;
    int status;

    ww->size.max_width = ww->size.max_height = 65536;
    ww->size.min_width = ww->size.min_height = ww->size.width_inc =
        ww->size.height_inc = 1;
    ww->size.base_width = ww->size.min_width;
    ww->size.base_height = ww->size.min_height;

    size_hints->flags = 0;
    XGrabServer(display);
    if (validatedrawable(ww->id))
        status = XGetWMNormalHints(display, ww->id, size_hints, &dummy);
    else WW_DELETED;
    XUngrabServer(display);

    if (size_hints->flags & PPosition || size_hints->flags & USPosition)
        ww->positioned_by_app = true;

    if (status) {
        if (size_hints->flags & PMaxSize) {
            ww->size.max_width = size_hints->max_width;
            ww->size.max_height = size_hints->max_height;
        }
        if (size_hints->flags & PMinSize) {
            ww->size.min_width = size_hints->min_width;
            ww->size.min_height = size_hints->min_height;
        }
        if (size_hints->flags & PResizeInc) {
            ww->size.width_inc = size_hints->width_inc;
            ww->size.height_inc = size_hints->height_inc;
        }
        if (size_hints->flags & PBaseSize) {
            ww->size.base_width = size_hints->base_width;
            ww->size.base_height = size_hints->base_height;
        }
        if (size_hints->flags & PWinGravity)
            ww->size.win_gravity = size_hints->win_gravity;

        if (ww->size.width_inc == 0) {
            ww->size.base_width = 0;
            ww->size.width_inc = 1;
        }
        if (ww->size.height_inc == 0) {
            ww->size.base_height = 0;
            ww->size.height_inc = 1;
        }
    }

    ww->size.min_width = max(3, ww->size.min_width);
    ww->size.min_height = max(3, ww->size.min_height);
    ww->size.base_width = max(3, ww->size.base_width);
    ww->size.base_height = max(3, ww->size.base_height);
}

bool NetHandler::DetermineMappable(Window * win) {

    // this stuff confuses me a lot, i tried to make sense of the stuff in
    // metacity and adapt it to here, but the metacity code is very different
    // then i'm used to, and i had a hard time wrapping my brain around it,
    // plus the fact that i just really don't understand X stuff that much yet.
    //
    // and it turns out i don't even think this function serves any useful
    // purpose :)

    CARD32 * state;

    XGrabServer(display);

    // argh, i hate trying to format stuff like this.  so we'll just do one
    // thing at a time..

    bool mappable;

    if (validatedrawable(*win)) {
        mappable = (bool) (XGetWindowProperty(display, *win, wm_state, 0L,
                           1L, false, wm_state, &real_type, &real_format,
                           &items_read, &items_left,
                           (unsigned char **) &state) == Success);
    } else {
        if (state)
            XFree(state);
        return false;
    }

    XUngrabServer(display);

    mappable = (mappable && (state));

    mappable = (mappable && real_format == 32 && wm_state == real_type &&
                                                             items_read>0);
    mappable = (mappable && (*state == IconicState || *state == NormalState));

    if (state)
        XFree(state);

    return mappable;
}

/**
 * @fn    GetState(KaWindow *ww)
 * @brief Read state hint
 *
 * Reads KaWindows state hint.
 *
 * @param ww KaWindow object
 */
void NetHandler::GetState(KaWindow *ww) {
    CARD32 *data;

    ww->state = WithdrawnState;
    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        if (XGetWindowProperty(display, ww->id, wm_state, 0L, 1L, false,
                               wm_state, &real_type, &real_format, &items_read,
                               &items_left, (unsigned char **) &data) ==
            Success && items_read) {
            ww->state = *data;
            XFree(data);
        }
    } else ww->deleted = true;
    XUngrabServer(display);
}

/**
 * @fn    SetState(KaWindow *ww, int newstate)
 * @brief Write state hint.
 *
 * Changes the state of the window and writes state hint.
 *
 * @param ww KaWindow object
 * @param newstate State to change to
 */
void NetHandler::SetState(KaWindow *ww, int newstate) {
    CARD32 data[2];

    ww->state = newstate;
    switch (ww->state) {
        case IconicState:
            ww->flags.hidden = true;
            ww->Hide();
            SetWmState(ww);
            break;
        case NormalState:
            ww->flags.hidden = false;
            if (! ww->mapped) ww->MapWindow();
            else if (ww->desktop_mask &
                     (1L << ww->kascreen->current_desktop->number))
                ww->Show();
    }
    if (ww->want_focus && ww->mapped && !ww->hidden) {
        XGrabServer(display);
        if (validatedrawable(ww->id))
            XSetInputFocus(display, ww->id, RevertToPointerRoot, CurrentTime);
        else WW_DELETED;
        XUngrabServer(display);
    }

    ww->want_focus = false;

    data[0] = ww->state;
    data[1] = None;

    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        XChangeProperty(display, ww->id, wm_state, wm_state,
                        32, PropModeReplace, (unsigned char *) data, 2);
    } else WW_DELETED;
    XUngrabServer(display);
    ww->SendConfig();
}

/**
 * @fn    GetWmState(KaWindow *ww)
 * @brief Reads net state hint
 *
 * Reads _NET_WM_STATE hint to get the current state of the window.
 *
 * @param ww KaWindow object
 */
void NetHandler::GetWmState(KaWindow *ww) {
    CARD32 *data;
    bool vert = false, horz = false, shaded = false, title = false,
        handle = false, border = false, decor = false, restore_handlebar=false;
    unsigned int i;
    int status;

    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        status = XGetWindowProperty(display, ww->id, net_wm_state, 0L, 10L,
                                    false, XA_ATOM, &real_type,
                                    &real_format, &items_read, &items_left, 
                                    (unsigned char **) &data);
    } else WW_DELETED;
    XUngrabServer(display);
        
    if (status == Success && items_read) {
        for (i = 0; i < items_read; i++) {
            if (data[i] == net_wm_state_sticky) ww->flags.sticky = true;
            else if (data[i] == net_wm_state_shaded) shaded = true;
            else if (data[i] == net_wm_maximized_vert) vert = true;
            else if (data[i] == net_wm_maximized_horz) horz = true;
            else if (data[i] == net_wm_state_hidden) ww->flags.hidden = true;
            else if (data[i] == net_wm_state_skip_taskbar)
                ww->flags.tasklist = false;
            else if (data[i] == net_wm_state_above ||
                     data[i] == net_wm_state_stays_on_top)
                ww->AlwaysontopOn();
            else if (data[i] == net_wm_state_below ||
                     data[i] == net_wm_state_stays_at_bottom)
                ww->AlwaysatbottomOn();
            else if (data[i] == net_wm_state_fullscreen)
                ww->flags.fullscreen = true;
            else if (data[i] == kahakai_net_wm_state_decor) decor = true;
            else if (data[i] == kahakai_net_wm_state_decortitle) title = true;
            else if (data[i] == kahakai_net_wm_state_decorhandle) handle = true;
            else if (data[i] == kahakai_net_wm_state_decorborder) border = true;
            else if (data[i] == kahakai_net_wm_maximize_restore_handlebar)
                restore_handlebar = true;
        }
    }
    if (decor) {
        ww->flags.title = ww->flags.handle = ww->flags.border = false;
        if (title) ww->flags.title = true;
        if (handle) ww->flags.handle = true;
        if (border) ww->flags.border = true;
        ww->flags.all = ww->flags.title && ww->flags.handle &&
            ww->flags.border;
    }
    XFree(data);
    
    if (vert || horz) {
        XGrabServer(display);
        if (validatedrawable(ww->id)) {
            status = XGetWindowProperty(display, ww->id,
                                        kahakai_net_maximized_restore,
                                        0L, 6L, false, XA_CARDINAL, &real_type,
                                        &real_format, &items_read,
                                        &items_left, (unsigned char **) &data);
        } else WW_DELETED;
        XUngrabServer(display);

        if (status == Success && items_read >= 6) {
            ww->_Maximize(data[4], data[5], horz, vert);
            ww->restore_max_attrib.x = data[0];
            ww->restore_max_attrib.y = data[1];
            ww->restore_max_attrib.width = data[2];
            ww->restore_max_attrib.height = data[3];
            ww->restore_max_flags.handle = restore_handlebar;
            XFree(data);
        }
    }

    if (shaded)
        ww->flags.shaded = true;
}

/**
 * @fn    SetWmState(KaWindow *ww)
 * @brief Sets net state hint
 *
 * Sets _NET_WM_STATE hint to the state of the window.
 *
 * @param ww KaWindow object
 */
void NetHandler::SetWmState(KaWindow *ww) {
    int i = 0;
    CARD32 data[15];
    CARD32 data2[6];

    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        if (ww->flags.sticky) data[i++] = net_wm_state_sticky;
        if (ww->flags.shaded) data[i++] = net_wm_state_shaded;
        if (ww->flags.alwaysontop) {
            data[i++] = net_wm_state_above;
            data[i++] = net_wm_state_stays_on_top;
        }
        if (ww->flags.alwaysatbottom) {
            data[i++] = net_wm_state_below;
            data[i++] = net_wm_state_stays_at_bottom;
        }
        if (!ww->flags.tasklist)
            data[i++] = net_wm_state_skip_taskbar;

        if (ww->flags.hidden) data[i++] = net_wm_state_hidden;
        if (ww->flags.fullscreen) data[i++] = net_wm_state_fullscreen;

        if (ww->flags.max) {
            if (ww->flags.vmax)
                data[i++] = net_wm_maximized_vert;
            if (ww->flags.hmax)
                data[i++] = net_wm_maximized_horz;
            
            data2[0] = ww->restore_max_attrib.x;
            data2[1] = ww->restore_max_attrib.y;
            data2[2] = ww->restore_max_attrib.width;
            data2[3] = ww->restore_max_attrib.height;
            data2[4] = ww->restore_max_attrib.misc0;
            data2[5] = ww->restore_max_attrib.misc1;
            XChangeProperty(display, ww->id, kahakai_net_maximized_restore,
                            XA_CARDINAL, 32, PropModeReplace,
                            (unsigned char *) data2, 6);
        } else
            XDeleteProperty(display, ww->id, kahakai_net_maximized_restore);
        
        data[i++] = kahakai_net_wm_state_decor;

        if (ww->flags.title) data[i++] = kahakai_net_wm_state_decortitle;
        if (ww->flags.handle) data[i++] = kahakai_net_wm_state_decorhandle;
        if (ww->flags.border) data[i++] = kahakai_net_wm_state_decorborder;

        if (ww->restore_max_flags.handle)
            data[i++] = kahakai_net_wm_maximize_restore_handlebar;
        
        XChangeProperty(display, ww->id, net_wm_state, XA_ATOM, 32,
                        PropModeReplace, (unsigned char *) data, i);
    } else ww->deleted = true;
    XUngrabServer(display);
}

/**
 * @fn    SetSupported(KaScreen *ws)
 * @brief Writes supported hint
 *
 * Sets _NET_SUPPORTED hint to the atoms Kahakai supports.
 *
 * @param ws KaScreen object
 */
void NetHandler::SetSupported(KaScreen *ws) {
    CARD32 data[52];
    int i = 0;
    
    data[i++] = net_supported;
    data[i++] = net_supported_wm_check;
    
    data[i++] = net_client_list;
    data[i++] = net_client_list_stacking;
    data[i++] = net_active_window;
    
    data[i++] = net_desktop_viewport;
    data[i++] = net_desktop_geometry;
    data[i++] = net_current_desktop;
    data[i++] = net_number_of_desktops;
    data[i++] = net_desktop_names;
    data[i++] = net_workarea;
    
    data[i++] = net_wm_desktop;
    data[i++] = net_wm_name;
    data[i++] = net_wm_visible_name;
    data[i++] = net_wm_strut;
    data[i++] = net_wm_pid;
    
    data[i++] = net_wm_state;
    data[i++] = net_wm_state_sticky;
    data[i++] = net_wm_state_shaded;
    data[i++] = net_wm_state_hidden;
    data[i++] = net_wm_maximized_vert;
    data[i++] = net_wm_maximized_horz;
    data[i++] = net_wm_state_above;
    data[i++] = net_wm_state_below;
    data[i++] = net_wm_state_stays_on_top;
    data[i++] = net_wm_state_stays_at_bottom;
    data[i++] = net_wm_state_fullscreen;
    data[i++] = net_wm_state_skip_taskbar;

    data[i++] = net_wm_allowed_actions;
    data[i++] = net_wm_action_move;
    data[i++] = net_wm_action_resize;
    data[i++] = net_wm_action_minimize;
    data[i++] = net_wm_action_shade;
    data[i++] = net_wm_action_stick;
    data[i++] = net_wm_action_maximize_horz;
    data[i++] = net_wm_action_maximize_vert;
    data[i++] = net_wm_action_fullscreen;
    data[i++] = net_wm_action_change_desktop;
    data[i++] = net_wm_action_close;
    
    data[i++] = net_wm_window_type;
    data[i++] = net_wm_window_type_desktop;
    data[i++] = net_wm_window_type_dock;
    data[i++] = net_wm_window_type_toolbar;
    data[i++] = net_wm_window_type_menu;
    data[i++] = net_wm_window_type_splash;
    data[i++] = net_wm_window_type_dialog;
    data[i++] = net_wm_window_type_utility;
    data[i++] = net_wm_window_type_normal;
    
    data[i++] = net_close_window;
    data[i++] = net_moveresize_window;
    data[i++] = net_wm_moveresize;
    
    XChangeProperty(display, ws->id, net_supported, XA_ATOM, 32,
                    PropModeReplace, (unsigned char *) data, i);
}

/**
 * @fn    SetSupportedCheck(KaScreen *ws, Window child)
 * @brief Writes _NET_SUPPORTED_WM_CHECK hint
 *
 * Sets _NET_SUPPORTED_WM_CHECK to child window and sets child windows
 * _NET_WM_NAME hint to the window manager name.
 *
 * @param ws KaScreen object
 & @param child Window to use as child window
*/
void NetHandler::SetSupportedWMCheck(KaScreen *ws, Window child) {
    XChangeProperty(display, ws->id, net_supported_wm_check, XA_WINDOW, 32,
                    PropModeReplace, (unsigned char *) &child, 1);

    XChangeProperty(display, child, net_supported_wm_check, XA_WINDOW, 32,
                    PropModeReplace, (unsigned char *) &child, 1);

    XChangeProperty(display, child, net_wm_name, utf8_string, 8,
                    PropModeReplace, (unsigned char *) PACKAGE,
                    strlen(PACKAGE));
}

/**
 * @fn    SetClientList(KaScreen *ws)
 * @brief Writes _NET_CLIENT_LIST hint
 *
 * Updates _NET_CLIENT_LIST hint to the current window list.
 *
 * @param ws KaScreen object
 */
void NetHandler::SetClientList(KaScreen *ws) {
    CARD32 *data;
    int i = 0;

    data = new CARD32[ws->kawindow_list_map_order.size() + 1];

    list<KaWindow *>::iterator it =
        ws->kawindow_list_map_order.begin();
    for (; it != ws->kawindow_list_map_order.end(); ++it)
        data[i++] = (*it)->id;

    XChangeProperty(display, ws->id, net_client_list, XA_WINDOW, 32,
                    PropModeReplace, (unsigned char *) data, i);

    delete [] data;
}

/**
 * @fn    SetClientListStacking(KaScreen *ws)
 * @brief Writes _NET_CLIENT_LIST_STACKING hint
 *
 * Updates _NET_CLIENT_LIST_STACKING hint to the current stacking order.
 *
 * @param ws KaScreen object
 */
void NetHandler::SetClientListStacking(KaScreen *ws) {
    typedef CARD32 *PCARD32;
    CARD32 *data;
    PCARD32 *desk_data;
    int i = 0, *desk_i;
    KaChildWindow *wc;
    list<Window> stack;

    desk_i = new int[16];
    data = new CARD32[ws->kawindow_list.size() + 1];
    desk_data = new PCARD32[16];
    for (int j = 0; j < 16; j++) {
        desk_i[j] = 0;
        desk_data[j] = new CARD32[ws->kawindow_list.size() + 1];
    }
        
    // Fold the stacking lists for the different desktops into one big
    // stacking list.
    list<Window>::reverse_iterator it;
    DESKTOP_LOOP(ws) {
        it = (*__dit)->aab_stacking_list.rbegin();
        for (; it != (*__dit)->aab_stacking_list.rend(); ++it) {
            wc = (KaChildWindow *) kahakai->FindWin(*it, FrameType);
            if (wc) {
                desk_data[(*__dit)->number][desk_i[(*__dit)->number]++] = wc->ka->id;
                if (find(stack.begin(), stack.end(), wc->ka->id) == stack.end()) {
                    data[i++] = wc->ka->id;
                    stack.push_back(wc->ka->id);
                }
            }
        }
    }
    DESKTOP_LOOP(ws) {
        it = (*__dit)->stacking_list.rbegin();
        for (; it != (*__dit)->stacking_list.rend(); ++it) {
            wc = (KaChildWindow *) kahakai->FindWin(*it, FrameType);
            if (wc) {
                desk_data[(*__dit)->number][desk_i[(*__dit)->number]++] = wc->ka->id;
                if (find(stack.begin(), stack.end(), wc->ka->id) == stack.end()) {
                    data[i++] = wc->ka->id;
                    stack.push_back(wc->ka->id);
                }
            }
        }
    }
    DESKTOP_LOOP(ws) {
        it = (*__dit)->aot_stacking_list.rbegin();
        for (; it != (*__dit)->aot_stacking_list.rend(); ++it) {
            wc = (KaChildWindow *) kahakai->FindWin(*it, FrameType);
            if (wc) {
                desk_data[(*__dit)->number][desk_i[(*__dit)->number]++] = wc->ka->id;
                if (find(stack.begin(), stack.end(), wc->ka->id) == stack.end()) {
                    data[i++] = wc->ka->id;
                    stack.push_back(wc->ka->id);
                }
            }
        }
    }
    XChangeProperty(display, ws->id, net_client_list_stacking, XA_WINDOW,
                    32, PropModeReplace, (unsigned char *) data, i);

    for (int j = 0; j < 16; j++) {
        XChangeProperty(display, ws->id, kahakai_net_client_list_stacking_desktop[j],
                        XA_WINDOW, 32, PropModeReplace,
                        (unsigned char *) desk_data[j], desk_i[j]);
        delete [] desk_data[j];
    }
    
    delete [] data;
    delete [] desk_data;
    delete [] desk_i;
}

/**
 * @fn    GetClientList(KaScreen *ws)
 * @brief Reads _NET_CLIENT_LIST_STACKING hint
 *
 * Updates window stacking order to order in _NET_CLIENT_LIST_STACKING hint.
 *
 * @param ws KaScreen object
 */
void NetHandler::GetClientListStacking(KaScreen *ws) {
    CARD32 *data;
    unsigned int i;
    
    if (XGetWindowProperty(display, ws->id, net_client_list_stacking,
                           0L, 8192L,
                           false, XA_WINDOW, &real_type,
                           &real_format, &items_read, &items_left, 
                           (unsigned char **) &data) == Success && 
        items_read) {
        for (i = 0; i < items_read; i++) {
            KaWindow *ww = (KaWindow *) kahakai->FindWin(data[i], WindowType);
            if (ww) ws->RaiseWindow(ww->frame->id);
        }
        free(data);
        SetClientListStacking(ws);
    }
}

/**
 * @fn    SetActiveWindow(KaScreen *ws, KaWindow *ww)
 * @brief Writes _NET_ACTIVE_CLIENT hint
 *
 * Updates _NET_ACTIVE_CLIENT hint to the current list of last active
 * windows. The currently active window at the front.
 *
 * @param ws KaScreen object
 * @param ww KaWindow that was set active or NULL if root window.
 */
void NetHandler::SetActiveWindow(KaScreen *ws, KaWindow *ww) {
    CARD32 *data;
    vector<KaWindow *>::iterator it;
    int i = 0;

    data = new CARD32[ws->kawindow_list.size() + 1];
    
    if (ww) {
        ws->focus = false;
        ws->kawindow_list.erase(remove(ws->kawindow_list.begin(), 
                                       ws->kawindow_list.end(),
                                       ww), ws->kawindow_list.end());
        ws->kawindow_list.insert(ws->kawindow_list.begin(), ww);
    } else
        data[i++] = None;

    it = ws->kawindow_list.begin();
    for (; it != ws->kawindow_list.end(); ++it)
        data[i++] = (*it)->id;
    
    XChangeProperty(display, ws->id, net_active_window, XA_WINDOW, 32,
                    PropModeReplace, (unsigned char *) data, i);

    delete [] data;
}

/**
 * @fn    GetActiveWindow(KaScreen *ws)
 * @brief Reads active client hint
 *
 * Sorts active window list after _NET_ACTIVE_CLIENT hint and sets input
 * focus to first window in _NET_ACTIVE_CLIENT list.
 *
 * @param ws KaScreen object
 */
void NetHandler::GetActiveWindow(KaScreen *ws) {
    KaWindow *ww;
    CARD32 *data;
    int i;
    
    if (XGetWindowProperty(display, ws->id, net_active_window, 0L,
                           ws->kawindow_list.size(), 
                           false, XA_WINDOW, &real_type, &real_format, 
                           &items_read, &items_left, 
                           (unsigned char **) &data) == Success && 
        items_read) {
        for (i = items_read - 1; i >= 0; i--) {
            if (i == 0 && data[0] == None) {
                ws->Focus(NULL, NULL);
                break;
            }
            ww = (KaWindow *) kahakai->FindWin(data[i], WindowType);
            if (ww) {                
                ws->kawindow_list.erase(remove(ws->kawindow_list.begin(), 
                                               ws->kawindow_list.end(),
                                               ww), ws->kawindow_list.end());
                ws->kawindow_list.insert(ws->kawindow_list.begin(), ww);

              //                ws->kawindow_list.remove(ww);
              //                ws->kawindow_list.push_front(ww);
                if (i == 0)
                    ww->Focus(false);
            }
        }
        XFree(data);
    }
}

/**
 * @fn    GetVirtualPos(KaWindow *ww)
 * @brief Reads virtual position hint
 *
 * Reads KaWindows virtual position hint.
 *
 * @param ww KaWindow object
 */
void NetHandler::GetVirtualPos(KaWindow *ww) {
    int *data;
    
    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        if (XGetWindowProperty(display, ww->id, kahakai_net_virtual_pos,
                               0L, 2L, false, XA_INTEGER, &real_type,
                               &real_format, &items_read, &items_left, 
                               (unsigned char **) &data) == Success && 
            items_read >= 2) {
            ww->attrib.x = data[0] - ww->kascreen->v_x;
            ww->attrib.y = data[1] - ww->kascreen->v_y;
            if (ww->flags.sticky) {
                ww->attrib.x = ww->attrib.x % ww->kascreen->width;
                ww->attrib.y = ww->attrib.y % ww->kascreen->height;
            }
            XFree(data);
        }
    } else ww->deleted = true;
    XUngrabServer(display);
}

/**
 * @fn    GetXaName(KaWindow *ww)
 * @brief Reads window title
 *
 * Reads WM_NAME hint and if hint exists the window title is set to this
 * and _NET_WM_VISIBLE_HINT is updated.
 *
 * @param ww KaWindow object
 */
void NetHandler::GetXaName(KaWindow *ww) {
    char *data = NULL;
    int status;
    
    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        status = XFetchName(display, ww->id, &data);
    } else ww->deleted = true;
    XUngrabServer(display);

    if (status && data) {
        ww->labeltext = ww->name = ww->realname = data;
        ww->realnamelen = ww->name.length();
        XFree(data);
        ww->SetActionLists();

        ww->kascreen->SmartName(ww);
        ww->nameUpdated();
    }
        
    SetVisibleName(ww);
}

/**
 * @fn    GetNetName(KaWindow *ww)
 * @brief Reads window title
 *
 * Reads _NET_WM_NAME hint and if hint exists the window title is set to this
 * and _NET_WM_VISIBLE_HINT is updated.
 *
 * @param ww KaWindow object
 *
 * @return True if _NET_WM_NAME hint did exist otherwise false
 */
bool NetHandler::GetNetName(KaWindow *ww) {
    char *data;
    int status;
    
    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        status = XGetWindowProperty(display, ww->id, net_wm_name, 0L, 8192L, 
                                    false, utf8_string, &real_type,
                                    &real_format, &items_read, &items_left, 
                                    (unsigned char **) &data);
    } else ww->deleted = true;
    XUngrabServer(display);
    
    if (status == Success && items_read) {
        ww->labeltext = ww->name = ww->realname = data;
        ww->realnamelen = ww->name.length();
        ww->SetActionLists();
        XFree(data);

        ww->kascreen->SmartName(ww);
        ww->nameUpdated();

        SetVisibleName(ww);
        
        return true;
    }
    return false;
}

/**
 * @fn    SetVisibleName(KaWindow *ww)
 * @brief Writes visible name hint
 *
 * Sets _NET_WM_VISIBLE_HINT to the current visible window name.
 *
 * @param ww KaWindow object
 */
void NetHandler::SetVisibleName(KaWindow *ww) {
    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        XChangeProperty(display, ww->id, net_wm_visible_name,
                        utf8_string, 8, PropModeReplace,
                        (unsigned char *) ww->name.c_str(), ww->name.length());
    } else ww->deleted = true;
    XUngrabServer(display);
}

/**
 * @fn    RemoveVisibleName(KaWindow *ww)
 * @brief Removes visible name hint
 *
 * Deletes _NET_WM_VISIBLE_HINT.
 *
 * @param ww KaWindow object
 */
void NetHandler::RemoveVisibleName(KaWindow *ww) {
    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        XDeleteProperty(display, ww->id, net_wm_visible_name);
    }
    XUngrabServer(display);
}


/**
 * @fn    SetVirtualPos(KaWindow *ww)
 * @brief Writes virtual position hint
 *
 * Writes KaWindows virtual position hint.
 *
 * @param ww KaWindow object
 */
void NetHandler::SetVirtualPos(KaWindow *ww) {
    int data[2];

    ww->Gravitate(RemoveGravity);
    data[0] = ww->kascreen->v_x + ww->attrib.x;
    data[1] = ww->kascreen->v_y + ww->attrib.y;
    ww->Gravitate(ApplyGravity);

    XGrabServer(display);
    if (validatedrawable(ww->id))
        XChangeProperty(display, ww->id, kahakai_net_virtual_pos, XA_INTEGER,
                        32, PropModeReplace, (unsigned char *) data, 2);
    else ww->deleted = true;
    XUngrabServer(display);
}

/**
 * @fn    GetWmStrut(KaWindow *ww)
 * @brief Reads strut hint
 *
 * Reads windows _NET_WM_STRUT hint, adds it to the strut list and updates
 * workarea.
 *
 * @param ww KaWindow object
 */
void NetHandler::GetWmStrut(KaWindow *ww) {
    CARD32 *data;
    WMstrut *wm_strut;
    bool found = false;
    int status;

    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        status = XGetWindowProperty(display, ww->id, net_wm_strut, 0L, 4L, 
                                    false, XA_CARDINAL, &real_type,
                                    &real_format, &items_read, &items_left, 
                                    (unsigned char **) &data);
    } else WW_DELETED;
    XUngrabServer(display);
    
    if (status == Success && items_read >= 4) {
        list<WMstrut *>::iterator it = ww->kascreen->strut_list.begin();
        for (; it != ww->kascreen->strut_list.end(); ++it) {
            if ((*it)->window == ww->id) {
                (*it)->left = data[0];
                (*it)->right = data[1];
                (*it)->top = data[2];
                (*it)->bottom = data[3];
                found = true;
                ww->kascreen->UpdateWorkarea();
            }
        }
        if (! found) {
            wm_strut = new WMstrut;
            wm_strut->window = ww->id;
            wm_strut->left = data[0];
            wm_strut->right = data[1];
            wm_strut->top = data[2];
            wm_strut->bottom = data[3];
            ww->wm_strut = wm_strut;
            ww->kascreen->strut_list.push_back(wm_strut);
            ww->kascreen->UpdateWorkarea();
        }
        XFree(data);
    }
}

/**
 * @fn    GetWmPid(KaWindow *ww)
 * @brief Reads _NET_WM_PID hint
 *
 * Reads windows _NET_WM_PID hint and stores it in KaWindows pid variable.
 *
 * @param ww KaWindow object
 */
void NetHandler::GetWmPid(KaWindow *ww) {
    char tmp[32];
    CARD32 *data;

    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        if (XGetWindowProperty(ww->display, ww->id, net_wm_pid, 0L, 1L, 
                               false, XA_CARDINAL, &real_type,
                               &real_format, &items_read, &items_left, 
                               (unsigned char **) &data) == Success && 
            items_read) {
            sprintf(tmp, "%d" , (unsigned int) *data);
            ww->pid = new char[strlen(tmp) + 1];
            sprintf(ww->pid, "%s" , tmp);
            XFree(data);
        }
    } else ww->deleted = true;
    XUngrabServer(display);
}

/**
 * @fn    GetDesktopViewPort(KaScreen *ws)
 * @brief Reads viewport hint
 *
 * Reads KaScreens viewport hint.
 *
 * @param ws KaScreen object
 */
void NetHandler::GetDesktopViewPort(KaScreen *ws) {
    CARD32 *data;
    
    if (XGetWindowProperty(display, ws->id, net_desktop_viewport, 0L, 2L, 
                           false, XA_CARDINAL, &real_type,
                           &real_format, &items_read, &items_left, 
                           (unsigned char **) &data) == Success && 
        items_read >= 2) {
        ws->MoveViewportTo(data[0], data[1]);
        XFree(data);
    }
}

/**
 * @fn    SetDesktopViewPort(KaScreen *ws)
 * @brief Writes viewport hint
 *
 * Writes KaScreens viewport hint.
 *
 * @param ws KaScreen object
 */
void NetHandler::SetDesktopViewPort(KaScreen *ws) {
    CARD32 data[2];

    data[0] = ws->v_x;
    data[1] = ws->v_y;

    XChangeProperty(display, ws->id, net_desktop_viewport, XA_CARDINAL, 32,
    PropModeReplace, (unsigned char *) data, 2);
}

/**
 * @fn    SetDesktopGeometry(KaScreen *ws)
 * @brief Writes desktop geometry hint
 *
 * Sets _NET_DESKTOP_GEOMETRY hint.
 *
 * @param ws KaScreen object
 */
void NetHandler::SetDesktopGeometry(KaScreen *ws) {
    CARD32 data[2];

    data[0] = ws->v_xmax + ws->width;
    data[1] = ws->v_ymax + ws->height;
    XChangeProperty(display, ws->id, net_desktop_geometry, XA_CARDINAL, 32,
                    PropModeReplace, (unsigned char *) data, 2);
}

/**
 * @fn    SetNumberOfDesktops(KaScreen *ws)
 * @brief Writes number of desktops hint
 *
 * Sets _NET_NUMBER_OF_DESKTOPS hint.
 *
 * @param ws KaScreen object
 */
void NetHandler::SetNumberOfDesktops(KaScreen *ws) {
    CARD32 data[1];

    data[0] = ws->desktop_list.size();
    XChangeProperty(display, ws->id, net_number_of_desktops, XA_CARDINAL, 32,
                    PropModeReplace, (unsigned char *) data, 1);
}
    
/**
 * @fn    SetCurrentDesktop(KaScreen *ws)
 * @brief Writes current desktop hint
 *
 * Sets _NET_CURRENT_DESKTOP hint.
 *
 * @param ws KaScreen object
 */
void NetHandler::SetCurrentDesktop(KaScreen *ws) {
    CARD32 data[1];
    
    data[0] = ws->current_desktop->number;
    XChangeProperty(display, ws->id, net_current_desktop, XA_CARDINAL, 32,
                    PropModeReplace, (unsigned char *) data, 1);
}

/**
 * @fn    GetCurrentDesktop(KaScreen *ws)
 * @brief Reads current desktop hint
 *
 * Reads _NET_CURRENT_DESKTOP hint and sets desktop accordingly.
 *
 * @param ws KaScreen object
 */
void NetHandler::GetCurrentDesktop(KaScreen *ws) {
    CARD32 *data;

    if (XGetWindowProperty(display, ws->id, net_current_desktop, 0L, 1L, 
                           false, XA_CARDINAL, &real_type, &real_format, 
                           &items_read, &items_left, 
                           (unsigned char **) &data) == Success && 
        items_read) {
        ws->GoToDesktop(data[0]);
        XFree(data);
    }
}


/**
 * @fn    SetDesktopNames(KaScreen *ws, char *names)
 * @brief Writes desktop names hint
 *
 * Sets _NET_DESKTOP_NAMES.
 *
 * @param ws KaScreen object
 * @param names String to parse desktop names from
 */
void NetHandler::SetDesktopNames(KaScreen *ws, char *names) {
    int i;

    for (i = 0; i < 8192; i++) {
        if (names[i] == ',') names[i] = '\0';
        else if (names[i] == '\0') break;
    }
    names[i] = '\0';

    if (i)
        XChangeProperty(display, ws->id, net_desktop_names, utf8_string, 8,
                        PropModeReplace, (unsigned char *) names, i + 1);
}

/**
 * @fn    SetWorkarea(KaScreen *ws)
 * @brief Sets window manager workarea
 *
 * Sets the window manager workarea, used for placing icons and maximizing
 * windows.
 *
 * @param ws KaScreen object
 */
void NetHandler::SetWorkarea(KaScreen *ws) {
    CARD32 data[4 * 16];
    int i = 0;
    
    vector<Desktop *>::iterator it = ws->desktop_list.begin();
    for (; it != ws->desktop_list.end(); ++it) {
        data[i++] = (*it)->workarea.x;
        data[i++] = (*it)->workarea.y;
        data[i++] = (*it)->workarea.width;
        data[i++] = (*it)->workarea.height;
    }
    
    XChangeProperty(kahakai->display, ws->id, net_workarea, XA_CARDINAL,
                    32, PropModeReplace, (unsigned char *) data, i);
}

/**
 * @fn    wXDNDMakeAwareness(Window window)
 * @brief Make window DND aware
 *
 * Makes drag'n'drop awareness for window.
 *
 * @param window Window to make DND aware
 */
void NetHandler::wXDNDMakeAwareness(Window window) {
    long int xdnd_version = 3;

    XChangeProperty(kahakai->display, window, xdndaware, XA_ATOM,
                    32, PropModeReplace, (unsigned char *) &xdnd_version, 1);
}

/**
 * @fn    wXDNDClearAwareness(Window window)
 * @brief Removes DND awareness from window
 *
 * Removes drag'n'drop awareness for window.
 *
 * @param window Window to remove DND awareness from
 */
void NetHandler::wXDNDClearAwareness(Window window) {
    XDeleteProperty (kahakai->display, window, xdndaware);
}

/**
 * @fn    DeleteSupported(KaScreen *ws)
 * @brief Deletes supported hints
 *
 * Deletes all hints set on the root window that we don't have any use
 * of next time Kahakai starts.
 *
 * @param ws KaScreen object
 */
void NetHandler::DeleteSupported(KaScreen *ws) {
    XDeleteProperty(display, ws->id, net_desktop_geometry);
    XDeleteProperty(display, ws->id, net_workarea);
    XDeleteProperty(display, ws->id, net_supported_wm_check);
    XDeleteProperty(display, ws->id, net_supported);
}

/**
 * @fn    GetXRootPMapId(KaScreen *ws)
 * @brief Reads XROOTPMAPID hint
 *
 * Reads _XROOTPMAP_ID hint, which if it exist is the pixmap ID for the root
 * window.
 *
 * @param ws KaScreen object
 */
void NetHandler::GetXRootPMapId(KaScreen *ws) {
    CARD32 *data;

    XSync(ws->display, false);

#if 0
    if (ws->xrootpmap_id)
        XFreePixmap(ws->ic->getDisplay(), ws->xrootpmap_id);

    Pixmap pixmap = XCreatePixmap(ws->ic->getDisplay(),
                                  ws->ic->getDrawable(), ws->width,
                                  ws->height, ws->ic->getDepth());
    XGCValues gcv;
    gcv.fill_style = FillSolid;
    
    GC gc = XCreateGC(ws->ic->getDisplay(), 
                      pixmap,
                      GCForeground, &gcv);

    XImage* img = XGetImage(ws->pdisplay, ws->id, 0, 0, 
                            ws->width, ws->height,
                            (1 << ws->ic->getDepth()) - 1, ZPixmap);

    XPutImage(ws->pdisplay, pixmap, gc, img, 0, 0, 0, 0, ws->width,
              ws->height);

    ws->xrootpmap_id = pixmap;
    XFree(img);
    XFree(gc);
#endif
        
    if (XGetWindowProperty(ws->pdisplay, ws->id, xrootpmap_id, 0L, 1L, 
                           false, XA_PIXMAP, &real_type,
                           &real_format, &items_read, &items_left, 
                           (unsigned char **) &data) == Success && 
        items_read) { 
        ws->xrootpmap_id = (Pixmap) (*data);
        XFree(data);
    }
    else
        ws->xrootpmap_id = (Pixmap) 0;

    XSync(ws->display, false);
    XSync(ws->pdisplay, false);
}

/**
 * @fn    GetType(KaWindow *ww)
 * @brief Read type hint
 *
 * Reads _NET_WM_WINDOW_TYPE hint and sets window properties accordingly.
 *
 * @param ww KaWindow object
 */
void NetHandler::GetWmType(KaWindow *ww) {
    CARD32 *data;
    int status;
    
    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        status = XGetWindowProperty(display, ww->id, net_wm_window_type,
                                    0L, 8L, false, XA_ATOM,
                                    &real_type, &real_format, &items_read,
                                    &items_left, (unsigned char **) &data);
    } else WW_DELETED;
    XUngrabServer(display);

    ww->windowType = OtherWindowType;
 
    if (status == Success && items_read) {
        for (unsigned int i = 0; i < items_read; ++i) {
            
            // check for dialog type

            if (data[i] == net_wm_window_type_desktop) {
                ww->windowType = DesktopWindowType;
                ww->desktop_mask = (1L << 16) - 1;
                ww->flags.tasklist = false;
                ww->flags.sticky = true;
                ww->flags.border = ww->flags.title = ww->flags.handle =
                    ww->flags.all = false;
                ww->size.max_width = ww->kascreen->width;
                ww->size.min_width = ww->kascreen->width;
                ww->size.max_height = ww->kascreen->height;
                ww->size.min_height = ww->kascreen->height;
                ww->attrib.x = 0;
                ww->attrib.y = 0;
                ww->AlwaysatbottomOn();
            }
            else if (data[i] == net_wm_window_type_toolbar ||
                     data[i] == net_wm_window_type_dock) {
                ww->windowType = (data[i] == net_wm_window_type_toolbar) ?
                    ToolbarWindowType : DockWindowType;

                ww->desktop_mask = (1L << 16) - 1;
                ww->flags.tasklist = false;
                ww->flags.sticky = true;
                ww->flags.focusable = false;
                ww->flags.border = ww->flags.title = ww->flags.handle =
                    ww->flags.all = false;
                if (!ww->flags.alwaysatbottom) {
                    ww->AlwaysontopOn();
                }
            }
            else if (data[i] == net_wm_window_type_splash ||
                     data[i] == net_wm_window_type_menu) {

                ww->windowType = (data[i] == net_wm_window_type_splash) ?
                    SplashWindowType : MenuWindowType;
                ww->flags.tasklist = false;
                ww->flags.border = ww->flags.title = ww->flags.handle =
                    ww->flags.all = false;
                ww->AlwaysontopOn();
            }
            else {
                if (ww->attrib.x == 0) {
                    if (ww->kascreen->current_desktop->workarea.x >
                        ww->attrib.x)
                        ww->attrib.x =
                            ww->kascreen->current_desktop->workarea.x;
                }
                if (ww->attrib.y == 0) {
                    if (ww->kascreen->current_desktop->workarea.y >
                        ww->attrib.y)
                        ww->attrib.y =
                            ww->kascreen->current_desktop->workarea.y;
                }
            }
        }
        XFree(data);
    } else {
        if (ww->attrib.x == 0) {
            if (ww->kascreen->current_desktop->workarea.x > ww->attrib.x)
                ww->attrib.x = ww->kascreen->current_desktop->workarea.x;
        }
        if (ww->attrib.y == 0) {
            if (ww->kascreen->current_desktop->workarea.y > ww->attrib.y)
                ww->attrib.y = ww->kascreen->current_desktop->workarea.y;
        }
    }
}

/**
 * @fn    SetDesktop(KaWindow *ww)
 * @brief Write net_wm_desktop hint
 *
 * Sets _NET_WM_DESKTOP hint to current desktop, if window is not a member
 * of the current desktop then hint is set to the lowest desktop number that
 * the window is a member of. If window is a member of all desktops then
 * hint is set to 0xffffffff.
 *
 * @param ww KaWindow object
 */
void NetHandler::SetDesktop(KaWindow *ww) {

    CARD32 data[1];

    data[0] = 0;
    if (ww->desktop_mask & (1L << ww->kascreen->current_desktop->number))
        data[0] = ww->kascreen->current_desktop->number;
    else {
        int i;
        for (i = 0; i < 16; i++)
            if (ww->desktop_mask & (1L << i)) {
                data[0] = i;
                break;
            }
    }
    if (ww->desktop_mask == ((1L << 16) - 1))
        data[0] = 0xffffffff;

    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        XChangeProperty(display, ww->id, net_wm_desktop, XA_CARDINAL, 32,
                        PropModeReplace, (unsigned char *) data, 1);
    } else ww->deleted = true;
    XUngrabServer(display);
}

/**
 * @fn    SetDesktopMask(KaWindow *ww)
 * @brief Write net_wm_desktop_mask hint
 *
 * Sets _WAIMEA_NET_WM_DESKTOP_MASK hint to the current desktop mask
 * of the window.
 *
 * @param ww KaWindow object
 */
void NetHandler::SetDesktopMask(KaWindow *ww) {

    CARD32 data[1];

    data[0] = ww->desktop_mask;

    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        XChangeProperty(display, ww->id, kahakai_net_wm_desktop_mask,
                        XA_CARDINAL, 32, PropModeReplace,
                        (unsigned char *) data, 1);
    } else ww->deleted = true;
    XUngrabServer(display);
}

/**
 * @fn    GetDesktop(KaWindow *ww)
 * @brief Reads net_wm_desktop and net_desktop_mask hints
 *
 * Reads _NET_WM_DESKTOP and _WAIMEA_NET_WM_DESKTOP_MASK hints and sets
 * desktop_mask for window accordingly.
 *
 * @param ww KaWindow object
 */
void NetHandler::GetDesktop(KaWindow *ww) {
    CARD32 *data;

    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        if (XGetWindowProperty(display, ww->id, net_wm_desktop, 0L, 1L, 
                               false, XA_CARDINAL, &real_type, &real_format, 
                               &items_read, &items_left,
                               (unsigned char **) &data) == Success && 
            items_read) {
            if (data[0] == 0xffffffff || data[0] == 0xfffffffe)
                ww->desktop_mask = (1L << 16) - 1;
            else if (data[0] < 15 && data[0] >= 0) {
                ww->desktop_mask = 1L << data[0];
            }
            XFree(data);
        }
        if (XGetWindowProperty(display, ww->id, kahakai_net_wm_desktop_mask,
                               0L, 1L, false, XA_CARDINAL, &real_type,
                               &real_format, &items_read, &items_left, 
                               (unsigned char **) &data) == Success && 
            items_read) {
            ww->desktop_mask = data[0];
            XFree(data);
        }
        ww->UpdateStack();
    } else ww->deleted = true;
    XUngrabServer(display);
}

/**
 * @fn    IsSystrayWindow(Window w)
 * @brief Checks if window is systray window
 *
 * Reads _KDE_NET_WM_SYSTEM_TRAY_WINDOW_FOR hint to see if window is a systray
 * window.
 *
 * @param w Window to check
 *
 * @return True if window is systray window, otherwise false
 */
bool NetHandler::IsSystrayWindow(Window w) {
    CARD32 *data;

    items_read = 0;
    XGrabServer(display);
    if (validatedrawable(w)) {
        if (XGetWindowProperty(display, w, kde_net_wm_system_tray_window_for,
                               0L, 1L, false, XA_WINDOW, &real_type,
                               &real_format, &items_read, &items_left,
                               (unsigned char **) &data) != Success) {
            items_read = 0;
        }
    }
    XUngrabServer(display);

    return ((items_read)? true: false);
}

/**
 * @fn    SetSystrayWindows(KaScreen *ws)
 * @brief Write kde_net_system_tray_windows hint
 *
 * Sets _KDE_NET_SYSTEM_TRAY_WINDOWS hint to the current list of systray
 * windows.
 *
 * @param ws KaScreen object
 */
void NetHandler::SetSystrayWindows(KaScreen *ws) {
    CARD32 *data;
    int i = 0;

    data = new CARD32[ws->systray_window_list.size() + 1];

    list<Window>::iterator it = ws->systray_window_list.begin();
    for (; it != ws->systray_window_list.end(); it++)
        data[i++] = *it;

    XChangeProperty(display, ws->id, kde_net_system_tray_windows, XA_WINDOW,
                    32, PropModeReplace, (unsigned char *) data, i);

    delete [] data;
}

/**
 * @fn    SetAllowedActions(KaWindow *ww)
 * @brief Writes allowed actions hint
 *
 * Sets _NET_WM_ALLOWED_ACTIONS hint to indicate which actions that are
 * allowed.
 *
 * @param ww KaWindow object
 */
void NetHandler::SetAllowedActions(KaWindow *ww) {
    CARD32 data[10];
    int i = 0;

    if (ww->flags.tasklist) {
        data[i++] = net_wm_action_move;
        if (ww->size.max_width != ww->size.min_width ||
            ww->size.max_height != ww->size.min_height) {
            data[i++] = net_wm_action_resize;
            data[i++] = net_wm_action_maximize_horz;
            data[i++] = net_wm_action_maximize_vert;
            data[i++] = net_wm_action_fullscreen;
        }
        data[i++] = net_wm_action_minimize;
        if (ww->flags.title)
            data[i++] = net_wm_action_shade;
        data[i++] = net_wm_action_stick;
        data[i++] = net_wm_action_change_desktop;
        data[i++] = net_wm_action_close;
    }
    
    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        XChangeProperty(display, ww->id, net_wm_allowed_actions,
                        XA_ATOM, 32, PropModeReplace,
                        (unsigned char *) data, i);
    } else
        ww->deleted = true;
    XUngrabServer(display);
}

/**
 * @fn    RemoveAllowedActions(KaWindow *ww)
 * @brief Removes allowed actions hint
 *
 * Removes _NET_WM_ALLOWED_ACTIONS hint from window.
 *
 * @param ww KaWindow object
 */
void NetHandler::RemoveAllowedActions(KaWindow *ww) {
    XGrabServer(display);
    if (validatedrawable(ww->id)) {
        XDeleteProperty(display, ww->id, net_wm_allowed_actions);
    }
    XUngrabServer(display);
}

