/* $Id: NewStyleParser.cc,v 1.5 2003/11/28 23:47:07 pervago Exp $ */

/**
 * @file   NewStyleParser.cc
 * @date   3 Nov 2003
 *
 * @brief Implementation of the style parser for the new (.kastyle) styles.
 *
 */

#include "NewStyleParser.hh"
#include "ExpressionEvaluator.hh"
#include "Logger.hh"
#include "Kahakai.hh"

#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <utility>
#include <sstream>
#include <map>
#include <stdlib.h>

using std::max;
using std::map;
using std::cout;
using std::endl;
using std::pair;
using std::string;
using std::vector;
using std::ifstream;
using std::stringstream;


/** 
 * Searches for node using dot-separated path. Returns NULL if not found.
 * 
 * @param path A dot-separated path like 'window.focused.title.draw'
 * 
 * @return Reference to that node on success, NULL if not found.
 */
TreeNode* TreeNode::findNode(const string& path) {
    size_t dot_idx = path.find('.');
    string immediateChild;
    string restOfPath;

    if (dot_idx == string::npos)
        immediateChild = path;
    else
        immediateChild = path.substr(0, dot_idx);
    
    TreeNode* c = child(immediateChild);
    if (!c) 
        return NULL;
    else if (dot_idx == string::npos)
        return c;
    else
        return c->findNode(path.substr(dot_idx+1));
}


NewStyleParser::NewStyleParser()
{
    root = new TreeNode();
}

NewStyleParser::~NewStyleParser()
{
    delete root;
}

string joinBlocks(vector<pair<int, string> >& blocks)
{
    string result = "";
    for (vector<pair<int, string> >::iterator it=blocks.begin(); it != blocks.end(); ++it) {
        if (it != blocks.begin()) result += ".";
        result += it->second;
    }
    return result;
}

string NewStyleParser::processValue(const string& value)
{
    string result = value;
    map<string, string>::iterator it;

    map<string, float> values;

    // Try to convert all numeric valus we can.
    for (it = defs.begin(); it != defs.end(); ++it) {
        char* endptr;
        const char* nptr = it->second.c_str();
        float floatval = strtod(nptr, &endptr);
        if (floatval == 0 && nptr == endptr) continue;
        values[it->first] = floatval;
    }

    ExpressionEvaluator evaluator(values);

    // $[] is used (not $()) because $() can (and does) produce string values
    // while $[] and the expression evaluator itself is limited strictly to 
    // numeric values.
    size_t offset = 0;
    size_t start;
    while ((start = value.find("$[", offset)) != string::npos) {
        size_t end = value.find("]", start);
        if (end == string::npos) 
            break;

        string expression = value.substr(start + 2, end - start - 2);

        float r = 0;
        try {
            r = evaluator.evaluate(expression);
        } catch (parse_exception e) {
            logger << warn << e.what() << endl;
        }

        result.replace(start, end - start + 1, to_string(r));

        offset = end + 1;
    }

    for (it = defs.begin(); it != defs.end(); ++it) {
        result = replaceAll(result, "$(" + it->first + ")", it->second);
    }
    return result;
}

void printTree(TreeNode* root, int level=0)
{
    cout << string(2*level, ' ') << root->name << ":" << endl;
    printf("[%p]\n", root);
    unsigned int i;
    for (i=0;i<root->subBlocks.size();i++) {
        printTree(root->subBlocks[i], level+1);
    }
    for (i=0;i<root->leaves.size();i++) {
        cout << string(2*(level+1), ' ') << root->leaves[i].name << endl;
    }
    for (i=0;i<root->drawLeaves.size();i++) {
        cout << string(2*(level+1), ' ') << "#" << 
            root->drawLeaves[i].name << endl;
    }

}

void NewStyleParser::parse(const char* filename)
{
    ifstream fs(filename);

    if (!fs.is_open()) {
        logger << warn << "Couldn't open '" << filename << "' for reading" << endl;
        return;
    }

    blocks.clear();
    currentBlockNode = root;
    last_block_indent = -1;

    // Draw mode specifies whether to interpret statements as 
    // assignments or as draw operations.

    // It's triggered by draw: block.

    drawMode = false;

    int    current_indent    = 0;
    string line;
    while (std::getline(fs, line, '\n')) {
        int indent = line.find_first_not_of(" \t");
        if (indent < 0) continue; // Just whitespace
        if (line[indent] == '#') continue; // Comments

        if (indent > current_indent) {
            if (last_block_indent == -1 || last_block_indent < current_indent)
                {
                    logger << warn << "Indentation error" << endl;
                    return;
                }
            current_indent = indent;
        }

        if (indent < current_indent) {
            current_indent = indent;
            while (!blocks.empty() && (blocks.back().first >= indent)) {
                blocks.pop_back();
                currentBlockNode = currentBlockNode->parent;
            }

            drawMode = false; // In any case, exiting a block resets draw mode
        }

        line = trim(line);

        if (line[0] == '@') {
            string blockName  = line.substr(1);
            includeBlock(blockName, joinBlocks(blocks));
        } else if (line[line.size()-1] == ':') {
            processBlock(line, indent);
        } else {
            if (drawMode) {
                processDrawOperation(line);
            } else {
                processAssignment(line);
            }
        }
    }
    //    printTree(root);
    fs.close();
}

/** 
 * Processes draw operation e.g.  
 * rect #336699, 0, 1, 400, 300
 * 
 * @param line Trimmed line in the above format
 */
void NewStyleParser::processDrawOperation(const string& line) {
    // Format is <name> <param1>, <param2>
    size_t space_idx = line.find(' ');
    if (space_idx == string::npos || space_idx == line.size() - 1) {
        logger << warn << "Invalid drawing operation: " << line << endl;
        return;
    }
    size_t idx = space_idx + 1;

    TreeDrawLeaf::Parameters params;

    string name = line.substr(0, space_idx);

    do {
        size_t next_comma = line.find(',', idx);
        string value;
        if (next_comma == string::npos) {
            value = trim(line.substr(idx));
        } else {
            value = trim(line.substr(idx, next_comma - idx));
        }

        params.push_back(processValue(value));
        
        idx = line.find_first_not_of(", ", line.find(',', idx));
    } while (idx != string::npos);

    currentBlockNode->drawLeaves.push_back(TreeDrawLeaf(name, params));
}

/** 
 * Processes assignment e.g. 'borderColor = #336699'
 * 
 * @param line A trimmed line in the format 'setting = value'
 */
void NewStyleParser::processAssignment(const string& line) {
    size_t equal_idx = line.find_first_of(" =");
    if (equal_idx == string::npos) {
        logger << warn << "Invalid non-empty line: " << line << endl;
        return;
    }

    string name  = line.substr(0, equal_idx);
    string value = line.substr(line.find_first_not_of(" =", 
                                                      equal_idx));
            
    string fullName;
    if (blocks.empty())
        fullName = name;
    else
        fullName = joinBlocks(blocks) + "." + name;
            
    defs[fullName] = processValue(value);
    currentBlockNode->leaves.push_back(TreeLeaf(fullName, 
                                                defs[fullName]));

}

/** 
 * Processes block line (e.g. 'aBlock:')  with specified indent
 * 
 * @param line   Trimmed line which ends in ':'
 * @param indent Indent of the line
 */

void NewStyleParser::processBlock(const string& line, int indent) {
    string blockName = line.substr(0, line.length()-1);

    last_block_indent = indent;

    // Drawing operation define
    if (blockName.substr(0, 4) == "def ") {
        blockName = trim(blockName.substr(4));
        drawMode = true;
    }

    blocks.push_back(make_pair(indent, blockName));

    TreeNode *newBlockNode = new TreeNode();
    newBlockNode->name   = blocks.back().second;
    newBlockNode->parent = currentBlockNode;
    currentBlockNode->subBlocks.push_back(newBlockNode);
    currentBlockNode = newBlockNode;

    if (blockName == "draw") {
        drawMode = true;
    }
}

/** 
 * Includes block inside current block (copying all values starting
 * with blockName into current block).
 *  
 * In other words if called with blockName='A' and currentBlock='B' will change
 * A.whatever to B.whatever and put that into definitions map.
 *
 * If in draw mode, copy all draw operations from the block into current block.
 * 
 * @param blockName Dot-separated name of block to include
 * @param currentBlock Dot-separated name of current black
 */
void NewStyleParser::includeBlock(const string& blockName,
                                  const string& currentBlock) {

    if (drawMode) {
        TreeNode* block = getRoot()->findNode(blockName);
        if (!block) {
            logger << warn << "Drawing block " << blockName << " not found" 
                   << endl;
            return;
        } 
        copy(block->drawLeaves.begin(), block->drawLeaves.end(),
             back_inserter(currentBlockNode->drawLeaves));
    } else {
        map<string,string>::iterator it;
        for (it=defs.begin();it !=defs.end(); ++it) {
            if (it->first.find(blockName + ".") == 0) {
                string newName = currentBlock + 
                it->first.substr(blockName.length()); 
                defs[newName] = it->second;
            }
        }
    }
}

bool NewStyleParser::getValue(const string& rname, string& value)
{
    if (rname.find(".style.") != string::npos) {
        string new_name = rname;
        new_name.replace(new_name.find(".style."), 6, "");
        return getValue(new_name, value);
    }
    if (defs.count(rname)) {
        value = defs[rname];
        return true;
    } else {
        return false;
    }
}
