/* $Id: Resources.cc,v 1.53 2004/02/07 22:54:59 slide_o_mix Exp $ */

/**
 * @file   Resources.cc
 * @date   18-Jul-2001 00:31:22
 *
 * @brief Implementation of ResourceHandler class.
 *
 * ResourceHandler class is used for reading window manager settings.
 * Most settings are retrieved from X resource files.
 *
 */

#include "Resources.hh"
#include "Screen.hh"
#include "Net.hh"
#include "NewStyleLoader.hh"
#include "XrmStyleLoader.hh"
#include "MenuLoader.hh"
#include "Logger.hh"
#include "SimpleMatch.hh"

extern "C" {
#include <X11/Xlib.h>

#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#ifdef    HAVE_STDIO_H
#  include <stdio.h>
#endif // HAVE_STDIO_H

#ifdef    STDC_HEADERS
#  include <stdlib.h>
#  include <string.h>
#endif // STDC_HEADERS

#ifdef    HAVE_CTYPE_H
#  include <ctype.h>
#endif // HAVE_CTYPE_H

#ifdef    HAVE_LIBGEN_H
#  include <libgen.h>
#else
inline char *basename(char *name) {
    int i = strlen(name);
    for (; i >= 0 && name[i] != '/'; i--);
    if (name[i] == '/') i++;
    return &name[i];
}
#endif // HAVE_LIBGEN_H

#ifdef    PIXMAP
#  include <Imlib2.h>
#endif // PIXMAP
}

#include <iostream>
#include <string>
#include <sstream>
#include <strings.h>
#include <boost/tokenizer.hpp>

using boost::tokenizer;

using std::cerr;
using std::cout;
using std::endl;
using std::string;

/**
 * @fn    ResourceHandler(void)
 * @brief Constructor for ResourceHandler class
 *
 * Sets config file variables. Creates lists with function pointers and
 * lists for actions.
 */
ResourceHandler::ResourceHandler(Kahakai *ka, struct kaoptions *options) {
    char *__m_wastrdup_tmp;

    kahakai = ka;
    display = kahakai->display;

    linenr = 0;
    
    homedir = getenv("HOME");

    style_file = __m_wastrdup((char *) DEFAULTSTYLE);
    menu_file = __m_wastrdup((char *) DEFAULTMENU);
    script_config = __m_wastrdup((char *) DEFAULTSCRIPTCONFIG);

    rc_forced = style_forced = menu_forced = false;
    script_config_forced = false;

    if (options->rcfile) {
        rc_file = __m_wastrdup(options->rcfile);
        rc_forced = true;
    } else {
        rc_file = new char[strlen(homedir) + strlen("/.kahakai/config") + 1];
        sprintf(rc_file, "%s/.kahakai/config", homedir);
    }
    if (options->stylefile) {
        style_file = __m_wastrdup(options->stylefile);
        style_forced = true;
    }
    if (options->menufile) {
        menu_file = __m_wastrdup(options->menufile);
        menu_forced = true;
    }
    if (options->scriptconfig) {
        script_config = __m_wastrdup(options->scriptconfig);
        script_config_forced = true;
    }
}

/**
 * @fn    ~ResourceHandler(void)
 * @brief Destructor for ResourceHandler class
 *
 * Deletes all action lists and all KaActions in them.
 */
ResourceHandler::~ResourceHandler(void) {
    delete [] rc_file;
    delete [] style_file;
    delete [] menu_file;
    delete [] script_config;
}

void safeCopy(string source, string destination) {
    if (!fork()) {
        setsid();
        // -u is a GNU option for update - don't overwrite newer files.
        cout << "Copy: " << source << " Destination: " << destination << endl;
        execl("/bin/cp", "/bin/cp", 
              source.c_str(), 
              destination.c_str(), NULL);
        exit(0);
    }
}

void ResourceHandler::CheckConfigDirectory() {
    string homedir = getenv("HOME");
    string configDirectory = homedir + "/.kahakai";
    
    struct stat st;
    
    // Do nothing if directory already exists
    if (stat(configDirectory.c_str(), &st) == 0) return;
    
    // Create the directory
    mkdir(configDirectory.c_str(), 0755);

    string pkgDataDir = string(PKGDATADIR);

    safeCopy(pkgDataDir + "/home_config",    configDirectory + "/config");
    safeCopy(pkgDataDir + "/home_autostart", configDirectory + "/autostart");
    safeCopy(DEFAULTMENU,                    configDirectory + "/menu");
    safeCopy(pkgDataDir + "/home_userconfig.py",
             configDirectory + "/userconfig.py");
}

/**
 * @fn    LoadConfig(Kahakai *kahakai)
 * @brief Reads config file
 *
 * Reads, from the config file, all configuration resources which are common to 
 * all screens.
 *
 * @param kahakai Pointer to Kahakai object
 */
void ResourceHandler::LoadConfig(Kahakai *kahakai) {
    XrmValue value;
    char *value_type;
    char rc_name[30], rc_class[30];
    char *__m_wastrdup_tmp;

    CheckConfigDirectory();

    database = (XrmDatabase) 0;
    if (! (database = XrmGetFileDatabase(rc_file))) {
        if (rc_forced) logger << warn << "can't open rcfile `" << rc_file <<
                           "' for reading" << endl;
        else
            if (! (database = XrmGetFileDatabase(DEFAULTRCFILE)))
                logger << warn << "can't open system default rcfile `" << 
                    DEFAULTRCFILE << "' for reading" << endl;
    }

    kahakai->screenmask = 0;
    sprintf(rc_name, "screenMask");
    sprintf(rc_class, "ScreenMask");
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        char *token = strtok(value.addr, ", \t");
		for(; token; token = strtok(NULL, ", \t"))
            kahakai->screenmask |= (1L << atoi(token));
    } else
        kahakai->screenmask = (1L << 0) | (1L << 2) | (1L << 3);

    char *path = getenv("PATH");
    sprintf(rc_name, "scriptDir");
    sprintf(rc_class, "ScriptDir");
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
		//There was a possible buffer overflow here, user inputted data was being copied into a
		//string with sprintf!  This fix ugly, but it will work.
		char * tmp = __m_wastrdup(value.addr);
        char * sdir = environment_expansion(tmp);
		int len = strlen(sdir);

		if(strlen(tmp) > SCRIPT_DIR_MAX_LENGTH) {
                logger << warn << "script filename length exceeds maximum allowed (" 
					<< SCRIPT_DIR_MAX_LENGTH << "), it will be truncated to the maximum length" << endl;
				len = SCRIPT_DIR_MAX_LENGTH;
		}


		//What is the magic 7 for?  
        kahakai->pathenv = new char[strlen(path) + len + 7];
        snprintf(kahakai->pathenv, strlen(path) + len + 7, "PATH=%s:%s", sdir, path);
        delete [] sdir;
    }
    else {
        kahakai->pathenv =
            new char[strlen(path) + strlen(DEFAULTSCRIPTDIR) + 7];
        sprintf(kahakai->pathenv, "PATH=%s:%s", DEFAULTSCRIPTDIR, path);
    }

    sprintf(rc_name, "doubleClickInterval");
    sprintf(rc_class, "DoubleClickInterval");
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        if((sscanf(value.addr, "%lu", &kahakai->double_click)) != 1)
            kahakai->double_click = 300;
    } else
        kahakai->double_click = 300;

    if (kahakai->double_click > 999) kahakai->double_click = 999;

    XrmDestroyDatabase(database);
}

/**
 * @fn    LoadConfig(KaScreen *kascreen)
 * @brief Reads config file
 *
 * Reads all configuration resources for this screen from the config file.
 *
 * @param kascreen Screen to read resources for
 */
void ResourceHandler::LoadConfig(KaScreen *kascreen) {
    XrmValue value;
    char *value_type;
    char rc_name[30], rc_class[30];
    int sn = kascreen->screen_number;
    ScreenConfig *sc = &kascreen->config;
    char *__m_wastrdup_tmp;

    database = (XrmDatabase) 0;
    if (! (database = XrmGetFileDatabase(rc_file)))
        if (! rc_forced) database = XrmGetFileDatabase(DEFAULTRCFILE);

    sc->style_file = __m_wastrdup(style_file);
    if (! style_forced) {
        sprintf(rc_name, "screen%d.styleFile", sn);
        sprintf(rc_class, "Screen%d.StyleFile", sn);
        if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
            delete [] sc->style_file;
            sc->style_file = environment_expansion(__m_wastrdup(value.addr));
        }
    }

    sc->script_config = __m_wastrdup(script_config);
    if (! script_config_forced) {
        sprintf(rc_name, "screen%d.scriptConfig", sn);
        sprintf(rc_class, "Screen%d.ScriptConfig", sn);
        if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
            delete [] sc->script_config;
            sc->script_config = environment_expansion(__m_wastrdup(value.addr));
        }
    }

    sc->menu_file = __m_wastrdup(menu_file);
    if (! menu_forced) {
        sprintf(rc_name, "screen%d.menuFile", sn);
        sprintf(rc_class, "Screen%d.MenuFile", sn);
        if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
            delete [] sc->menu_file;
            sc->menu_file = environment_expansion(__m_wastrdup(value.addr));
        }
    }

    sprintf(rc_name, "screen%d.reloadMenu", sn);
    sprintf(rc_class, "Screen%d.ReloadMenu", sn);
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        if (! strncasecmp("true", value.addr, value.size))
            sc->menu_file_mtime = 0;
        else
            sc->menu_file_mtime = -1;
    } else
        sc->menu_file_mtime = 0;

    sprintf(rc_name, "screen%d.numberOfDesktops", sn);
    sprintf(rc_class, "Screen%d.NumberOfDesktops", sn);
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        if (sscanf(value.addr, "%u", &sc->desktops) != 1) {
            sc->desktops = 1;
        } else {
            if (sc->desktops < 1) sc->desktops = 1;
            if (sc->desktops > 16) sc->desktops = 16;
        }
    } else
        sc->desktops = 1;

    sprintf(rc_name, "screen%d.desktopNames", sn);
    sprintf(rc_class, "Screen%d.DesktopNames", sn);
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        std::string name, names = (char*)value.addr;
	logger << "Desktop Names: " << names << endl;
	int startpos = 0, endpos = 0;
	name = names.substr(0, names.find(","));
	sc->desktop_names.push_back(name);
	startpos = names.find(",");
	while(startpos != std::string::npos) {
	    startpos++;
	    endpos = names.find(",", startpos);
	    if(endpos == std::string::npos) {
	        name = names.substr(startpos);
	    } else {
	        name = names.substr(startpos, endpos - startpos);
	    }
	    logger << "Current Name: " << name << endl;
	    sc->desktop_names.push_back(trim(name));
	    startpos = endpos;
	}
        kascreen->net->SetDesktopNames(kascreen, value.addr);
    }

    sprintf(rc_name, "screen%d.virtualSize", sn);
    sprintf(rc_class, "Screen%d.VirtualSize", sn);
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        if (sscanf(value.addr, "%ux%u", &sc->virtual_x, &sc->virtual_y) != 2) {
            sc->virtual_x = sc->virtual_y = 3;
        }
    } else
        sc->virtual_x = sc->virtual_y = 3;
    if (sc->virtual_x > 20) sc->virtual_x = 20;
    if (sc->virtual_y > 20) sc->virtual_y = 20;
    if (sc->virtual_x < 1) sc->virtual_x = 1;
    if (sc->virtual_y < 1) sc->virtual_y = 1;

    sprintf(rc_name, "screen%d.doubleBufferedText", sn);
    sprintf(rc_class, "Screen%d.DoubleBufferedText", sn);
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        if (! strncasecmp("true", value.addr, value.size))
            sc->db = true;
        else
            sc->db = false;
    } else
        sc->db = true;

    sprintf(rc_name, "screen%d.lazyTransparency", sn);
    sprintf(rc_class, "Screen%d.LazyTransparency", sn);
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        if (! strncasecmp("true", value.addr, value.size))
            sc->lazy_trans = true;
        else
            sc->lazy_trans = false;
    } else
        sc->lazy_trans = true;

    sprintf(rc_name, "screen%d.colorsPerChannel", sn);
    sprintf(rc_class, "Screen%d.ColorsPerChannel", sn);
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        if (sscanf(value.addr, "%d", &sc->colors_per_channel) != 1) {
            sc->colors_per_channel = 4;
        } else {
            if (sc->colors_per_channel < 2) sc->colors_per_channel = 2;
            if (sc->colors_per_channel > 6) sc->colors_per_channel = 6;
        }
    } else
        sc->colors_per_channel = 4;

    sprintf(rc_name, "screen%d.cacheMax", sn);
    sprintf(rc_class, "Screen%d.CacheMax", sn);
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        if (sscanf(value.addr, "%lu", &sc->cache_max) != 1)
            sc->cache_max = 200;
    } else
        sc->cache_max = 200;

    sprintf(rc_name, "screen%d.imageDither", sn);
    sprintf(rc_class, "screen%d.ImageDither", sn);
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        if (! strncasecmp("true", value.addr, value.size))
            sc->image_dither = true;
        else
            sc->image_dither = false;
    } else
        sc->image_dither = true;

    sprintf(rc_name, "screen%d.menuStacking", sn);
    sprintf(rc_class, "Screen%d.MenuStacking", sn);
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        if (! strncasecmp("AlwaysAtBottom", value.addr, value.size))
            sc->menu_stacking = AlwaysAtBottom;
        else if (! strncasecmp("AlwaysOnTop", value.addr, value.size))
            sc->menu_stacking = AlwaysOnTop;
    } else
        sc->menu_stacking = NormalStacking;

    sprintf(rc_name, "screen%d.transientAbove", sn);
    sprintf(rc_class, "Screen%d.TransientAbove", sn);
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        if (! strncasecmp("true", value.addr, value.size))
            sc->transient_above = true;
        else
            sc->transient_above = false;
    } else
        sc->transient_above = true;

    sprintf(rc_name, "screen%d.focusRevertTo", sn);
    sprintf(rc_class, "Screen%d.focusRevertTo", sn);
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        if (! strncasecmp("Root", value.addr, value.size))
            sc->revert_to_window = false;
        else
            sc->revert_to_window = true;
    } else
        sc->revert_to_window = true;
    
    sprintf(rc_name, "screen%d.rootCommand", sn);
    sprintf(rc_class, "screen%d.rootCommand", sn);
    if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
        printf("Found screen root command: %s\n", value.addr);
        kascreen->config.rootCommand = value.addr;
    }
    
    unsigned int dummy;
    int dock_num;
    bool d_exists = true;
    DockStyle *dockstyle;

    for (dock_num = 0; d_exists && dock_num < 100; ++dock_num) {
        d_exists = false;
        dockstyle = new DockStyle;
    
        sprintf(rc_name, "screen%d.dock%d.geometry", sn, dock_num);
        sprintf(rc_class, "Screen%d.Dock%d.Geometry", sn, dock_num);
        if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
            dockstyle->geometry = XParseGeometry(value.addr, &dockstyle->x,
                                                 &dockstyle->y, &dummy,
                                                 &dummy);
            d_exists = true;
        } else
            dockstyle->geometry = XParseGeometry("-0+0", &dockstyle->x,
                                                 &dockstyle->y, &dummy,
                                                 &dummy);

        sprintf(rc_name, "screen%d.dock%d.order", sn, dock_num);
        sprintf(rc_class, "Screen%d.Dock%d.Order", sn, dock_num);
        if (XrmGetResource(database, rc_name, rc_class, &value_type, &value)) {
            d_exists = true;
            tokenizer<> tok(string(value.addr));
            for(tokenizer<>::iterator it=tok.begin(); it != tok.end(); ++it) {
                dockstyle->order.push_back(new SimpleMatch(*it));
            }
        }
        
        sprintf(rc_name, "screen%d.dock%d.desktopMask", sn, dock_num);
        sprintf(rc_class, "Screen%d.Dock%d.DesktopMask", sn, dock_num);
        if (XrmGetResource(database, rc_name, rc_class, &value_type,
                           &value)) {
            d_exists = true;
            if (! strncasecmp("all", value.addr, 3))
                dockstyle->desktop_mask = (1L << 16) - 1;
            else {
                dockstyle->desktop_mask = 0;
                char *token = strtok(value.addr, " \t");
                while (token) {
                    int desk = (unsigned int) atoi(token);
                    if (desk < 16) dockstyle->desktop_mask |= (1L << desk);
                    token = strtok(NULL, " \t");
                }
            }
        } else
            dockstyle->desktop_mask = (1L << 16) - 1;

        sprintf(rc_name, "screen%d.dock%d.centered", sn, dock_num);
        sprintf(rc_class, "Screen%d.Dock%d.Centered", sn, dock_num);
        if (XrmGetResource(database, rc_name, rc_class, &value_type,
                           &value)) {
            d_exists = true;
            if (! strncasecmp("true", value.addr, value.size))
                dockstyle->centered = true;
            else
                dockstyle->centered = false;
        } else
            dockstyle->centered = false;

        sprintf(rc_name, "screen%d.dock%d.inworkspace", sn, dock_num);
        sprintf(rc_class, "Screen%d.Dock%d.Inworkspace", sn, dock_num);
        if (XrmGetResource(database, rc_name, rc_class, &value_type,
                           &value)) {
            d_exists = true;
            if (! strncasecmp("true", value.addr, value.size))
                dockstyle->inworkspace = true;
            else
                dockstyle->inworkspace = false;
        } else
            dockstyle->inworkspace = false;

        sprintf(rc_name, "screen%d.dock%d.direction", sn, dock_num);
        sprintf(rc_class, "Screen%d.Dock%d.Direction", sn, dock_num);
        if (XrmGetResource(database, rc_name, rc_class, &value_type,
                           &value)) {
            d_exists = true;
            if (! strncasecmp("Horizontal", value.addr, value.size))
                dockstyle->direction = HorizontalDock;
            else
                dockstyle->direction = VerticalDock;
        } else
            dockstyle->direction = VerticalDock;

        sprintf(rc_name, "screen%d.dock%d.gridSpace", sn, dock_num);
        sprintf(rc_class, "Screen%d.Dock%d.GridSpace", sn, dock_num);
        if (XrmGetResource(database, rc_name, rc_class, &value_type,
                           &value)) {
            d_exists = true;
            if (sscanf(value.addr, "%u", &dockstyle->gridspace) != 1)
                dockstyle->gridspace = 2;
        } else
            dockstyle->gridspace = 2;
        
        if (dockstyle->gridspace > 50) dockstyle->gridspace = 50;

        sprintf(rc_name, "screen%d.dock%d.stacking", sn, dock_num);
        sprintf(rc_class, "Screen%d.Dock%d.Stacking", sn, dock_num);
        if (XrmGetResource(database, rc_name, rc_class, &value_type,
                           &value)) {
            d_exists = true;
            if (! strncasecmp("AlwaysAtBottom", value.addr, value.size))
                dockstyle->stacking = AlwaysAtBottom;
            else
                dockstyle->stacking = AlwaysOnTop;
        } else
            dockstyle->stacking = AlwaysOnTop;
        
        if (d_exists || ! dock_num)
            kascreen->style.dockStyles.push_back(*dockstyle);
        else
            delete dockstyle;
    }

    XrmDestroyDatabase(database);
}

/**
 * @fn    LoadStyle(KaScreen *kascreen)
 * @brief Reads style file
 *
 * Reads a style resources from a style file.
 *
 * @param kascreen KaScreen to load style for
 */
void ResourceHandler::LoadStyle(KaScreen *kascreen) {
    StyleLoader* styleLoader;
    if (strstr(kascreen->config.style_file, ".kastyle")) {
        styleLoader = new NewStyleLoader;
    } else {
        styleLoader = new XrmStyleLoader;
    }
    styleLoader->loadStyle(kascreen);

    delete styleLoader;
}

/**
 * @fn    LoadMenus(KaScreen *kascreen)
 * @brief Reads menu file
 *
 * Creates menus by parsing the menu file.
 *
 * @param kascreen KaScreen to load menus for
 */
void ResourceHandler::LoadMenus(KaScreen *kascreen) {
    MenuLoader menuLoader(kascreen->config.menu_file);

    menuLoader.loadMenus(kascreen);
}

/**
 * @fn    ParseMenu(KaMenu &menu, FILE *file, KaScreen &kascreen)
 * @brief Parses a menu file
 *
 * Parses a menu section of the menu file and creates a menu object for the
 * menu. If a [start] or [begin] statement is found when parsing a menu, we
 * make a recursive function call to this function. This makes it possible to
 * to define a submenu within a the menu itself.
 *
 * @param menu Menu to add items to
 * @param file File descriptor for menu file
 *
 * @return Pointer to new menu if menu was sucessfully parsed,
 *         otherwise NULL
 */
KaMenu *ResourceHandler::ParseMenu(KaMenu *menu, FILE *file,
                                   KaScreen* kascreen) {
    MenuLoader menuLoader(kascreen->config.menu_file);

    return menuLoader.parseMenu(menu, file, kascreen);
}        

/**
 * @fn    strtrim(char *s)
 * @brief Trims a string
 *
 * Removes leading and trailing spaces, tabs and newlines.
 *
 * @param s String we want to trim
 *
 * @return Trimmed string
 */
char *strtrim(char *s) {
    for (; *s != '\0' && *s == ' ' || *s == '\t' || *s == '\n'; s++);
    if (*s == '\0') return s;
    while (s[strlen(s) - 1] == ' ' ||
           s[strlen(s) - 1] == '\t' ||
           s[strlen(s) - 1] == '\n')
        s[strlen(s) - 1] = '\0';
    return s;
}

/**
 * @fn    environment_expansion(char *s)
 * @brief Expand '~' and environment variables
 *
 * Replaces all '~' characters with HOME environment variable and all
 * $ENVIRONMENT_VARIABLE with environment variable value.
 *
 * @param s String to expand
 * @return Pointer to expanded string
 */
char *environment_expansion(char *s) {
    char *tmp, *env, *env_name;
    int i, tmp_char;

    for (i = 0; s[i] != '\0'; i++) {
        switch (s[i]) {
            case '\\':
                if (s[i + 1] != '\0') i++;
                break;
            case '$':
                if (IS_ENV_CHAR(s[i + 1])) {
                    s[i] = '\0';
                    env_name = &s[++i];
                    for (; IS_ENV_CHAR(s[i]); i++);
                    tmp_char = s[i];
                    s[i] = '\0';
                    if ((env = getenv(env_name)) == NULL) env = "";
                    s[i] = tmp_char;
                    tmp = new char[strlen(s) + strlen(env) +
                                   strlen(&s[i]) + 1];
                    sprintf(tmp, "%s%s%s", s, env, &s[i]);
                    i = strlen(s) + strlen(env);
                    delete [] s;
                    s = tmp;
                }
                break;
            case '~': 
                s[i] = '\0';
                if ((env = getenv("HOME")) == NULL) env = "~";
                tmp = new char[strlen(s) + strlen(env) +
                               strlen(&s[i + 1]) + 1];
                sprintf(tmp, "%s%s%s", s, env, &s[i + 1]);
                i = strlen(s) + strlen(env);
                delete [] s;
                s = tmp;
                break;
        }
    }
    return s;
}

