/* $Id: Screen.cc,v 1.100 2004/02/07 22:43:44 slide_o_mix Exp $ */

/**
 * @file   Screen.cc
 * @date   25-Jul-2001 23:26:22
 *
 * @brief Implementation of KaScreen and ScreenEdge classes
 *
 * A KaScreen object handles one X server screen. A ScreenEdge is a
 * transperant window placed at the edge of the screen, good to use for
 * virtual screen scrolling.
 *
 */

#include "Action.hh"
#include "ChildWindow.hh"
#include "DesktopMenu.hh"
#include "Dockapp.hh"
#include "Event.hh"
#include "ImageControl.hh"
#include "Logger.hh"
#include "Menu.hh"
#include "MenuItem.hh"
#include "Net.hh"
#include "Resources.hh"
#include "Screen.hh"
#include "Script.hh"
#include "ScriptFactory.hh"
#include "SimpleMatch.hh"
#include "Timer.hh"
#include "WindowMenu.hh"

extern "C" {
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#ifdef    XRANDR
#  include <X11/extensions/Xrandr.h>
#endif // XRANDR

#ifdef    HAVE_STDIO_H
#  include <stdio.h>
#endif // HAVE_STDIO_H

#ifdef    HAVE_UNISTD_H
#  include <sys/types.h>
#  include <sys/stat.h>
#  include <sys/wait.h>
#  include <unistd.h>
#endif // HAVE_UNISTD_H

#ifdef    STDC_HEADERS
#  include <stdlib.h>
#endif // STDC_HEADERS

#ifdef    HAVE_SIGNAL_H
#  include <signal.h>
#endif // HAVE_SIGNAL_H
}

#include <algorithm>
#include <sstream>
#include <iostream>
#include <iomanip>

using std::cout;
using std::cerr;
using std::endl;
using std::list;
using std::vector;
using std::make_pair;
using std::setprecision;
using std::stringstream;
using std::min;

#define EVENT_FUNCTION(name) void ScreenEv##name::operator()(KaScreen* self, XEvent* event, KaAction* action)

EVENT_FUNCTION(GoToDesktop) {
    if (action->param) self->GoToDesktop((unsigned int) atoi(action->param));
}

EVENT_FUNCTION(ViewportFixedMove) {
    int x, y, mask;
    unsigned int w = 0, h = 0;

    if (! action->param) return;

    mask = XParseGeometry(action->param, &x, &y, &w, &h);
    if (mask & XNegative) x = self->v_xmax + x;
    if (mask & YNegative) y = self->v_ymax + y;
    self->MoveViewportTo(x, y);
}


/**
 * @fn    KaScreen(Display *d, int scrn_number, Kahakai *ka)
 * @brief Constructor for KaScreen class
 *
 * Sets root window input mask. Creates new image control object and reads
 * style file. Then we create fonts, colors and renders common images.
 * Last thing we do in this functions is to create KaWindows for all windows
 * that should be managed.
 *
 * @param d The display
 * @param scrn_number Screen to manage
 * @param ka Kahakai object
 */
KaScreen::KaScreen(Display *d, int scrn_number, Kahakai *ka) :
    WindowObject(0, RootType) {
    int eventmask;
    XSetWindowAttributes attrib_set;
    struct timeval tv_start, tv_end;
    long elapsed_millisec;

    gettimeofday(&tv_start, NULL);

    display = d;
    screen_number = scrn_number;
    id = RootWindow(display, screen_number);
    visual = DefaultVisual(display, screen_number);
    colormap = DefaultColormap(display, screen_number);
    screen_depth = DefaultDepth(display, screen_number);
    width = DisplayWidth(display, screen_number);
    height = DisplayHeight(display, screen_number);

    kahakai = ka;
    net = kahakai->net;
    rh = ka->rh;
    focus = true;
    shutdown = false;
    imod_mask = 0;

    focus_freeze = false;
    user_focus_freeze = false;

    default_font.xft = false;
    default_font.fontNames.push_back("fixed");

    config.menu_stacking = NormalStacking;
    config.snapping = false;
    config.resist = false;
    config.snapDistance = 15;
    config.resistDistance = 30;
    config.moveresize_crosshair = false;

    XSync(display, false);
	//how is ka->options->display different from 'display'?
    if (! (pdisplay = XOpenDisplay(ka->options->display))) {
        logger << err << "can't open display: " << ka->options->display << endl;
        exit(1);
    }

    net->GetXRootPMapId(this);
    ic->setXRootPMapId((xrootpmap_id)? true: false);
    
    eventmask = SubstructureRedirectMask | StructureNotifyMask |
        PropertyChangeMask | ColormapChangeMask | KeyPressMask |
        KeyReleaseMask | ButtonPressMask | ButtonReleaseMask |
        EnterWindowMask | LeaveWindowMask | FocusChangeMask;

    sprintf(displaystring, "DISPLAY=%s", DisplayString(display));
    sprintf(displaystring + strlen(displaystring) - 1, "%d", screen_number);

    XSetErrorHandler((XErrorHandler) wmrunningerror);
    XSelectInput(display, id, eventmask);
    XSync(display, false);
    XSync(pdisplay, false);
    XSetErrorHandler((XErrorHandler) xerrorhandler);
    if (kahakai->wmerr) {
        logger << warn << "another window manager is running on " <<
            displaystring + 8 << endl;
        return;
    }

    v_x = v_y = 0;

    int event_basep, error_basep;
    render_extension =
        XRenderQueryExtension(pdisplay, &event_basep, &error_basep);

#ifdef XRANDR
    XRRSelectInput(display, id, RRScreenChangeNotifyMask);
#endif // XRANDR

    rh->LoadConfig(this);
    for(unsigned int i = config.desktop_names.size(); config.desktops > i; i++) {
        char buf[20];
	sprintf(buf, "Desktop %d", i + 1);
	config.desktop_names.push_back(string(buf));
    }

    current_desktop = new Desktop(0, width, height, config.desktop_names[0]);
    desktop_list.push_back(current_desktop);
    net->SetWorkarea(this);

    for (unsigned int i = 1; config.desktops > i; i++)
        desktop_list.push_back(new Desktop(i, width, height, config.desktop_names[i]));

    kahakai->window_table.insert(make_pair(id, this));

    attrib_set.override_redirect = true;
    wm_check = XCreateWindow(display, id, 0, 0, 1, 1, 0,
                             CopyFromParent, InputOnly, CopyFromParent,
                             CWOverrideRedirect, &attrib_set);
    net->SetSupportedWMCheck(this, wm_check);
    net->SetSupported(this);

    ic = new KaImageControl(pdisplay, this, config.image_dither,
                            config.colors_per_channel, config.cache_max);
    ic->installRootColormap();

    rh->LoadStyle(this);

    if (config.rootCommand != "")
        kaexec(config.rootCommand.c_str(), displaystring);
    
    CreateFonts();
    CreateColors();
    XDefineCursor(display, id, kahakai->session_cursor);

    v_xmax = (config.virtual_x - 1) * width;
    v_ymax = (config.virtual_y - 1) * height;

    edgemgr = new ScreenEdgeManager(this);

    net->SetDesktopGeometry(this);
    net->SetNumberOfDesktops(this);
    net->GetCurrentDesktop(this);
    net->SetCurrentDesktop(this);
    net->GetDesktopViewPort(this);
    net->SetDesktopViewPort(this);

    control = new ImageControl(this, display, 
                               screen_number, 
                               ic->getVisual(), 
                               ic->getColormap());    


    gettimeofday(&tv_end, NULL);
    elapsed_millisec = (tv_end.tv_sec - tv_start.tv_sec)*1000 + 
                       (tv_end.tv_usec - tv_start.tv_usec)/1000;

    logger << "Before script configuration: "
           << setprecision(1) << elapsed_millisec/1000.0
           << " seconds" << endl;

    LoadScriptConfig();

    gettimeofday(&tv_end, NULL);
    elapsed_millisec = (tv_end.tv_sec - tv_start.tv_sec)*1000 + 
                       (tv_end.tv_usec - tv_start.tv_usec)/1000;

    logger << "After script configuration: "
           << setprecision(3) << elapsed_millisec/1000.0
           << " seconds" << endl;

    delete [] config.style_file;

    vector<DockStyle>::iterator dit = style.dockStyles.begin();
    for (; dit != style.dockStyles.end(); ++dit) {
        docks.push_back(new DockappHandler(this, &*dit));
    }

    CreateWindows();
    CreateMenus();

    LISTDEL(mreqs);
    net->GetClientListStacking(this);
    net->SetClientList(this);
    net->GetActiveWindow(this);

    actionlist = &config.rootacts;

    gettimeofday(&tv_end, NULL);
    elapsed_millisec = (tv_end.tv_sec - tv_start.tv_sec)*1000 +
                       (tv_end.tv_usec - tv_start.tv_usec)/1000;

    logger << "Screen " << scrn_number << " loaded in " << setprecision(3)
           << elapsed_millisec/1000.0 << " seconds" << endl;

    // Check if edges have any actions assigned, if not unmap them
    if (west->actionlist->empty())  XUnmapWindow(display,  west->id);
    if (east->actionlist->empty())  XUnmapWindow(display,  east->id);
    if (north->actionlist->empty()) XUnmapWindow(display, north->id);
    if (south->actionlist->empty()) XUnmapWindow(display, south->id);
    
}

void KaScreen::CreateMenus() {
    window_menu = new WindowMenu();
    kamenu_list.push_back(window_menu);
    kamenu_list.push_back(new DesktopMenu(new ScreenEvGoToDesktop(),
                                          "__gotodesktoplist__"));
    kamenu_list.push_back(new DesktopMenu(new WindowEvPartCurrentJoinDesktop(),
                                          "__senddesktoplist__"));
    kamenu_list.push_back(new DesktopMenu(new WindowEvJoinDesktop(), 
                                          "__joindesktoplist__"));
    kamenu_list.push_back(new DesktopMenu(new WindowEvPartDesktop(), 
                                          "__partdesktoplist__", true));
    kamenu_list.push_back(new MinimizedWindowMenu());


    rh->LoadMenus(this);
        
    vector<KaMenu *>::iterator mit = kamenu_list.begin();
    for (; mit != kamenu_list.end(); ++mit)
        (*mit)->Build(this);
}

void KaScreen::LoadScriptConfig() {

    logger << "Loading script configuration from "
           << config.script_config << endl;

    script = ScriptFactory::newScript(config.script_config);
    kahakai->script_current_screen = this;

    if (script == NULL) {
        logger << warn << "Couldn't create script: "
               << config.script_config << endl;

    } else if (! script->run()) {

        logger << warn << "Couldn't execute script:"
               << script->getFilename() << endl;

        logger << warn << "Loading default configuration " << endl;

        delete script;
        script = ScriptFactory::newScript(DEFAULTSCRIPTCONFIG);
        if (!script->run()) {
            logger << warn << "Loading default configuration failed. Aborting."
                   << endl;
            exit(1);
        }
    }

    logger << "Finished loading script configuration" << endl;
}

void KaScreen::CreateWindows() {
    Window ro, pa, * children;
    XWindowAttributes attr;
    unsigned int nchild;
    XQueryTree(display, id, &ro, &pa, &children, &nchild);

    for (int i = 0; i < (int) nchild; ++i) {
        bool status = false;

        XGrabServer(display);

        if (validatedrawable(id)) {
            XGetWindowAttributes(display, children[i], &attr);
            status = true;
        }

        XUngrabServer(display);

        if (!status)
            continue;

        if (attr.override_redirect)
            continue;

        if (net->IsSystrayWindow(children[i])) {
            if (! (kahakai->FindWin(children[i], SystrayType))) {
                XGrabServer(display);
                if (validatedrawable(children[i])) {
                    XSelectInput(display, children[i], StructureNotifyMask);
                }
                XUngrabServer(display);
                SystrayWindow *stw = new SystrayWindow(children[i], this);
                kahakai->window_table.insert(make_pair(children[i], stw));
                systray_window_list.push_back(children[i]);
                net->SetSystrayWindows(this);
            }
            continue;
        }

        XWMHints *wm_hints = NULL;
        XGrabServer(display);

        if (validatedrawable(children[i]))
            wm_hints = XGetWMHints(display, children[i]);

        XUngrabServer(display);

        // check if it's hidden and if so, skip it

        bool hidden = false;

        if (attr.map_state != IsViewable && wm_hints &&
            wm_hints->initial_state != IconicState)
            hidden = true;

        else if (attr.map_state != IsViewable && !wm_hints)
            hidden = true;

        if (hidden || !wm_hints) {
            // window that doesn't want to be mapped
            if (wm_hints)
                XFree(wm_hints);
            continue;
        }

        // add it to a dock, if it's a dockapp

        if (attr.map_state == IsViewable && wm_hints &&
            (wm_hints->initial_state == WithdrawnState))
            AddDockapp(children[i]);

        // else it's probably a normal window, map it

        else if (kahakai->window_table.find(children[i])
                   == kahakai->window_table.end()) {

            KaWindow *newwin = new KaWindow(children[i], this);

            if (kahakai->FindWin(children[i], WindowType)) {
                bool minimize = false;

                if ((wm_hints) && wm_hints->initial_state == IconicState)
                    minimize = true;

                if (newwin->flags.hidden)
                    minimize = true;

                if (minimize)
                    newwin->Minimize(true);
                else
                    newwin->net->SetState(newwin, NormalState);
            }
        }
        if (wm_hints) XFree(wm_hints);
    }
    XFree(children);
}

ScreenEdgeManager::ScreenEdgeManager(KaScreen * screen) {

    n = new ScreenEdge(screen, 0, 0, screen->width, 2, NEdgeType);
    s = new ScreenEdge(screen, 0, screen->height-2, screen->width, 2,SEdgeType);
    e = new ScreenEdge(screen, screen->width-2, 0, 2, screen->height,EEdgeType);
    w = new ScreenEdge(screen, 0, 0, 2, screen->height, WEdgeType);

    n->actionlist = &screen->config.neacts;
    s->actionlist = &screen->config.seacts;
    e->actionlist = &screen->config.eeacts;
    w->actionlist = &screen->config.weacts;

    screen->north = n;
    screen->south = s;
    screen->east = e;
    screen->west = w;

    screen->RestackWindows(0);
}

ScreenEdgeManager::~ScreenEdgeManager() {
    delete n;
    delete s;
    delete e;
    delete w;
}

/**
 * @fn    ~KaScreen(void)
 * @brief Destructor for KaScreen class
 *
 * Deletes all created colors and fonts.
 */
KaScreen::~KaScreen(void) {
    KaChildWindow *wc;
    shutdown = true;
    XSelectInput(display, id, NoEventMask);
    net->DeleteSupported(this);
    XDestroyWindow(display, wm_check);

    if (script) delete script;

    LISTDEL(docks);

    list<KaWindow *> delstack;

    DESKTOP_LOOP(this) {
        list<Window>::reverse_iterator it = (*__dit)->aab_stacking_list.rbegin();
        for (; it != (*__dit)->aab_stacking_list.rend(); ++it) {
            wc = (KaChildWindow *) kahakai->FindWin(*it, FrameType);
            if (wc && ((find(delstack.begin(), delstack.end(), wc->ka) == delstack.end())))
                delstack.push_back(wc->ka);
        }
        it = (*__dit)->stacking_list.rbegin();
        for (; it != (*__dit)->stacking_list.rend(); ++it) {
            wc = (KaChildWindow *) kahakai->FindWin(*it, FrameType);
            if (wc && ((find(delstack.begin(), delstack.end(), wc->ka) == delstack.end())))
                delstack.push_back(wc->ka);
        }
        it = (*__dit)->aot_stacking_list.rbegin();
        for (; it != (*__dit)->aot_stacking_list.rend(); ++it) {
            wc = (KaChildWindow *) kahakai->FindWin(*it, FrameType);
            if (wc && ((find(delstack.begin(), delstack.end(), wc->ka) == delstack.end())))
                delstack.push_back(wc->ka);
        }
    }

    list<KaWindow *>::iterator kit = delstack.begin();
    for (; kit != delstack.end(); ++kit)
        delete *kit;
    
    LISTCLEAR(kawindow_list);
    LISTCLEAR(kawindow_list_map_order);
    
    LISTDEL(strut_list);
    
    vector<ButtonStyle>::iterator bit = style.window.buttonStyles.begin();
    for (; bit != style.window.buttonStyles.end(); ++bit) {
        ButtonStyle& bs = *bit;
 
        if (bs.fg) {
            XFreeGC(display, bs.focused.gc);
            XFreeGC(display, bs.focusedActive.gc);
            XFreeGC(display, bs.unfocused.gc);
            XFreeGC(display, bs.unfocusedActive.gc);
            XFreeGC(display, bs.pressed.gc);
            XFreeGC(display, bs.pressedActive.gc);
        }
    }

#ifdef PIXMAP
    //    imlib_context_free(imlib_context);
#endif // PIXMAP
       
    delete ic;
    delete edgemgr;

    delete [] config.menu_file;
   
    XSync(display, false);
    XSync(pdisplay, false);
    XCloseDisplay(pdisplay);
    kahakai->window_table.erase(id);
}

/**
 * @fn    raiseWindowInList(Window win,
                            list<Window>& stacking,
                            Desktop *desktop)
 * @brief Raises window specified by 'win' in the stacking list 'stacking'.
 *
 * Auxiliary function for RaiseWindow.
 *
 * @param win      Window either in 'stacking' list or not
 * @parma stacking Usually one of aot_stacking_list, aab_stacking_list, 
 *                 stacking_list
 *
 * @return true if window was found, false if not.
 */

bool KaScreen::raiseWindowInList(Window win, list<Window>& stacking, Desktop *desktop)
{
    list<Window>::iterator it = find(stacking.begin(), stacking.end(), win);

    if (it == stacking.end()) return false;

    stacking.erase(it);
    stacking.push_front(*it);

    KaChildWindow *wc = (KaChildWindow *) kahakai->FindWin(*it, FrameType);

    // We return true even if wc is NULL, because other stacking lists contain
    // the same Window so the result will be the same if they try again.
    if (!wc) return true;

    KaWindow *ww = wc->ka;
    if (! ww->transients.empty()) {
        list<Window>::iterator tit = ww->transients.begin();
        for (; tit != ww->transients.end(); ++tit) {
            KaWindow *wt = (KaWindow *)  kahakai->FindWin(*tit, WindowType);
            if (wt) {
                desktop->stacking_list.remove(wt->frame->id);
                desktop->aab_stacking_list.remove(wt->frame->id);
                desktop->aot_stacking_list.remove(wt->frame->id);
                stacking.push_front(wt->frame->id);
            }
            else {
                ww->transients.erase(tit);
                tit = ww->transients.begin();
            }
        }
    }

    return true;
}

/**
 * @fn    RaiseWindow(Window win)
 * @brief Raises window on the current desktop
 *
 * Raises a window in the display stack, still keeping it in the same stacking
 * layer. Used when a user (or kahakai) requests a window raise.
 *
 * @param win Window to raise, win equal to zero will restacks all windows
 */
void KaScreen::RaiseWindow(Window win) {
    raiseWindowInList(win, current_desktop->aot_stacking_list, current_desktop) || 
        raiseWindowInList(win, current_desktop->stacking_list, current_desktop) || 
        raiseWindowInList(win, current_desktop->aab_stacking_list, current_desktop);

    RestackWindows(win);
    net->SetClientListStacking(this);
}

/**
 * @fn    RaiseWindowAllDesktops(KaWindow *win)
 * @brief Raises window on all desktops
 *
 * Raises a window in the display stack, still keeping it in the same stacking
 * layer. Used when an application request a window raise.
 *
 * @param win KaWindow to raise
 */
void KaScreen::RaiseWindowAllDesktops(KaWindow *win) {
    DESKTOP_LOOP_MASK(this, win->desktop_mask) {
        raiseWindowInList(win->frame->id, (*__dit)->aot_stacking_list, *__dit) || 
          raiseWindowInList(win->frame->id, (*__dit)->stacking_list, *__dit) || 
          raiseWindowInList(win->frame->id, (*__dit)->aab_stacking_list, *__dit);
    }
    if (win->desktop_mask & (1L << current_desktop->number))
        RestackWindows(win->frame->id);
    net->SetClientListStacking(this);
}

/**
 * @fn    lowerWindowInList(Window win, Desktop *desktop)
 * @brief Lowers window specified by 'win' in the specified desktop
 *
 * Auxiliary function for LowerWindow.
 *
 * @param win      Window either in 'stacking' list or not
 * @param desktop  A pointer to a desktop
 */
void KaScreen::lowerWindowOnDesktop(Window win, Desktop *desktop)
{
    list<Window>::iterator it = find(desktop->aot_stacking_list.begin(),
                                    desktop->aot_stacking_list.end(), win);
    if (it != desktop->aot_stacking_list.end()) {
        desktop->aot_stacking_list.erase(it);
        desktop->aot_stacking_list.push_back(win);
        return;
    }
    it = find(desktop->stacking_list.begin(),
              desktop->stacking_list.end(), win);
    if (it != desktop->stacking_list.end()) {
        desktop->stacking_list.erase(it);
        desktop->stacking_list.push_back(win);
        return;
    }
    it = find(desktop->aab_stacking_list.begin(),
             desktop->aab_stacking_list.end(), win);
    if (it != desktop->aab_stacking_list.end()) {
        desktop->aab_stacking_list.erase(it);
        desktop->aab_stacking_list.push_back(win);
    }
}

/**
 * @fn    LowerWindow(Window win)
 * @brief Lowers window on current desktop
 *
 * Lowers a window in the display stack, still keeping it in the same stacking
 * layer.
 *
 * @param win Window to lower, win equal to zero will restacks all windows
 */
void KaScreen::LowerWindow(Window win) {
    lowerWindowOnDesktop(win, current_desktop);
    RestackWindows(win);
    net->SetClientListStacking(this);
}

/**
 * @fn    LowerWindowAllDesktops(KaWindow *win)
 * @brief Lowers window on all desktops
 *
 * Lowers a window in the display stack, still keeping it in the same stacking
 * layer.
 *
 * @param win KaWindow to lower
 */
void KaScreen::LowerWindowAllDesktops(KaWindow *win) {
    DESKTOP_LOOP_MASK(this, win->desktop_mask)
        lowerWindowOnDesktop(win->frame->id, *__dit);
    if (win->desktop_mask & (1L << current_desktop->number))
        RestackWindows(win->frame->id);
    net->SetClientListStacking(this);
}

/**
 * @fn    RestackWindows(Window win)
 * @brief Updates window stacking
 *
 * Update the display stacking order above window win. If win is equal to zero
 * then the complete stacking order is updated.
 *
 * @param win Window to to update stacking order above, win equal to zero will
 *            restacks all windows
 */
void KaScreen::RestackWindows(Window win) {
    int i = 0;
    bool end = false;
    
    Window *stack = new Window[current_desktop->aot_stacking_list.size() +
                              current_desktop->stacking_list.size() +
                              current_desktop->aab_stacking_list.size() + 4];

    stack[i++] = west->id;
    stack[i++] = east->id;
    stack[i++] = north->id;
    stack[i++] = south->id;
    
    list<Window>::iterator it = current_desktop->aot_stacking_list.begin();
    for (; it != current_desktop->aot_stacking_list.end(); ++it) {
        stack[i++] = *it;
        if (*it == win) { end = true; break; }
    }
    if (! end) {
        it = current_desktop->stacking_list.begin();
        for (; it != current_desktop->stacking_list.end(); ++it) {
            stack[i++] = *it;
            if (*it == win) { end = true; break; }
        }
    }
    if (! end) {
        it = current_desktop->aab_stacking_list.begin();
        for (; it != current_desktop->aab_stacking_list.end(); ++it) {
            stack[i++] = *it;
            if (*it == win) { end = true; break; }
        }
    }
    XRaiseWindow(display, stack[0]);
    XRestackWindows(display, stack, i);
        
    delete [] stack;
}

/**
 * @fn    UpdateCheckboxes(int type)
 * @brief Updates menu checkboxes
 *
 * Redraws all checkbox menu items to make sure they are synchronized with
 * their given flag.
 *
 * @param type Type of checkboxes to update
 */
void KaScreen::UpdateCheckboxes(int type) {
    list<KaMenuItem *>::iterator miit;

    if (! kahakai->eh) return;

    vector<KaMenu *>::iterator mit = kamenu_list.begin();
    for (; mit != kamenu_list.end(); ++mit) {
        if ((*mit)->mapped) (*mit)->Render();
    }
}

/** 
 * @fn   ReloadMenu(void)
 * @brief Reload the menu file
 *
 * Reloads the menu file (the one specified in the config file)
 */
void KaScreen::ReloadMenu(void) {
    // Make a list of menus we should keep, and destroy all others
    KaMenu **keepstack = new KaMenu*[kamenu_list.size()];
    int i, stackp = 0;
    vector<KaMenu *>::iterator menu_it = kamenu_list.begin();
    for (; menu_it != kamenu_list.end(); ++menu_it)
        if (((*menu_it)->ext_type != NoExtMenuType))
            keepstack[stackp++] = *menu_it;
        else {
            if ((*menu_it)->mapped)
                (*menu_it)->Unmap(true);
            delete (*menu_it);
        }
    // Clear the menu list
    kamenu_list.clear();
        
    // Load the menus
    rh->LoadMenus(this);

    // Re-add the menus we kept
    for (i = 0; i < stackp; ++i)
       kamenu_list.push_back(keepstack[i]);

    delete [] keepstack;

    for (menu_it = kamenu_list.begin(); menu_it != kamenu_list.end(); ++menu_it)
        if (! (*menu_it)->built)
            (*menu_it)->Build(this);

    return;
}

/**
 * @fn    CheckMenuMTime(void)
 * @brief Check if the menu needs to be reloaded, and if so, do so
 *
 * Compares the mtime of the menu file against the stored mtime, and if
 * necessary, reload the menus.
 *
 * @return true if menu has been reloaded, false if not
 */
bool KaScreen::CheckMenuMTime(void) {
    if (config.menu_file_mtime == -1)
        return false;
    
    struct stat _s;

    // Check the mtime of the menu file
    if (stat(config.menu_file, &_s)) {
        logger << warn << "`" << config.menu_file << "` couldn't check mtime" << endl;
        config.menu_file_mtime = -1;
        return false;
    }
    if (_s.st_mtime == config.menu_file_mtime)
        return false;

    config.menu_file_mtime = _s.st_mtime;

    ReloadMenu();

    return true;
}

/**
 * @fn    GetMenuNamed(string menuName)
 * @brief Find a menu
 *
 * Searches through menu list after a menu named as 'menu' parameter.
 *
 * @param menu menu name to use for search
 *
 * @return Pointer to menu object if a menu was found, if no menu was found
 *         NULL is returned
 */
KaMenu *KaScreen::GetMenuNamed(string menuName) {
    if (menuName == "") return NULL;
   
    vector<KaMenu *>::iterator menu_it = kamenu_list.begin();
    for (; menu_it != kamenu_list.end(); ++menu_it)
        if ((*menu_it)->name == menuName) {
            if ((*menu_it)->built && (! (*menu_it)->mapped) && CheckMenuMTime())
                return GetMenuNamed(menuName);
            return *menu_it;
    }

    menuName = trim(menuName);
    if (menuName[0] == '!' && menuName.length() > 1) {
        return CreateDynamicMenu(menuName);
    }
   
    logger << warn << "`" << menuName << "' unknown menu" << endl;
    return NULL;
} 

/**
 * @fn    CreateDynamicMenu(char *name)
 * @brief Creates a dynamic menu
 *
 * Executes command line and parses standard out as a menu file.
 *
 * @param name Name of dynamic menu to create
 *
 * @return Created menu, NULL if error occured inmenu parsing
 */
KaMenu *KaScreen::CreateDynamicMenu(string name) {
    char *tmp_argv[128];
    int m_pipe[2];
    KaMenu *dmenu;
    int pid, status, i;
    char *allocname = NULL;
    char *__m_wastrdup_tmp;

    for (i = 0; name[i] != '\0' && name[i] != '!'; i++);
    if (name[i] == '!' && name[i + 1] != '\0') {
        allocname = __m_wastrdup(&name[i + 1]);
        commandline_to_argv(allocname, tmp_argv);
    } else
        return NULL;

    if (pipe(m_pipe) < 0) {
        logger << warn << "pipe: " << strerror(errno) << endl;
    }
    else {
        struct sigaction action;

        action.sa_handler = SIG_DFL;
        action.sa_mask = sigset_t();
        action.sa_flags = 0;
        sigaction(SIGCHLD, &action, NULL);
        pid = fork();
        if (pid == 0) {
            dup2(m_pipe[1], STDOUT_FILENO);
            close(m_pipe[0]);
            close(m_pipe[1]);
            putenv(kahakai->pathenv);
            if (execvp(*tmp_argv, tmp_argv) < 0)
                logger << warn << *tmp_argv << ": command not found" << endl;
            close(STDOUT_FILENO);
            exit(127);
        }
        close(m_pipe[1]);
        rh->linenr = 0;
        char *old_menu_file = config.menu_file;
        config.menu_file = new char[strlen(*tmp_argv) + 8];
        sprintf(config.menu_file, "%s:STDOUT", *tmp_argv);
        dmenu = new KaMenu(name);
        dmenu->dynamic = dmenu->dynamic_root = true;
        FILE *fd = fdopen(m_pipe[0], "r");
        dmenu = rh->ParseMenu(dmenu, fd, this);
        fclose(fd);
        if (waitpid(pid, &status, 0) == -1)
            logger << warn << "waitpid: " << strerror(errno) << endl;
        delete [] config.menu_file;
        config.menu_file = old_menu_file;
        action.sa_handler = signalhandler;
        action.sa_flags = SA_NOCLDSTOP | SA_NODEFER;
        sigaction(SIGCHLD, &action, NULL);
        if (dmenu != NULL) {
            dmenu->Build(this);
            if (allocname) delete [] allocname;

            return dmenu;
        }
    }
    if (allocname) delete [] allocname;
    return NULL;
}

/**
 * @fn    CreateFonts(void)
 * @brief Open fonts
 *
 * Opens all fonts and sets frame height.
 */
void KaScreen::CreateFonts(void) {
    bool set_mih;

    if (!style.menu.itemHeight)
        set_mih = true;
    else
        set_mih = false;

    if (default_font.Open(display, screen_number, NULL) == -1) {
        logger << err << "failed loading default font" << endl;
        exit(1);
    }

    int height;
    height = style.window.focused.font.Open(display, screen_number, 
                                            &default_font);
//    if (!style.window.titleHeight) style.window.titleHeight = height + 4;

    int tmp_sx = style.window.unfocused.font.shadow_off_x;
    int tmp_sy = style.window.unfocused.font.shadow_off_y;
    style.window.unfocused.font = style.window.focused.font;
    style.window.unfocused.font.shadow_off_x = tmp_sx;
    style.window.unfocused.font.shadow_off_y = tmp_sy;

    height = style.menu.frame.font.Open(display, screen_number, &default_font);
    if (set_mih) style.menu.itemHeight = height + 2;

    tmp_sx = style.menu.frame.highlightFont.shadow_off_x;
    tmp_sy = style.menu.frame.highlightFont.shadow_off_y;
    style.menu.frame.highlightFont = style.menu.frame.font;
    style.menu.frame.highlightFont.shadow_off_x = tmp_sx;
    style.menu.frame.highlightFont.shadow_off_y = tmp_sy;

    height = style.menu.bullet.font.Open(display, screen_number, &default_font);
    if (set_mih && style.menu.itemHeight < (unsigned int) (height + 2))
        style.menu.itemHeight = height + 2;

    style.menu.bullet.highlightFont = style.menu.bullet.font;
    style.menu.bullet.highlightFont.shadow_off_x = tmp_sx;
    style.menu.bullet.highlightFont.shadow_off_y = tmp_sy;

    height = style.menu.checkboxTrue.font.Open(display, screen_number, 
                                               &default_font);
    if (set_mih && style.menu.itemHeight < (unsigned int) (height + 2))
        style.menu.itemHeight = height + 2;

    style.menu.checkboxTrue.highlightFont = style.menu.checkboxTrue.font;   
    style.menu.checkboxTrue.highlightFont.shadow_off_x = tmp_sx;
    style.menu.checkboxTrue.highlightFont.shadow_off_y = tmp_sy;


    height = style.menu.checkboxFalse.font.Open(display, screen_number, 
                                                &default_font);
    if (set_mih && style.menu.itemHeight < (unsigned int) (height + 2))
        style.menu.itemHeight = height + 2;

    style.menu.checkboxFalse.highlightFont = style.menu.checkboxFalse.font;   
    style.menu.checkboxFalse.highlightFont.shadow_off_x = tmp_sx;
    style.menu.checkboxFalse.highlightFont.shadow_off_y = tmp_sy;

    height = style.menu.title.font.Open(display, screen_number, &default_font);
    if (! style.menu.titleHeight) 
        style.menu.titleHeight = height + 2;
    
//    if (style.window.titleHeight < 10) style.window.titleHeight = 10;
    if (style.menu.titleHeight < 4) style.menu.titleHeight = 4;
    if (style.menu.itemHeight < 4) style.menu.itemHeight = 4;
    
    style.window.yPosition = (style.window.titleHeight / 2 - 2) +
        style.window.focused.font.diff / 2 + style.window.focused.font.diff % 2;
    style.menu.frame.yPosition = (style.menu.itemHeight / 2) +
        style.menu.frame.font.diff / 2 + style.menu.frame.font.diff % 2;
    style.menu.title.yPosition = (style.menu.titleHeight / 2) +
        style.menu.title.font.diff / 2 + style.menu.title.font.diff % 2;
    style.menu.bullet.yPosition = (style.menu.itemHeight / 2) +
        style.menu.bullet.font.diff / 2 + style.menu.bullet.font.diff % 2;
    style.menu.checkboxTrue.yPosition = (style.menu.itemHeight / 2) +
        style.menu.checkboxTrue.font.diff / 2 + style.menu.checkboxTrue.font.diff % 2;
    style.menu.checkboxFalse.yPosition = (style.menu.itemHeight / 2) +
        style.menu.checkboxFalse.font.diff / 2 + style.menu.checkboxFalse.font.diff % 2;
}

/**
 * @fn    CreateColors(void)
 * @brief Creates all colors
 *
 * Creates all color GCs.
 */
void KaScreen::CreateColors(void) {
    XGCValues gcv;
    
    vector<ButtonStyle>::iterator bit = style.window.buttonStyles.begin();
    for (; bit != style.window.buttonStyles.end(); ++bit) {
        ButtonStyle& bs = *bit;
        if (bs.fg) {
            ButtonStateStyle* bss[] = { &bs.focused,   &bs.focusedActive,
                                       &bs.unfocused, &bs.unfocusedActive,
                                       &bs.pressed,   &bs.pressedActive };
            
            for (int i=0;i<6;i++) {
                ButtonStateStyle* bs = bss[i]; 
                gcv.foreground = bs->color.getPixel();
                bs->gc = XCreateGC(display, id, GCForeground, &gcv);
            }
        }
    }

    WindowStateStyle* windowStates[] = {&style.window.focused, 
                                        &style.window.unfocused};
    for (int i=0;i<2;i++) {
        WindowStateStyle* wstyle = windowStates[i];
        wstyle->font.AllocColor(display, id, 
                                &wstyle->labelColor, 
                                &wstyle->labelShadowColor);
    }

    style.menu.title.font.AllocColor(display, id, 
                                     &style.menu.title.color,
                                     &style.menu.title.shadowColor);

    style.menu.frame.font.AllocColor(display, id, 
                                     &style.menu.frame.color,
                                     &style.menu.frame.shadowColor);

    style.menu.frame.highlightFont.AllocColor(display, id, 
                                              &style.menu.frame.highlightColor,
                                              &style.menu.frame.highlightShadowColor);
    

    MenuItemStyle* menuParts[] = { &style.menu.bullet,
                                   &style.menu.checkboxTrue,
                                   &style.menu.checkboxFalse };
    
    for (int i=0;i<3;i++) {
        MenuItemStyle& mstyle = *menuParts[i];
        mstyle.font.AllocColor(display, id, &style.menu.frame.color,
                                          &style.menu.frame.highlightColor);
        mstyle.highlightFont.AllocColor(display, id, 
                                        &style.menu.frame.highlightColor,
                                        &style.menu.frame.highlightShadowColor);
    }
}

/**
 * @fn    UpdateWorkarea(void)
 * @brief Update workarea
 *
 * Updates workarea, all maximized windows are maximizied to the new
 * workarea.
 */
void KaScreen::UpdateWorkarea(void) {
    int old_x = current_desktop->workarea.x,
        old_y = current_desktop->workarea.y,
        old_width = current_desktop->workarea.width,
        old_height = current_desktop->workarea.height;
    
    current_desktop->workarea.x = current_desktop->workarea.y = 0;
    current_desktop->workarea.width = width;
    current_desktop->workarea.height = height;

    list<WMstrut *>::iterator it = strut_list.begin();
    for (; it != strut_list.end(); ++it) {
        WindowObject *wo = kahakai->FindWin((*it)->window,
                                           WindowType | DockHandlerType);
        if (wo) {
            if (wo->type == WindowType) {
                if (! (((KaWindow *) wo)->desktop_mask &
                       (1L << current_desktop->number)))
                    continue;
            } else if (wo->type == DockHandlerType) {
                if (! (((DockappHandler *) wo)->style->desktop_mask &
                       (1L << current_desktop->number)))
                    continue;
            }
        } else
            continue;

        if ((*it)->left > current_desktop->workarea.x)
            current_desktop->workarea.x = (*it)->left;
        if ((*it)->top > current_desktop->workarea.y)
            current_desktop->workarea.y = (*it)->top;
        if ((width - (*it)->right) < current_desktop->workarea.width)
            current_desktop->workarea.width = width - (*it)->right;
        if ((height - (*it)->bottom) < current_desktop->workarea.height)
            current_desktop->workarea.height = height - (*it)->bottom;
    }
    current_desktop->workarea.width = current_desktop->workarea.width -
        current_desktop->workarea.x;
    current_desktop->workarea.height = current_desktop->workarea.height -
        current_desktop->workarea.y;

    int res_x, res_y, res_w, res_h;
    if (old_x != current_desktop->workarea.x ||
        old_y != current_desktop->workarea.y ||
        old_width != current_desktop->workarea.width ||
        old_height != current_desktop->workarea.height) {
        net->SetWorkarea(this);

        vector<KaWindow *>::iterator ka_it = kawindow_list.begin();
        for (; ka_it != kawindow_list.end(); ++ka_it) {
            if (! ((*ka_it)->desktop_mask &
                   (1L << current_desktop->number))) break;
            if ((*ka_it)->flags.max) {
                (*ka_it)->flags.max = false;
                res_x = (*ka_it)->restore_max_attrib.getX();
                res_y = (*ka_it)->restore_max_attrib.getY();
                res_w = (*ka_it)->restore_max_attrib.getWidth();
                res_h = (*ka_it)->restore_max_attrib.getHeight();
                (*ka_it)->_Maximize((*ka_it)->restore_max_attrib.getMisc0(),
                                    (*ka_it)->restore_max_attrib.getMisc1(),
                                    (*ka_it)->flags.hmax,
                                    (*ka_it)->flags.vmax);
                (*ka_it)->restore_max_attrib.setX(res_x);
                (*ka_it)->restore_max_attrib.setY(res_y);
                (*ka_it)->restore_max_attrib.setWidth(res_w);
                (*ka_it)->restore_max_attrib.setHeight(res_h);
            }
        }
    }
}

/**
 * @fn    GetWorkareaSize(int *x, int *y, int *w, int *h)
 * @brief Calculates real workarea size
 *
 * Returns the real workarea size. Calculated from the xinerama area where the
 * pointer is currently located and the current desktops workarea. If xinerama
 * isn't used then the real workarea is the current desktops workarea.
 *
 * @param x Returns the real workarea X value
 * @param y Returns the workarea Y value
 * @param w Returns the workarea width value
 * @param h Returns the workarea height value
 */
void KaScreen::GetWorkareaSize(int *x, int *y, int *w, int *h) const {
    *x = current_desktop->workarea.x;
    *y = current_desktop->workarea.y;
    *w = current_desktop->workarea.width;
    *h = current_desktop->workarea.height;
#if 0 
#ifdef XINERAMA
    int px, py, i;

    if (kahakai->xinerama->active && kahakai->xinerama_info) {
        QueryPointer(&px, &py);
        for (i = 0; i < kahakai->xinerama_info_num; ++i) {
            if (px > kahakai->xinerama_info[i].x_org &&
                px < (kahakai->xinerama_info[i].x_org +
                      kahakai->xinerama_info[i].width) &&
                py > kahakai->xinerama_info[i].y_org &&
                py < (kahakai->xinerama_info[i].y_org +
                      kahakai->xinerama_info[i].height)) {
                int diff = kahakai->xinerama_info[i].x_org - *x;
                int xt = kahakai->xinerama_info[i].width;
                if (diff > 0) {
                    *w -= diff;
                    *x = kahakai->xinerama_info[i].x_org;
                } else xt += diff;
                if (*w > xt) *w = xt;

                diff = kahakai->xinerama_info[i].y_org - *y;
                xt = kahakai->xinerama_info[i].height;
                if (diff > 0) {
                    *h -= diff;
                    *y = kahakai->xinerama_info[i].y_org;
                } else xt += diff;
                if (*h > xt) *h = xt;
                break;
            }
        }
    }
#endif // XINERAMA
#endif
    
}

/**
 * @fn    MoveViewportTo(int x, int y)
 * @brief Move viewport to position
 *
 * Moves the virtual viewport to position (x, y). This is done by moving all
 * windows relative to the viewport change.
 *
 * @param x New x viewport
 * @param y New y viewport
 */
void KaScreen::MoveViewportTo(int x, int y) {
    if (x > v_xmax) x = v_xmax;
    else if (x < 0) x = 0;
    if (y > v_ymax) y = v_ymax;
    else if (y < 0) y = 0;
    
    int x_move = - (x - v_x);
    int y_move = - (y - v_y);
    v_x = x;
    v_y = y;

    //FocusFreezeOn();

    vector<KaWindow *>::iterator it = kawindow_list.begin();
    for (; it != kawindow_list.end(); ++it) {
        if ((*it)->flags.sticky) {
            net->SetVirtualPos(*it);
            continue;
        }

        // move it
        int old_x = (*it)->attrib.getX();
        int old_y = (*it)->attrib.getY();
        (*it)->attrib.setX(old_x + x_move);
        (*it)->attrib.setY(old_y + y_move);
        
        // not on screen anymore
        if (!(*it)->IsHorizontallyVisible() && !(*it)->IsVerticallyVisible()) {
            (*it)->dontsend = true;
            (*it)->RedrawWindow();
            (*it)->dontsend = false;
            net->SetVirtualPos(*it);
        }
    }

    for (it = kawindow_list.begin(); it != kawindow_list.end(); it++) {
        // if it's moved into the viewport, redraw it to make it appear
        if (!(*it)->flags.sticky && ((*it)->IsHorizontallyVisible() ||
                                     (*it)->IsVerticallyVisible()))
            (*it)->RedrawWindow(true);
    }

    vector<KaMenu *>::iterator it2 = kamenu_list.begin();
    for (; it2 != kamenu_list.end(); ++it2) {
        if ((*it2)->mapped && (! (*it2)->root_menu))
            (*it2)->Move(x_move, y_move);
    }
    net->SetDesktopViewPort(this);

    //FocusFreezeOff();
}

/**
 * @fn    MoveViewport(int direction, bool wrap)
 * @brief Move viewport one screen in specified direction
 *
 * Moves viewport one screen in direction specified by direction parameter.
 * Direction can be one of WestDirection, EastDirection, NorthDirection and
 * SouthDirection.
 *
 * @param direction Direction to move viewport
 * @param wrap Whether or not to "wrap" around when at the desktop edges
 */
void KaScreen::MoveViewport(int direction, bool wrap) {
    int vd;
    
    switch (direction) {
        case WestDirection:
            if (v_x > 0) {
                if ((v_x - width) < 0) vd = v_x;
                else vd = width;
                XWarpPointer(display, None, None, 0, 0, 0, 0, vd - 6, 0);
                MoveViewportTo(v_x - vd, v_y);
            } else if (wrap) {
                XWarpPointer(display, None, None, 0, 0, 0, 0, width - 6, 0);
                MoveViewportTo(v_xmax, v_y);
            }
            break;
        case EastDirection:
            if (v_x < v_xmax) {
                if ((v_x + width) > v_xmax) vd = v_xmax - v_x;
                else vd = width;
                XWarpPointer(display, None, None, 0, 0, 0, 0, 6 - vd, 0);
                MoveViewportTo(v_x + vd, v_y);
            } else if (wrap) {
                XWarpPointer(display, None, None, 0, 0, 0, 0, 6 - width, 0);
                MoveViewportTo(0, v_y);
            }
            break;
        case NorthDirection:
            if (v_y > 0) {
                if ((v_y - height) < 0) vd = v_y;
                else vd = height;
                XWarpPointer(display, None, None, 0, 0, 0, 0, 0, vd - 6);
                MoveViewportTo(v_x, v_y - vd);
            } else if (wrap) {
                XWarpPointer(display, None, None, 0, 0, 0, 0, 0, height - 6);
                MoveViewportTo(v_x, v_ymax);
            }
            break;
        case SouthDirection:
            if (v_y < v_ymax) {
                if ((v_y + height) > v_ymax) vd = v_ymax - v_y;
                else vd = height;
                XWarpPointer(display, None, None, 0, 0, 0, 0, 0, 6 - vd);
                MoveViewportTo(v_x, v_y + vd);
            } else if (wrap) {
                XWarpPointer(display, None, None, 0, 0, 0, 0, 0, 6 - height);
                MoveViewportTo(v_x, 0);
            }
    }
}

/** 
 * @fn    ViewportFixedMove(XEvent *, KaAction *ac)
 * @brief Fixed viewport move
 *
 * Parses the action parameter as an X offset string. The X and Y
 * offset values defines the fixed pixel position to move the
 * viewport to.
 *
 * @param ac KaAction object
 */ 
void KaScreen::ViewportFixedMove(XEvent *e, KaAction *ac) {
    ScreenEvViewportFixedMove func;
    func(this, e, ac);
    
    /*    int x, y, mask; 
    unsigned int w = 0, h = 0;
                            
    if (! ac->param) return;
   
    
    mask = XParseGeometry(ac->param, &x, &y, &w, &h);
    if (mask & XNegative) x = v_xmax + x;    
    if (mask & YNegative) y = v_ymax + y;
    MoveViewportTo(x, y); */
}

/**
 * @fn    ViewportRelativeMove(XEvent *, KaAction *ac)
 * @brief Relative viewport move
 *
 * Parses the action parameter as an X offset string. The X and Y
 * offset values defines the number of pixels to move the viewport.
 *
 * @param ac KaAction object
 */
void KaScreen::ViewportRelativeMove(XEvent *, KaAction *ac) {
    int x, y, mask;
    unsigned int w = 0, h = 0;
                  
    if (! ac->param) return;
   
    mask = XParseGeometry(ac->param, &x, &y, &w, &h);
    MoveViewportTo(v_x + x, v_y + y);
}

/**
 * @fn    ViewportMove(XEvent *e, KaAction *)
 * @brief Move viewport after mouse movement
 *
 * Moves viewport after mouse motion events.
 *
 * @param e XEvent causing function call
 */
void KaScreen::ViewportMove(XEvent *e, KaAction *) {
    XEvent event;
    int px, py;
    list<XEvent *> *maprequest_list;
    vector<KaWindow *>::iterator it;

    if (kahakai->eh->move_resize != EndMoveResizeType) return;
    kahakai->eh->move_resize = MoveOpaqueType;

    QueryPointer(&px, &py);
 
    maprequest_list = new list<XEvent *>;
    XGrabPointer(display, id, true, ButtonReleaseMask | ButtonPressMask |
                 PointerMotionMask | EnterWindowMask | LeaveWindowMask,
                 GrabModeAsync, GrabModeAsync, None, kahakai->move_cursor,
                 CurrentTime);
    XGrabKeyboard(display, id, true, GrabModeAsync, GrabModeAsync,
                  CurrentTime);
    for (;;) {
        it = kawindow_list.begin();
        for (; it != kawindow_list.end(); ++it) {
            (*it)->dontsend = true;
        }
        kahakai->eh->EventLoop(kahakai->eh->menu_viewport_move_return_mask,
                              &event);
        switch (event.type) {
            case MotionNotify: {
                while (XCheckTypedWindowEvent(display, event.xmotion.window,
                                              MotionNotify, &event));
                int x = v_x - (event.xmotion.x_root - px);
                int y = v_y - (event.xmotion.y_root - py);
                
                if (x > v_xmax) x = v_xmax;
                else if (x < 0) x = 0;
                if (y > v_ymax) y = v_ymax;
                else if (y < 0) y = 0;
                
                int x_move = - (x - v_x);
                int y_move = - (y - v_y);
                v_x = x;
                v_y = y;

                vector<KaWindow *>::iterator it = kawindow_list.begin();
                for (; it != kawindow_list.end(); ++it) {
                    if (! (*it)->flags.sticky) {
                        int old_x = (*it)->attrib.getX();
                        int old_y = (*it)->attrib.getY();
                        (*it)->attrib.setX((*it)->attrib.getX() + x_move);
                        (*it)->attrib.setY((*it)->attrib.getY() + y_move);
            
                        if ((((*it)->attrib.getX() + (*it)->attrib.getWidth()) > 0 &&
                             (*it)->attrib.getX() < width) && 
                            (((*it)->attrib.getY() + (*it)->attrib.getHeight()) > 0 &&
                             (*it)->attrib.getY() < height))
                            (*it)->RedrawWindow();
                        else {
                            if (((old_x + (*it)->attrib.getWidth()) > 0 &&
                                 old_x < width) && 
                                ((old_y + (*it)->attrib.getHeight()) > 0 &&
                                 old_y < height))
                                (*it)->RedrawWindow();
                        }
                    }
                }
                vector<KaMenu *>::iterator it2 = kamenu_list.begin();
                for (; it2 != kamenu_list.end(); ++it2) {
                    if ((*it2)->mapped && (! (*it2)->root_menu))
                        (*it2)->Move(x_move, y_move, !config.lazy_trans);
                }
                px = event.xmotion.x_root;
                py = event.xmotion.y_root;
            } break;
            case LeaveNotify:
            case EnterNotify:
                break;
            case MapRequest:
                maprequest_list->push_front(&event); break;
            case ButtonPress:
            case ButtonRelease:
                event.xbutton.window = id;
            case KeyPress:
            case KeyRelease:
                if (event.type == KeyPress || event.type == KeyRelease)
                    event.xkey.window = id;
                kahakai->eh->HandleEvent(&event);
                if (kahakai->eh->move_resize != EndMoveResizeType) break;
                while (! maprequest_list->empty()) {
                    XPutBackEvent(display, maprequest_list->front());
                    maprequest_list->pop_front();
                }
                delete maprequest_list;
                it = kawindow_list.begin();
                for (; it != kawindow_list.end(); ++it) {
                    (*it)->dontsend = false;
                    net->SetVirtualPos(*it);
                    if ((((*it)->attrib.getX() + (*it)->attrib.getWidth()) > 0 &&
                         (*it)->attrib.getX() < width) && 
                        (((*it)->attrib.getY() + (*it)->attrib.getHeight()) > 0 &&
                         (*it)->attrib.getY() < height)) {
                        
                        if (config.lazy_trans) {
                            (*it)->render_if_transparency = true;
                            (*it)->DrawTitlebar();
                            (*it)->DrawHandlebar();
                            (*it)->render_if_transparency = false;
                        }

                        (*it)->SendConfig();
                    }
                }

                if (config.lazy_trans) {
                    vector<KaMenu *>::iterator it2 = kamenu_list.begin();
                    for (; it2 != kamenu_list.end(); ++it2) {
                        if ((*it2)->mapped && (! (*it2)->root_menu))
                            (*it2)->Move(0, 0, true);
                    }
                }

                net->SetDesktopViewPort(this);
                XUngrabKeyboard(display, CurrentTime);
                XUngrabPointer(display, CurrentTime);
                return;
        }
    }
}

/**
 * @fn    EndMoveResize(XEvent *e, KaAction *)
 * @brief Ends move
 *
 * Ends viewport moving process.
 */
void KaScreen::EndMoveResize(XEvent *, KaAction *) {
    kahakai->eh->move_resize = EndMoveResizeType;
}

/**
 * @fn    Focus(XEvent *, KaAction *)
 * @brief Set input focus to root image
 *
 * Sets the keyboard input focus to the KaScreens root window.
 */
void KaScreen::Focus(XEvent *, KaAction *) {
    focus = true;
    XSetInputFocus(display, id, RevertToPointerRoot, CurrentTime);
}

/**
 * @fn    MenuMap(XEvent *e, KaAction *ac)
 * @brief Maps a menu
 *
 * Maps a menu at the current mouse position.
 *
 * @param ac KaAction object
 * @param bool True if we should focus first item in menu
 */
void KaScreen::MenuMap(XEvent *, KaAction *ac, bool focus) {
    int x, y;
    KaMenu *menu = GetMenuNamed(ac->param);

    if (! menu) return;
    if (kahakai->eh->move_resize != EndMoveResizeType) return;

    int workx, worky, workw, workh;
    GetWorkareaSize(&workx, &worky, &workw, &workh);

    if (QueryPointer(&x, &y)) {
        if (menu->ext_type) menu->Build(this);
        menu->rf = this;
        menu->ftype = MenuRFuncMask;
        if ((y + menu->height + style.menu.borderThickness * 2) >
            (unsigned int) (workh + worky))
           y -= (menu->height + style.menu.borderThickness * 2);
        if ((x + menu->width + style.menu.borderThickness * 2) >
            (unsigned int) (workw + workx))
            x -= (menu->width + style.menu.borderThickness * 2);
        menu->Map(x, y);
        if (focus) menu->FocusFirst();
    }
}

/**
 * @fn    MenuRemap(XEvent *e, KaAction *ac)
 * @brief Maps a menu
 *
 * Remaps a menu at the current mouse position.
 *
 * @param ac KaAction object
 * @param bool True if we should focus first item in menu
 */
void KaScreen::MenuRemap(XEvent *, KaAction *ac, bool focus) {
    int x, y;
    KaMenu *menu = GetMenuNamed(ac->param);

    if (! menu) return;
    if (menu->dynamic && menu->mapped) {
        menu->Unmap(menu->has_focus);
        if (! (menu = CreateDynamicMenu(ac->param))) return;
    }
    if (kahakai->eh->move_resize != EndMoveResizeType) return;

    int workx, worky, workw, workh;
    GetWorkareaSize(&workx, &worky, &workw, &workh);

    if (QueryPointer(&x, &y)) {
        if (menu->ext_type) menu->Build(this);
        menu->rf = this;
        menu->ftype = MenuRFuncMask;
        if ((y + menu->height + style.menu.borderThickness * 2) >
            (unsigned int) (workh + worky))
           y -= (menu->height + style.menu.borderThickness * 2);
        if ((x + menu->width + style.menu.borderThickness * 2) >
            (unsigned int) (workw + workx))
            x -= (menu->width + style.menu.borderThickness * 2);
        menu->ignore = true;
        menu->ReMap(x, y);
        menu->ignore = false;
        if (focus) menu->FocusFirst();
    }
}

/**
 * @fn    MenuUnmap(XEvent *, KaAction *ac, bool focus)
 * @brief Unmaps menu
 *
 * Unmaps a menu and its linked submenus.
 *
 * @param ac KaAction object
 * @param bool True if we should focus root item
 */
void KaScreen::MenuUnmap(XEvent *, KaAction *ac, bool focus)
{
    KaMenu *menu = GetMenuNamed(ac->param);

    if (!menu || kahakai->eh->move_resize != EndMoveResizeType)
      return;

    menu->Unmap(focus);
}

/**
 * @fn    Restart(XEvent *, KaAction *)
 * @brief Restarts window manager
 *
 * Restarts window manager by deleting all objects and executing the program
 * file again.
 *
 * @param ac KaAction object
 */
void KaScreen::Restart(XEvent *, KaAction *ac) {
    if (ac->param)
        restart(ac->param);
    else
        restart(NULL);
}

/**
 * @fn    Exit(XEvent *, KaAction *)
 * @brief Shutdowns window manager
 *
 * Shutdowns window manager. Returning successful status.
 */
void KaScreen::Exit(XEvent *, KaAction *) {
    quit(EXIT_SUCCESS);
}

/**
 * @fn    TaskSwitcher(XEvent *, KaAction *)
 * @brief Maps task switcher menu
 *
 * Maps task switcher menu at middle of screen and sets input focus to
 * first window in list.
 */
void KaScreen::TaskSwitcher(XEvent *, KaAction *) {
    if (kahakai->eh->move_resize != EndMoveResizeType) return;
    if (kawindow_list.empty()) return;
    if (window_menu->mapped) return;

    vector<KaWindow *>::iterator it = kawindow_list.begin();
    KaWindow * lastontasklist;
    int tasklistsize = 0;

    for(; it != kawindow_list.end(); it++) {
        if((*it)->flags.tasklist && ((*it)->desktop_mask &
                                     (1L << current_desktop->number))) {
            tasklistsize++;
            lastontasklist = *it;
        }
    }

    if((tasklistsize == 1 && lastontasklist->has_focus) || tasklistsize == 0)
        return;

    int workx, worky, workw, workh;
    GetWorkareaSize(&workx, &worky, &workw, &workh);
    
    window_menu->Build(this);
    window_menu->ReMap(workx + (workw / 2 - window_menu->width / 2),
                       worky + (workh / 2 - window_menu->height / 2));

    if(tasklistsize == 1) {
        // we know from above that it doesn't have focus
        window_menu->FocusFirst();
    } else
        window_menu->FocusItem(1);
}

/**
 * @fn    PreviousTask(XEvent *e, KaAction *ac)
 * @brief Switches to previous task
 *
 * Switches to the previous focused window.
 *
 * @param e X event that have occurred
 * @param ac KaAction object
 */
void KaScreen::PreviousTask(XEvent *e, KaAction *ac) {
    if (kahakai->eh->move_resize != EndMoveResizeType) return;
    if (kawindow_list.size() < 2) return;
    
    vector<KaWindow *>::iterator it = kawindow_list.begin();
    it++; // kawindow_list.begin() is the current window, so we skip it

    // Search a window on current desktop front to back
    for (; it != kawindow_list.end(); ++it) {
        if ((*it)->OnCurrentDesktop() && (*it)->flags.tasklist) {
            (*it)->Raise();
            (*it)->FocusVis();
            return;
        }
    }
}

/**
 * @fn    NextTask(XEvent *e, KaAction *ac)
 * @brief Switches to next task
 *
 * Switches to the window that haven't had focus for longest time.
 *
 * @param e X event that have occurred
 * @param ac KaAction object
 */
void KaScreen::NextTask(XEvent *e, KaAction *ac) {
    if (kahakai->eh->move_resize != EndMoveResizeType) return;
    if (kawindow_list.size() < 2) return;
    
    vector<KaWindow *>::reverse_iterator rit = kawindow_list.rbegin();

    // Search a window on current desktop from the back 
    for (; rit != kawindow_list.rend(); ++rit) {
        if ((*rit)->OnCurrentDesktop() && (*rit)->flags.tasklist) {
            (*rit)->Raise();
            (*rit)->FocusVis();
            return;
        }
    }
}

/**
 * @fn    PointerFixedWarp(XEvent *e, KaAction *ac)
 * @brief Warps pointer to fixed position
 *
 * Parses the action parameter as an X offset string. The X and Y
 * offset values defines the fixed pixel position to warp the 
 * pointer to.
 *
 * @param ac KaAction object
 */
void KaScreen::PointerFixedWarp(XEvent *, KaAction *ac) {
    int x, y, mask, o_x, o_y;
    unsigned int w, h;
    
    mask = XParseGeometry(ac->param, &x, &y, &w, &h);
    if (mask & XNegative) x = width + x;
    if (mask & YNegative) y = height + y;

    QueryPointer(&o_x, &o_y);
    x = x - o_x;
    y = y - o_y;
    XWarpPointer(display, None, None, 0, 0, 0, 0, x, y);
}

/** 
 * @fn    PointerRelativeWarp(XEvent *e, KaAction *ac)
 * @brief Relative pointer warp
 *
 * Parses the action parameter as an X offset string. The X and Y
 * offset values defines the number of pixels to warp the pointer.
 * 
 * @param ac KaAction object
 */
void KaScreen::PointerRelativeWarp(XEvent *, KaAction *ac) {
    int x, y, mask;
    unsigned int w, h;
                    
    mask = XParseGeometry(ac->param, &x, &y, &w, &h);
    XWarpPointer(display, None, None, 0, 0, 0, 0, x, y);
}

/** 
 * @fn    GoToDesktop(unsigned int number)
 * @brief Go to desktop
 *
 * Sets current desktop to the one specified by number parameter.
 * 
 * @param number Desktop number to go to
 */
void KaScreen::GoToDesktop(unsigned int number) {
    vector<Desktop *>::iterator dit = desktop_list.begin();

    for (; dit != desktop_list.end(); dit++)
        if ((unsigned int) (*dit)->number == number) break;

    if (dit != desktop_list.end() && *dit != current_desktop) {

        Window oldf = (Window) 0;
        bool moving = false;

        if (kahakai->eh)
            oldf = kahakai->eh->focused;

        //XSetInputFocus(display, id, RevertToPointerRoot, CurrentTime);
        //FocusFreezeOn();

        (*dit)->workarea.x = current_desktop->workarea.x;
        (*dit)->workarea.y = current_desktop->workarea.y;
        (*dit)->workarea.width = current_desktop->workarea.width;
        (*dit)->workarea.height = current_desktop->workarea.height;
        Desktop* old_desktop = current_desktop;
        old_desktop->oldfocus = oldf;
        current_desktop = (*dit);

        vector<KaWindow *>::iterator it = kawindow_list.begin();
        for (; it != kawindow_list.end(); ++it) {
            // If we are currently moving a window that was only on old desktop
            // (not on more than one desktop) move it to the new desktop.
            if ((*it)->has_focus && (*it)->move_resize &&
                (*it)->desktop_mask ==
                (unsigned int)(1L << old_desktop->number)) {

                (*it)->PartAllDesktopsExceptCurrent();
                moving = true;
            }

            if ((*it)->OnCurrentDesktop()) {
                (*it)->Show();
                net->SetDesktop(*it);
            } else
                (*it)->Hide();
        }

        RestackWindows(0);
        net->SetClientListStacking(this);

        //FocusFreezeOff();
        DiscardEnterNotifies();

        if (oldf) {
            KaWindow *ww = (KaWindow *) kahakai->FindWin(oldf, WindowType);
            if (ww && ww->OnCurrentDesktop())
                    ww->Focus(false);
        }

        // Redefine the function of oldf, it's not needed anymore
        oldf = 0;
        if ((! moving) && current_desktop->oldfocus) {
            KaWindow *ww = (KaWindow *) kahakai->FindWin(current_desktop->oldfocus,
                                                        WindowType);
            if (ww && ww->OnCurrentDesktop()) {
                    ww->Focus();
                    oldf = current_desktop->oldfocus;
            }
        }

        // Focus the last selected window that is only on current desktop
        /* We restrict the windows to only the current (new) desktop 
           because otherwise desktop-sticky windows focused on the old desktop
           will be automatically selected as the new focused window
           because they are first in the list to satisfy the OnCurrentDesktop
           predicate.
         */
        // Do this only if we couldn't focus last window on workspace
        if ((! moving) && (!oldf))
          for (it = kawindow_list.begin(); it != kawindow_list.end(); ++it) {
              if ((*it)->desktop_mask ==
                  (unsigned int)(1L << current_desktop->number)) {
                  (*it)->Focus();
                  break;
              }
          }

        list<DockappHandler *>::iterator dock_it = docks.begin();
        for (; dock_it != docks.end(); ++dock_it) {
            if ((*dock_it)->style->desktop_mask &
                (1L << current_desktop->number)) {
                if ((*dock_it)->hidden) {
                    (*dock_it)->Show();
                }
            } else if (! (*dock_it)->hidden) {
                (*dock_it)->Hide();
            }
        }
        UpdateWorkarea();
        net->SetCurrentDesktop(this);
    } else
        if (dit == desktop_list.end())
            logger << warn << "bad desktop id `" << number << "', desktop "
                   << number << " doesn't exist" << endl;
}

/** 
 * @fn    GoToDesktop(XEvent *, KaAction *ac)
 * @brief Go to desktop
 *
 * Sets current desktop to the one specified by action parameter.
 * 
 * @param ac KaAction object
 */
void KaScreen::GoToDesktop(XEvent *event, KaAction *action) {
    ScreenEvGoToDesktop func;
    func(this, event, action);
}

/** 
 * @fn    NextDesktop(XEvent *, KaAction *)
 * @brief Go to next desktop
 *
 * Sets current desktop to desktop with ID number current ID + 1. If
 * current ID + 1 doesn't exist then current desktop is set to desktop with
 * ID 0.
 * 
 * @param ac KaAction object
 */
void KaScreen::NextDesktop(XEvent *, KaAction *) {
    if (current_desktop->number + 1 == config.desktops)
        GoToDesktop(0);
    else
        GoToDesktop(current_desktop->number + 1);
}

/** 
 * @fn    PreviousDesktop(XEvent *, KaAction *)
 * @brief Go to previous desktop
 *
 * Sets current desktop to desktop with ID number current ID - 1. If
 * current ID - 1 is negative then current desktop is set to desktop with
 * highest desktop number.
 * 
 * @param ac KaAction object
 */
void KaScreen::PreviousDesktop(XEvent *, KaAction *) {
    if (current_desktop->number == 0)
        GoToDesktop(config.desktops - 1);
    else
        GoToDesktop(current_desktop->number - 1);
}

/**
 * @fn    EvAct(XEvent *e, EventDetail *ed, vector<KaAction *> *acts)
 * @brief Calls KaScreen function
 *
 * Tries to match an occurred X event with the actions in an action list.
 * If we have a match then we execute that action.
 *
 * @param e X event that have occurred
 * @param ed Event details
 * @param acts List with actions to match event with
 */
void KaScreen::EvAct(XEvent *e, EventDetail *ed, vector<KaAction *> *acts) {
    if (kahakai->eh->move_resize != EndMoveResizeType)
        ed->mod |= MoveResizeMask;

    vector<KaAction *>::iterator it = acts->begin();

    for (; it != acts->end(); ++it) {
        if (eventmatch(kahakai->display, *it, ed)) {

            if ((*it)->delay.tv_sec || (*it)->delay.tv_usec) {
                Interrupt *i = new Interrupt(*it, e, id);
                kahakai->timer->AddInterrupt(i);

            } else {
                if ((*it)->exec)
                    kaexec((*it)->exec, displaystring);
                else
                    (*((*it)->rootfunc))(this, e, *it);
            }

        }
    }
}

/**
 * @fn    AddDockapp(Window window)
 * @brief Add dockapp to dockapp system
 *
 * Inserts the dockapp into the right dockapp holder.
 *
 * @param window Window ID of dockapp window
 */
void KaScreen::AddDockapp(Window window) {
    Dockapp *da;
    XClassHint *c_hint = XAllocClassHint();
    int have_hints = XGetClassHint(display, window, c_hint);
    char *title;
    bool found = false;

    if (! XFetchName(display, window, &title))
        title = NULL;

    if (have_hints) {
        vector<SimpleMatch *>::iterator match_it;
        list<DockappHandler *>::iterator dock_it = docks.begin();

        for (; dock_it != docks.end(); dock_it++) {
            match_it = (*dock_it)->style->order.begin();
            for (; match_it != (*dock_it)->style->order.end(); match_it++) {
                if ((*match_it)->matches(c_hint->res_name)) {
                    da = new Dockapp(window, *dock_it);
                    da->c_hint = c_hint;
                    da->title = title;
                    (*dock_it)->Update();
                    found = true;
                }
            }
        }
    }

    if (!found) {
        DockappHandler *lastd = docks.back();
        da = new Dockapp(window, lastd);
        da->c_hint = NULL;
        da->title = NULL;
        lastd->Update();
    }
}

#ifdef XRANDR
/**
 * @fn    RRUpdate(void)
 * @brief Update screen size
 *
 * Updates screen edge positions/sizes, dockappholder positions and
 * workarea size.
 */
void KaScreen::RRUpdate(void) {
    v_xmax = (config.virtual_x - 1) * width;
    v_ymax = (config.virtual_y - 1) * height;

    XMoveResizeWindow(display, west->id, 0, 0, 2, height);
    XMoveResizeWindow(display, east->id, width - 2, 0, 2, height);
    XMoveResizeWindow(display, north->id, 0, 0, width, 2);
    XMoveResizeWindow(display, south->id, 0, height - 2, width, 2);

    list<DockappHandler *>::iterator dit = docks.begin();
    for (; dit != docks.end(); ++dit)
        (*dit)->Update();

    UpdateWorkarea();

    net->SetDesktopGeometry(this);
}
#endif // XRANDR

inline void KaScreen::SetSmartName(KaWindow * win, int smartnum) {

    win->smartnum = smartnum;

    if (win->smartnum) {
        stringstream tmps;
        tmps << win->realname << " <" << win->smartnum << ">";
        win->name = tmps.str();
    } else
        win->name = win->realname;

    win->labeltext = win->name;

    net->SetVisibleName(win);
}

/**
 * @fn    SmartName(KaWindow *ww)
 * @brief Sets viewable title name for a window
 *
 * Checks if a window with the same name exists and adds a unique suffix to
 * the window name if this is the case. Viewable titles of old matching
 * windows are also updated so that they have unique suffixes.
 *
 * @param ww Window to set viewable name for
 */
void KaScreen::SmartName(KaWindow *ww) {
    int match = 1;
    list<KaWindow *>::iterator it = kawindow_list_map_order.begin();

    for (; it != kawindow_list_map_order.end(); it++) {
        if (*it != ww && ww->realname == (*it)->realname) {

            SetSmartName(*it, match++);
            (*it)->UpdateTitlebar();

            if (match >= 989) return;
        }
    }

    SetSmartName(ww, (match>1 ? match : 0));
}

/**
 * @fn    SmartNameRemove(KaWindow *ww)
 * @brief Indicate that a title name no longer exists
 *
 * Updates viewable title names for windows with same title name as the title
 * name that no longer exists.
 *
 * @param ww Window that have a title name that no longer exists
 */
void KaScreen::SmartNameRemove(KaWindow *ww) {
    int match = 1;
    KaWindow * first = NULL;
    bool foundsecond = false;
    list<KaWindow *>::iterator it = kawindow_list_map_order.begin();

    for (; it != kawindow_list_map_order.end(); it++) {
        if (*it != ww && ww->realname == (*it)->realname) {
            if(!first) {
                first = *it;
                continue;
            }

            if(first && !foundsecond) {
                SetSmartName(first, match++);
                (first)->UpdateTitlebar();
                foundsecond = true;
            }

            SetSmartName(*it, match++);
            (*it)->UpdateTitlebar();

            if (match >= 998) return;
        }
    }

    if(first && !foundsecond) {
        SetSmartName(first, 0);
        (first)->UpdateTitlebar();
    }

    SetSmartName(ww, 0);
}

void KaScreen::DiscardEnterNotifies() {
    DiscardEventsOfType(EnterNotify);
}

void KaScreen::DiscardEventsOfType(int type) {
    XEvent junk;
    XSync(display, false);
    while (XCheckTypedEvent(display, type, &junk));
}

bool KaScreen::QueryPointer(Window w,  int *root_x_return, int *root_y_return,
                                       unsigned int *mask_return) {
    Window discard_w;
    int    discard_i;

    return XQueryPointer(display, w,  &discard_w, &discard_w,
                         root_x_return, root_y_return,
                         &discard_i, &discard_i, mask_return);
}


bool KaScreen::QueryPointer(Window w,  int *root_x_return,  int *root_y_return) {

    Window discard_w;
    int    discard_i;
    unsigned int discard_ui;

    return XQueryPointer(display, w,  &discard_w, &discard_w,
                         root_x_return, root_y_return,
                         &discard_i, &discard_i, &discard_ui);
}

void KaScreen::ImodAdd(unsigned int mask) {
    imod_list.push_back(mask);
    imod_mask |= mask;
}

/**
 * @fn    ScreenEdge(KaScreen *kascrn, int x, int y, int width, int height,
 *                   int type) : WindowObject(0, type)
 * @brief Constructor for ScreenEdge class
 *
 * Creates an always on top screen edge window.
 *
 * @param kascrn KaScreen Object
 * @param x X position
 * @param y Y position
 * @param width Width of ScreenEdge window
 * @param height Height of ScreenEdge window
 * @param type Type of WindowObject
 */
ScreenEdge::ScreenEdge(KaScreen *kascrn, int x, int y, int width, int height,
                       int type) : WindowObject(0, type) {
    XSetWindowAttributes attrib_set;
    
    ka = kascrn;
    attrib_set.override_redirect = true;
    attrib_set.event_mask = EnterWindowMask | LeaveWindowMask |
        ButtonPressMask | ButtonReleaseMask;
    
    id = XCreateWindow(ka->display, ka->id, x, y, width, height, 0,
                       CopyFromParent, InputOnly, CopyFromParent,
                       CWOverrideRedirect | CWEventMask, &attrib_set);

    ka->kahakai->net->wXDNDMakeAwareness(id);

    XMapWindow(ka->display, id);
    ka->kahakai->window_table.insert(make_pair(id, this));
}

/**
 * @fn    ~ScreenEdge(void)
 * @brief Destructor for ScreenEdge class
 *
 * Destroys ScreenEdge window
 */
ScreenEdge::~ScreenEdge(void) {
    ka->kahakai->window_table.erase(id);
    XDestroyWindow(ka->display, id);
}

