/* $Id: Script.cc,v 1.23 2003/09/11 22:42:21 faugn Exp $ */

/**
 * @file   Script.cc
 * @date   27-May-2003
 *
 * @brief Implementation of an interface to scripting languages
 *
 * Extensible interface to scripting languages (Python/Ruby/etc.)
 *
 */

#include "Script.hh"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <iostream>
#include <string>
#include <list>

using std::string;
using std::list;
using std::cerr;
using std::endl;

/**
 * Used to initialize _scriptDirs.
 */
static list<string> makeScriptDirs() {
    list<string> dirs;

    string homedir = getenv("HOME");

    dirs.push_back(homedir + "/.kahakai");
    dirs.push_back(homedir + "/.kahakai/scripts");

    dirs.push_back(DEFAULTSCRIPTDIR);

    return dirs;
}

/*
 * static initialization
 */

const list<string> Script::_scriptDirs = makeScriptDirs();

/*
 * public
 */

Script::~Script() {
}

const string & Script::getFilename() const {
    return _filename;
}

string Script::getSuffix(const string &filename) {

    // suffixes exist after the last '.' character
    unsigned index = filename.find_last_of('.');
    if (index != string::npos) {
        // be sure to start *after* the '.'
        return filename.substr(index + 1);
    } else {
        // didnt find one. claim no suffix
        return "";
    }
}

/*
 * protected
 */

Script::Script(const list<string> &suffixes, const string &filename)
    : _filename(findScript(filename)), _suffixes(suffixes)
{
    // findScript will have done all file existence and readability error
    // checking and might have thrown an InvalidFile exception along the way

    // prevent construction if the filename is unacceptable
    if (! checkSuffix(_filename)) {
        throw InvalidFileType(string("bad script suffix: ") +
                string(getSuffix(_filename)));
    }

    // Postcondition: suffix should be valid and _filename should be readable
}

bool Script::checkSuffix(const string &filename) const {
    return find(_suffixes.begin(), _suffixes.end(),  
                getSuffix(filename)) != _suffixes.end();
}

const list<string> & Script::scriptDirs() {
    return _scriptDirs;
}

/*
 * private
 */

string Script::findScript(const string &name) const {
    string filename = name;

    // create list of file paths to check:
    // if filename begins with '/', just check it;
    // else, check for filename in each of our script directories
    list<string> paths;
    if (filename[0] == '/') {
        paths.push_back(filename);
    } else {
        list<string>::const_iterator i;
        for (i = _scriptDirs.begin(); i != _scriptDirs.end(); i++) {
            paths.push_back((*i) + "/" + filename);
        }
    }

    // check each element of paths for an existent, readable filename.
    // warn for each existent, non-readable file found,
    // error if not existent, readable file is found
    bool found = false;
    list<string>::iterator i;
    for (i = paths.begin(); i != paths.end(); i++) {
        if (exists(*i)) {
            if (readable(*i)) {
                filename = (*i);
                found = true;
                break;
            } else {
                cerr << "WARNING: script found but not readable: "
                     << *i << endl;
            }
        }
    }

    // if we didn't find it, be loud
    if (found)
        return filename;
    else
        throw InvalidFile("script not found: " + filename);
}

bool Script::exists(const string &filename) {
    struct stat buf;
    return (stat(filename.c_str(), &buf) == 0);
}

bool Script::readable(const string &filename) {
    // WARNING: this will break with suid programs
    return (access(filename.c_str(), R_OK) == 0);
}
