/* $Id: Script.hh,v 1.21 2003/11/20 22:34:28 pervago Exp $ */

/**
 * @file   Script.hh
 * @date   11-May-2001 11:48:03
 *
 * @brief Definition of an interface to scripting languages
 *
 * Extensible interface to scripting languages (Python/Ruby/etc.)
 *
 */

#ifndef __SCRIPT_HH
# define __SCRIPT_HH

# include "Types.hh"

# include <list>
# include <string>

// feel free to push this back up into some general exception include file
class KahakaiException : public std::exception {
public:
    KahakaiException(const std::string &whatString) throw()
        : _whatString(whatString) {}
    virtual ~KahakaiException() throw() {}
    virtual const char * what() const throw() {
        return _whatString.c_str();
    }
    virtual operator std::string () const {
        return what();
    }
private:
        const std::string _whatString;
};

// this one needs to stay here
class InvalidFileType : public KahakaiException {
public:
    InvalidFileType(const std::string &whatString) throw()
        : KahakaiException(whatString) {}
    virtual ~InvalidFileType() throw() {}
};

class InvalidFile : public KahakaiException {
public:
    InvalidFile(const std::string &whatString) throw()
        : KahakaiException(whatString) {}
    virtual ~InvalidFile() throw () {}
};

class Script {

public:

    // Destroy this object.
    virtual ~Script();

    // Get the file name this Script represents.
    // @return The file name
    virtual const std::string & getFilename() const;

    // Execute the file represented by this Script.
    // @return true if executed successfully, false otherwise
    virtual bool run() = 0;

    /* Get the suffix of a file name by extracting the string that follows
    * the last '.' character. If no '.' character exists within the file
    * name, return an empty string.
    * @param filename The file name
    * @return the suffix */
    static std::string getSuffix(const std::string &filename);

protected:

    /* Create a Script object from a file name and list of acceptable
    * suffixes.
    * @param suffixes Acceptable suffixes
    * @param filename The file name
    * @throw InvalidFileType if file name's suffix is not in suffixes */
    Script(const std::list<std::string> &suffixes,
           const std::string &filename);

    /* Check the acceptability of a file name's suffix, based on the list of
     * stored suffixes. Used by run()
     * @param suffix The file name
     * @return true if acceptable, false otherwise */
    virtual bool checkSuffix(const std::string &filename) const;

     // Get the list of directory names where we search for scripts.
     // @return the list of dir names
    static const std::list<std::string> & scriptDirs();

private:

     // Script objects should never be created without a file name and list
     // of acceptable suffixes.
    Script() {};

     // Helper to the constructor.
    std::string findScript(const std::string &filename) const;

     // Test file's existence.
    static bool exists(const std::string &filename);

     // Test file's readability.
    static bool readable(const std::string &filename);

    // The directories in which we should look for script files.
    static const std::list<std::string> _scriptDirs;

    // The name of the file we represent. Children shouldn't have direct
    // access to this; they should use getFilename.
    const std::string _filename;

    // A list of acceptable file name suffixes. Must be populated by
    // children classes in the constructor.
    const std::list<std::string> _suffixes;

};

#endif // __Script_hh
