/* $Id: StyleLoader.cc,v 1.18 2004/02/08 00:38:06 pervago Exp $ */

/**
 * @file   StyleLoader.cc
 * @date   21 Jun 2003
 *
 * @brief Definition of StyleLoader class which handles all of the
 * style loading
 *
 */

#include "StyleLoader.hh"
#include "Image.hh"
#include "Screen.hh"
#include "Resources.hh"
#include "Logger.hh"
#include "DrawOp.hh"
#include "Texture.hh"

extern "C" {
#ifdef    HAVE_STDIO_H
#  include <stdio.h>
#endif // HAVE_STDIO_H

#ifdef    STDC_HEADERS
#  include <string.h>
#endif // STDC_HEADERS
}

#include <algorithm>
#include <iostream>
#include <utility>
#include <sstream>
#include <string>
#include <vector>

using namespace std;

StyleLoader::StyleLoader()
{
}

inline void limit(int& value, int minValue, int maxValue)
{
    if (value < minValue) value = minValue;
    if (value > maxValue) value = maxValue;
}

inline void limit(unsigned int& value,
                  unsigned int minValue, unsigned int maxValue)
{
    if (value < minValue) value = minValue;
    if (value > maxValue) value = maxValue;
}

/**
 * @fn    readDatabaseColor(string rname,
 *                          KaColor *color,
 *                          unsigned long default_pixel,
 *                          KaImageControl *ic)
 * @brief Reads a color
 *
 * Reads a color from resource database.
 *
 * @param rname Resource name to use
 * @param default_pixel Pixel value to use if resource doesn't exist
 * @param ic KaImageControl to use for parsing color
 */
void StyleLoader::readDatabaseColor(const std::string& rname,
                                    KaColor *color,
                                    unsigned long default_pixel,
                                    KaImageControl *ic) {
    string value;
    int transparency;

    if (getValue(rname, value)) {
        value = trim(value);
        ic->parseColor(color, value.c_str());
    } else {
        ic->parseColor(color);
        color->setPixel(default_pixel);
    }

    // By some bonehead decision waimea styles use 'opacity' to mean
    // 'transparency'
    if (getValue(string(rname) + ".opacity", value) ||
        getValue(string(rname) + ".transparency", value))
        transparency = atoi(value.c_str());
    else
        transparency = 0;

    if (transparency > 100) transparency = 100;
    else if (transparency < 0) transparency = 0;

    color->setXftTransparency(transparency);
}

ImColor StyleLoader::parseColor(string colorString,
                                int transparency) {
    XColor color;
    color.pixel = 0;

    size_t slash_idx = colorString.find('/');

    if (slash_idx != string::npos) {
        transparency = atoi(colorString.substr(slash_idx + 1).c_str());
        colorString = colorString.substr(0, slash_idx);
    }

    if (! XParseColor(kascreen->display, kascreen->ic->getColormap(),
                      colorString.c_str(), &color))
        logger << warn << "color parse error: \"" << colorString << "\"" << endl;
    else if (! XAllocColor(kascreen->display, kascreen->ic->getColormap(),
                           &color))
        logger << warn << "color alloc error: \"" << colorString << "\"" << endl;

    return ImColor(color.red >> 8, color.green >> 8, color.blue >> 8,
                   255 - (255 * transparency / 100));    
}

ImColor StyleLoader::getColor(const std::string& prefix,
                              const std::string& suffix) {
    string colorString;

    if (!getValue(prefix + "." + suffix, colorString))
        return ImColor(255, 255, 255);
    else {
        int transparency = 0;
        getIntegerValue(prefix + ".transparency", transparency,
                        0, 100, 0);
        // Backward compatibility
        getIntegerValue(prefix + ".opacity", transparency,
                        0, 100, transparency);

        return parseColor(colorString, transparency);        
    }
}

void StyleLoader::readTexture(KaScreen* kascreen,
                              const std::string& rname,
                              DrawOps& ops) {
    string tex, colorValue;

    if (!getValue(rname, tex)) {
        ops.push_back(
                      shared_ptr<DrawOp>(new DrawOpFillRect(
                                                            ImColor(255, 255, 255),
                                                            "0", "0", "width", "height")));
        return;
    }

    tex = lower(tex);

    stringstream ss(tex);
    string token;

    if (tex == "parentrelative") {
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpFillRect(ImColor(0, 0, 0, 0),
                                          "0", "0", "width", "height")));
        return;
    }

    bool background = false;

    while (ss >> token) {
        if (token == "solid") {
            background = true;
            ops.push_back(shared_ptr<DrawOp>
                          (new DrawOpFillRect(getColor(rname, "color"),
                                              "0", "0", "width", "height")));
        } else if (token == "gradient") {
            background = true;

            vector<pair<ImColor, int> > gradient;
            gradient.push_back(make_pair(getColor(rname, "color"), 1));
            gradient.push_back(make_pair(getColor(rname, "colorTo"), 1));

            double angle = -90;
            if (tex.find("vertical") != string::npos) {
                angle = 0;
            } else if (tex.find("crossdiagonal") != string::npos) {
                angle = 45;
            }

            ops.push_back(shared_ptr<DrawOp>
                          (new DrawOpGradient(gradient,
                                              "0", "0", "width", "height",
                                              angle)));
        } else if (token == "pixmap") {
            string filename;
            if (getValue(rname + ".pixmap", filename)) {
                if (filename.find('/') == string::npos)
                    filename = string(kascreen->config.style_file) + "/" +
                        filename;
                background = true;

                int transparency = 0;
                getIntegerValue(rname + ".transparency", transparency,
                                0, 100, 0);
                // Backward compatibility
                getIntegerValue(rname + ".opacity", transparency,
                                0, 100, transparency);

                unsigned char alpha = 255 - (255 * transparency / 100);

                unsigned int bLeft, bTop, bRight, bBottom;
                bLeft = bTop = bRight = bBottom = 0;

                ImageDrawType drawType = DrawTiled;

                if (tex.find("stretched") != string::npos) {
                    drawType = DrawStretched;

                    string borderString;
                    if (getValue(rname + ".border", borderString)) {
                        sscanf(borderString.c_str(), "{ %u, %u, %u, %u }",
                               (unsigned int *) &bLeft,
                               (unsigned int *) &bRight,
                               (unsigned int *) &bTop,
                               (unsigned int *) &bBottom);
                    }
                } else if (tex.find("scaled") != string::npos) {
                    drawType = DrawScaled;
                }
                ops.push_back(shared_ptr<DrawOp>
                              (new DrawOpImage(filename, 0, 0, alpha, 
                                               drawType,
                                               bLeft,
                                               bTop, bRight, bBottom))); 
            }
        }
    }
    
    // If we have some kind of style specified but nothing to fill the 
    // background with, assume "solid". 
    if (!background && !ops.empty()) {
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpFillRect(getColor(rname, "color"),
                                          "0", "0", "width", "height")));
    }

    bool raised = tex.find("raised") != string::npos;
    bool sunken = tex.find("sunken") != string::npos;

    if (raised || sunken) {
        int bevelType = raised ? 1 : 0;

        ImColor color(getColor(rname, "color"));
        color.alpha = 192;

        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpBevel(color, 1, bevelType)));
    }

    if (tex.find("interlaced") != string::npos) {
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpInterlace(64, 64, getColor(rname, "color"))));
    }

    if (ops.empty() && tex.find("flat") != string::npos)
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpFillRect(getColor(rname, "color"),
                                          "0", "0", "width", "height")));
    if (ops.empty()) {
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpFillRect(ImColor(255, 255, 255),
                                          "0", "0", "width", "height")));
    }
}

void StyleLoader::readTexture(KaScreen* kascreen,
                              const std::string& rname,
                              shared_ptr<Texture>& texture) {
    DrawOps ops;
    readTexture(kascreen, rname, ops);
    texture = shared_ptr<Texture>(new Texture(ops));
}


/**
 * @fn    readDatabaseFont(string rname,
 *                         KaFont *font, KaFont *defaultfont)
 * @brief Reads a font
 *
 * Reads a font from resource database.
 *
 * @param rname Resource name to use
 * @param font Pointer to KaFont structure
 * @param defaultfont Font to use if resource doesn't exist
 * @return true if font resource was found, false if defaultfont was used
 */
bool StyleLoader::readDatabaseFont(const string& rname,
                                   KaFont& font, KaFont& defaultfont) {
    string value;

    if (getValue(rname, value)) {
        font.fontNames.push_back(value);
        
        for (int i=2;i<=10;i++) {
            if (!getValue(rname + to_string(i), value)) break;
            font.fontNames.push_back(value);
        }
        return true;
    } else {
        font.xft = defaultfont.xft;
        font.fontNames = defaultfont.fontNames;
        return false;
    }
}

void StyleLoader::readRoundedCorners(const string& name,
                                     bool& topLeft, bool& topRight,
                                     bool& bottomLeft, bool& bottomRight)
{
    string value;
    if (!getValue(name, value)) return;

    string tok;
    stringstream ss(value);
    while (ss >> tok) {
        tok = lower(tok);

        if (tok == "topleft")
            topLeft = true;
        else if (tok == "topright")
            topRight = true;
        else if (tok == "bottomleft")
            bottomLeft = true;
        else if (tok == "bottomright")
            bottomRight = true;
    }
}

void StyleLoader::getIntegerValue(const string& name,
                                  int& value,
                                  int min, int max,
                                  int defaultValue) {
    string str;
    if (!(getValue(name, str) &&
          sscanf(str.c_str(), "%d", &value) == 1))
        value = defaultValue;
    limit(value, min, max);
}

void StyleLoader::getIntegerValue(const string& name,
                                  unsigned int& value,
                                  int min, int max,
                                  int defaultValue) {
    string str;
    if (!(getValue(name, str) &&
          sscanf(str.c_str(), "%d", &value) == 1))
        value = defaultValue;
    limit(value, min, max);
}

int StyleLoader::readJustify(const string& name, int defaultValue) {
    string value;

    if (getValue(name, value)) {
        value = lower(value);
        if (value.find("right") != string::npos)
            return RightJustify;
        else if (value.find("center") != string::npos)
            return CenterJustify;
        else
            return LeftJustify;
    } else {
        return defaultValue;
    }    
}

void StyleLoader::readFonts() {
    WindowStyle *wstyle = &kascreen->style.window;
    MenuStyle   *mstyle = &kascreen->style.menu;

    KaFont default_font;

    default_font.fontNames.push_back("arial:pixelsize=12 [xft]");

    char* resnames[] = {
        "window.font",
        "menu.frame.font",
        "menu.title.font",
        "menu.bullet.font",
        "menu.checkbox.true.font",
        "menu.checkbox.false.font"
    };

    KaFont* fonts[] = {
        &wstyle->focused.font,
        &mstyle->frame.font,
        &mstyle->title.font,
        &mstyle->bullet.font,
        &mstyle->checkboxTrue.font,
        &mstyle->checkboxFalse.font
    };

    KaFont* defaults[] = {
        &default_font,
        &wstyle->focused.font,
        &mstyle->frame.font,
        &mstyle->frame.font,
        &mstyle->frame.font,
        &mstyle->checkboxTrue.font
    };

    for (unsigned int i=0;i<sizeof(resnames)/sizeof(char*);i++) {
        KaFont* font        = fonts[i];
        KaFont* defaultFont = defaults[i];

        readDatabaseFont(resnames[i], *font, *defaultFont);
    }    
}

void StyleLoader::readPadding(const std::string& name, Padding& padding) {
    string value;
    if (getValue(name, value)) {
        stringstream ss(value);
        ss >> padding.left >> padding.top >> padding.right >> padding.bottom;
    }
}
