/* $Id: TitleBar.cc,v 1.10 2004/01/18 20:58:19 pervago Exp $ */
/**
 * @file   TitleBar.cc
 * @date   Sun Nov  9 11:42:49 2003
 * 
 * @brief Implements TitleBar class which handles drawing window's
 * title bar, label and buttons.
 * 
 * 
 */


#include "ChildWindow.hh"
#include "TitleBar.hh"
#include "Window.hh"
#include "Screen.hh"
#include "Resources.hh"

#include "Imlib2pp.hh"

#include "DrawOp.hh"
#include "ImageControl.hh"
#include "Texture.hh"

#include <iostream>

using std::make_pair;

TexturePtr closeButton(ImColor color) {
    DrawOps ops;

    ops.push_back(shared_ptr<DrawOp>
                  (new DrawOpLine(color, 
                                 "3", "3", 
                                  "width-3", "height-3")));
    // /
    ops.push_back(shared_ptr<DrawOp>
                  (new DrawOpLine(color, 
                                 "width-3","3",
                                  "3", "height-3")));

    return TexturePtr(new Texture(ops));
}

TexturePtr minimizeButton(ImColor color) {
    DrawOps ops;
    ops.push_back(shared_ptr<DrawOp>
                  (new DrawOpLine(color, 
                                 "3", "height-3",
                                  "width-3", "height-3")));

    return TexturePtr(new Texture(ops));
}

TexturePtr maximizeButton(ImColor color) {
    DrawOps ops;
    ops.push_back(shared_ptr<DrawOp>
                  (new DrawOpRect(color, 
                                 "3", "3",
                                  "height-5", "height-5")));

    return TexturePtr(new Texture(ops));
}


TitleBar::TitleBar(KaWindow* kawin, Window parent) 
    : KaChildWindow(kawin, parent, TitleType)
{   
    createWindow(parent);

    xftdraw = XftDrawCreate(display, (Drawable) id,
                            kascreen->visual, kascreen->colormap);

    focusedTexture = kawin->kascreen->style.window.focused.titleTexture;
    unfocusedTexture = kawin->kascreen->style.window.unfocused.titleTexture;

    pixmap = None;
}

TitleBar::~TitleBar() {
    XftDrawDestroy(xftdraw);
}

void TitleBar::drawLabel(Pixmap pixmap, unsigned int width, 
                              unsigned int height) {
    int x = ka->label->attrib.getX() + kascreen->style.window.borderThickness;
    int y = kascreen->style.window.yPosition + ka->label->attrib.getY() + 
        kascreen->style.window.borderThickness;

    KaFont *kafont = (ka->has_focus)?
        &kascreen->style.window.focused.font:
        &kascreen->style.window.unfocused.font;

    int textWidth = kafont->Width(display, ka->labeltext);
    int numChars = ka->labeltext.length();

    // find the optimum number of characters
    while (numChars > 0 && textWidth > (ka->label->attrib.getWidth() - 10)) {
        numChars--;
        textWidth = kafont->Width(display, ka->labeltext.substr(0, numChars));
    }
    
    switch (kascreen->style.window.justify) {    
    case LeftJustify: 
        x += 2; 
        break;
    case CenterJustify:
        x += (ka->label->attrib.getWidth() / 2) - (textWidth / 2);
        break;
    case RightJustify:
        x += (ka->label->attrib.getWidth() - textWidth) - 2;
        break;
    }

    // draw the title bar with only the number of characters that will fit
    XftDrawChange(xftdraw, pixmap);
    kafont->Draw(display, pixmap, xftdraw, x, y, ka->labeltext, numChars);
}

void TitleBar::drawButton(DrawOps& add_to_ops, KaWindowButton* button,
                               bool focused) {
    char buf[1024];

    const ButtonStateStyle* bss;
    if (button->isActive()) {
        if (button->pressed)
            bss = &button->bstyle->pressedActive;
        else if (focused)
            bss = &button->bstyle->focusedActive;
        else
            bss = &button->bstyle->unfocusedActive;
    } else {
        if (button->pressed)
            bss = &button->bstyle->pressed;
        else if (focused)
            bss = &button->bstyle->focused;
        else
            bss = &button->bstyle->unfocused;        
    }

    if (buttonCache.count(bss)) {
        const DrawOps& cached_ops = buttonCache[bss];

        for (unsigned int i=0;i<cached_ops.size();i++)
            add_to_ops.push_back(cached_ops[i]);
        return;
    }

    DrawOps& ops = buttonCache[bss];

    const Padding& padding = kascreen->style.window.buttonPadding;

    int bw = kascreen->style.window.borderThickness;

    std::string buttonSize = "height - (" + 
        to_string(2 + 2 * bw + padding.top + padding.bottom) + ")";

    if (button->g_x < 0)
        sprintf(buf, "width - (%d + %s) + %d - %d", 
                -button->g_x, buttonSize.c_str(), bw, padding.right);
    else
        sprintf(buf, "%d + %d + %d", button->g_x, bw, padding.left);

    
    char y_buf[64];
    sprintf(y_buf, "2 + %d + %d", bw, padding.top);


    ops.push_back(shared_ptr<DrawOp>
                  (new DrawOpTexture(bss->texture, buf,
                                     y_buf, buttonSize, buttonSize)));    

    if (button->bstyle->fg) {
        TexturePtr texture;
        ImColor color(bss->color.getRed(), bss->color.getGreen(), 
                      bss->color.getBlue());
        switch (button->bstyle->cb) {
        case MaxCBoxType: 
            texture = maximizeButton(color); 
            break;
        case MinCBoxType:
        case ShadeCBoxType:
            texture = minimizeButton(color);
            break;
        case CloseCBoxType:
            texture = closeButton(color);
            break;        
        default:
            texture = TexturePtr(new Texture(DrawOps()));
        }

        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpTexture(texture,
                                        buf,
                                         y_buf, buttonSize, buttonSize)));
    }

    for (unsigned int i=0;i<ops.size();i++)
        add_to_ops.push_back(ops[i]);
}

TexturePtr TitleBar::draw(bool focused) {
    DrawOps ops;

    unsigned int border_w = kascreen->style.window.borderThickness;

    char border_buf[64];
    sprintf(border_buf, "%d", border_w);

    char y_buf[64];
    sprintf(y_buf, "2 + %d", border_w);


    if (kascreen->style.window.usesDrawing) {
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpTexture(focused ? 
                                         focusedTexture : unfocusedTexture, 
                                         "0", "0", 
                                         "width", "height")));
    } else {
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpTexture(focused ? 
                                         focusedTexture : unfocusedTexture, 
                                         border_buf, border_buf, 
                                         "width", "height")));
    }

    TexturePtr labelTexture = focused ? 
        kascreen->style.window.focused.labelTexture :
        kascreen->style.window.unfocused.labelTexture;

    ops.push_back(shared_ptr<DrawOp>
                  (new DrawOpTexture(labelTexture, 
                                    to_string(ka->label->g_x), 
                                    to_string(border_w + 2),
                                    "width - " + 
                                    to_string(ka->label->g_x - 
                                              ka->label->g_x2), 
                                    "height - " + 
                                     to_string(2 * border_w + 4))));

    std::list<KaWindowButton*>::iterator bit = ka->buttons.begin();
    for (; bit != ka->buttons.end(); ++bit) {
        drawButton(ops, *bit, focused);
    }

    if (!kascreen->style.window.usesDrawing && ka->border_w >= 1) {
        drawRoundedBorders(ops);
    }


    return TexturePtr(new Texture(ops));
}

void TitleBar::render() {
    if (!isVisible()) return;

    DrawOps ops;
    if (pixmap != None && pixmap != ParentRelative) 
        imlib_free_pixmap_and_mask(pixmap);

    // Delaying drawing until here because we need to know which buttons
    // do we have
    if (!focusedRenderedTexture)
        focusedRenderedTexture = draw(true);

    if (!unfocusedRenderedTexture)
        unfocusedRenderedTexture = draw(false);

    unsigned int height = attrib.getHeight();
    unsigned int width = attrib.getWidth();
    if (height <= 1 || width <= 1) return;

    struct timeval tv, tv2;

    gettimeofday(&tv, NULL);

    pixmap = control->render(id, ka->has_focus ? 
                             focusedRenderedTexture : unfocusedRenderedTexture,
                             ka->GetFrameLeft(), 
                             ka->GetFrameTop(), width, height);
    
    drawLabel(pixmap, width, height);

    XSetWindowBackgroundPixmap(display, id, pixmap);
    XClearWindow(display, id);

    ka->label->render();
        
    gettimeofday(&tv2, NULL);
    XSync(display, false);
    
    double elapsed = 
        (tv2.tv_sec + tv2.tv_usec * 1e-6) - 
        (tv.tv_sec + tv.tv_usec * 1e-6);

    printf("Rendered in %.4f second (%.2f fps)\n", elapsed, 1.0/elapsed);
}

void TitleBar::drawRoundedBorders(DrawOps& ops) {
    bool rounded_topleft = kascreen->style.window.roundedTopLeft, 
        rounded_topright = kascreen->style.window.roundedTopRight;
    KaColor borderKaColor = kascreen->style.window.borderColor;

    ImColor borderColor(borderKaColor.getRed(), 
                        borderKaColor.getGreen(), 
                        borderKaColor.getBlue());

    std::string thicknessStr = to_string(ka->border_w);

    if (rounded_topleft || rounded_topright) {
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpLine(borderColor,
                                      "0", "0", "width", "0")));
    } else {
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpFillRect(borderColor,
                                          "0", "0", "width", 
                                          thicknessStr)));
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpFillRect(borderColor,
                                          "0", "0", thicknessStr, 
                                          "height")));
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpFillRect(borderColor,
                                          "width -" + thicknessStr, "0",
                                          "width", "height")));
    }

    // Border along the bottom
    ops.push_back(shared_ptr<DrawOp>
                  (new DrawOpFillRect(borderColor,
                                      to_string(ka->left_w - ka->border_w), 
                                      "height-" + thicknessStr,
                                      "width - (" + 
                                      to_string(ka->left_w + ka->right_w - 
                                                2 * ka->border_w) + ")", 
                                      "height")));

    if (rounded_topleft) {
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpLine(borderColor,
                                      3, 1, 4, 1)));
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpLine(borderColor,
                                      2, 2, 2, 2)));
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpLine(borderColor,
                                      1, 3, 1, 4)));
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpLine(borderColor,
                                      "0", "4", "0", "height")));
    }
    if (rounded_topright) {
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpLine(borderColor,
                                      "width - 5", "1", 
                                      "width - 4", "1")));
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpLine(borderColor,
                                      "width - 3", "2", 
                                      "width - 3", "2")));
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpLine(borderColor,
                                      "width - 2", "3", 
                                      "width - 2", "4")));
        ops.push_back(shared_ptr<DrawOp>
                      (new DrawOpLine(borderColor,
                                      "width - 1", "4", 
                                      "width", "height")));
    }
}
