/* $Id: Window.cc,v 1.153 2004/02/05 11:37:29 hyriand Exp $ */

/**
 * @file   Window.cc
 * @date   02-May-2001 21:43:03
 *
 * @brief Implementation of KaWindow and KaChildWindow classes
 *
 * An instance if this class manages one window. Contains functions for
 * creating window decorations, reading window hints and controlling the
 * window.
 *
 */

#include "Window.hh"
#include "Net.hh"
#include "ChildWindow.hh"
#include "TitleBar.hh"
#include "FrameGeometry.hh"
#include "Event.hh"
#include "Dockapp.hh"
#include "Menu.hh"
#include "MenuItem.hh"
#include "Timer.hh"
#include "Logger.hh"
#include "Kahakai.hh"

extern "C" {
#include <X11/extensions/shape.h>
#ifdef    HAVE_STDIO_H
# include <stdio.h>
#endif // HAVE_STDIO_H

#ifdef    STDC_HEADERS
# include <string.h>
#endif // STDC_HEADERS
}

#include <assert.h>

#include <algorithm>
#include <iostream>
#include <sstream>
#include <list>

using namespace std;

/**
 * @fn    KaWindow(Window win_id, KaScreen *scrn) :
 *        WindowObject(win_id, WindowType)
 * @brief Constructor for KaWindow class
 *
 * Reparents the window, reads window hints and creates decorations ...
 *
 * @param win_id Resource ID of window to manage
 * @param scrn The screen the window should be displayed on
 */
KaWindow::KaWindow(Window win_id, KaScreen *scrn) :
    WindowObject(win_id, WindowType) {

    id = win_id;
    kascreen = scrn;

    InitVars(0);

    GetAttributes();

    InitVars(1);

    net->GetWMHints(this);
    net->GetMWMHints(this);
    net->GetWMNormalHints(this);
    net->GetWmPid(this);

    Gravitate(ApplyGravity);
    InitPosition();

    BuildDecorations();

    if (deleted) {
        delete this;
        return;
    }

    net->GetWmState(this);
    net->GetWmType(this);
    net->GetVirtualPos(this);
    net->GetWmStrut(this);
    net->GetDesktop(this);
    net->SetDesktop(this);
    net->SetDesktopMask(this);

    ReparentWin();

    if (! net->GetNetName(this))
        net->GetXaName(this);

    if (name.empty())
        SetActionLists();

    UpdateGrabs();

    if (deleted) {
        delete this;
        return;
    }

    UpdateAllAttributes();

    if (shaped)
        ShapeEvent(id, false);

    if (flags.shaded)
        Shade(true);

    kahakai->window_table.insert(make_pair(id, this));
    kascreen->kawindow_list.push_back(this);
    kascreen->kawindow_list_map_order.push_back(this);

    if (deleted) {
        delete this;
        return;
    }

    kascreen->RaiseWindow(frame->id);
    net->SetAllowedActions(this);
    net->SetWmState(this);

    this->geometry = new WindowFrameGeometry(this);

    if (flags.hidden)
        Minimize(true);
}

/**
 * @fn    ~KaWindow(void)
 * @brief Destructor of KaWindow class
 *
 * Reparents the window back to the root window if it still exists. Destroys
 * all windows used for decorations.
 */
KaWindow::~KaWindow(void) {
    kahakai->window_table.erase(id);

    if (transient_for) {
        if (transient_for == kascreen->id) {
            vector<KaWindow *>::iterator it =
                kascreen->kawindow_list.begin();
            for (; it != kascreen->kawindow_list.end(); ++it)
                (*it)->transients.remove(id);
        }
        else {
            map<Window, WindowObject *>::iterator hit;
            if ((hit = kahakai->window_table.find(transient_for)) !=
                kahakai->window_table.end()) {
                if (((*hit).second)->type == WindowType)
                    ((KaWindow *) (*hit).second)->transients.remove(id);
            }
        }
    }

    XGrabServer(display);
    if (validatedrawable(id)) {
        XRemoveFromSaveSet(display, id);
        Gravitate(RemoveGravity);
        if (flags.shaded) attrib.setHeight(restore_shade);
        if (attrib.getX() >= kascreen->width)
            attrib.setX(attrib.getX() % kascreen->width);
        if (attrib.getY() >= kascreen->height)
            attrib.setY(attrib.getY() % kascreen->height);

        if (attrib.getX() + attrib.getWidth() <= 0)
            attrib.setX(kascreen->width + (attrib.getX() % kascreen->width));
        if (attrib.getY() + attrib.getHeight() <= 0)
            attrib.setY(kascreen->height + (attrib.getY() % kascreen->height));

        XReparentWindow(display, id, kascreen->id, attrib.getX(), attrib.getY());
    }
    XUngrabServer(display);

    net->RemoveAllowedActions(this);
    net->RemoveVisibleName(this);

    list<KaWindowButton *>::iterator bit = buttons.begin();
    for (; bit != buttons.end(); ++bit)
        delete *bit;

    delete grip_l;
    delete grip_r;
    delete handle;
    delete label;
    delete title;
    delete leftEdge;
    delete rightEdge;

    kascreen->SmartNameRemove(this);
    if (host) delete [] host;
    if (pid) delete [] pid;
    if (classhint && classhint->res_name) XFree(classhint->res_name);
    if (classhint && classhint->res_class) XFree(classhint->res_class);

    kascreen->kawindow_list.erase(remove(kascreen->kawindow_list.begin(),
                                         kascreen->kawindow_list.end(),
                                         this), kascreen->kawindow_list.end());
    kascreen->kawindow_list_map_order.remove(this);

    desktop_mask = 0;
    UpdateStack();
    if (wm_strut) {
        kascreen->strut_list.remove(wm_strut);
        delete wm_strut;
        if (! kascreen->shutdown) kascreen->UpdateWorkarea();
    }

    delete frame;

    if (has_focus && kascreen->config.revert_to_window) {
        vector<KaWindow *>::iterator it = kascreen->kawindow_list.begin();
        for (; it != kascreen->kawindow_list.end(); ++it) {
            
            if (*it != this && (*it)->has_focus)
                break;

            if (! (*it)->hidden && ! (*it)->flags.hidden) {
                (*it)->Focus(false);
                break;
            }
        }
    }

    if (! kascreen->shutdown) {
        kascreen->net->SetClientList(kascreen);
        kascreen->net->SetClientListStacking(kascreen);
    }

    if (geometry)
        delete geometry;
}

void KaWindow::InitVars(int phase) {

    if (phase == 0) {

        screen_number = kascreen->screen_number;
        display = kascreen->display;
        kahakai = kascreen->kahakai;
        net = kahakai->net;
        ic = kascreen->ic;

        classhint = NULL;
        geometry = NULL;
        wm_strut = NULL;

        realnamelen =
        smartnum = 0;

        move_resize = false;

    } else if (phase == 1) {

        handle_w =
        border_w =
        left_w =
        right_w =     
        title_w = 0;

        render_if_transparency =
        shadeOnUnmaximize =
        ign_config_req =
        been_shaped =
        want_focus =
        has_focus =
        dontsend =
        deleted =
        hidden =
        mapped =
        shaped = false;

        gone_past_bottom =
        gone_past_right =
        gone_past_left =
        gone_past_top = false;

        desktop_mask = 1L << kascreen->current_desktop->number;

        flags.alwaysatbottom =
        flags.alwaysontop =
        flags.fullscreen =
        flags.sticky =
        flags.shaded =
        flags.handle =
        flags.hidden =
        flags.border =
        flags.title =
        flags.max =
        flags.hmax =
        flags.vmax =
        flags.all = false;

        flags.focusable =
        flags.tasklist = true;

        transient_for = (Window) 0;

        host = NULL;
        pid = NULL;
    }
}

void KaWindow::GetAttributes() {
    XWindowAttributes init_attrib;

    XGrabServer(display);

    if (validatedrawable(id))
        XGetWindowAttributes(display, id, &init_attrib);
    else
        deleted = true;

    XUngrabServer(display);

    attrib = WindowAttributes(init_attrib.x, init_attrib.y,
                              init_attrib.width, init_attrib.height,
                              0, 0, init_attrib.colormap);

    size.win_gravity = init_attrib.win_gravity;

    positioned_by_app = attrib.getX() || attrib.getY();
}

void KaWindow::BuildDecorations() {
    frame  = new KaWindowFrame(this, kascreen->id);
    handle = new KaWindowHandle(this, frame->id);
    grip_l = new KaWindowGrip(this, frame->id, Left);
    grip_r = new KaWindowGrip(this, frame->id, Right);
    leftEdge = new KaWindowEdge(this, frame->id, Left);
    rightEdge = new KaWindowEdge(this, frame->id, Right);
    title  = new TitleBar(this, frame->id);

    int left_end = 2, right_end = -2;

    BuildButtons(&left_end, &right_end);

    label = new KaWindowLabel(this, title->id);
    label->g_x = left_end + 2;
    label->g_x2 = right_end - 2;
}

void KaWindow::BuildButtons(int * left_end, int * right_end) {
    KaWindowButton *button;
    int tw = (signed) kascreen->style.window.titleHeight;
    
    vector<ButtonStyle>::iterator bit =
        kascreen->style.window.buttonStyles.begin();

    for (int i=0; bit != kascreen->style.window.buttonStyles.end(); ++bit,++i) {

        button = new KaWindowButton(this, title->id);
        button->bstyle = &*bit;

        if ((*bit).autoplace == WestType)
            button->g_x = *left_end;
        else if ((*bit).autoplace == EastType)
            button->g_x = *right_end;
        else
            button->g_x = (*bit).x;

        if (button->g_x > 0 && (button->g_x + (tw - 2)) > *left_end)
            *left_end = button->g_x + (tw - 2);
        else if ((button->g_x - (tw - 2)) < *right_end)
            *right_end = button->g_x - (tw - 2);

        buttons.push_back(button);
    }
}

/* Position functions */

int KaWindow::GetFrameLeft() {
    return attrib.getLeft() - left_w;
}

int KaWindow::GetFrameTop() {
    int top = attrib.getTop();

    if (flags.title)
        top = top - title_w;


    return top;
}

int KaWindow::GetFrameBottom() {
    int bottom = attrib.getBottom();
    if (flags.handle)
        bottom += handle_w;

    return bottom;
}

int KaWindow::GetFrameRight() {
    return attrib.getRight() + right_w;
}

int KaWindow::GetFrameWidth() {
    return GetFrameRight() - GetFrameLeft();
}

int KaWindow::GetFrameHeight() {
    return GetFrameBottom() - GetFrameTop();
}

void KaWindow::SetFrameLeft(int x) {
    attrib.setX(x + left_w);
}

void KaWindow::SetFrameRight(int x) {
    SetFrameLeft(x - GetFrameWidth());
}

void KaWindow::SetFrameTop(int y) {
    attrib.setY(y);
    if (flags.title)
        attrib.setY(attrib.getY() + title_w);
}

void KaWindow::SetFrameBottom(int y) {
    SetFrameTop(y - GetFrameHeight());
}

void KaWindow::SetFrameWidth(int w) {
    attrib.setWidth(attrib.getWidth() + (w - GetFrameWidth()));
}

void KaWindow::SetFrameHeight(int h) {
    attrib.setHeight(attrib.getHeight() + (h - GetFrameHeight()));
}

void KaWindow::SkipTaskList() {
    flags.tasklist = false;
    net->SetWmState(this);
}


/**
 * @fn    SetActionLists(void)
 * @brief Set all actions lists
 *
 * Updates all action lists for the window.
 */
void KaWindow::SetActionLists(void) {
    if (has_focus)
        actionlist = &kascreen->config.awinacts;
    else
        actionlist = &kascreen->config.pwinacts;

    frame->actionlist = &kascreen->config.frameacts;
    title->actionlist = &kascreen->config.titleacts;
    label->actionlist = &kascreen->config.labelacts;
    handle->actionlist = &kascreen->config.handleacts;
    grip_l->actionlist = &kascreen->config.lgacts;
    grip_r->actionlist = &kascreen->config.rgacts;

    leftEdge->actionlist = &kascreen->config.lweacts;
    rightEdge->actionlist = &kascreen->config.rweacts;

    list<KaWindowButton *>::iterator it = buttons.begin();
    for (; it != buttons.end(); ++it) {
        (*it)->actionlist = &kascreen->config.bacts[(*it)->bstyle->action];
    }
}

/**
 * @fn    Gravitate(int multiplier)
 * @brief Applies or removes window gravity
 *
 * If multiplier parameter is RemoveGravity when we set the window
 * attributes so that the frame will match the windows position.
 * If multiplier parameter is ApplyGravity when we set the window
 * attributes so that the window will match the frames position.
 *
 * @param multiplier ApplyGravity or RemoveGravity
 */
void KaWindow::Gravitate(int multiplier) {
    int hdir = 0, vdir = 0; // (-1, -1) is north-west, (1, 1) is south-east

    switch (size.win_gravity) {
    case NorthWestGravity:
        hdir = -1; vdir = -1;
        break;
    case NorthEastGravity:
        hdir = +1; vdir = -1;
        break;
        //        attrib.setX(attrib.getX() - multiplier * left_w);
    case NorthGravity:
        hdir = 0; vdir = -1;
        break;
    case SouthWestGravity:
        hdir = -1; vdir = +1;
        break;
    case SouthEastGravity:
        hdir = +1; vdir = +1;
        break;
    case SouthGravity:
        hdir = 0; vdir = +1;
        break;
    case CenterGravity:
        attrib.setX(attrib.getX() + multiplier * (left_w / 2));
        //       attrib.setY(attrib.getY() + multiplier * (border_w / 2));
        if (title_w) attrib.setY(attrib.getY() + multiplier * ((title_w) / 2));
        break;
    case StaticGravity:
        break;
    }

    if (hdir == -1) {
        attrib.setX(attrib.getX() + multiplier * left_w);
    } else if (hdir == 1) {
        attrib.setX(attrib.getX() - multiplier * left_w / 2);
    }

    if (vdir == -1) {
        if (title_w)
            attrib.setY(attrib.getY() + multiplier * (title_w));
    } else if (vdir == 1) {
        if (handle_w) attrib.setY(attrib.getY() - multiplier * (handle_w));
    }
}

/**
 * @fn    InitPosition(void)
 * @brief Sets window position
 *
 * Initializes position for the window.
 */
void KaWindow::InitPosition(void) {

    // XXX shaded windows get stuck at the top of the screen

    if (size.min_width > attrib.getWidth()) attrib.setWidth(size.min_width);
    if (size.min_height > attrib.getHeight()) attrib.setHeight(size.min_height);
    restore_max_attrib.setX(attrib.getX());
    restore_max_attrib.setY(attrib.getY());
    restore_max_attrib.setWidth(attrib.getWidth());
    restore_shade = attrib.getHeight();
    restore_max_attrib.setHeight(attrib.getHeight());
    restore_max_attrib.setMisc0(0);
    restore_max_attrib.setMisc1(0);
    old_attrib.setX(- 0xffff);
    old_attrib.setY(- 0xffff);
    old_attrib.setHeight(- 0xffff);
    old_attrib.setWidth(- 0xffff);
}

/**
 * @fn    MapWindow(void)
 * @brief Map window
 *
 * Map client window and all child windows.
 */
void KaWindow::MapWindow(void) {
    XGrabServer(display);
    if (validatedrawable(id)) {
        XMapWindow(display, id);
        RedrawWindow();
    } else DELETED;
    XUngrabServer(display);

    if (flags.border && left_w | right_w) {
        if(left_w)
          XMapRaised(display, leftEdge->id);
        if(right_w)
          XMapRaised(display, rightEdge->id);
    } else {
        XUnmapWindow(display, leftEdge->id);
        XUnmapWindow(display, rightEdge->id);
    }

    if (flags.handle && handle_w) {
        XMapRaised(display, grip_l->id);
        XMapRaised(display, handle->id);
        XMapRaised(display, grip_r->id);
    } else {
        XUnmapWindow(display, grip_l->id);
        XUnmapWindow(display, handle->id);
        XUnmapWindow(display, grip_r->id);
    }
    if (flags.title && title_w) {
        XMapRaised(display, title->id);
        XMapRaised(display, label->id);
        list<KaWindowButton *>::iterator bit = buttons.begin();
        for (; bit != buttons.end(); ++bit)
            XMapRaised(display, (*bit)->id);
    } else {
        XUnmapWindow(display, title->id);
        XUnmapWindow(display, label->id);
        list<KaWindowButton *>::iterator bit = buttons.begin();
        for (; bit != buttons.end(); ++bit)
            XUnmapWindow(display, (*bit)->id);
    }    

    UpdateStack();
    if (desktop_mask & (1L << kascreen->current_desktop->number)) {
        XMapWindow(display, frame->id);
    } else {
        hidden = true;
    }
    mapped = true;
}

/**
 * @fn    Show(void)
 * @brief Show window
 *
 * Map frame window if not mapped.
 */
void KaWindow::Show(void) {
    if (! flags.hidden && hidden && mapped) {
        XMapWindow(display, frame->id);
        hidden = false;
    }
}

/**
 * @fn    Hide(void)
 * @brief Hide window
 *
 * Unmap frame window if mapped.
 */
void KaWindow::Hide(void) {
    if (! hidden) {
        if (has_focus && kascreen->config.revert_to_window) {
            vector<KaWindow *>::iterator it = kascreen->kawindow_list.begin();
            for (; it != kascreen->kawindow_list.end(); ++it)
                if (! (*it)->hidden && ! (*it)->flags.hidden) {
                    (*it)->Focus(false);
                    break;
                }
        }
        XUnmapWindow(display, frame->id);
        hidden = true;
    }
}

void KaWindow::nameUpdated() {
    if (kascreen->style.window.fitText) {
        UpdateTitlebar();
        Shape();
    }
}

/**
 * @fn    UpdateTitlebar(void)
 * @brief Updates window titlebar
 *
 * Calculates all titlebar posisitons and sizes.
 */
void KaWindow::UpdateTitlebar(void) {
    if (! flags.title || ! title_w) return;
    
    titleBarWidth = frame->attrib.getWidth();

    if (kascreen->style.window.titleBarWidthPercent > 0) {
        titleBarWidth = 
            int(titleBarWidth * 
                (kascreen->style.window.titleBarWidthPercent)/100.0);
    } else if (kascreen->style.window.titleBarWidth > 0)
         titleBarWidth = kascreen->style.window.titleBarWidth;

    if (kascreen->style.window.fitText) {
        // Fit the titlebar to be the required width but no more
        // than titleBarWidth
        KaFont *kafont = has_focus ?
            &kascreen->style.window.focused.font:
            &kascreen->style.window.unfocused.font;
        
        int width = kafont->Width(kascreen->display, labeltext.c_str()) + border_w +
            buttons.size() * (title_w + border_w);

        if (width < titleBarWidth) titleBarWidth = width;
    }

    title->attrib.setWidth(titleBarWidth);   
    title->attrib.setX(0);
    title->attrib.setY(0);
    title->attrib.setHeight(title_w);
    //    XSetWindowBorderWidth(display, title->id, border_w);
    title->updateAttributes();

    KaChildWindow *l = label;
    l->attrib.setX(l->g_x);
    l->attrib.setY(2);
    l->attrib.setWidth(title->attrib.getWidth() + l->g_x2 - label->g_x);
    if (l->attrib.getWidth() < 1) l->attrib.setWidth(1);
    l->attrib.setHeight(title_w - 4);
    l->updateAttributes();

    list<KaWindowButton *>::iterator bit = buttons.begin();
    for (; bit != buttons.end(); ++bit) {
        if((*bit)->g_x > 0)
            (*bit)->attrib.setX((*bit)->g_x);
        else
            (*bit)->attrib.setX((title->attrib.getWidth() + (*bit)->g_x - (title_w - 4)));
        (*bit)->attrib.setY(2);
        (*bit)->attrib.setWidth(title_w - 4);
        (*bit)->attrib.setHeight(title_w - 4);
        (*bit)->updateAttributes();
    }

    if (kascreen->config.db) {
        Region region = XCreateRegion();
        XRectangle xrect;
        xrect.x = l->g_x;
        xrect.y = 2;
        xrect.width = l->attrib.getWidth();
        xrect.height = l->attrib.getHeight();
        XUnionRectWithRegion(&xrect, region, region);
        XftDrawSetClip(title->xftdraw, region);
        XDestroyRegion(region);
    }

    DrawTitlebar(true);
}

/**
 * @fn    UpdateAllAttributes()
 * @brief Updates all window attrubutes
 *
 * Updates all positions and sizes of all windows in the frame.
 */
void KaWindow::UpdateAllAttributes() {

    Gravitate(RemoveGravity);
    border_w = (flags.border * kascreen->style.window.borderThickness);
    title_w  = (flags.title * 
                (kascreen->style.window.titleHeight + 2 * border_w));
    handle_w = (flags.handle * kascreen->style.window.handleThickness);
    left_w   = (flags.border * kascreen->style.window.leftEdgeWidth);
    right_w  = (flags.border * kascreen->style.window.rightEdgeWidth);

    Gravitate(ApplyGravity);

    heightChanged = true;

    frame->attrib.setX(GetFrameLeft());
    frame->attrib.setY(GetFrameTop());
    frame->attrib.setWidth(attrib.getWidth() + left_w);
    frame->attrib.setHeight(attrib.getHeight() + right_w);
    
    if (! flags.shaded) {
        XResizeWindow(display, frame->id, frame->attrib.getWidth(),
                      frame->attrib.getHeight());
    }

    XMoveWindow(display, frame->id, frame->attrib.getX(), 
                frame->attrib.getY());

    if (flags.title)
        UpdateTitlebar();

    // Edge
    if (left_w <= 0) 
        XUnmapWindow(display, leftEdge->id);
    else {
        leftEdge->attrib.setX(0);
        leftEdge->attrib.setY(title_w);
        leftEdge->attrib.setWidth(left_w);
        leftEdge->attrib.setHeight(frame->attrib.getHeight());
        leftEdge->updateAttributes();
        leftEdge->render();
    }
    
    if (right_w <= 0) 
        XUnmapWindow(display, rightEdge->id);
    else {
        rightEdge->attrib.setX(frame->attrib.getWidth() - right_w);
        rightEdge->attrib.setY(title_w);
        rightEdge->attrib.setWidth(right_w);
        rightEdge->attrib.setHeight(frame->attrib.getHeight());
        rightEdge->updateAttributes();
        rightEdge->render();
    }
    
    if (kascreen->style.window.handleThickness <= 0)
        XUnmapWindow(display, handle->id);
    else if (flags.handle) {
        handle->attrib.setX(25);
        handle->attrib.setY(attrib.getHeight() + title_w);
        handle->attrib.setWidth(frame->attrib.getWidth() - 50);
        if (handle->attrib.getWidth() < 1) handle->attrib.setWidth(1);
        handle->attrib.setHeight(kascreen->style.window.handleThickness);
        handle->updateAttributes();

        grip_l->attrib.setX(0);
        grip_l->attrib.setY(handle->attrib.getY());
        grip_l->attrib.setWidth(25);
        grip_l->attrib.setHeight(kascreen->style.window.handleThickness);
        grip_l->updateAttributes();

        grip_r->attrib.setX(handle->attrib.getRight());            
        grip_r->attrib.setY(handle->attrib.getY());            
        grip_r->attrib.setWidth(25);
        grip_r->attrib.setHeight(kascreen->style.window.handleThickness);
        grip_r->updateAttributes();
        DrawHandlebar(true);
    }

    int t_height = title_w;
    XGrabServer(display);
    if (validatedrawable(id)) {
        XMoveWindow(display, id, left_w, t_height);
    } else DELETED;
    XUngrabServer(display);

    int m_x, m_y, m_w, m_h;
    if (flags.max) {
        m_x = restore_max_attrib.getX();
        m_y = restore_max_attrib.getY();
        m_w = restore_max_attrib.getWidth();
        m_h = restore_max_attrib.getHeight();
        flags.max = false;
        _Maximize(restore_max_attrib.getMisc0(), restore_max_attrib.getMisc1(), flags.hmax, flags.vmax);
        restore_max_attrib.setX(m_x);
        restore_max_attrib.setY(m_y);
        restore_max_attrib.setWidth(m_w);
        restore_max_attrib.setHeight(m_h);
    } else
        RedrawWindow();


    Shape();
}

/**
 * @fn    RedrawWindow(bool force_if_viewable)
 * @brief Redraws Window
 *
 * Redraws the window at it's correct position with it's correct size.
 * We only redraw those things that need to be redrawn.
 *
 * @param force_if_viewable Force redraw if window is viewable
 */
void KaWindow::RedrawWindow(bool force_if_viewable, bool do_shape) {

    Bool move = false, resize = false;
    bool vresize = false, hresize = false;

    if (!flags.max)
        flags.hmax = flags.vmax = false;

    if (old_attrib.getX() != attrib.getX()) {
        frame->attrib.setX(attrib.getX() - left_w);        
        //        attrib.setX(attrib.getX() + left_w);

        old_attrib.setX(attrib.getX());

        move = true;
    }
    if (old_attrib.getY() != attrib.getY()) {
        frame->attrib.setY(attrib.getY());
        if (flags.title)
            frame->attrib.setY(frame->attrib.getY() - title_w);
        old_attrib.setY(attrib.getY());

        move = true;
    }
    if (old_attrib.getWidth() != attrib.getWidth()) {
        frame->attrib.setWidth(attrib.getWidth() + left_w + right_w);
        old_attrib.setWidth(attrib.getWidth());

        resize = true;
        hresize = true;

        if (flags.title) {
            if (! force_if_viewable) UpdateTitlebar();
        }
        if (flags.handle) {
            handle->attrib.setWidth(frame->attrib.getWidth() - 50);
            if (handle->attrib.getWidth() < 1) handle->attrib.setWidth(1);
            grip_r->attrib.setX(handle->attrib.getRight());

            XMoveWindow(display, grip_r->id, grip_r->attrib.getX(),
                        grip_r->attrib.getY());
            XResizeWindow(display, handle->id, handle->attrib.getWidth(),
                          handle->attrib.getHeight());

            if (! force_if_viewable) DrawHandlebar();
        }
        if (flags.border) {
            leftEdge->updateAttributes();
            rightEdge->attrib.setX(frame->attrib.getWidth() - right_w);
            rightEdge->updateAttributes();
            if (! force_if_viewable) DrawEdges();
        }
    }
    if (heightChanged || old_attrib.getHeight() != attrib.getHeight()) {
        heightChanged = false;
        frame->attrib.setHeight(attrib.getHeight());
        if (flags.title)
            frame->attrib.setHeight(frame->attrib.getHeight() + title_w);
        if (flags.handle)
            frame->attrib.setHeight(frame->attrib.getHeight() + handle_w);
        old_attrib.setHeight(attrib.getHeight());

        resize = true;
        vresize = true;

        if (flags.handle) {
            handle->attrib.setY(attrib.getHeight() + title_w);
            grip_l->attrib.setY(handle->attrib.getY());
            grip_r->attrib.setY(handle->attrib.getY());

            XMoveWindow(display, handle->id, handle->attrib.getX(),
                        handle->attrib.getY());
            XMoveWindow(display, grip_l->id, grip_l->attrib.getX(),
                        grip_l->attrib.getY());
            XMoveWindow(display, grip_r->id, grip_r->attrib.getX(),
                        grip_r->attrib.getY());

            render_if_transparency = true;
            if (! force_if_viewable && ! resize) DrawHandlebar();
            render_if_transparency = false;
        }
        if (flags.border) {
            leftEdge->attrib.setHeight(max(1, frame->attrib.getHeight() - 
                                           handle_w - title_w));            
            rightEdge->attrib.setHeight(max(1, frame->attrib.getHeight() - 
                                            handle_w - title_w));
            leftEdge->updateAttributes();
            rightEdge->updateAttributes();
            leftEdge->render();
            rightEdge->render();
        }

        net->SetVirtualPos(this);
        SendConfig();
    }
    if (move) {
        if (flags.max) {
            restore_max_attrib.setMisc0(kascreen->v_x + frame->attrib.getX());
            restore_max_attrib.setMisc1(kascreen->v_y + frame->attrib.getY());
            net->SetWmState(this);
        }

        XMoveWindow(display, frame->id, frame->attrib.getX(), frame->attrib.getY());

        if (! resize && ! force_if_viewable) {
            if (! kascreen->config.lazy_trans) {
                render_if_transparency = true;
                DrawTitlebar();
                DrawHandlebar();
                DrawEdges();
                render_if_transparency = false;
            }
        }
    }
    if (force_if_viewable) {
        DrawTitlebar();
        DrawHandlebar();
        DrawEdges();
    }
    if (resize) {
        if (flags.max) {
            if (flags.hmax && hresize) {
                flags.hmax = false;
                if (!flags.hmax && !flags.vmax)
                    flags.max = false;

                net->SetWmState(this);
                UpdateButtons(MaxCBoxType|HMaxCBoxType);
                kascreen->UpdateCheckboxes(MaxCBoxType|HMaxCBoxType);
            }
            if (flags.vmax && vresize) {
                flags.vmax = false;
                if (!flags.hmax && !flags.vmax)
                    flags.max = false;

                net->SetWmState(this);
                UpdateButtons(MaxCBoxType|VMaxCBoxType);
                kascreen->UpdateCheckboxes(MaxCBoxType|VMaxCBoxType);
            }
        }
        XGrabServer(display);
        if (validatedrawable(id)) {
            //            XSetWindowBorderWidth(display, id, border_w);
            if (flags.shaded)
                XResizeWindow(display, id, attrib.getWidth(), restore_shade);
            else
                XResizeWindow(display, id, attrib.getWidth(), attrib.getHeight());
            XResizeWindow(display, frame->id, frame->attrib.getWidth(),
                          frame->attrib.getHeight());
        } else DELETED;
        XUngrabServer(display);
    }

    if (do_shape)
        Shape();
}

/**
 * @fn    ReparentWin(void)
 * @brief Reparents the window into the frame
 *
 * Sets the input mask for the managed window so that it will report
 * FocusChange and PropertyChange events. Then we reparent the window into
 * our frame and activates needed grab buttons.
 */
void KaWindow::ReparentWin(void) {
    XSetWindowAttributes attrib_set;

    XGrabServer(display);
    if (validatedrawable(id)) {
        XSelectInput(display, id, NoEventMask);
        XSetWindowBorderWidth(display, id, 0);
        XReparentWindow(display, id, frame->id, -border_w, title_w);
        XChangeSaveSet(display, id, SetModeInsert);
        XFlush(display);

        attrib_set.event_mask =
            PropertyChangeMask | StructureNotifyMask | FocusChangeMask |
            EnterWindowMask | LeaveWindowMask;
        attrib_set.do_not_propagate_mask =
            ButtonPressMask | ButtonReleaseMask | ButtonMotionMask;

        XChangeWindowAttributes(display, id, CWEventMask | CWDontPropagate,
                                &attrib_set);


        XRectangle *dummy = NULL;
        int n, order;
        if (kahakai->shape) {
            XShapeSelectInput(display, id, ShapeNotifyMask);
            dummy = XShapeGetRectangles(display, id, ShapeBounding, &n,
                                        &order);
            if (n > 1) shaped = true;
        }
        XFree(dummy);
    } else DELETED;
    XUngrabServer(display);
}

/**
 * @fn    UpdateGrabs(void)
 * @brief Update window grabs
 *
 * Updates passive window grabs for the window.
 */
void KaWindow::UpdateGrabs(void) {
    XGrabServer(display);

    if (! validateclient_mapped(id))
        DELETED;

    // remove all current grabs

    XUngrabButton(display, AnyButton, AnyModifier, id);
    XUngrabKey(display, AnyKey, AnyModifier, id);

    // go through list and grab what we need

    vector<KaAction *>::iterator it = actionlist->begin();
    for (; it != actionlist->end(); ++it) {
        if (kahakai->eh->IsButtonEventType((*it)->type)) {

            XGrabButton(display, (*it)->detail ? (*it)->detail : AnyButton,
                        AnyModifier, id, true,
                        ButtonPressMask | ButtonReleaseMask | ButtonMotionMask,
                        GrabModeSync, GrabModeSync, None, None);

        } else if (kahakai->eh->IsKeyEventType((*it)->type)) {
            KeyCode keycode = XKeysymToKeycode(display, (*it)->detail);

            if (keycode) {
                vector<unsigned int>::iterator imod_it =
                    kascreen->imod_list.begin();
                for (; imod_it != kascreen->imod_list.end(); ++imod_it)
                    XGrabKey(display, keycode, (*it)->mod | (*imod_it), id,
                             True, GrabModeSync, GrabModeSync);

                XGrabKey(display, keycode, (*it)->mod, id, True,
                        GrabModeSync, GrabModeSync);
                //AnyModifier & ~((*it)->nmod & 0x1fff), id, True,
                //GrabModeSync, GrabModeSync);
            }
        }
    }
    XUngrabServer(display);
}

/**
 * @fn    Shape(void)
 * @brief Set Shape of frame window
 *
 * Shapes frame window after shaped clients, also handles rounded corners
 * and partial titlebars, all of which affect the window's shape
 */
void KaWindow::Shape(void) {

    unsigned int emptyRight = title_w ? (frame->attrib.getWidth() - titleBarWidth) : 0;

    // A lot of credit to metacity 

    int new_window_width = GetFrameWidth();
    int new_window_height = GetFrameHeight();

    Region window_xregion;

    bool rounded_topleft = kascreen->style.window.roundedTopLeft, 
         rounded_topright = kascreen->style.window.roundedTopRight,
         rounded_bottomleft = kascreen->style.window.roundedBottomLeft, 
         rounded_bottomright = kascreen->style.window.roundedBottomRight;

    bool any_rounded = rounded_topleft || rounded_topright || 
                       rounded_bottomleft || rounded_bottomright;

    // Don't shape toolbars/splash screens/docks/panels
    if (windowType != OtherWindowType)
        return;

    if (!any_rounded && !shaped && !been_shaped &&
        !(title_w && emptyRight > 0)) {
        return;
    }

    if (!flags.title)
        rounded_topleft = rounded_topright = false;

    if (!flags.handle) 
        rounded_bottomleft = rounded_bottomright = false;

    Region corners_xregion = XCreateRegion();

    XRectangle xrect;
    if (rounded_topleft) {
            xrect.x = 0;
            xrect.y = 0;
            xrect.width = 5;
            xrect.height = 1;
      
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);

            xrect.y = 1;
            xrect.width = 3;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
      
            xrect.y = 2;
            xrect.width = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);

            xrect.y = 3;
            xrect.width = 1;
            xrect.height = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
        }

    if (rounded_topright) {            
            xrect.x = new_window_width - emptyRight - 5;
            xrect.y = 0;
            xrect.width = 5;
            xrect.height = 1;
      
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);

            xrect.y = 1;
            xrect.x = new_window_width - emptyRight - 3;
            xrect.width = 3;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
      
            xrect.y = 2;
            xrect.x = new_window_width - emptyRight - 2;
            xrect.width = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);

            xrect.y = 3;
            xrect.x = new_window_width - emptyRight - 1;
            xrect.width = 1;
            xrect.height = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
        }

    if (rounded_bottomleft) {
            xrect.x = 0;
            xrect.y = new_window_height - 1;
            xrect.width = 5;
            xrect.height = 1;
      
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
      
            xrect.y = new_window_height - 2;
            xrect.width = 3;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
      
            xrect.y = new_window_height - 3;
            xrect.width = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);

            xrect.y = new_window_height - 5;
            xrect.width = 1;
            xrect.height = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
        }

    if (rounded_bottomright) {
            xrect.x = new_window_width - 5;
            xrect.y = new_window_height - 1;
            xrect.width = 5;
            xrect.height = 1;
      
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
      
            xrect.y = new_window_height - 2;
            xrect.x = new_window_width - 3;
            xrect.width = 3;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
      
            xrect.y = new_window_height - 3;
            xrect.x = new_window_width - 2;
            xrect.width = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);

            xrect.y = new_window_height - 5;
            xrect.x = new_window_width - 1;
            xrect.width = 1;
            xrect.height = 2;
            XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
        }

    window_xregion = XCreateRegion();

    xrect.x = 0;
    xrect.y = 0;
    xrect.width = GetFrameWidth();
    xrect.height = GetFrameHeight();

    XUnionRectWithRegion (&xrect, window_xregion, window_xregion);

    if (title_w && emptyRight > 0) {
        // Cut out the part not covered by titlebar
        XRectangle title_xrect;
        title_xrect.x = titleBarWidth;
        title_xrect.y = -border_w;
        title_xrect.width = frame->attrib.getWidth() -
                            titleBarWidth + border_w*2;
        title_xrect.height = title_w + border_w;

        // Remove the bottom border on shaded windows
        if (flags.shaded)
            title_xrect.height += border_w;

        XUnionRectWithRegion(&title_xrect, corners_xregion, corners_xregion);
    }

    XSubtractRegion (window_xregion, corners_xregion, window_xregion);

    XDestroyRegion (corners_xregion);

    if (shaped) {
        XSetWindowAttributes attrs;
        been_shaped = true;

        // Credit: metacity
        /* The client window is oclock or something and has a shape
         * mask. To avoid a round trip to get its shape region, we
         * create a fake window that's never mapped, build up our shape
         * on that, then combine. Wasting the window is assumed cheaper
         * than a round trip, but who really knows for sure.
         */
        Window shape_window;
        Region client_xregion = XCreateRegion();

        attrs.override_redirect = True;

        shape_window = XCreateWindow (kascreen->display,
                                      kascreen->id,
                                      -5000, -5000,
                                      new_window_width,
                                      new_window_height,
                                      0,
                                      CopyFromParent,
                                      CopyFromParent,
                                      CopyFromParent,
                                      CWOverrideRedirect,
                                      &attrs);

        int decor_left = attrib.getLeft() - GetFrameLeft();
        int decor_top  = attrib.getTop() - GetFrameTop();

        XShapeCombineShape(kascreen->display, shape_window, ShapeBounding,
                           decor_left, decor_top, id, ShapeBounding, ShapeSet);
        
        /* Punch the client area out of the normal frame shape,
         * then union it with the shape_window's existing shape
         */
        xrect.x = decor_left;
        xrect.y = decor_top;
        xrect.width = attrib.getWidth();
        xrect.height = attrib.getHeight();

        XUnionRectWithRegion (&xrect, client_xregion, client_xregion);

        XSubtractRegion (window_xregion, client_xregion, window_xregion);

        XDestroyRegion (client_xregion);

        XShapeCombineRegion (kascreen->display, shape_window,
                             ShapeBounding, 0, 0, window_xregion, ShapeUnion);

        /* Now copy shape_window shape to the real frame */
        XShapeCombineShape (kascreen->display, frame->id, ShapeBounding,
                            0, 0, shape_window, ShapeBounding, ShapeSet);

        XDestroyWindow (kascreen->display, shape_window);

    } else if (been_shaped) {
        been_shaped = false;
        XShapeCombineMask(display, frame->id, ShapeBounding, 0, 0,
                          (Pixmap) 0, ShapeSet);
    } else {
        XShapeCombineRegion(display, frame->id, ShapeBounding, 0, 0,
                            window_xregion, ShapeSet);
    }
    XDestroyRegion(window_xregion);
}


/**
 * @fn    ShapeEvent(Window id)
 * @brief Update shape of frame window
 *
 * Shapes frame window in the same way as shape of client.
 *
 * @param id Window id of window that have changed shape
 * @param updateShaped Specifies whether to update the 'shaped' flag
 *                     to true if we get that event (false for the initial one)
 */
void KaWindow::ShapeEvent(Window id, bool updateShapedFlag) {
    if (updateShapedFlag) 
        shaped = true;
    Shape();
}

/**
 * @fn    SendConfig(void)
 * @brief Sends ConfigureNotify to window
 *
 * Sends a ConfigureNotify event to the window with all the current
 * window attributes.
 */
void KaWindow::SendConfig(void) {
    XConfigureEvent ce;

    ce.type              = ConfigureNotify;
    ce.event             = id;
    ce.window            = id;
    ce.x                 = attrib.getX();
    ce.y                 = attrib.getY();
    ce.width             = attrib.getWidth();
    ce.border_width      = border_w;
    ce.override_redirect = false;

    ce.above = frame->id;

    if (flags.shaded)
        ce.height = restore_shade;
    else
        ce.height = attrib.getHeight();

    sendcf = true;

    XGrabServer(display);
    if (validatedrawable(id)) {
        XSendEvent(display, id, false, StructureNotifyMask, (XEvent *) &ce);
        XSendEvent(display, kascreen->id, false, StructureNotifyMask,
                   (XEvent *) &ce);
    }
    else DELETED;
    XUngrabServer(display);
}

/**
 * @fn    CreateOutline(void)
 * @brief Creates window outline
 *
 * Creates four windows used for displaying an outline when doing
 * non opaque moving and resizing of the window.
 */
void KaWindow::CreateOutline(void) {
    XSetWindowAttributes attrib_set;

    int create_mask = CWOverrideRedirect | CWBackPixel | CWEventMask |
        CWColormap;
    attrib_set.background_pixel = kascreen->style.window.outlineColor.getPixel();
    attrib_set.colormap = kascreen->colormap;
    attrib_set.override_redirect = true;
    attrib_set.event_mask = NoEventMask;

    o_west = XCreateWindow(display, kascreen->id, -10, -10, 1, 1, 0,
                           CopyFromParent, CopyFromParent, kascreen->visual,
                           create_mask, &attrib_set);
    o_east = XCreateWindow(display, kascreen->id, -10, -10, 1, 1, 0,
                           CopyFromParent, CopyFromParent, kascreen->visual,
                           create_mask, &attrib_set);
    o_north = XCreateWindow(display, kascreen->id, -10, -10, 1, 1, 0,
                            CopyFromParent, CopyFromParent, kascreen->visual,
                            create_mask, &attrib_set);
    o_south = XCreateWindow(display, kascreen->id, -10, -10, 1, 1, 0,
                            CopyFromParent, CopyFromParent, kascreen->visual,
                            create_mask, &attrib_set);
    DESKTOP_LOOP(kascreen) {
        (*__dit)->aot_stacking_list.push_front(o_west);
        (*__dit)->aot_stacking_list.push_front(o_east);
        (*__dit)->aot_stacking_list.push_front(o_north);
        (*__dit)->aot_stacking_list.push_front(o_south);
    }
    XMapWindow(display, o_west);
    XMapWindow(display, o_east);
    XMapWindow(display, o_north);
    XMapWindow(display, o_south);
    kascreen->RestackWindows(o_west);
}

/**
 * @fn    DestroyOutline(void)
 * @brief Destroys window outline
 *
 * Destorys the four outline windows.
 */
void KaWindow::DestroyOutline(void) {
    DESKTOP_LOOP(kascreen) {
        (*__dit)->aot_stacking_list.remove(o_west);
        (*__dit)->aot_stacking_list.remove(o_east);
        (*__dit)->aot_stacking_list.remove(o_north);
        (*__dit)->aot_stacking_list.remove(o_south);
    }
    XDestroyWindow(display, o_west);
    XDestroyWindow(display, o_east);
    XDestroyWindow(display, o_north);
    XDestroyWindow(display, o_south);
}

/**
 * @fn    DrawOutline(int x, int y, int width, int height)
 * @brief Draw window outline
 *
 * Draws an outer line for a window with the parameters given.
 *
 * @param x The x position
 * @param y The y position
 * @param width The width
 * @param height The height
 */
void KaWindow::DrawOutline(int x, int y, int width, int height) {
    int bw = (border_w) ? border_w: 2;

    XResizeWindow(display, o_west, bw, bw * 2 + title_w + handle_w + height +
                  border_w * 2);
    XResizeWindow(display, o_east, bw, bw * 2 + title_w + handle_w + height +
                  border_w * 2);
    XResizeWindow(display, o_north, width + bw * 2, bw);
    XResizeWindow(display, o_south, width + bw * 2, bw);

    XMoveWindow(display, o_west, x - bw, y - title_w - border_w - bw);
    XMoveWindow(display, o_east, x + width, y - title_w - border_w - bw);
    XMoveWindow(display, o_north, x - bw, y - title_w - border_w - bw);
    XMoveWindow(display, o_south, x - bw, y + height + handle_w + border_w);
}


/**
 * @fn    DrawFrameOutline(int x, int y, int width, int height)
 * @brief Draw window outline
 *
 * Draws a frame outline for the window using the coordinates given
   (without adding/removing decor)
 *
 * @param x The x position
 * @param y The y position
 * @param width The width
 * @param height The height
 */
void KaWindow::DrawFrameOutline(int x, int y, int width, int height) {
    int bw = (border_w) ? border_w: 2;

    int linewidth, lineheight;
    int linex, liney;

    if (kascreen->config.moveresize_crosshair) {
        linewidth = kascreen->width;
        lineheight = kascreen->height;
        linex = 0;
        liney = 0;
    } else {
        linewidth = width;
        lineheight = height;
        linex = x;
        liney = y;
    }

    XResizeWindow(display, o_west, bw, lineheight);
    XResizeWindow(display, o_east, bw, lineheight);
    XResizeWindow(display, o_north, linewidth, bw);
    XResizeWindow(display, o_south, linewidth, bw);

    XMoveWindow(display, o_west, x, liney);
    XMoveWindow(display, o_east, x + width - bw, liney);
    XMoveWindow(display, o_north, linex, y);
    XMoveWindow(display, o_south, linex, y + height - bw);
}

/**
 * @fn    DrawTitlebar(bool force)
 * @brief Draw window titlebar
 *
 * Renders titlebar pixmaps and draws titlebar foreground.
 *
 * @param force True if rendering should be forced
 */
void KaWindow::DrawTitlebar(bool force) {

    bool title_below_scrn_top = ((attrib.getY()) > 0);
    bool title_above_scrn_bottom = (attrib.getY() - title_w) < kascreen->height;

    if (force || (title_w && IsHorizontallyVisible() &&
                  title_below_scrn_top && title_above_scrn_bottom)) {
        title->render();
        label->render();
        list<KaWindowButton *>::iterator bit = buttons.begin();
        for (; bit != buttons.end(); ++bit)
            (*bit)->render();
    }
}

/**
 * @fn    DrawHandlebar(bool force)
 * @brief Draw window handlebar
 *
 * Renders handlebar pixmaps and draws handlebar foreground.
 *
 * @param force True if rendering should be forced
 */
void KaWindow::DrawHandlebar(bool force) {
    bool handle_below_scrn_top =  (attrib.getY() + attrib.getHeight() + 
                                   handle_w) > 0;
    bool handle_above_scrn_bottom =  (attrib.getY() + attrib.getHeight()) < 
        kascreen->height;

    if (force || (handle_w && IsHorizontallyVisible() &&
                  handle_below_scrn_top && handle_above_scrn_bottom)) {
        handle->render();
        grip_r->render();
        grip_l->render();
    }
}

/**
 * @fn    DrawEdges()
 * @brief Draw window edges
 *
 * Draws left/right edges.
 *
 */
void KaWindow::DrawEdges() {
    if(left_w)
      leftEdge->render();
    if(right_w)
      rightEdge->render();
}

/** 
 * Re-renders decorations (clears the cache and deletes rendered textures)
 * 
 */
void KaWindow::updateDecorations() {
    title->clearCache();
    handle->clearCache();
    leftEdge->clearCache();
    rightEdge->clearCache();
    grip_l->clearCache();
    grip_r->clearCache();
}

/**
 * @fn    FocusWin(void)
 * @brief Sets window to have the look of a focused window
 *
 * Sets window decoration pointers to point so that window decorations will
 * represent a focused window. Redraws needed graphics.
 */
void KaWindow::FocusWin(void) {
    if (has_focus) return;
    has_focus = true;

    if (title_w) DrawTitlebar(true);
    if (handle_w) DrawHandlebar(true);

    if (left_w || right_w)
        DrawEdges();
}

/**
 * @fn    UnFocusWin(void)
 * @brief Sets window to have the look of a unfocused window
 *
 * Sets window decoration pointers to point so that window decorations will
 * represent an unfocused window. Redraws needed graphics.
 */
void KaWindow::UnFocusWin(void) {
    if (! has_focus) return;
    has_focus = false;

    if (title_w)  DrawTitlebar(true);
    if (handle_w) DrawHandlebar(true);

    if (left_w || right_w) DrawEdges();
}

/**
 * @fn    ButtonPressed(KaChildWindow *button)
 * @brief Titlebar buttonpress
 *
 * Performes button press animation on one of the titlebar buttons.
 *
 * @param button KaChildWindow pointer to button that should be pressed
 */
void KaWindow::ButtonPressed(KaWindowButton *button) {
    XEvent e;
    bool in_window = true;

    if (kahakai->eh->move_resize != EndMoveResizeType) return;

    XUngrabButton(display, AnyButton, AnyModifier, id);
    XUngrabKey(display, AnyKey, AnyModifier, id);

    button->pressed = true;
    title->update();
    button->render();
    for (;;) {
        XMaskEvent(display, ButtonReleaseMask | EnterWindowMask |
                   LeaveWindowMask, &e);

        switch (e.type) {
            case EnterNotify:
                in_window = true;
                break;
            case LeaveNotify:
                in_window = false;
                break;
            case ButtonRelease:
                // this is wrong.  try pressing a window button, holding down,
                // moving your pointer outside of the button/window, then
                // pressing/releasing another button, and then letting go
                // of the first button.  the obvious solution (to me) didn't
                // work out very well though..
                button->pressed = false;
                title->update();
                button->render();
                if (in_window)
                    XPutBackEvent(display, &e);
                UpdateGrabs();
                kahakai->eh->move_resize = EndMoveResizeType;
                return;
        }
    }
}

/***
 * @fn IncSizeEnforce()
 * @brief Enforces window increment sizes
 *
 * If the window is not respecting the increment sizes properly, shrink it down
 * to the next smallest increments.
 */
void KaWindow::IncSizeEnforce() {
    int n_w, n_h;

    if(IncFrameSizeCheck(GetFrameWidth(), GetFrameHeight(), &n_w, &n_h)) {
        SetFrameWidth(n_w);
        SetFrameHeight(n_h);
    }
}


/**
 * @fn IncFrameSizeCheck(int width, int height, int *n_w, int *n_h)
 * @brief IncSizeCheck but using frame (not client) sizes.
 */
bool KaWindow::IncFrameSizeCheck(int width, int height, int *n_w, int *n_h)
{
    int decor_width  = GetFrameWidth()  - attrib.getWidth();
    int decor_height = GetFrameHeight() - attrib.getHeight();

    bool result = IncSizeCheck(width-decor_width, height-decor_height,
                               n_w, n_h);
    *n_w += decor_width;
    *n_h += decor_height;

    return result;
}

/**
 * @fn    IncSizeCheck(int width, int height, int *n_w, int *n_h)
 * @brief Calculate increasement sizes
 *
 * Given a new width and height this functions calculates if the windows
 * increasement sizes allows a resize of the window. The n_w parameter
 * is used for returning allowed width and the n_h parameter is used for
 * returning allowed height.
 *
 * @param width Width we want to resize to
 * @param height Height we want to resize to
 * @param n_w Return of allowed width
 * @param n_h Return of allowed height
 *
 * @return True if a resize is allowed otherwise false
 */
bool KaWindow::IncSizeCheck(int width, int height, int *n_w, int *n_h) {

    if (width % size.width_inc == (size.base_width % size.width_inc) &&
        height % size.height_inc == (size.base_height % size.height_inc)) {
        *n_w = width;
        *n_h = height;
        return true;
    }

    bool canresize = false;

    *n_w = attrib.getWidth();
    *n_h = attrib.getHeight();

    if ((width >= (attrib.getWidth() + size.width_inc)) ||
        (width <= (attrib.getWidth() - size.width_inc)) ||
        attrib.getWidth() == width) {

        if (width >= size.min_width && width <= size.max_width) {
            *n_w = width - ((width - size.base_width) % size.width_inc);
            canresize = true;
        } else if (width < size.min_width && attrib.getWidth() > size.min_width) {
            *n_w = size.min_width;
            canresize = true;
        } else if (width > size.max_width && attrib.getWidth() < size.max_width) {
            *n_w = size.max_width;
            canresize = true;
        }
    }

    // ????
    // why would we ever want to make a window a negative size?

    /*if ((height <= -(handle_w + border_w * 2)) && title_w) {
        if (! flags.shaded) {
            flags.shaded = true;
            restore_shade = attrib.getHeight();
            net->SetWmState(this);
            UpdateButtons(ShadeCBoxType);
            kascreen->UpdateCheckboxes(ShadeCBoxType);
        }
        *n_h = -(handle_w);
        //        if (handle_w) *n_h -= border_w;
        return canresize;
        } */
    
    if ((height >= (attrib.getHeight() + size.height_inc)) ||
        (height <= (attrib.getHeight() - size.height_inc)) ||
        attrib.getHeight() == height) {

        // resizing a window really small shouldn't make it be shaded.
        // it's just small.

        /*
        if ((height < 1) && (size.min_height <= 1) && title_w) {
            canresize = true;
            if (! flags.shaded) {
                flags.shaded = true;
                restore_shade = attrib.getHeight();
                net->SetWmState(this);
                UpdateButtons(ShadeCBoxType);
                kascreen->UpdateCheckboxes(ShadeCBoxType);
            }
            *n_h = height - ((height - size.base_height) % size.height_inc);
        }
        else */

        bool unshade = false;
        
        if (height >= size.min_height && height <= size.max_height) {
            *n_h = height - ((height - size.base_height) % size.height_inc);
            unshade = canresize = true;
        } else if (height < size.min_height && attrib.getHeight() > size.min_height) {
            *n_h = size.min_height;
            unshade = canresize = true;
        } else if (height > size.max_height && attrib.getHeight() < size.max_height) {
            *n_h = size.max_height;
            unshade = canresize = true;
        }
        
        if (unshade && flags.shaded) {
            flags.shaded = false;
            net->SetWmState(this);
            UpdateButtons(ShadeCBoxType);
            kascreen->UpdateCheckboxes(ShadeCBoxType);
        }
    }

    return canresize;
}

/**
 * @fn    Raise()
 * @brief Raises the window
 *
 * Raises the window to the top of the stacking layer.
 */
void KaWindow::Raise() {
    kascreen->RaiseWindow(frame->id);
}

/**
 * @fn    Lower()
 * @brief Lowers the window
 *
 * Lowers the window to the bottom of the stacking layer.
 */
void KaWindow::Lower() {
    kascreen->LowerWindow(frame->id);
}

/**
 * @fn    Focus(bool vis)
 * @brief Sets input focus to the window
 *
 * Gives window keyboard input focus. If vis parameter is true we make sure
 * the window is viewable.
 *
 * @param vis True if we should make sure the window is viewable
 */
void KaWindow::Focus(bool vis) {
    int newvx, newvy, x, y;

    if (! flags.focusable || ((! vis) && hidden))
        return;

    if (kascreen->FocusFreeze() || kascreen->UserFocusFreeze())
        return;

    if (vis) {

        if (!mapped)
            MapWindow();

        if (flags.hidden)
            UnMinimize();

        if (! (desktop_mask &
               (1L << kascreen->current_desktop->number))) {
            vector<Desktop *>::iterator dit =
                kascreen->desktop_list.begin();
            for (; dit != kascreen->desktop_list.end(); dit++)
                if (desktop_mask & (1L << (*dit)->number)) {
                    kascreen->GoToDesktop((*dit)->number);
                    break;
                }
        }

        bool x_toolarge = false, y_toolarge = false;

        if ((kascreen->v_x + attrib.getX()) >=
            (kascreen->v_xmax + kascreen->width))
            x_toolarge = true;

        if ((kascreen->v_y + attrib.getY()) >=
            (kascreen->v_ymax + kascreen->height))
            y_toolarge = true;

        if (x_toolarge || y_toolarge) {
            int bw = flags.border * border_w;
            int handleh = handle_w + flags.handle * bw;

            int th = attrib.getHeight() + bw + handleh;
            int tw = attrib.getWidth() + bw;

            if (x_toolarge)
                attrib.setX((kascreen->v_xmax + kascreen->width - kascreen->v_x) - tw);
            if (y_toolarge)
                attrib.setY((kascreen->v_ymax + kascreen->height - kascreen->v_y) - th);
            RedrawWindow();
        }

        // if not on screen
        if (!IsHorizontallyVisible() || !IsVerticallyVisible()) {

            x = kascreen->v_x + attrib.getX();
            y = kascreen->v_y + attrib.getY();
            newvx = (x / kascreen->width) * kascreen->width;
            newvy = (y / kascreen->height) * kascreen->height;
            kascreen->MoveViewportTo(newvx, newvy);
            kascreen->DiscardEnterNotifies();
        }
    }


    XInstallColormap(display, attrib.getColormap());
    XGrabServer(display);

    if (validateclient_mapped(id))
        XSetInputFocus(display, id, RevertToPointerRoot, CurrentTime);
    else DELETED;

    XUngrabServer(display);

    want_focus = true;
}

void KaWindow::MoveWindowToCurrentViewport() {

    // shift this thing into view by increments of viewport size

    int top    = GetFrameTop();
    int bottom = GetFrameBottom();
    int left   = GetFrameLeft();
    int right  = GetFrameRight();

    while(top > kascreen->height) {
        top -= kascreen->height;
        bottom -= kascreen->height;
    }

    while(left > kascreen->width) {
        left -= kascreen->width;
        right -= kascreen->width;
    }

    while(bottom < 0) {
        top += kascreen->height;
        bottom += kascreen->height;
    }

    while(right < 0) {
        left += kascreen->width;
        right += kascreen->width;
    }

    SetFrameTop(top);
    SetFrameBottom(bottom);
    SetFrameLeft(left);
    SetFrameRight(right);

    int workx, worky, workw, workh;
    kascreen->GetWorkareaSize(&workx, &worky, &workw, &workh);

    /* now we've gotten it closest to "on this viewport", but it still could be
     * hanging off the screen edge more than we want.  so now we pull it within
     * the workarea boundaries, if it's not already in them. */

    if (GetFrameTop()+title_w+border_w*2 > worky + workh)
        SetFrameTop(worky + workh - title_w - border_w*2);
    else if (GetFrameBottom() < worky)
        SetFrameBottom(worky + handle_w + border_w*2);

    if (GetFrameLeft() > workx + workw - 4)
        SetFrameLeft(workx + workw - 4);
    else if (GetFrameRight() < workx + 4)
        SetFrameLeft(workx + 4);

    RedrawWindow();
}

void KaWindow::RestrictMaxMovement(FrameGeometry * geom) { 
    if (!flags.max)
        return; 

    if (!geom) 
        geom = this->geometry;

    int workx, worky, workw, workh;
    kascreen->GetWorkareaSize(&workx, &worky, &workw, &workh);

    int center_padding = (workw - geom->getWidth()) / 2;

    if (flags.hmax)
        geom->setLeft(workx + center_padding);

    if (flags.vmax)
        geom->setTop(worky);
}


// Some ideas for snapping implementation from pekwm

/**
 * @fn    CheckSnapAgainst()
 * @brief Tries to snap the window to a rectangle (a window, an edge, dockapp)
 */
bool KaWindow::CheckSnapAgainst(FrameGeometry* geom,
                                int wleft, int wtop, int wright, int wbottom)
{
    int snap = kascreen->config.snapDistance;

    if (!geom) geom = this->geometry;

    const Workarea & work = kascreen->GetWorkarea();

    int top       = geom->getTop();
    int bottom    = geom->getBottom();
    int left      = geom->getLeft();
    int right     = geom->getRight();

    bool x_snapped = false, y_snapped = false;
    bool x_overlaps = false, y_overlaps = false;

    y_overlaps = (bool) overlaps(left, right, wleft, wright);
    x_overlaps = (bool) overlaps(top, bottom, wtop, wbottom);

    if (!y_overlaps && !x_overlaps)
        return false;

    // the last thing in each of these if's ensures that the window stays within
    // the workarea, and also 2 extra pixels inside of that, to account for
    // screen edges (so they can still grab at the window)

    // Snap our right side to w's left side
    if (x_overlaps && right >= wleft - snap &&
        right <= wleft + snap && wleft > work.x+2) {
        geom->setRight(wleft);
        x_snapped = true;
    }

    // Snap our bottom to w's top
    if (y_overlaps && bottom >= wtop - snap &&
        bottom <= wtop + snap && wtop > work.y+2) {
        geom->setBottom(wtop);
        y_snapped = true;
    }

    // And our left side to w's right side
    if (x_overlaps && left >= wright - snap &&
        left <= wright + snap && wright < work.x+work.width-2) {
        geom->setLeft(wright);
        x_snapped = true;
    }

    // Snap our top to w's bottom
    if (y_overlaps && top >= wbottom - snap &&
        top <= wbottom + snap && wbottom < work.y+work.height-2) {
        geom->setTop(wbottom);
        y_snapped = true;
    }


    return x_snapped || y_snapped;
}

/**
 * @fn    CheckWindowSnap()
 * @brief Tries to snap the window to nearby windows
 */
bool KaWindow::CheckWindowSnap(FrameGeometry* geom)
{
    vector<KaWindow *>::iterator it = kascreen->kawindow_list.begin();

    bool snapped = false;

    if (!geom) geom = this->geometry;

    for (; it != kascreen->kawindow_list.end(); it++) {
        if (*it == this || (*it)->hidden || (*it)->deleted)
            continue;

        if (!(*it)->desktop_mask & (1L << kascreen->current_desktop->number))
            continue;

        KaWindow* w = *it;
        int wleft   = w->GetFrameLeft();
        int wtop    = w->GetFrameTop();
        int wright  = w->GetFrameRight();
        int wbottom = w->GetFrameBottom();

        bool didSnap = CheckSnapAgainst(geom, wleft, wtop, wright, wbottom);
        snapped = snapped || didSnap;
    }

    return snapped;
}

/**
 * @fn    CheckSnapAgainstEdges()
 * @brief Tries to snap the window to a set of edges (screen, Xinerama screen)
 */

bool KaWindow::CheckSnapAgainstEdges(FrameGeometry* geom,
                                     int left_edge, 
                                     int top_edge, 
                                     int right_edge, 
                                     int bottom_edge) {

    int insnap, outsnap;
    bool resist = kascreen->config.resist;

    if (resist) {
        insnap = 0;
        outsnap = kascreen->config.resistDistance;
    } else
        insnap = outsnap = kascreen->config.snapDistance;

    if (!geom) geom = this->geometry;

    int top       = geom->getTop();
    int bottom    = geom->getBottom();
    int left      = geom->getLeft();
    int right     = geom->getRight();
    
    bool x_snapped = false, y_snapped = false;

    // once the window is pushed past the resistance size, we shouldn't snap it
    // to the edge as it is coming back, so we have to save the state.

    if (resist) {
        if (right >= right_edge + outsnap)
            gone_past_right = true;
        else if (right <= right_edge)
            gone_past_right = false;

        if (bottom >= bottom_edge + outsnap)
            gone_past_bottom = true;
        else if (bottom <= bottom_edge)
            gone_past_bottom = false;

        if (left <= left_edge - outsnap)
            gone_past_left = true;
        else if (left >= left_edge)
            gone_past_left = false;

        if (top <= top_edge - outsnap)
            gone_past_top = true;
        else if (top >= top_edge)
            gone_past_top = false;
    }

    ////////

    if (right >= right_edge - insnap && right <= right_edge + outsnap &&
        (!resist || !gone_past_right)) {
        geom->setRight(right_edge);
        x_snapped = true;
    }

    if (bottom >= bottom_edge - insnap && bottom <= bottom_edge + outsnap &&
        (!resist || !gone_past_bottom)) {
        geom->setBottom(bottom_edge);
        y_snapped = true;
    }    

    if (left <= left_edge + insnap && left >= left_edge - outsnap &&
        (!resist || !gone_past_left)) {
        geom->setLeft(left_edge);
        x_snapped = true;
    }

    if (top <= top_edge + insnap && top >= top_edge - outsnap &&
        (!resist || !gone_past_top)) {
        geom->setTop(top_edge);
        y_snapped = true;
    }

    return x_snapped || y_snapped;
}

/**
 * @fn    CheckEdgeSnap()
 * @brief Tries to snap the window to nearby screen edges
 */
bool KaWindow::CheckEdgeSnap(FrameGeometry* geom)
{
    bool snapped = false;

    int workx, worky, workw, workh;
    kascreen->GetWorkareaSize(&workx, &worky, &workw, &workh);

    snapped = CheckSnapAgainstEdges(geom, workx, worky,
                                          workx+workw, worky+workh);

#ifdef XINERAMA
    // thanks valaxer
    //sf.net/tracker/index.php?func=detail&aid=762966&group_id=82102&atid=565050
    if (kahakai->xinerama->active) {
        vector<XineramaScreenInfo>::iterator it;
        for (it = kahakai->xinerama->screens.begin(); 
             it != kahakai->xinerama->screens.end(); ++it) {

            snapped = snapped || CheckSnapAgainstEdges(geom,
                                                       it->x_org,  it->y_org,
                                                       it->x_org + it->width, 
                                                       it->y_org + it->height);
        }
    }
#endif // XINERAMA

    return snapped;
}

/**
 * @fn    CheckDockSnap()
 * @brief Tries to snap the window to nearby dockapps
 */

bool KaWindow::CheckDockSnap(FrameGeometry* geom)
{
    list<DockappHandler*>::iterator it;

    int i = 0;

    for (it = kascreen->docks.begin(); it != kascreen->docks.end(); ++it) {
        DockappHandler* dock = *it;
        
        if (dock->hidden)
            continue;
        
        int wleft   = dock->map_x;
        int wtop    = dock->map_y;
        int wright  = dock->map_x + dock->width;
        int wbottom = dock->map_y + dock->height;

        // Add dock borders
        DockStyle  dockStyle = kascreen->style.dockStyles[i];
        wright  += dockStyle.gridspace;
        wbottom += dockStyle.gridspace;

        if (CheckSnapAgainst(geom, wleft, wtop, wright, wbottom)) return true;

        i++;
    }

    return false;
}


/**
 * @fn    CheckSnap()
 * @brief Tries to snap the frame to nearby frames & screen borders
 */

void KaWindow::CheckSnap(FrameGeometry* geom) {
    if (kascreen->config.snapping) {
        CheckDockSnap(geom);
        CheckWindowSnap(geom);
    }


    if (kascreen->config.snapping || kascreen->config.resist)
        CheckEdgeSnap(geom);
}

/**
 * @fn    Move(XEvent *e)
 * @brief Moves the window
 *
 * Moves the window through displaying an outline of the window while dragging
 * the mouse.
 *
 * @param e XEvent causing start of move
 */
void KaWindow::Move(XEvent *e) {
    XEvent event, *map_ev;
    int px, py, nx, ny;
    list<XEvent *> *maprequest_list;
    bool started = false;

    kascreen->QueryPointer(&px, &py);

    if (kahakai->eh->move_resize != EndMoveResizeType) return;
    nx = attrib.getX();
    ny = attrib.getY();
    kahakai->eh->move_resize = MoveType;
    move_resize = true;

    if (e->type == MapRequest) {
        attrib.setX(px + left_w);
        nx = attrib.getX();
        attrib.setY(py + title_w);
        ny = attrib.getY();
        CreateOutline();
        DrawOutline(nx, ny, attrib.getWidth(), attrib.getHeight());
        started = true;
    }

    maprequest_list = new list<XEvent *>;
    XGrabServer(display);
    if (validatedrawable(id)) {
        if (XGrabPointer(display, (mapped && !hidden) ? id:
                         kascreen->id, true,
                         ButtonReleaseMask | ButtonPressMask |
                         PointerMotionMask | EnterWindowMask |
                         LeaveWindowMask, GrabModeAsync, GrabModeAsync,
                         None, kahakai->move_cursor, CurrentTime)
            != GrabSuccess) {
            move_resize = false;
            kahakai->eh->move_resize = EndMoveResizeType;
            return;
        }
        if (XGrabKeyboard(display, (mapped && !hidden) ? id:
                          kascreen->id,
                          true, GrabModeAsync, GrabModeAsync, CurrentTime)
            != GrabSuccess) {
            move_resize = false;
            kahakai->eh->move_resize = EndMoveResizeType;
            return;
        }
    } else DELETED;
    XUngrabServer(display);

    int old_x = GetFrameLeft();
    int old_y = GetFrameTop();

    int spx, spy;
    kascreen->QueryPointer(&spx, &spy);

    int width = GetFrameWidth(), height = GetFrameHeight();

    MoveFrameGeometry geom(GetFrameLeft(),  GetFrameTop(),
                           width, height);
    for (;;) {
        kahakai->eh->EventLoop(kahakai->eh->moveresize_return_mask, &event);
        switch (event.type) {
        case MotionNotify:
            while (XCheckTypedWindowEvent(display, event.xmotion.window,
                                          MotionNotify, &event));

            geom.setLeft(old_x + event.xmotion.x_root - spx);
            geom.setTop (old_y + event.xmotion.y_root - spy);

            CheckSnap(&geom);
            RestrictMaxMovement(&geom);

            if (! started) {
                CreateOutline();
                started = true;
            }

            DrawFrameOutline(geom.getLeft(),
                           geom.getTop(),
                           width, height);
            break;
        case LeaveNotify:
        case EnterNotify:
            if (kascreen->west->id == event.xcrossing.window ||
                kascreen->east->id == event.xcrossing.window ||
                kascreen->north->id == event.xcrossing.window ||
                kascreen->south->id == event.xcrossing.window) {
                kahakai->eh->HandleEvent(&event);
            } else if (event.type == LeaveNotify) {
                int cx, cy;
                kascreen->QueryPointer(&cx, &cy);

                geom.setLeft(old_x + cx - spx);
                geom.setTop(old_y + cy - spy);

                CheckSnap(&geom);
            }
            break;
        case DestroyNotify:
        case UnmapNotify:
            if ((((event.type == UnmapNotify)? event.xunmap.window:
                  event.xdestroywindow.window) == id)) {
                while (! maprequest_list->empty()) {
                    XPutBackEvent(display, maprequest_list->front());
                    delete maprequest_list->front();
                    maprequest_list->pop_front();
                }
                delete maprequest_list;
                XPutBackEvent(display, &event);
                if (started) DestroyOutline();
                XUngrabKeyboard(display, CurrentTime);
                XUngrabPointer(display, CurrentTime);
                kahakai->eh->move_resize = EndMoveResizeType;
                move_resize = false;
                return;
            }
            kahakai->eh->EvUnmapDestroy(&event);
            break;
        case ConfigureRequest:
            if (event.xconfigurerequest.window != id)
                kahakai->eh->EvConfigureRequest(&event.xconfigurerequest);
            break;
        case MapRequest:
            map_ev = new XEvent;
            *map_ev = event;
            maprequest_list->push_front(map_ev); break;
        case ButtonPress:
        case ButtonRelease:
            event.xbutton.window = id;
        case KeyPress:
        case KeyRelease:
            if (event.type == KeyPress || event.type == KeyRelease)
                event.xkey.window = id;

            kahakai->eh->HandleEvent(&event);

            //if (! started) {
                //CreateOutline();
                //started = true;
            //}

            if (started)
                DrawFrameOutline(geom.getLeft(),
                                 geom.getTop(),
                                 width, height);

            if (kahakai->eh->move_resize != EndMoveResizeType)
                break;

            if (started)
                DestroyOutline();

            SetFrameLeft(geom.getLeft());
            SetFrameTop(geom.getTop());

            while (! maprequest_list->empty()) {
                XPutBackEvent(display, maprequest_list->front());
                delete maprequest_list->front();
                maprequest_list->pop_front();
            }

            delete maprequest_list;
            move_resize = false;
            XUngrabKeyboard(display, CurrentTime);
            XUngrabPointer(display, CurrentTime);
            UpdateTitlebar();
            RedrawWindow();
            SendConfig();
            net->SetVirtualPos(this);
            return;
        }
    }
}

void KaWindow::restoreDecorationsFromResize() {
    if (!flags.max)
        return;

    if (flags.hmax && !flags.vmax) {
        flags.max = flags.hmax = false;
        shadeOnUnmaximize = false;
        return;
    }

    if (flags.vmax && !flags.hmax) {
        flags.max = flags.vmax = false;
        shadeOnUnmaximize = false;
        return;
    }

    if (flags.hmax && flags.vmax) {
        flags.max = flags.vmax = flags.hmax = false;
        shadeOnUnmaximize = false;

        int junk, workw, workh, n_w, n_h;
        kascreen->GetWorkareaSize(&junk, &junk, &workw, &workh);

        if (! flags.handle && restore_max_flags.handle) {
            DecorHandleOn();
            restore_max_flags.handle = false;

            if (IncFrameSizeCheck(GetFrameWidth(), workh, &junk, &n_h)) {
                // make sure the whole window is still in the workarea
                while (n_h > workh)
                    n_h -= size.height_inc;
                SetFrameHeight(n_h);
                RedrawWindow();
            }
        }

        if (! flags.border && restore_max_flags.border) {
            DecorBorderOn();
            restore_max_flags.border = false;

            if (IncFrameSizeCheck(workw, GetFrameHeight(), &n_w, &junk)) {
                // make sure the whole window is still in the workarea
                while (n_w > workw)
                    n_w -= size.width_inc;
                SetFrameWidth(n_w);
                RedrawWindow();
            }
        }

        return;
    }

    logger << warn << endl << "  KaWindow::_RestoreHandlebarFromResize():"
                   << endl << "    invalid combination of flags:"
                   << "max:" << flags.max
                   << " vmax:" << flags.vmax
                   << " hmax:" << flags.hmax << endl
                   << "  this is probably a bug!" << endl;
}

/**
 * @fn    MoveOpaque(XEvent *e, KaAction *)
 * @brief Moves the window
 *
 * Moves the window using the opaque moving process, which means that the
 * window is moved after the mouse pointers every motion event.
 *
 * @param e XEvent causing start of move

*/
void KaWindow::MoveOpaque(XEvent *e) {
    int px, py, prelx, prely;
    XEvent ed;
    list<XEvent *> maprequest_list;

    kascreen->QueryPointer(&px, &py);

    prelx = px - attrib.getX();
    prely = py - attrib.getY();
    int pposx = px;
    int pposy = py;

    for (;;) {
        bool status = _MoveOpaque(e, pposx, pposy, &maprequest_list);
        if (! status) return;

        // nothing past here seems to ever be reached

        kascreen->QueryPointer(&px, &py);

        pposx = attrib.getX() + prelx;
        pposy = attrib.getY() + prely;

        XWarpPointer(display, None, None, 0, 0, 0, 0, pposx - px, pposy - py);
        XSync(display, false);
        while (XCheckTypedEvent(display, MotionNotify, &ed));
        while (XCheckTypedEvent(display, EnterNotify, &ed));
        while (XCheckTypedEvent(display, LeaveNotify, &ed));
    }
}

bool KaWindow::_MoveOpaque(XEvent *e, int px, int py,
                           list<XEvent *> *maprequest_list) {
    XEvent event, *map_ev;
    int sx, sy, nx, ny;

    if (kahakai->eh->move_resize != EndMoveResizeType) return false;
    sx = nx = attrib.getX();
    sy = ny = attrib.getY();
    kahakai->eh->move_resize = MoveOpaqueType;
    move_resize = true;

    if (e && e->type == MapRequest) {
        attrib.setX(px + left_w);
        nx = attrib.getX();
        attrib.setY(py + title_w);
        ny = attrib.getY();
        RedrawWindow();
        net->SetState(this, NormalState);
        net->SetVirtualPos(this);
    }

    dontsend = true;
    XGrabServer(display);
    if (validatedrawable(id)) {
        if (XGrabPointer(display,
                         (mapped && !hidden) ? id: kascreen->id, true,
                         ButtonReleaseMask | ButtonPressMask |
                         PointerMotionMask | EnterWindowMask |
                         LeaveWindowMask, GrabModeAsync, GrabModeAsync,
                         None, kahakai->move_cursor, CurrentTime)
            != GrabSuccess) {
            move_resize = false;
            kahakai->eh->move_resize = EndMoveResizeType;
            return false;
        }
        if (XGrabKeyboard(display,
                          (mapped && !hidden) ? id: kascreen->id,
                          true, GrabModeAsync, GrabModeAsync, CurrentTime)
            != GrabSuccess) {
            move_resize = false;
            kahakai->eh->move_resize = EndMoveResizeType;
            return false;
        }
    } else { deleted = true; XUngrabServer(display); return false; }
    XUngrabServer(display);

    int spx, spy;
    kascreen->QueryPointer(&spx, &spy);

    int old_x = attrib.getX();
    int old_y = attrib.getY();

    int lastx = GetFrameLeft();
    int lasty = GetFrameTop();

    int currentx, currenty;

    bool xmoved = false, ymoved = false;

    stringstream tmps;
    for (;;) {
        kahakai->eh->EventLoop(kahakai->eh->moveresize_return_mask, &event);
        switch (event.type) {
        case MotionNotify:
            while (XCheckTypedWindowEvent(display, event.xmotion.window,
                                          MotionNotify, &event));

            nx = old_x + event.xmotion.x_root - spx;
            ny = old_y + event.xmotion.y_root - spy;

            attrib.setX(nx);
            attrib.setY(ny);

            CheckSnap();
            RestrictMaxMovement();

            /* due to snapping, a mouse movement doesn't always equal a
             * window movement */

            currentx = GetFrameLeft();
            currenty = GetFrameTop();

            xmoved = (lastx != currentx);
            ymoved = (lasty != currenty);

            lastx = currentx;
            lasty = currenty;

            if (flags.title && (xmoved || ymoved)) {
                tmps.str("");
                tmps << name << " [moving] "
                     << (currentx >= 0 ? "+" : "") << currentx
                     << (currenty >= 0 ? "+" : "") << currenty;

                labeltext = tmps.str();

                UpdateTitlebar();
            }

            RedrawWindow(false, false);

            SendConfig();
            net->SetVirtualPos(this);
            break;
        case LeaveNotify:
        case EnterNotify:
            if (kascreen->west->id == event.xcrossing.window ||
                kascreen->east->id == event.xcrossing.window ||
                kascreen->north->id == event.xcrossing.window ||
                kascreen->south->id == event.xcrossing.window) {
                kahakai->eh->HandleEvent(&event);
            } else if (event.type == LeaveNotify) {
                int cx, cy;
                kascreen->QueryPointer(&cx, &cy);

                nx = old_x + cx - spx;
                ny = old_y + cy - spy;
                attrib.setX(nx);
                attrib.setY(ny);
                CheckSnap();
                RestrictMaxMovement();
                RedrawWindow(false, false);
            }
            break;
        case DestroyNotify:
        case UnmapNotify:
            if ((((event.type == UnmapNotify)? event.xunmap.window:
                  event.xdestroywindow.window) == id)) {
                while (! maprequest_list->empty()) {
                    XPutBackEvent(display, maprequest_list->front());
                    delete maprequest_list->front();
                    maprequest_list->pop_front();
                }
                XPutBackEvent(display, &event);
                XUngrabKeyboard(display, CurrentTime);
                XUngrabPointer(display, CurrentTime);
                kahakai->eh->move_resize = EndMoveResizeType;
                dontsend = move_resize = false;
                labeltext = name;
                UpdateTitlebar();
                return false;
            }
            kahakai->eh->EvUnmapDestroy(&event);
            break;
        case ConfigureRequest:
            if (event.xconfigurerequest.window != id)
                kahakai->eh->EvConfigureRequest(&event.xconfigurerequest);
            break;
        case MapRequest:
            map_ev = new XEvent;
            *map_ev = event;
            maprequest_list->push_front(map_ev); break;
        case ButtonPress:
        case ButtonRelease:
            event.xbutton.window = id;
        case KeyPress:
        case KeyRelease:
            if (event.type == KeyPress || event.type == KeyRelease)
                event.xkey.window = id;
            kahakai->eh->HandleEvent(&event);
            if (kahakai->eh->move_resize != EndMoveResizeType) break;
            if (attrib.getX() != sx || attrib.getY() != sy) {

                if (kascreen->config.lazy_trans) {
                    render_if_transparency = true;
                    DrawTitlebar();
                    DrawHandlebar();
                    DrawEdges();
                    render_if_transparency = false;
                }

                SendConfig();
                net->SetVirtualPos(this);
            }
            while (! maprequest_list->empty()) {
                XPutBackEvent(display, maprequest_list->front());
                delete maprequest_list->front();
                maprequest_list->pop_front();
            }
            dontsend = move_resize = false;
            XUngrabKeyboard(display, CurrentTime);
            XUngrabPointer(display, CurrentTime);
            labeltext = name;
            UpdateTitlebar();
            return false;
        }
    }
}

/**
 * @fn    Resize(XEvent *e, int hdir, int vdir)
 * @brief Resizes the window
 *
 * Resizes the window through displaying a outline of the window while dragging
 * the mouse. If how parameter is RightType when the window is resized in
 * south-east direction and if how parameter is LeftType the resize is
 * being performed in south-west direction.
 *
 * @param e XEvent causing start of resize
 * @param hdir If EastType resize to the right, WestType resize to the left
 * @param vdir If SouthType resize down, NorthType resize up
 */
void KaWindow::Resize(XEvent *e, int hdir, int vdir) {
    _Resize(e, hdir, vdir, false);
}

/**
 * @fn    SmartResize(XEvent *e)
 * @brief Resizes the window determining direction from cursor position
 *
 * Resizes the window through displaying a outline of the window while
 * dragging the mouse. Determines direction by the relative position
 * of the mouse cursor in one of four quadrants of the window (NW, NE,
 * SW, SE).
 *
 * @param e XEvent causing start of resize
 * @param opaque If resizing is opaque (false by default)
 */
void KaWindow::SmartResize(XEvent *e, bool opaque) {
    int px, py;
    kascreen->QueryPointer(&px, &py);

    int cx = attrib.getX() + attrib.getWidth()/2;
    int cy = attrib.getY() + attrib.getHeight()/2;

    int hdir = (px < cx) ? WestType  : EastType;
    int vdir = (py < cy) ? NorthType : SouthType;

    _Resize(e, hdir, vdir, opaque);
}

/**
 * @fn    _Resize(XEvent *e, int how, bool opaque)
 * @brief Resizes the window
 *
 * Resizes the window either using the opaque resizing process, which
 * means that the window is resized after the mouse pointers every
 * motion event or using an outline.
 *
 * @param e XEvent causing start of resize
 * @param hdir If EastType resize to the right, WestType resize to the left
 * @param vdir If SouthType resize down, NorthType resize up
 * @param opaque If true, use opaque resizing, outline otherwise
 */
void KaWindow::_Resize(XEvent *e, int hdir, int vdir, bool opaque) {
    XEvent event, *map_ev;
    int px, py, width, height, n_w, n_h;
    bool started = false;
    list<XEvent *> *maprequest_list;

    restoreDecorationsFromResize();

    ResizeFrameGeometry geom(this,
                             GetFrameLeft(), GetFrameTop(),
                             GetFrameWidth(), GetFrameHeight(),
                             hdir, vdir);

    kascreen->QueryPointer(&px, &py);

    if (kahakai->eh->move_resize != EndMoveResizeType) return;
    dontsend = true;

    int orig_width = GetFrameWidth(), orig_height = GetFrameHeight();

    kahakai->eh->move_resize = opaque ? ResizeOpaqueType : ResizeType;
    move_resize = true;

    if (e && e->type == MapRequest) {
        if (hdir > 0)
            SetFrameRight(px);
        else
            SetFrameLeft(px);

        if (vdir > 0)
            SetFrameBottom(py);
        else
            SetFrameTop(py);

        geom.copyPositionFromWindow(this);

        if (opaque)
            RedrawWindow();
        else {
            CreateOutline();
            DrawFrameOutline(geom.getLeft(), 
                             geom.getTop(), 
                             geom.getWidth(),
                             geom.getHeight());
            started = true;
        }
        net->SetState(this, NormalState);
        net->SetVirtualPos(this);
    }

    maprequest_list = new list<XEvent *>;
    XGrabServer(display);
    if (validatedrawable(id)) {
        Cursor cursor;
        if (hdir == EastType) {
            cursor = vdir == NorthType ? 
                kahakai->resizene_cursor : kahakai->resizese_cursor;
        } else {
            cursor = vdir == NorthType ? 
                kahakai->resizenw_cursor : kahakai->resizesw_cursor;
        }

        if (XGrabPointer(display, (mapped && !hidden) ? id: kascreen->id, true,
                         ButtonReleaseMask | ButtonPressMask |
                         PointerMotionMask | EnterWindowMask |
                         LeaveWindowMask, GrabModeAsync, GrabModeAsync,
                         None, cursor, CurrentTime)
            != GrabSuccess) {
            move_resize = false;
            kahakai->eh->move_resize = EndMoveResizeType;
            return;
        }
        if (XGrabKeyboard(display, (mapped && !hidden) ? id: kascreen->id,
                          true, GrabModeAsync, GrabModeAsync, CurrentTime)
            != GrabSuccess) {
            move_resize = false;
            kahakai->eh->move_resize = EndMoveResizeType;
            XUngrabPointer(display, CurrentTime);
            return;
        }
    } else DELETED;

    XUngrabServer(display);

    int lastw = GetFrameWidth();
    int lasth = GetFrameHeight();

    int currentw, currenth;

    bool wchanged = false, hchanged = false;

    stringstream tmps;

    for (;;) {
        kahakai->eh->EventLoop(kahakai->eh->moveresize_return_mask, &event);
        switch (event.type) {
        case MotionNotify:
            while (XCheckTypedWindowEvent(display, event.xmotion.window,
                                          MotionNotify, &event));

            width  = orig_width  + (event.xmotion.x_root - px) * hdir;
            height = orig_height + (event.xmotion.y_root - py) * vdir;

            if (IncFrameSizeCheck(width, height, &n_w, &n_h)) {
                geom.setWidth(n_w);
                geom.setHeight(n_h);

                CheckSnap(&geom);

                /* due to snapping, a mouse movement doesn't always equal a
                 * window resize */

                currentw = geom.getWidth();
                currenth = geom.getHeight();

                wchanged = (lastw != currentw);
                hchanged = (lasth != currenth);

                lastw = currentw;
                lasth = currenth;

                if (opaque) {
                    if (wchanged || hchanged) {

                        geom.copyToWindow(this);

                        tmps.str("");

                        /* if the increment is set, then we want the size
                         * displayed to only reflect the window contents, so
                         * that a 80x25 xterm is not mysteriously something
                         * other than 80x25.  Otherwise, we are showing pixel
                         * count, and we should include the decorations in that
                         * measurement. */

                        tmps << name << " [resizing] "
                             << (size.width_inc == 1 ? currentw
                                 : attrib.getWidth() / size.width_inc)
                             << "x"
                             << (size.height_inc == 1 ? currenth
                                 : attrib.getHeight() / size.height_inc);

                        labeltext = tmps.str();
                        UpdateTitlebar();
                    }
                    RedrawWindow();

                } else {
                    if (! started) {
                        CreateOutline();
                        started = true;
                    }

                    DrawFrameOutline(geom.getLeft(), geom.getTop(),
                                     geom.getWidth(), geom.getHeight());
                }
            }
            break;
        case LeaveNotify:
        case EnterNotify:
            if (kascreen->west->id == event.xcrossing.window ||
                kascreen->east->id == event.xcrossing.window ||
                kascreen->north->id == event.xcrossing.window ||
                kascreen->south->id == event.xcrossing.window) {
                int old_vx = kascreen->v_x;
                int old_vy = kascreen->v_y;
                kahakai->eh->HandleEvent(&event);
                px -= (kascreen->v_x - old_vx);
                py -= (kascreen->v_y - old_vy);
                if (!opaque) {
                    geom.copyPositionFromWindow(this);

                    DrawFrameOutline(geom.getLeft(), 
                                        geom.getTop(), 
                                        geom.getWidth(),
                                        geom.getHeight());
                }
            } else if (event.type == LeaveNotify) {
                int cx, cy;
                kascreen->QueryPointer(&cx, &cy);

                width  = orig_width  + (cx - px) * hdir;
                height = orig_height + (cy - py) * vdir;

                if (IncFrameSizeCheck(width, height, &n_w, &n_h)) {
                    geom.setWidth(n_w);
                    geom.setHeight(n_h);
                    CheckSnap(&geom);

                    if (opaque) {
                        geom.copyToWindow(this);
                        RedrawWindow();
                    } else {
                        if (! started) {
                            CreateOutline();
                            started = true;
                        }
                        DrawFrameOutline(geom.getLeft(),
                                            geom.getTop(), 
                                            geom.getWidth(),
                                            geom.getHeight());
                    }
                }
            }
            break;
        case DestroyNotify:
        case UnmapNotify:
            if ((((event.type == UnmapNotify)? event.xunmap.window:
                  event.xdestroywindow.window) == id)) {
                while (! maprequest_list->empty()) {
                    XPutBackEvent(display, maprequest_list->front());
                    delete maprequest_list->front();
                    maprequest_list->pop_front();
                }
                delete maprequest_list;
                if (!opaque && started) DestroyOutline();
                XPutBackEvent(display, &event);
                XUngrabKeyboard(display, CurrentTime);
                XUngrabPointer(display, CurrentTime);
                kahakai->eh->move_resize = EndMoveResizeType;
                dontsend = move_resize = false;
                labeltext = name;
                UpdateTitlebar();
                return;
            }
            kahakai->eh->EvUnmapDestroy(&event);
            break;
        case ConfigureRequest:
            if (event.xconfigurerequest.window != id)
                kahakai->eh->EvConfigureRequest(&event.xconfigurerequest);
            break;
        case MapRequest:
            map_ev = new XEvent;
            *map_ev = event;
            maprequest_list->push_front(map_ev); break;
        case ButtonPress:
        case ButtonRelease:
            event.xbutton.window = id;
        case KeyPress:
        case KeyRelease:
            if (event.type == KeyPress || event.type == KeyRelease)
                event.xkey.window = id;
            kahakai->eh->HandleEvent(&event);

            if (opaque) {
                width = attrib.getWidth();
                height = attrib.getHeight();
                if (kahakai->eh->move_resize != EndMoveResizeType) break;
                if (GetFrameWidth() != orig_width ||
                    GetFrameHeight() != orig_height) {

                    SendConfig();
                    net->SetVirtualPos(this);
                }
            } else {
                if (kahakai->eh->move_resize != EndMoveResizeType) break;
                if (started) DestroyOutline();

                SetFrameLeft(geom.getLeft());
                SetFrameTop(geom.getTop());
                SetFrameWidth(geom.getWidth());
                SetFrameHeight(geom.getHeight());

                if (!opaque && !flags.handle && restore_max_flags.handle) {
                    DecorHandleOn();
                    restore_max_flags.handle = false;
                }

                RedrawWindow();
            }
            while (! maprequest_list->empty()) {
                XPutBackEvent(display, maprequest_list->front());
                delete maprequest_list->front();
                maprequest_list->pop_front();
            }
            delete maprequest_list;
            dontsend = move_resize = false;
            XUngrabKeyboard(display, CurrentTime);
            XUngrabPointer(display, CurrentTime);
            labeltext = name;
            UpdateTitlebar();
            return;
        }
    }
}

/**
 * @fn    ResizeOpaque(XEvent *e, int how)
 * @brief Resizes the window
 *
 * Resizes the window using the opaque resizing process, which means that
 * the window is resized after the mouse pointers every motion event.
 *
 * Parameters hdir (EastType/WestType) and vdir (SouthType/NorthType) specify
 * the direction of resizing.
 *
 * @param e XEvent causing start of resize
 */
void KaWindow::ResizeOpaque(XEvent *e, int hdir, int vdir) {
    _Resize(e, hdir, vdir, true);
}

/**
 * @fn    EndMoveResize()
 * @brief Ends move/resize
 *
 * Ends moving and resizing process.
 */
void KaWindow::EndMoveResize() {
    kahakai->eh->move_resize = EndMoveResizeType;
    labeltext = name;
    UpdateTitlebar();
}

/**
 * @fn    _Maximize(int x, int y)
 * @brief Maximize the window
 *
 * Maximizes the window so that the window with its decorations fills up
 * the whole workarea.
 *
 * @param x Virtual x pos where window should be maximized, ignored if negative
 * @param y Virtual y pos where window should be maximized, ignored if negative
 * @param horizontal Maximize horizontally. Always true if flags.fullscreen
 * @param vertical   Maximize vertically.   Always true if flags.fullscreen
 */
void KaWindow::_Maximize(int x, int y, bool horizontal, bool vertical) {

    if (flags.max)
        return;

    if (horizontal && vertical && flags.handle) {
        DecorHandleOff();
        restore_max_flags.handle = true;
    }

    if (horizontal && vertical && flags.border) {
        DecorBorderOff();
        restore_max_flags.border = true;
    }

    flags.hmax = flags.vmax = false;

    int n_w, n_h, new_width, new_height;

    int workx, worky, workw, workh;
    kascreen->GetWorkareaSize(&workx, &worky, &workw, &workh);

#ifdef XINERAMA
    const XineramaScreenInfo* xinerama_screen = 
        kahakai->xinerama->determineScreen(this);

    if (xinerama_screen) {
        kahakai->xinerama->getWorkArea(xinerama_screen, kascreen,
                                      workx, worky, workw, workh);
    }
#endif

    new_width  = attrib.getWidth();
    new_height = attrib.getHeight();

    if (flags.fullscreen) {
        if (horizontal) new_width  = kascreen->width;
        if (vertical)   new_height = kascreen->height;
    } else {
        if (horizontal) new_width  = workw;
        if (vertical)   new_height = workh;
    }

    restore_max_attrib.setWidth(attrib.getWidth());
    restore_max_attrib.setHeight(attrib.getHeight());

    if (vertical && flags.shaded) {
        restore_max_attrib.setHeight(restore_shade);
        restore_shade = new_height;
        new_height = attrib.getHeight();
        shadeOnUnmaximize = true;
    }

    int rest_x = attrib.getX();
    int rest_y = attrib.getY();

    if (IncFrameSizeCheck(new_width, new_height, &n_w, &n_h)) {
        restore_max_attrib.setX(rest_x);
        restore_max_attrib.setY(rest_y);

        if (x > -100000 && y > -100000) {
            restore_max_attrib.setMisc0(kascreen->v_x);
            restore_max_attrib.setMisc1(kascreen->v_y);
        } else {
            restore_max_attrib.setMisc0(attrib.getX() - (attrib.getX() % kascreen->width));
            restore_max_attrib.setMisc1(attrib.getY() - (attrib.getY() % kascreen->height));
        }

        // decide what viewport this window should maximize to

        int voffsetx = 0, voffsety = 0;

        if (!IsHorizontallyVisible() || GetFrameTop() > kascreen->height
                                     || GetFrameBottom() < 0) {

            // the window is currently not visible at all, and thus not on this
            // viewport. figure out which viewport it should maximize to.

            voffsetx = GetFrameLeft() - (GetFrameLeft() % kascreen->width);
            voffsety = GetFrameTop() - (GetFrameTop() % kascreen->height);

        } // else, it's visible and thus in the current viewport

        int center_padding = (workw - n_w) / 2;

        if (flags.fullscreen) {
            SetFrameLeft(0 - voffsetx + center_padding);
            SetFrameTop(0 - voffsety);
        } else {
            if (horizontal) SetFrameLeft(workx + voffsetx + center_padding);
            if (vertical)   SetFrameTop(worky + voffsety);
        }

        SetFrameWidth(n_w);
        SetFrameHeight(n_h);

        RedrawWindow();

        flags.max = true;
        flags.hmax = horizontal;
        flags.vmax = vertical;

        flags.shaded = shadeOnUnmaximize;

        UpdateButtons(MaxCBoxType);
        UpdateTitlebar();

        net->SetWmState(this);
        kascreen->UpdateCheckboxes(MaxCBoxType);
    }
}

/**
 * @fn    UnMaximize()
 * @brief Unmaximize the window
 *
 * Restores size and position of maximized window.
 */
void KaWindow::UnMaximize() {

    if (!flags.max) {
        flags.hmax = flags.vmax = false;
        return;
    }

    int n_w, n_h, restore_height, tmp_shade_height = 0;

    // shadeOnUnmaximize flag is set by _Maximize if window was previously
    // shaded so we get shaded window again
    if (shadeOnUnmaximize) {
        restore_height = attrib.getHeight();
        tmp_shade_height = restore_max_attrib.getHeight();
        flags.shaded = true;
        shadeOnUnmaximize = false;

    } else
        restore_height = restore_max_attrib.getHeight();

    if (IncSizeCheck(restore_max_attrib.getWidth(), restore_height, &n_w, &n_h)) {
        attrib.setX(restore_max_attrib.getX());
        attrib.setY(restore_max_attrib.getY());
        attrib.setWidth(n_w);
        attrib.setHeight(n_h);

        flags.max = flags.vmax = flags.hmax = false;

        if (flags.shaded)
            restore_shade = tmp_shade_height;

        UpdateButtons(MaxCBoxType | ShadeCBoxType | VMaxCBoxType | HMaxCBoxType);
        RedrawWindow();

        net->SetWmState(this);
        kascreen->UpdateCheckboxes(MaxCBoxType | ShadeCBoxType |
                                   HMaxCBoxType | VMaxCBoxType);
    }

    if (! flags.handle && restore_max_flags.handle) {
        DecorHandleOn();
        restore_max_flags.handle = false;
        RedrawWindow();
    }

    if (! flags.border && restore_max_flags.border) {
        DecorBorderOn();
        restore_max_flags.border = false;
        RedrawWindow();
    }
}

/**
 * @fn    ToggleMaximize()
 * @brief Maximizes or unmaximize window
 *
 * If window isn't maximized this function maximizes it and if it is already
 * maximized then function will unmaximized window.
 */
void KaWindow::ToggleMaximize() {

    if (! flags.max)
        Maximize();
    else if (flags.hmax && flags.vmax)
        UnMaximize();
}

void KaWindow::ToggleMaximizeVertically() {
    if (! flags.max)
        MaximizeVertically();
    else if (flags.vmax && !flags.hmax)
        UnMaximize();
}

void KaWindow::ToggleMaximizeHorizontally() {
    if (! flags.max)
        MaximizeHorizontally();
    else if (flags.hmax && !flags.vmax)
        UnMaximize();
}

/**
 * @fn    Close()
 * @brief Close the window
 *
 * Sends a delete message to the client window. A normal running X window
 * should accept this event and destroy itself.
 */
void KaWindow::Close() {
    XEvent ev;

    ev.type = ClientMessage;
    ev.xclient.window = id;
    ev.xclient.message_type = XInternAtom(display, "WM_PROTOCOLS", false);
    ev.xclient.format = 32;
    ev.xclient.data.l[0] = XInternAtom(display, "WM_DELETE_WINDOW", false);
    ev.xclient.data.l[1] = CurrentTime;

    XGrabServer(display);
    if (validatedrawable(id))
        XSendEvent(display, id, false, NoEventMask, &ev);
    else DELETED;
    XUngrabServer(display);
    XSync(display, false);
}

/**
 * @fn    Kill()
 * @brief Kill the window
 *
 * Tells the the X server to remove the window from the screen through
 * killing the process that created it.
 */
void KaWindow::Kill() {
    XGrabServer(display);
    if (validatedrawable(id))
        XKillClient(display, id);
    else DELETED;
    XUngrabServer(display);
}

/**
 * @fn    CloseKill(XEvent *e, KaAction *ac)
 * @brief Close/Kill the window
 *
 * Checks if the window will accept a delete message. If it will, then we
 * use that method for closing the window otherwise we use the kill method.
 *
 * @param e XEvent causing close/kill of window
 * @param ac KaAction object
 */
void KaWindow::CloseKill() {
    int i, n;
    bool close = false;
    Atom *protocols;
    Atom del_atom = XInternAtom(display, "WM_DELETE_WINDOW", false);

    XGrabServer(display);
    if (validatedrawable(id)) {
        if (XGetWMProtocols(display, id, &protocols, &n)) {
            for (i = 0; i < n; i++) if (protocols[i] == del_atom) close = true;
            XFree(protocols);
        }
    } else DELETED;
    XUngrabServer(display);

    if (close)
        Close();
    else
        Kill();
}

/**
 * @fn    MenuMap(const char* name, bool focus)
 * @brief Maps a menu
 *
 * Links the window to the menu and maps it at the position of the button
 * event causing mapping. If menu is already mapped nothing is done.
 *
 * @param name Menu name
 * @param bool True if we should focus first item in menu
 */
void KaWindow::MenuMap(const char* name, bool focus) {
    int x, y, exp;
    KaMenu *menu = kascreen->GetMenuNamed(name);

    if (! menu) return;
    if (kahakai->eh->move_resize != EndMoveResizeType) return;

    int workx, worky, workw, workh;
    kascreen->GetWorkareaSize(&workx, &worky, &workw, &workh);

    if (kascreen->QueryPointer(&x, &y)) {
        if (menu->ext_type) menu->Build(kascreen);
        menu->wf = id;
        menu->ftype = MenuWFuncMask;
        list<KaMenuItem *>::iterator it = menu->item_list.begin();
        for (exp = 0; it != menu->item_list.end(); ++it)
            exp += (*it)->ExpandAll(this);
        if (exp) menu->Build(kascreen);
        if ((y + menu->height + kascreen->style.menu.borderThickness * 2) >
            (unsigned int) (workh + worky))
            y -= (menu->height + kascreen->style.menu.borderThickness * 2);
        if ((x + menu->width + kascreen->style.menu.borderThickness * 2) >
            (unsigned int) (workw + workx))
            x -= (menu->width + kascreen->style.menu.borderThickness * 2);
        menu->Map(x, y);
        if (focus) menu->FocusFirst();
    }
}

/**
 * @fn    MenuRemap(const char* name, bool focus)
 * @brief Remaps a menu
 *
 * Links the window to the menu and maps it at the position of the button
 * event causing mapping. If the window is already mapped then we just move
 * it to the new position.
 *
 * @param name Menu name
 * @param bool True if we should focus first item in menu
 */
void KaWindow::MenuRemap(const char* name, bool focus) {
    int x, y, exp;
    KaMenu *menu = kascreen->GetMenuNamed(name);

    if (! menu) return;
    if (menu->dynamic && menu->mapped) {
        menu->Unmap(menu->has_focus);
        if (! (menu = kascreen->CreateDynamicMenu(name))) return;
    }
    if (kahakai->eh->move_resize != EndMoveResizeType) return;

    int workx, worky, workw, workh;
    kascreen->GetWorkareaSize(&workx, &worky, &workw, &workh);

    if (kascreen->QueryPointer(&x, &y)) {
        if (menu->ext_type) menu->Build(kascreen);
        menu->wf = id;
        menu->ftype = MenuWFuncMask;
        list<KaMenuItem *>::iterator it = menu->item_list.begin();
        for (exp = 0; it != menu->item_list.end(); ++it)
            exp += (*it)->ExpandAll(this);
        if (exp) menu->Build(kascreen);
        if ((y + menu->height + kascreen->style.menu.borderThickness * 2) >
            (unsigned int) (workh + worky))
            y -= (menu->height + kascreen->style.menu.borderThickness * 2);
        if ((x + menu->width + kascreen->style.menu.borderThickness * 2) >
            (unsigned int) (workw + workx))
            x -= (menu->width + kascreen->style.menu.borderThickness * 2);
        menu->ignore = true;
        menu->ReMap(x, y);
        menu->ignore = false;
        if (focus) menu->FocusFirst();
    }
}

/**
 * @fn    MenuUnmap(const char* name, bool focus)
 * @brief Unmaps a menu
 *
 * Unmaps a menu and all its submenus.
 *
 * @param name Menu name
 * @param bool True if we should focus root item
 */
void KaWindow::MenuUnmap(const char* name, bool focus) {
    KaMenu *menu = kascreen->GetMenuNamed(name);

    if ((! menu) || (! menu->mapped)) return;
    if (kahakai->eh->move_resize != EndMoveResizeType) return;

    menu->Unmap(focus);
    menu->UnmapSubmenus(focus);
}

/**
 * @fn    Shade()
 * @brief Shade window
 *
 * Resizes window to so that only the titlebar is shown. Remembers previous
 * height of window that will be restored when unshading.
 */
void KaWindow::Shade(bool force) {

    if (flags.shaded && !force)
        return;

    int n_w, n_h_unused;
    int height = title_w;

    if (IncSizeCheck(attrib.getWidth(), height, &n_w, &n_h_unused)) {
        flags.shaded = true;
        restore_shade = attrib.getHeight();
        attrib.setWidth(n_w);
        SetFrameHeight(height);
        RedrawWindow();
        kascreen->UpdateCheckboxes(ShadeCBoxType);
        UpdateButtons(ShadeCBoxType);
    }
}

/**
 * @fn    UnShade()
 * @brief Unshade window
 *
 * Restores height of shaded window.
 */
void KaWindow::UnShade() {
    if (! flags.shaded) return;

    flags.shaded = false;
    attrib.setHeight(restore_shade);
    RedrawWindow();
    kascreen->UpdateCheckboxes(ShadeCBoxType);
    UpdateButtons(ShadeCBoxType);
    net->SetWmState(this);
}

/**
 * @fn    ToggleShade()
 * @brief Shades window or unshades it
 *
 * If window isn't shaded this function will shade it and if it is already
 * shaded function unshades it.
 */
void KaWindow::ToggleShade() {
    if (flags.shaded) UnShade();
    else Shade();
}

/**
 * @fn    Sticky()
 * @brief Makes window sticky
 *
 * Sets the sticky flag to true. This makes viewport moving functions
 * to ignore this window.
 */
void KaWindow::Sticky() {

    if (flags.sticky) return;

    flags.sticky = true;
    kascreen->UpdateCheckboxes(StickCBoxType);
}

/**
 * @fn    UnSticky()
 * @brief Makes window normal
 *
 * Sets the sticky flag to false. This makes viewport moving functions
 * treat this window as a normal window.
 */
void KaWindow::UnSticky() {

    if (! flags.sticky) return;

    flags.sticky = false;
    kascreen->UpdateCheckboxes(StickCBoxType);
}

/**
 * @fn    ToggleSticky()
 * @brief Toggles sticky flag
 *
 * Inverts the sticky flag.
 */
void KaWindow::ToggleSticky() {
    if (flags.sticky) UnSticky();
    else Sticky();
}

/**
 * @fn    Minimize()
 * @brief Minimize window
 *
 * Sets window in iconic state.
 */
void KaWindow::Minimize(bool force) {

    if (flags.hidden && !force) return;

    net->SetState(this, IconicState);
    net->SetWmState(this);
    UpdateButtons(MinCBoxType);
    kascreen->UpdateCheckboxes(MinCBoxType);

}

/**
 * @fn    UnMinimize()
 * @brief Restore window
 *
 * Sets window in normal state.
 */
void KaWindow::UnMinimize() {

    if (! flags.hidden) return;

    flags.hidden = false;
    net->SetState(this, NormalState);
    net->SetWmState(this);
    UpdateButtons(MinCBoxType);
    kascreen->UpdateCheckboxes(MinCBoxType);
    Focus();
}

/**
 * @fn    ToggleMinimize()
 * @brief Toggles minimization
 *
 * Toggles between iconic and normal state.
 */
void KaWindow::ToggleMinimize() {

    if (flags.hidden) UnMinimize();
    else Minimize();
}

/**
 * @fn    ToggleMinimizeActive()
 * @brief Toggles minimization, activating window if un-minimized.
 *
 * Toggles between iconic and normal state, activating (raise + focus) the
 * window if de-iconfiying.
 */
void KaWindow::ToggleMinimizeActive() {
    if (flags.hidden) { // just got raised
        UnMinimize();
        MoveWindowToCurrentViewport();
        Raise();
    } else
        Minimize();
}

/**
 * @fn    FullscreenOn()
 * @brief Turns on fullscreen flag
 *
 * Sets window in fullscreen mode. When window is maximized in fullscreen mode
 * it is maximized to the edges of the screen.
 */
void KaWindow::FullscreenOn() {

    if (flags.fullscreen) return;

    flags.fullscreen = true;

    if (flags.max) {
        flags.max = false;
        int res_x = restore_max_attrib.getX();
        int res_y = restore_max_attrib.getY();
        int res_w = restore_max_attrib.getWidth();
        int res_h = restore_max_attrib.getHeight();
        _Maximize(restore_max_attrib.getMisc0(), restore_max_attrib.getMisc1());
        restore_max_attrib.setX(res_x);
        restore_max_attrib.setY(res_y);
        restore_max_attrib.setWidth(res_w);
        restore_max_attrib.setHeight(res_h);
    }
    net->SetWmState(this);
    UpdateButtons(FsCBoxType);
    kascreen->UpdateCheckboxes(FsCBoxType);
}

/**
 * @fn    FullscreenOff()
 * @brief Turns off fullscreen flag
 *
 * Sets window in non fullscreen mode.
 */
void KaWindow::FullscreenOff() {

    if (! flags.fullscreen) return;

    flags.fullscreen = false;

    if (flags.max) {
        flags.max = false;
        int res_x = restore_max_attrib.getX();
        int res_y = restore_max_attrib.getY();
        int res_w = restore_max_attrib.getWidth();
        int res_h = restore_max_attrib.getHeight();
        _Maximize(restore_max_attrib.getMisc0(), restore_max_attrib.getMisc1());
        restore_max_attrib.setX(res_x);
        restore_max_attrib.setY(res_y);
        restore_max_attrib.setWidth(res_w);
        restore_max_attrib.setHeight(res_h);
    }
    net->SetWmState(this);
    UpdateButtons(FsCBoxType);
    kascreen->UpdateCheckboxes(FsCBoxType);
}

/**
 * @fn    FullscreenToggle()
 * @brief Toggles fullscreen flag
 *
 * Toggles fullscreen mode.
 */
void KaWindow::FullscreenToggle() {
    if (flags.fullscreen) FullscreenOff();
    else FullscreenOn();
}


/**
 * @fn    DecorTitleOn()
 * @brief Turn on title decoration
 *
 * Turn on titlebar decorations for the window.
 */
void KaWindow::DecorTitleOn() {

    if (flags.title) return;

    flags.title = true;
    flags.all = flags.title && flags.handle && flags.border;
    UpdateAllAttributes();
    MapWindow();
    net->SetAllowedActions(this);
    net->SetWmState(this);
    kascreen->UpdateCheckboxes((flags.all) ? TitleCBoxType | AllCBoxType
                                           : TitleCBoxType);
}

/**
 * @fn    DecorHandleOn()
 * @brief Turn on handle decoration
 *
 * Turn on handlebar decorations for the window.
 */
void KaWindow::DecorHandleOn() {

    if (flags.handle) return;

    flags.handle = true;
    flags.all = flags.title && flags.handle && flags.border;
    heightChanged = true;
    net->SetWmState(this);
    UpdateAllAttributes();
    MapWindow();
    kascreen->UpdateCheckboxes((flags.all) ? HandleCBoxType | AllCBoxType
                                           : HandleCBoxType);
}

/**
 * @fn    DecorBorderOn()
 * @brief Turn on border decoration
 *
 * Turn on border decorations for the window.
 */
void KaWindow::DecorBorderOn() {

    if (flags.border) return;

    flags.border = true;
    flags.all = flags.title && flags.handle && flags.border;
    net->SetWmState(this);
    UpdateAllAttributes();
    MapWindow();
    kascreen->UpdateCheckboxes((flags.all) ? BorderCBoxType | AllCBoxType
                                           : BorderCBoxType);
}

/**
 * @fn    DecorAllOn()
 * @brief Turn on all decorations
 *
 * Turn on all decorations for the window.
 */
void KaWindow::DecorAllOn() {

    if (flags.all) return;

    flags.all = true;
    flags.border = true;
    flags.title = true;
    flags.handle = true;
    net->SetWmState(this);
    UpdateAllAttributes();
    MapWindow();
    kascreen->UpdateCheckboxes(TitleCBoxType | HandleCBoxType |
                               BorderCBoxType | AllCBoxType);
}

/**
 * @fn    DecorTitleOff()
 * @brief Turn off title decoration
 *
 * Turn off title decorations for the window.
 */
void KaWindow::DecorTitleOff() {

    if (flags.shaded || ! flags.title) return;

    flags.title = false;
    flags.all = false;
    been_shaped = true;
    net->SetWmState(this);
    UpdateAllAttributes();
    MapWindow();
    net->SetAllowedActions(this);
    kascreen->UpdateCheckboxes(TitleCBoxType | AllCBoxType);
}

/**
 * @fn    DecorHandleOff()
 * @brief Turn off hendle decoration
 *
 * Turn off handlebar decorations for the window.
 */
void KaWindow::DecorHandleOff() {

    if (! flags.handle) return;

    flags.handle = false;
    flags.all = false;
    heightChanged = true;
    been_shaped = true;
    net->SetWmState(this);
    UpdateAllAttributes();
    MapWindow();
    kascreen->UpdateCheckboxes(HandleCBoxType | AllCBoxType);
}

/**
 * @fn    DecorBorderOff()
 * @brief Turn off border decoration
 *
 * Turn off border decorations for the window.
 */
void KaWindow::DecorBorderOff() {

    if (! flags.border) return;

    flags.border = false;
    flags.all = false;
    net->SetWmState(this);
    UpdateAllAttributes();
    MapWindow();
    kascreen->UpdateCheckboxes(BorderCBoxType | AllCBoxType);
}

/**
 * @fn    DecorAllOff()
 * @brief Turn off all decorations
 *
 * Turn off all decorations for the window.
 */
void KaWindow::DecorAllOff() {

    if (flags.shaded /* || ! flags.all */) return; // if a window is maximized, flags.all will be false

    flags.all = false;
    flags.border = false;
    flags.title = false;
    flags.handle = false;
    been_shaped = true;
    net->SetWmState(this);
    UpdateAllAttributes();
    MapWindow();
    kascreen->UpdateCheckboxes(TitleCBoxType | HandleCBoxType | BorderCBoxType |
                               AllCBoxType);
}

/**
 * @fn    DecorTitleToggle()
 * @brief Toggle title decoration
 *
 * Toggle titlebar decorations for the window.
 */
void KaWindow::DecorTitleToggle() {

    if (flags.title) DecorTitleOff();
    else DecorTitleOn();
}

/**
 * @fn    DecorHandleToggle()
 * @brief Toggle handle decoration
 *
 * Toggle handlebar decorations for the window.
 */
void KaWindow::DecorHandleToggle() {

    if (flags.handle) DecorHandleOff();
    else DecorHandleOn();
}

/**
 * @fn    DecorBorderToggle()
 * @brief Toggle border decoration
 *
 * Toggle border decorations for the window.
 */
void KaWindow::DecorBorderToggle() {

    if (flags.border) DecorBorderOff();
    else DecorBorderOn();
}

/**
 * @fn    DecorAllToggle()
 * @brief Toggle all decorations
 *
 * Toggle all decorations, if one Decor*Toggle is on, turn then all off,
 * else turn them all on.
 * - ibri
 */
void KaWindow::DecorAllToggle() {

    if (flags.border || flags.title || flags.handle) DecorAllOff();
    else DecorAllOn();
}

/**
 * @fn    UpdateButtons(int buttonmask)
 * @brief Updates the buttons in the titlebar
 *
 * Update the masked buttons in the titlebar.
 */
void KaWindow::UpdateButtons(unsigned int buttonmask) {
    if (title_w) {
        title->update();
        list<KaWindowButton *>::iterator bit = buttons.begin();
        for (; bit != buttons.end(); ++bit)
            (*bit)->render();
    }
}

/**
 * @fn    AlwaysontopOn()
 * @brief Make window always on top
 *
 * Add window to list of always on top windows and update always on top
 * windows.
 */
void KaWindow::AlwaysontopOn() {

    if (flags.alwaysontop) return;

    flags.alwaysontop = true;
    flags.alwaysatbottom = false;

    UpdateStack();
    UpdateButtons(AOTCBoxType | AABCBoxType);
    net->SetWmState(this);
    kascreen->RestackWindows(frame->id);
    kascreen->UpdateCheckboxes(AOTCBoxType | AABCBoxType);

    net->SetClientListStacking(kascreen);
}

/**
 * @fn    AlwaysatbottomOn()
 * @brief Make window always at bottom
 *
 * Add window to list of always at bottom windows and update always at bottom
 * windows.
 */
void KaWindow::AlwaysatbottomOn() {

    if (flags.alwaysatbottom) return;

    flags.alwaysontop = false;
    flags.alwaysatbottom = true;

    UpdateStack();
    UpdateButtons(AOTCBoxType | AABCBoxType);
    net->SetWmState(this);
    kascreen->RestackWindows(frame->id);
    kascreen->UpdateCheckboxes(AOTCBoxType | AABCBoxType);

    net->SetClientListStacking(kascreen);
}

/**
 * @fn    AlwaysontopOff()
 * @brief Make window not always on top
 *
 * Removes window from list of always on top windows.
 */
void KaWindow::AlwaysontopOff() {

    if (! flags.alwaysontop) return;

    flags.alwaysontop = false;
    UpdateStack();
    UpdateButtons(AOTCBoxType);
    net->SetWmState(this);
    kascreen->RestackWindows(frame->id);
    kascreen->UpdateCheckboxes(AOTCBoxType);
}

/**
 * @fn    AlwaysatbottomOff()
 * @brief Make window not always at bottom
 *
 * Removes window from list of always at bottom windows.
 */
void KaWindow::AlwaysatbottomOff() {

    if (! flags.alwaysatbottom) return;

    flags.alwaysatbottom = false;
    UpdateStack();
    UpdateButtons(AABCBoxType);
    net->SetWmState(this);
    kascreen->RestackWindows(frame->id);
    kascreen->UpdateCheckboxes(AABCBoxType);
}

/**
 * @fn    AlwaysontopToggle()
 * @brief Toggle always on top flag
 *
 * If window is always on top we removed it from always on top list, or if
 * window isn't always on top we add it to always on top list.
 */
void KaWindow::AlwaysontopToggle() {

    if (flags.alwaysontop) AlwaysontopOff();
    else AlwaysontopOn();
}

/**
 * @fn    AlwaysatbottomToggle()
 * @brief Toggle always at bottom flag
 *
 * If window is always at bottom we removed it from always at bottom list,
 * or if window isn't always at bottom we add it to always at bottom list.
 */
void KaWindow::AlwaysatbottomToggle() {

    if (flags.alwaysatbottom) AlwaysatbottomOff();
    else AlwaysatbottomOn();
}

/**
 * @fn    AcceptConfigRequestOn()
 * @brief Accept ConfigureRequests from this window
 *
 * Sets flag so the configure request events from this window are handled as
 * usual.
 */
void KaWindow::AcceptConfigRequestOn() {
    ign_config_req = false;
}

/**
 * @fn    AcceptConfigRequestOff()
 * @brief Don't accept ConfigureRequests from this window
 *
 * Sets flag so the configure request events from this window are ignored.
 */
void KaWindow::AcceptConfigRequestOff() {
    ign_config_req = true;
}

/**
 * @fn    AcceptConfigRequestToggle()
 * @brief Toggle ign_config_req flag
 *
 * Toggles configure request ignore flag.
 */
void KaWindow::AcceptConfigRequestToggle() {
    ign_config_req = !ign_config_req;
}

/**
 * @fn    MoveResize(const char* geometry)
 * @brief Moves and resizes window
 *
 * X geometry string. Moves and resizes window to
 * the size and position in the actual screen area specified by X geometry
 * string.
 *
 * @param geometry New position in X geometry format
 */
void KaWindow::MoveResize(const char* geometry_str) {
    int x, y, geometry;
    unsigned int width, height;

    if (kahakai->eh->move_resize != EndMoveResizeType || !geometry_str)
        return;

    if (flags.max) {
        flags.max = flags.vmax = flags.hmax = false;
        shadeOnUnmaximize = false;
        kascreen->UpdateCheckboxes(MaxCBoxType | ShadeCBoxType);
    }

    width = attrib.getWidth();
    height = attrib.getHeight();

    geometry = XParseGeometry(geometry_str, &x, &y, &width, &height);
    IncSizeCheck(width, height, &attrib.width, &attrib.height);

    if (geometry & XValue) {
        if (geometry & XNegative)
            attrib.setX(kascreen->width + x - attrib.getWidth());
        else attrib.setX(x);
    }
    if (geometry & YValue) {
        if (geometry & YNegative)
            attrib.setY(kascreen->height + y - attrib.getHeight());
        else attrib.setY(y);
    }

    RedrawWindow();
}

/**
 * @fn    MoveResizeVirtual(const char* geometry)
 * @brief Moves and resizes window
 *
 * X geometry string. Moves and resizes window to
 * the size and position in the virtual screen area specified by X geometry
 * string.
 *
 * @param geometry New position in X geometry format
 */
void KaWindow::MoveResizeVirtual(const char* geometry_str) {
    int x, y, geometry;
    unsigned int width, height;

    if (kahakai->eh->move_resize != EndMoveResizeType || !geometry_str) return;
    width = attrib.getWidth();
    height = attrib.getHeight();

    geometry = XParseGeometry(geometry_str, &x, &y, &width, &height);
    IncSizeCheck(width, height, &attrib.width, &attrib.height);

    Gravitate(RemoveGravity);
    if (geometry & XValue) {
        if (geometry & XNegative)
            attrib.setX(((kascreen->v_xmax + kascreen->width) + x -
                         attrib.getWidth()) - kascreen->v_x);
        else attrib.setX(x - kascreen->v_x);
    }
    if (geometry & YValue) {
        if (geometry & YNegative)
            attrib.setY(((kascreen->v_ymax + kascreen->height) + y - attrib.getHeight()) - kascreen->v_y);
        else attrib.setY(y - kascreen->v_y);
    }
    Gravitate(ApplyGravity);

    RedrawWindow();
}

/**
 * @fn    MoveWindowBy(const char* geometry_str)
 * @brief Moves window by x,y position
 *
 * Moves window by x,y position; +10-5 moves window 10 pixels right, 5 pixels up
 */
void KaWindow::MoveWindowBy(const char* geometry_str) {
    int x, y, mask;
    unsigned int w, h;

    if (flags.max) {
        flags.max = flags.hmax = flags.vmax = false;
        shadeOnUnmaximize = false;
        kascreen->UpdateCheckboxes(MaxCBoxType | ShadeCBoxType);
    }

    mask = XParseGeometry(geometry_str, &x, &y, &w, &h);
    attrib.setX(attrib.getX() + x);
    attrib.setY(attrib.getY() + y);

    RedrawWindow();
}

/**
 * @fn    MoveWindowToPointer(XEvent *e)
 * @brief Moves window to pointer position
 *
 * Moves window to mouse pointer position and makes sure that it isn't
 * moved outside screen.
 *
 * @param e XEvent causing function call
 */
void KaWindow::MoveWindowToPointer(XEvent *e) {

    attrib.setX(e->xbutton.x_root - attrib.getWidth() / 2);
    attrib.setY(e->xbutton.y_root - attrib.getHeight() / 2);

    int workx, worky, workw, workh;
    kascreen->GetWorkareaSize(&workx, &worky, &workw, &workh);

#ifdef XINERAMA
    kahakai->xinerama->adjustWindowPosition(this);
#endif    

    if (GetFrameRight() > workw)
        SetFrameRight(workw);
    else if (GetFrameLeft() < workx)
        SetFrameLeft(workx);

    if (GetFrameBottom() > workh)
        SetFrameBottom(workh);
    else if (GetFrameTop() < worky)
        SetFrameTop(worky);

    RedrawWindow();
}

bool KaWindow::OnCurrentDesktop() {
    return desktop_mask & (1L << kascreen->current_desktop->number);
}

/**
 * @fn    IsHorizontallyOnScreen()
 * @brief Returns whether or not the window is within the horizontal boundaries
 *        of the screen.
 */
bool KaWindow::IsHorizontallyOnScreen() {
    return overlaps(GetFrameLeft(), GetFrameRight(), 0, kascreen->width);
}

/**
 * @fn    IsVerticallyOnScreen()
 * @brief Returns whether or not the window is within the vertical boundaries
 *        of the screen.
 */
bool KaWindow::IsVerticallyOnScreen() { 
    return overlaps(GetFrameTop(), GetFrameBottom(), 0, kascreen->height);
}

/**
 * @fn    IsOnScreen()
 * @brief Returns whether or not the window is within both the horizontal
 *        and vertical boundaries of the screen.
 */
bool KaWindow::IsOnScreen() {
    return IsHorizontallyOnScreen() && IsVerticallyOnScreen();
}

/**
 * @fn    IsHorizontallyInWorkarea()
 * @brief Returns whether or not the window is within the horizontal
 *        boundaries of the workarea.
 */
bool KaWindow::IsHorizontallyInWorkarea() {
    int workx, workw, junk;
    kascreen->GetWorkareaSize(&workx, &junk, &workw, &junk);

    return overlaps(GetFrameLeft(), GetFrameRight(), workx, workx+workw);
}

/**
 * @fn    IsVerticallyInWorkarea()
 * @brief Returns whether or not the window is within the vertical
 *        boundaries of the workarea.
 */
bool KaWindow::IsVerticallyInWorkarea() {
    int worky, workh, junk;
    kascreen->GetWorkareaSize(&junk, &worky, &junk, &workh);

    return overlaps(GetFrameTop(), GetFrameBottom(), worky, worky+workh);
}

/**
 * @fn    IsInWorkarea()
 * @brief Returns whether or not the window is within both the vertical
 *        and horizontal boundaries of the workarea.
 */
bool KaWindow::IsInWorkarea() {
    return IsHorizontallyInWorkarea() && IsVerticallyInWorkarea();
}

/**
 * @fn    IsVisible()
 * @brief Returns true if window is on current desktop and is in workarea.
 */
bool KaWindow::IsVisible() {
    return OnCurrentDesktop() && IsInWorkarea();
}

/**
 * @fn    MoveWindowToSmartPlace(XEvent *)
 * @brief Moves window to smart position
 *
 * Moves window using Smart Placement algorithm.
 */
void KaWindow::MoveWindowToSmartPlace(XEvent *e) {

    if (positioned_by_app || windowType == DockWindowType) {
        RedrawWindow();
        return; 
    }

    // If transient, try to center over the parent, if not possible, continue
    if (transient_for && CenterOverParent())
        return; 

    // try to find an open place onscreen to put the window, if we can't find
    // one, then put it centered under the pointer
    if (!SmartPlaceLeftToRight())
        MoveWindowToPointer(e);
}

/**
 * @fn    SmartPlaceLeftToRight()
 * @brief Tries to find a place to put the window, left to right.
 *
 * Tries to find an empty spot in the workarea, sweeping from left to right and
 * then top to bottom.  Returns true if it found a spot, else false.
 */
bool KaWindow::SmartPlaceLeftToRight() {

    int workx, worky, workw, workh;
    kascreen->GetWorkareaSize(&workx, &worky, &workw, &workh);

    bool found_spot = false;

    while (IsVerticallyInWorkarea() && GetFrameBottom() < worky+workh
           && !found_spot) {
        SetFrameLeft(0);
        
        while (IsHorizontallyInWorkarea() && GetFrameRight() < workx+workw
               && !found_spot) {

            found_spot = true; 

            vector<KaWindow *>::iterator it = kascreen->kawindow_list.begin();
            for (; it != kascreen->kawindow_list.end(); it++) { 
                if (!found_spot)
                    break;  

                KaWindow* window = *it;

                if (window != this && window->flags.tasklist && window->IsVisible()) {
                    if (intersects(window->GetFrameLeft(), 
                                   window->GetFrameTop(), 
                                   window->GetFrameRight(), 
                                   window->GetFrameBottom(),
                                   GetFrameLeft(), GetFrameTop(),
                                   GetFrameRight(), GetFrameBottom())) {
                        found_spot = false;
                        SetFrameLeft(window->GetFrameRight());
                    }
                }
            }
            
            if (!found_spot) SetFrameLeft(GetFrameLeft()+1);
        }
        
        if (!found_spot) SetFrameTop(GetFrameTop()+1);
    }
    
    if (found_spot) {
#ifdef XINERAMA
        kahakai->xinerama->adjustWindowPosition(this);
#endif  
        RedrawWindow();
    }
    
    return found_spot;

}

/**
 * @fn    CenterOverParent()
 * @brief Centers transient/dialog windows over their parents
 *
 * Returns true if the window found its parent and centered itself,
 * otherwise window does not move and false is returned.
 * FIXME: Use the net_wm_type thing (net_wm_type_dialog)
 */
bool KaWindow::CenterOverParent() {

    KaWindow * parent = (KaWindow *)kahakai->FindWin(transient_for, WindowType);

    if (!parent || !parent->IsVisible())
        return false;

    int centerX = parent->GetFrameLeft() + parent->GetFrameWidth()/2;
    int centerY = parent->GetFrameTop() + parent->GetFrameHeight()/2;

    int workx, worky, workw, workh;
    kascreen->GetWorkareaSize(&workx, &worky, &workw, &workh);

    attrib.setX(min(workw-GetFrameWidth(),
                    max(workx, centerX - GetFrameWidth()/2)));
    attrib.setY(min(workh-attrib.getHeight(),
                    max(workx, centerY - GetFrameWidth()/2)));

    Gravitate(ApplyGravity);

#ifdef XINERAMA
    kahakai->xinerama->adjustWindowPosition(this);
#endif
    
    RedrawWindow();

    return true;
}

/**
 * @fn    MoveWindowToViewPort(const char* viewport_str)
 * @brief Moves window to Viewport
 *
 * Moves window to viewport specified. Viewport 2x3 would be as shown below in
 * a desktop with 3x3 viewports (2nd over, 3rd down):
 * _____________
 * |___|___|___|
 * |___|___|___|
 * |___|_x_|___|
 *
 * @param viewport_str String in the format described above
 */
void KaWindow::MoveWindowToViewport(const char* viewport_str) {

    if(viewport_str == NULL){
        logger << warn << "missing parameter for MoveToViewport" << std::endl;
        return;
    }

    int x, y;
    if (!viewportStrToXY(viewport_str, &x, &y)) {
        logger << warn << "invalid parameter `" << viewport_str
               << "' in MoveToViewport" << std::endl;
        return;
    }

    int full_x, full_y, dest_x, dest_y, move_x, move_y;
    full_x = attrib.getX() + kascreen->v_x;
    full_y = attrib.getY() + kascreen->v_y;
    dest_x = (x-1) * kascreen->width + attrib.getX();
    dest_y = (y-1) * kascreen->height + attrib.getY();
    move_x = dest_x - full_x;
    move_y = dest_y - full_y;
    attrib.setX(attrib.getX() + move_x);
    attrib.setY(attrib.getY() + move_y);

    RedrawWindow();
}

/**
 * @fn    UpdateStack(void)
 * @brief Update desktop stacking lists after a join/part desktop
 *
 * If we part or join a desktop, the desktops stacking_lists must be
 * updated.
 */

void KaWindow::UpdateStack(void) {
    DESKTOP_LOOP_MASK(kascreen, desktop_mask) {
        list<Window> *stacking_list;
        if (flags.alwaysontop)
            stacking_list = &((*__dit)->aot_stacking_list);
        else if (flags.alwaysatbottom)
            stacking_list = &(*__dit)->aab_stacking_list;
        else
            stacking_list = &(*__dit)->stacking_list;
        list<Window>::iterator it = find(stacking_list->begin(),
                                        stacking_list->end(), frame->id);
        if (it == stacking_list->end()) {
            (*__dit)->stacking_list.remove(frame->id);
            (*__dit)->aot_stacking_list.remove(frame->id);
            (*__dit)->aab_stacking_list.remove(frame->id);
            stacking_list->push_front(frame->id);
        }
    } else {
        (*__dit)->stacking_list.remove(frame->id);
        (*__dit)->aot_stacking_list.remove(frame->id);
        (*__dit)->aab_stacking_list.remove(frame->id);
    }
    net->SetClientListStacking(kascreen);
}

/**
 * @fn    DesktopMask(const char* mask_str)
 * @brief Set desktop mask
 *
 * Sets the mask for in which desktops the window should be a member.
 *
 * @param mask_str Either "all" else a tab/space separated list of desktops
 */
void KaWindow::DesktopMask(const char* mask_str) {

    unsigned int mask = desktopMaskStrToMask(mask_str, desktop_mask,
                                             kascreen->config.desktops);
    if (!mask) {
        logger << warn << "DesktopMask: bad input: " << mask_str << endl;
        return;
    }

    desktop_mask = mask;

    UpdateStack();

    if (OnCurrentDesktop())
        Show();
    else
        Hide();

    net->SetDesktop(this);
    net->SetDesktopMask(this);
}

/**
 * @fn    JoinDesktop(const char* desktop_str)
 * @brief Join desktop
 *
 * Join window to desktop specified by action parameter.
 *
 * @param desktop_str Desktop number (0-n) as a string
 */
void KaWindow::JoinDesktop(const char* desktop_str) {

    unsigned int desk = (unsigned int) desktopStrToNum(desktop_str);

    if ((int)desk == -1 || (int)desk == 16) {
        logger << warn << "JoinDesktop: bad input: " << desk << endl;
        return;
    }

    if (desk < kascreen->config.desktops) {
        desktop_mask |= 1L << desk;
        UpdateStack();

        if (OnCurrentDesktop())
            Show();

        net->SetDesktop(this);
        net->SetDesktopMask(this);
    }
}

/**
 * @fn    PartDesktop(const char* desktop_str)
 * @brief Part desktop
 *
 * Part window from desktop specified by action parameter.
 *
 * @param desktop_str Desktop number (0-n) as a string
 */
void KaWindow::PartDesktop(const char* desktop_str) {

    unsigned int desk = (unsigned int) desktopStrToNum(desktop_str);

    if ((int)desk == -1 || (int)desk == 16) {
        logger << warn << "PartDesktop: bad input: " << desk << endl;
        return;
    }

    if (desk < kascreen->config.desktops) {
        long int new_mask = desktop_mask & ~(1L << desk);
        if (new_mask) {
            desktop_mask = new_mask;
            UpdateStack();

            if (!OnCurrentDesktop())
                Hide();

            net->SetDesktop(this);
            net->SetDesktopMask(this);
        }
    }
}

/**
 * @fn    PartCurrentDesktop()
 * @brief Part current desktop
 *
 * Part window from current desktop if window is part of another desktop.
 */
void KaWindow::PartCurrentDesktop() {

    long int new_mask = desktop_mask &
        ~(1L << kascreen->current_desktop->number);
    if (new_mask) {
        desktop_mask = new_mask;
        UpdateStack();
        Hide();
        net->SetDesktop(this);
        net->SetDesktopMask(this);
    }
}

/**
 * @fn    JoinCurrentDesktop()
 * @brief Join current desktop
 *
 * Join window to current desktop. Used for unminimizing.
 */
void KaWindow::JoinCurrentDesktop() {

    desktop_mask |= 1L << kascreen->current_desktop->number;
    UpdateStack();
    Show();
    net->SetDesktop(this);
    net->SetDesktopMask(this);
}

/**
 * @fn    JoinAllDesktops()
 * @brief Join all desktops
 *
 * Join window to all desktops.
 */
void KaWindow::JoinAllDesktops() {

    desktop_mask = (1L << 16) - 1;
    UpdateStack();
    Show();
    net->SetDesktop(this);
    net->SetDesktopMask(this);
}

/**
 * @fn    PartAllDesktopsExceptCurrent()
 * @brief Part all desktops
 *
 * Part window from all desktops except current desktop.
 */
void KaWindow::PartAllDesktopsExceptCurrent() {

    desktop_mask = 1L << kascreen->current_desktop->number;
    UpdateStack();
    Show();
    net->SetDesktop(this);
    net->SetDesktopMask(this);
}

/**
 * @fn    PartCurrentJoinDesktop(const char* desktop_str)
 * @brief Part current desktop and join desktop
 *
 * Part current desktop and join desktop specified by parameter.
 *
 * @param desktop_str Desktop number (0-n) as a string
 */
void KaWindow::PartCurrentJoinDesktop(const char* desktop_str) {

    unsigned int desk = (unsigned int) desktopStrToNum(desktop_str);

    if((int)desk == -1 || (int)desk == 16) {
        logger << warn << "PartCurrentJoinDesktop: bad input: " << desk << endl;
        return;
    }

    if (desk < kascreen->config.desktops) {

        unsigned dm = desktop_mask;
    
        dm &= ~(1L << kascreen->current_desktop->number);
        dm |= 1L << desk;
        desktop_mask = dm;
        UpdateStack();
        if (dm & (1L << kascreen->current_desktop->number))
            Show();
        else
            Hide();

        net->SetDesktop(this);
        net->SetDesktopMask(this);
    }
}

/**
 * @fn    EvAct(XEvent *e, EventDetail *ed, vector<KaAction *> *acts,
 *              int etype)
 * @brief Calls KaWindow function
 *
 * Tries to match an occurred X event with the actions in an action list.
 * If we have a match then we execute that action.
 *
 * @param e X event that have occurred
 * @param ed Event details
 * @param acts List with actions to match event with
 * @param etype Type of window event occurred on
 */
void KaWindow::EvAct(XEvent *e, EventDetail *ed, vector<KaAction *> *acts,
                     int etype) {
    XEvent fev;
    bool replay = false, wait_release = false, match = false;

    // arrrrggh
    // produces seemingly random false positives 
    //kahakai->eh->PossiblyDemoteDoubleClick(this, ed);

    vector<KaAction *>::iterator it = acts->begin();

    if (kahakai->eh->move_resize != EndMoveResizeType)
        ed->mod |= MoveResizeMask;

    if (etype == WindowType) {
        if (ed->type == ButtonPress) {
            for (; it != acts->end(); ++it) {
                if ((*it)->type == ButtonRelease &&
                    (*it)->detail == ed->detail &&
                    (! ((*it)->mod & MoveResizeMask)))
                    wait_release = match = true;
            }
        }
/*
        else if (ed->type == KeyPress) {
            for (; it != acts->end(); ++it) {
                if ((*it)->type == KeyRelease &&
                    (*it)->detail == ed->detail &&
                    !((*it)->mod & MoveResizeMask)) {
                    wait_release = match = true;
                    XAutoRepeatOff(display);
                }
            }
        }
*/
    }

    for(it = acts->begin(); it != acts->end(); ++it) {
        if (eventmatch(display, *it, ed, kascreen->imod_mask)) {
            match = true;
            //XAutoRepeatOn(display);
            if ((*it)->replay && ! wait_release)
                replay = true;
            if ((*it)->delay.tv_sec || (*it)->delay.tv_usec) {
                Interrupt *i = new Interrupt(*it, e, id);
                kahakai->timer->AddInterrupt(i);
            } else {
                if ((*it)->exec)
                    kaexec((*it)->exec, kascreen->displaystring);
                else
                    (*((*it)->winfunc))(this, e, *it);
            }
        }
    }
    if (kahakai->eh->move_resize != EndMoveResizeType) {
        if (deleted) delete this;
        return;
    }

    XSync(display, false);
    while (XCheckTypedEvent(display, FocusOut, &fev))
        kahakai->eh->EvFocus(&fev.xfocus);
    while (XCheckTypedEvent(display, FocusIn, &fev))
        kahakai->eh->EvFocus(&fev.xfocus);

    if (! match)
        replay = true;

    if (etype == WindowType) {
        if (kahakai->eh->IsButtonEventType(ed->type)) {
            if (replay) {
                XAllowEvents(display, ReplayPointer, e->xbutton.time);
            } else
                XAllowEvents(display, AsyncPointer, e->xbutton.time);
        }
        else if (kahakai->eh->IsKeyEventType(ed->type)) {
            if (replay) {
                XAllowEvents(display, ReplayKeyboard, e->xbutton.time);
            } else
                XAllowEvents(display, AsyncKeyboard, e->xbutton.time);
        }
        else if (ed->type == MapRequest && ! mapped) {
            int _state = NormalState;
            if (flags.hidden || state == IconicState)
                _state = IconicState;
            net->SetState(this, _state);
            net->SetVirtualPos(this);
        }
    }
    if (deleted)
        delete this;
}
