/* $Id: WindowMenu.cc,v 1.1 2004/01/18 20:58:19 pervago Exp $ */

/**
 * @file   WindowMenu.cc
 * @date   Sun Jan  4 19:39:51 2004
 * 
 * @brief Menu listing either all windows on current desktops (task switcher)
 * or only minimized windows.
 * 
 */

#include "WindowMenu.hh"
#include "WindowIcons.hh"
#include "MenuItem.hh"

using namespace std;


/**
 * @fn    WindowMenu(void) : KaMenu("__windowlist__")
 * @brief Constructor for WindowMenu class
 *
 * Creates a WindowMenu object, used for fast switching between windows.
 * This is a KaMenu with some extra functionality.
 *
 */
WindowMenu::WindowMenu(char * name) : KaMenu(name) {

    ext_type = TaskExtMenuType;
    title_label = "Window List";

    KaMenuItem *m;
    m = new KaMenuItem(title_label);
    m->type = MenuTitleType;
    AddItem(m);
}

/**
 * @fn    Build(KaScreen *kascreen)
 * @brief Builds WindowMenu
 *
 * Overloaded Build function to make WindowMenu build a menu from
 * current windows.
 *
 * @param kascreen KaScreen to map menu on
 */
void WindowMenu::Build(KaScreen *kascreen) {

    LISTCLEAR(item_list);

    kawindow_list = &kascreen->kawindow_list;
    vector<KaWindow *>::iterator it = kawindow_list->begin();
    for (; it != kawindow_list->end() &&
             ((KaWindow *) *it)->flags.tasklist != true; ++it);

    if (it == kawindow_list->end()) return;

    KaMenuItem *m;
    m = new KaMenuItem(title_label);
    m->type = MenuTitleType;
    AddItem(m);

    KaWindow *ww;
    for ( ; it != kawindow_list->end(); ++it) {
        ww = (KaWindow *) *it;

        WindowIcons wi(ww);

        if (ItemShouldShow(ww, kascreen)) {

            m = new KaMenuItem(FormatItem(ww));

            m->type = MenuItemType;
            m->wfunc = new WindowEvRaiseFocus();
            m->func_mask |= MenuWFuncMask;
            m->func_mask1 |= MenuWFuncMask;
            m->wf = ww->id;

            if (wi.hasIcons()) {
                m->icon = wi.getIcon(16, 16);
            }

            AddItem(m);
        }
    }
    for(it = kawindow_list->begin(); it != kawindow_list->end() &&
             ((KaWindow *) *it)->flags.tasklist != true; ++it);

    KaMenu::Build(kascreen);
}

/**
 * @fn    ItemShouldShow(KaWindow * win)
 * @brief Decides whether a window should show up as an item in the menu.
 *
 * Returns true if the window is flagged to be on the tasklist, and is on
 * the current desktop.
 *
 */
bool WindowMenu::ItemShouldShow(KaWindow * win, KaScreen * screen) {
    return (bool) (win->flags.tasklist && win->OnCurrentDesktop());
}

/**
 * @fn    FormatItem(KaWindow * win)
 * @brief Returns the text to put in the menu to represent <win>
 *
 * The window's name is returned, and put in parentheses if minimized.
 *
 */
string WindowMenu::FormatItem(KaWindow * win) {
    return win->flags.hidden ? "("+win->name+")" : win->name;
}

/**
 * @fn    MinimizedWindowMenu(void) : WindowMenu("__minimizedwindowlist__")
 * @brief Constructor for MinimizedWindowMenu class
 *
 * Creates a MinimizedWindowMenu object, which lists currently minimized
 * windows.  This is a WindowMenu with rather tiny changes.
 *
 */
MinimizedWindowMenu::MinimizedWindowMenu(char * name) : WindowMenu(name) {

    ext_type = MinimizedExtMenuType;
    title_label = "Minimized Windows";

    KaMenuItem *m;
    m = new KaMenuItem(title_label);
    m->type = MenuTitleType;
    AddItem(m);
}

/**
 * @fn    ItemShouldShow(KaWindow * win)
 * @brief Decides whether a window should show up as an item in the menu.
 *
 * Same criteria as WindowMenu's version, but also filters out non-minimized
 * windows.
 *
 */
bool MinimizedWindowMenu::ItemShouldShow(KaWindow * win, KaScreen * screen) {
    return (bool) (win->flags.hidden && 
                   (WindowMenu::ItemShouldShow(win, screen)));
}

/**
 * @fn    FormatItem(KaWindow * win)
 * @brief Returns the text to put in the menu to represent <win>
 *
 * Since we already know that all of these windows are minimized, this simply
 * returns their names unmodified.
 *
 */
string MinimizedWindowMenu::FormatItem(KaWindow * win) {
    return win->name;
}


