/**
 * @file   Xinerama.cc
 * @date   30 Sep 2003
 *
 * @brief Handles Xinerama-specific functions.
 *
 */

#include "Xinerama.hh"
#include "Window.hh"

#include <algorithm>

using namespace std;

#include <X11/extensions/Xinerama.h>

// idea borrowed from sawfish: fake a couple xinerama screens
#ifdef TEST_XINERAMA
static XineramaScreenInfo debug_heads[] = {
    { 0, 0,    0, 640, 480 },
    { 0, 640,  0, 640, 320 }
};
#endif

Xinerama::Xinerama() {
# ifdef TEST_XINERAMA
    active = true;
    
    copy(debug_heads, 
         debug_heads + sizeof(debug_heads) / sizeof(XineramaScreenInfo),
         back_inserter(screens));
# else
    Kahakai* kahakai = Kahakai::Instance();
    int dummy;

    active = XineramaQueryExtension(kahakai->display, &dummy, &dummy);
    if (active)
        active = (XineramaIsActive(kahakai->display) == True);
    else
        active = false;

    if (active) {
        int xinerama_info_num;

        XineramaScreenInfo*  xinerama_info = 
            XineramaQueryScreens(kahakai->display, &xinerama_info_num);
        
        copy(xinerama_info, xinerama_info + xinerama_info_num, 
             back_inserter(screens));
    }
# endif
}


/** 
 * Return Xinerama screen that contains most of the window area relative to other screens.
 * 
 * @param window Window to determine screen for.
 * 
 * @return NULL if Xinerama is not active, screen containing most area otherwise.
 */
const XineramaScreenInfo* Xinerama::determineScreen(KaWindow* window) {
    if (!active) return NULL;
    
    int max_intersection = -1; // Not 0, to guarantee that we return something.

    vector<XineramaScreenInfo>::iterator it, max_it;
        
    for (max_it = it = screens.begin(); it != screens.end(); ++it) {
        int area = intersectionArea(window, &(*it));
        
        if (area > max_intersection) {
            max_intersection = area;
            max_it = it;
        }
    }

    return &(*max_it);
}

/** 
 * Return area of intersection between window and Xinerama screen.
 * 
 * @param window Window to check.
 * @param screen_info Xinerama screen.
 * 
 * @return Zero if no intersection was found, some value less than or
 * equal to total window area otherwise.
 */
int Xinerama::intersectionArea(KaWindow* window, const XineramaScreenInfo* screen_info) {
    int left_edge = screen_info->x_org;
    int right_edge = screen_info->x_org + screen_info->width;        
    int top_edge = screen_info->y_org;
    int bottom_edge = screen_info->y_org + screen_info->height;
    
    int intersectionWidth, intersectionHeight;
    if (window->GetFrameLeft() > left_edge) {
        intersectionWidth = right_edge - window->GetFrameLeft();
    } else {
        intersectionWidth = window->GetFrameRight() - left_edge;
    }
    
    if (window->GetFrameTop() > top_edge) {
        intersectionHeight = bottom_edge - window->GetFrameTop();
    } else {
        intersectionHeight = window->GetFrameBottom() - top_edge;
    }
        
    return max(0, intersectionWidth) * max(0, intersectionHeight);
}

/** 
 * 
 * 
 * @param screen_info Xinerama screen to determine work area for
 * @param kascreen KaScreen object (to get workarea)
 * @param x (out) Returns left of Xinerama-adjusted work area 
 * @param y (out) Top of work area
 * @param width (out) Width of work area
 * @param height (out) Height of work area
 */
void Xinerama::getWorkArea(const XineramaScreenInfo* screen_info,
                           const KaScreen* kascreen,
                           int& x, int& y, int& width, int& height) {

    int left, top, right, bottom;

    // Get Xinerama screen dimensions
    left   = screen_info->x_org;
    top    = screen_info->y_org;
    right  = screen_info->x_org + screen_info->width;
    bottom = screen_info->y_org + screen_info->height;

    int workx, worky, workw, workh;
    kascreen->GetWorkareaSize(&workx, &worky, &workw, &workh);


    // Clip xinerama screen using workarea
    x      = max(workx, left); 
    width  = min(workw, right-left);
    y      = max(worky, top);
    height = min(workh, bottom-top);
}

/** 
 * @fn adjustWindowPosition(KaWindow* window)
 * @brief Adjust window position according to Xinerama configuration.
 *
 * If window is between two xinerama screens, move it to the screen
 * containing most of its area. If window is in Xinerama's "dead area"
 * (caused by non-equal resolutions), move it outside of it.
 * 
 * @param window Window that is being adjusted.
 */
void Xinerama::adjustWindowPosition(KaWindow* window) {
    if (!active) return;

    if (crossesBoundary(window) || inDeadArea(window)) {
        const XineramaScreenInfo* screen_info = determineScreen(window);

        int left, top, right, bottom;
        
        // Get Xinerama screen dimensions
        left   = screen_info->x_org;
        top    = screen_info->y_org;
        right  = screen_info->x_org + screen_info->width;
        bottom = screen_info->y_org + screen_info->height;
        
        if (window->GetFrameRight() > right) {
            window->SetFrameRight(right);
        } else if (window->GetFrameLeft() < left) {
            window->SetFrameLeft(left);
        }

        if (window->GetFrameTop() < top) {
            window->SetFrameTop(top);
        } else if (window->GetFrameBottom() > bottom) {
            window->SetFrameBottom(bottom);
        }
    }
}

/** 
 * @fn crossesBoundary(KaWindow* window)
 * 
 * @param window Window to check.
 * 
 * @return True if window crosses any Xinerama screen boundary, false
 * otherwise.
 */
bool Xinerama::crossesBoundary(KaWindow* window) {
    vector<XineramaScreenInfo>::iterator it;
    int n = 0;

    for (it = screens.begin(); it != screens.end(); ++it) {
        int left_edge = it->x_org;
        int right_edge = it->x_org + it->width;        
        int top_edge = it->y_org;
        int bottom_edge = it->y_org + it->height;

        if (intersects(left_edge, top_edge, right_edge, bottom_edge,
                       window->GetFrameLeft(), window->GetFrameTop(),
                       window->GetFrameRight(), window->GetFrameBottom())) {
            n++;
            if (n > 1) return true;
        }
    }
    return false;
}

/** 
 * @fn inDeadArea(KaWindow* window)
 * 
 * @param window Window to chck.
 * 
 * @return True if any part of the window is in area not covered by
 * any Xinerama screen (dead area).
 */
bool Xinerama::inDeadArea(KaWindow* window) {
    // What we do to determine the dead area is check intersection area
    // with each of the screens and see if the sum is equal to total window area.
    //
    // If it's less than it, then some part of thet window does not belong to any screen.

    int total_area = 0;
    vector<XineramaScreenInfo>::iterator it;

    for (it = screens.begin(); it != screens.end(); ++it) {
        total_area += intersectionArea(window, &(*it));
    }

    return total_area < (window->GetFrameWidth() * window->GetFrameHeight());
}
