/* $Id: main.cc,v 1.43 2004/02/04 21:53:28 brailsmt Exp $ */

/**
 * @file            main.cc
 * @date            01-May-2001 17:58:24
 *
 * @brief Main body implementation
 *
 * Starts Kahakai window manager.
 *
 */

#ifdef    HAVE_CONFIG_H
#  include "config.h"
#endif // HAVE_CONFIG_H

#include "main.hh"
#include "Kahakai.hh"
#include "Script.hh"
#include "ScriptFactory.hh"
#include "Logger.hh"
//#include "ka_logger.hh"

//extern int log_level;

#ifdef    HAVE_STDIO_H
#  include <cstdio>
#endif // HAVE_STDIO_H
#include <iostream>
#include <string>

using std::exception;
using std::string;
using std::endl;
using std::cout;

#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ptrace.h>

void usage(const string &program_name);
void help(const string &program_name);

static char *program;

void on_sigsegv(int sig) {

    pid_t gdb_pid;
    char pid_str[20];
    char cmd_file[PATH_MAX];

    sprintf(cmd_file, "%s/%s", getenv("HOME"), ".kahakai/cmd.gdb");

    logger << "caugh a SIGSEGV, " << PACKAGE << " just crashed :(" << std::endl;

    if (open(cmd_file, O_RDONLY) == -1) {
        realpath(GDB_COMMANDS_FILE, cmd_file);
        if (open(cmd_file, O_RDONLY) == -1) {
            logger << "couldn't find gdb commands file ..." << std::endl;
            exit(127);
        }
    }

    sprintf(pid_str, "%u", getpid());
    logger << "executing : gdb -quiet -x " << GDB_COMMANDS_FILE << " "
           << program << " -quiet " << pid_str << std::endl;

    std::cerr << "see log (" << logger.log_file() << ") for gdb output"
              << std::endl;

    if (!(gdb_pid = fork())) {
        dup2(open(logger.log_file().c_str(), O_APPEND | O_WRONLY), 1);
        execlp("gdb", "gdb", "-quiet", "-x", cmd_file, program, pid_str, 0);
        logger
        << "could not exec gdb ..."
        << std::endl;
        exit(127);
    }
    waitpid(gdb_pid, 0, 0);
    exit(127);
}

/**
 * @fn    main(int argc, char **argv)
 * @brief Main function for Kahakai window manager
 *
 * Parses command line options. Creates a new kahakai object.
 * Enters the eventloop.
 *
 * @param argc Number of command line arguments
 * @param argv Vector with command line arguments
 * @return     0 on successful execution else >0
 */
int main(int argc, char **argv) {

    program = argv[0];
    signal(SIGSEGV, on_sigsegv);

    struct kaoptions options;
    const string program_name(argv[0]);
    /* Set all option strings to NULL for now. */
    options.menufile = options.actionfile = NULL;
    options.stylefile = options.rcfile = options.display = NULL;
    options.scriptconfig = NULL;


    /* Set all the options to NULL initially. */
    options.menufile = options.actionfile = options.stylefile = NULL;
    options.rcfile = options.display = NULL;

    // thank you aewm
    for(int i = 1; i < argc; i++) {

        if(ARG("usage", "u", 0)) {
            usage(program_name);
            return 0;
        }
        if(ARG("help", "h", 0)) {
            help(program_name);
            return 0;
        }
        if(ARG("version", "v", 0)) {
            cout << PACKAGE << " " << VERSION << endl;
            return 0;
        }
        if(ARG("display", "d", 1)) {
            options.display = argv[++i];
            setenv("DISPLAY", options.display, 1);
            continue;
        }
        if(ARG("rcfile", "r", 1)) {
            options.rcfile = argv[++i];
            continue;
        }
        if(ARG("stylefile", "s", 1)) {
            options.stylefile = argv[++i];
            continue;
        }
        if(ARG("menufile", "m", 1)) {
            options.menufile = optarg;
            continue;
        }
        if(ARG("scriptconfig", "x", 1)) {
            options.scriptconfig = optarg;
            continue;
        }
		if(ARG("log", "l", 0)) {
			//log_level = OFF;
		}

        /* if we get here, it must be a bad option */
        logger << "unknown option: " << argv[i] << endl;
        usage(program_name);
        return 1;
    }

    try {
        //open_log(); //open the log file
        Kahakai *kahakai = new Kahakai(argv, &options);
        kahakai->RunEventLoop();
    } catch (exception& e) {
        logger << "*** Caught an exception ***" << endl;
        logger << e.what() << endl;
    }

    return 1;
}

/**
 * @fn    usage(void)
 * @brief Display brief usage message
 *
 * Prints brief usage message on standard out.
 *
 */
void usage(const string &program_name) {
    cout << "Usage: " << program_name << " [--display DISPLAYNAME]"
         << " [--rcfile CONFIGFILE]" << endl << "\t[--stylefile STYLEFILE]"
         << " [--menufile MENUFILE]" << endl << "\t[--usage]" << " [--help]"
         << " [--version]" << endl;
}

/** 
 * @fn    help(void)
 * @brief Shows help message
 *
 * Prints full command line help message to standard out.
 * 
 */
void help(const string &program_name) {
    cout << "Kahakai - an X11 window manager designed for maximum efficiency"
         << endl << "Usage: " << program_name << " [OPTION...]" << endl
         << endl
         << endl
         << " -d, --display DISPLAYNAME   X server to contact" << endl
         << " -r, --rcfile CONFIGFILE     Config-file to use" << endl
         << " -s, --stylefile STYLEFILE   Style-file to use" << endl
         << " -m, --menufile MENUFILE     Menu-file to use" << endl
         << " -u, --usage                 Show usage message and exit" << endl
         << " -h, --help                  Show this message and exit" << endl
         << " -v, --version               Show version and exit" << endl
         << endl
         << "Report bugs at <http://sf.net/projects/kahakai>." << endl;
}

